local Constants = require('Constants')
local DFFile = require('DFCommon.File')
local DFMath = require('DFCommon.Math')
local TrackDefs = require('AudioTrackDefinitions')

local AudioManager = {}

AudioManager.IntroTrack = { 'TitleScreenLoop', 31.88 }
AudioManager.GameOverTrack = 'Radius/Music/GameOverLoop'
AudioManager.GameCompleteTrack = { 'GameCompleteNote', 7.5 }
AudioManager.OverrideSet = -1   --set this to test out specific tracks (-1 is normal play)

function AudioManager:Init()    
    AudioManager.mAdvanceGameWithoutMusic = false
    AudioManager.mAdvanceGameScalar = 1.0
    AudioManager.mShouldEnterBuildState = false
    AudioManager.mAmbientCue = nil
    AudioManager.mMusicSection = 1
    AudioManager.mMusicSectionIndex = 0
    AudioManager.mMusicSectionMaxIndex = 0
    AudioManager.mMusicTimer = 0.0
    AudioManager.mMusicSoundDuration = 0.0
    AudioManager.mBuildStateTotalDuration = 0.0
    AudioManager.mIntensityDelay = 0.0
    AudioManager.mHeartbeatUp = true
    
    local i = 1
    AudioManager.mNumPrevFullMixTracks = 4
    AudioManager.mPrevFullMixIndexes = {}
    for i = 1, AudioManager.mNumPrevFullMixTracks do
        AudioManager.mPrevFullMixIndexes[i] = -1
    end    
    AudioManager.mFullMixTrack = 1
    AudioManager.mFullMixSection = 1
    AudioManager.mFullMixIndex = 1
    
    if MOAIFmodEventMgr then
        local initParams = {
            androidLowPerformance = Constants.UseLowPerformanceAudio
        }
        
        MOAIFmodEventMgr.init( initParams )
        
        MOAIFmodEventMgr.loadProject( DFFile.getAudioPath('Radius/Radius.fev') ) 
        MOAIFmodEventMgr.loadGroup("Radius/SFX")    
        MOAIFmodEventMgr.loadGroup("Radius/UI")   
        MOAIFmodEventMgr.loadGroup("Radius/Music")   
      
        --set sound event for seamless queue
        MOAIFmodEventMgr.initSeamlessQueue( "Radius/Music/MultiDSPMobile" )
        --MOAIFmodEventMgr.setSoundCategoryVolume("master", 0.001)
        MOAIFmodEventMgr.setSoundCategoryVolume("master", 1.0)
        
        MOAIFmodEventMgr.initFilters()    
    end
end

function AudioManager:SetFullMixMode( bFullMixMode )
    AudioManager.mFullMixTrack = 1
    AudioManager.mFullMixSection = 1
    AudioManager.mFullMixIndex = 1
    
    Constants.FullMixMode = bFullMixMode
    AudioManager:PickFullMixTrack()
end

function AudioManager:PickFullMixTrack()
    local i = 1
    local iNumTracks = #TrackDefs.FullMixSets
    local iIndex = -1
    while true do
        iIndex = math.random(1.0, iNumTracks)   

        local bUsedPreviously = false
        for i = 1, AudioManager.mNumPrevFullMixTracks do
            if AudioManager.mPrevFullMixIndexes[i] == iIndex then
                bUsedPreviously = true
                break
            end
        end        
        
        if not bUsedPreviously then
            break
        end
    end 
 
    --push back indexes and set this one as our first
    i = AudioManager.mNumPrevFullMixTracks
    while i > 1 do
        AudioManager.mPrevFullMixIndexes[i] = AudioManager.mPrevFullMixIndexes[i-1]
        i = i - 1
    end 
    AudioManager.mPrevFullMixIndexes[1] = iIndex    
    AudioManager.mFullMixTrack = iIndex    
end

function AudioManager:GetTrackPercent()
    if AudioManager.mMusicSoundDuration == 0.0 then
        return 0.0
    end
    
    return 1.0 - (AudioManager.mMusicTimer / AudioManager.mMusicSoundDuration)
end

function AudioManager:SetAdvanceGameFlag( bAdvanceWithoutMusic )
    AudioManager.mAdvanceGameWithoutMusic = bAdvanceWithoutMusic
end

function AudioManager:SetAdvanceGameScalar( fScalar )
    AudioManager.mAdvanceGameScalar = fScalar
end

function AudioManager:UpdateMusicTrack()    
    Constants.BeatDuration = 0.46875
    Constants.BeatActive = false
    
    --hold the music timer if we're paused
    if not Constants.mPaused then
        if MOAIFmodEventMgr.getNumItemsInSeamlessQueue() > 0 and not AudioManager.mAdvanceGameWithoutMusic then        
            --ask fmod for the position of the audio track
            local fMusicTimer = MOAIFmodEventMgr.getTimeRemainingForCurrentItemInSeamlessQueue()
            
            --update our music timer with the frame step
            local fLuaTimer = AudioManager.mMusicTimer - Constants.FrameStep
            
            --if they deviate too much, use the audio position
            local fMaxDeviation = 0.5
            if math.abs(fMusicTimer - fLuaTimer) > fMaxDeviation then
                fLuaTimer = fMusicTimer
            end
            
            --take the average of the two timers to get a smooth result
            AudioManager.mMusicTimer = (fLuaTimer + fMusicTimer) * 0.5
        else
            --if we're not playing from the seamless queue, just update with frame step
            AudioManager.mMusicTimer = AudioManager.mMusicTimer - (Constants.FrameStep * AudioManager.mAdvanceGameScalar)
            if AudioManager.mMusicTimer < 0.0 then
                AudioManager.mMusicTimer = AudioManager.mMusicTimer + AudioManager.mMusicSoundDuration
            end
        end
    end

    local fScaledBeat = Constants.BeatDuration / Constants.BeatScale
    
    --figure out how far through the beat we are
    local fPrevProgress = Constants.BeatProgress
    local fTimeProgressed = AudioManager.mMusicSoundDuration - AudioManager.mMusicTimer
    local iNumBeatsIn = math.floor( fTimeProgressed / fScaledBeat )
    local fBeatProgress = fTimeProgressed - (iNumBeatsIn * fScaledBeat)
    Constants.BeatProgress = fBeatProgress / fScaledBeat 
        
    --if we started a new beat, flip our heartbeat movement flag
    if fPrevProgress > Constants.BeatProgress and fPrevProgress > 0.75 then
        AudioManager.mHeartbeatUp = not AudioManager.mHeartbeatUp
        Constants.BeatActive = true
    end
    
    --udpate heartbeat
    Constants.Heartbeat = Constants.BeatProgress
    if not AudioManager.mHeartbeatUp then
        Constants.Heartbeat = 1.0 - Constants.Heartbeat            
    end  
end

function AudioManager:GetLevelIntensityFromMusic()      
    if Constants.MusicState == Constants.MusicStateBuilding then       
        --total up current time    
        local iAudioSetIndex = AudioManager:GetCurrentMusicSet(0) 
        local iLastFinishedTrackIndex = AudioManager.mMusicSectionIndex - 1
        local fTotalTimeProgressed = 0.0
        for i = 1, iLastFinishedTrackIndex do
            fTotalTimeProgressed = fTotalTimeProgressed + TrackDefs.MusicSets[iAudioSetIndex][3][i][2]
        end            
        
        --add on time progressed through this track
        fTotalTimeProgressed = fTotalTimeProgressed + (AudioManager.mMusicSoundDuration - AudioManager.mMusicTimer)
        
        return fTotalTimeProgressed / AudioManager.mBuildStateTotalDuration
    elseif Constants.LevelState == Constants.LevelStateBuild then
        --if we're not in the music building state, but we ARE in the level build state, we just need to ride it out
        return (AudioManager.mBuildStateTotalDuration - AudioManager.mIntensityDelay) / AudioManager.mBuildStateTotalDuration        
    elseif Constants.LevelState == Constants.LevelStateBonus then        
        --bonus mode is always at 1.0
        return 1.0
    elseif Constants.MusicState == Constants.MusicStateLoop then
        --if we're in the loop state and not the first section, bump the intensity
        if AudioManager.mMusicSection ~= 1 and AudioManager.mIntensityDelay <= 0.0 then
            return 1.0
        end
    end
    
    return 0.0
end

function AudioManager:Update()
    AudioManager:UpdateMusicTrack()
    
    --update delay used for timed intensity setting
    AudioManager.mIntensityDelay = AudioManager.mIntensityDelay - Constants.FrameStep
    
    if Constants.UseAudio then
        local currentAlloc, maxAlloc = MOAIFmodEventMgr.getMemoryStats()
        currentAlloc = DFMath.round((currentAlloc / (1024 * 1024)), 3)
        if (Constants.ShowDebugInfo) then   
            g_DebugDisplay:SetValue("SND", currentAlloc)
        end
        
        AudioManager:UpdateFFT()
               
        --determine if we need to pick a new music cue
        local bQueuePaused = MOAIFmodEventMgr.isSeamlessQueuePaused()
        local bMenuState = Constants.MusicState == Constants.MusicStateEnd
        if not Constants.mPaused and not bQueuePaused and not bMenuState then
            local fReasonablyClose = 1.5
            local bPickNewCue = false
        
            local numQueuedTracks = MOAIFmodEventMgr.getNumItemsInSeamlessQueue()

            --if we ever end up with no tracks queued.. queue one up immediately
            if numQueuedTracks == 0 then
                bPickNewCue = true
                AudioManager.mMusicTimer = 0.0
            else
                local fTimeRemaining = MOAIFmodEventMgr.getTimeRemainingForCurrentItemInSeamlessQueue()
                --queue up a cue if we don't have a ton queued up, we're not getting bogus time stamps, and we're near the end
                bPickNewCue = (numQueuedTracks < 2) and (fTimeRemaining > 0.0) and (fTimeRemaining < fReasonablyClose)                        
            end
            
            if bPickNewCue then
                if Constants.FullMixMode then
                    AudioManager:AdvanceFullMixTrack()
                else
                    AudioManager:AdvanceMusicTrack()
                end
            end
        end
    end
end
  
function AudioManager:CalculateBuildStateTotalDuration()
    AudioManager.mBuildStateTotalDuration = 0.0
    
    --total up time of build tracks
    local iAudioSetIndex = AudioManager:GetCurrentMusicSet(0)   
    local i = 1
    for i = i, #TrackDefs.MusicSets[iAudioSetIndex][3] do
        AudioManager.mBuildStateTotalDuration = AudioManager.mBuildStateTotalDuration + TrackDefs.MusicSets[iAudioSetIndex][3][i][2]
    end
end

function AudioManager:UpdateFFT()
    if not MOAIFmodEventMgr then
        return 
    end
    
    local numSpectrumSamples = Constants.AudioSpectrumNumValues
    local result = MOAIFmodEventMgr.getSpectrum(numSpectrumSamples)
    if result then
        Constants.tAudioSpectrum = result
        
        -- let the C++ do the average for us
        result = MOAIFmodEventMgr.getSpectrum(1)
        
        local curAvg = result[1]
        Constants.AudioSpectrumAverage = curAvg
        
        local sampleVal = Constants.tAudioSpectrum[1]
        Constants.AudioSpectrumSmoothValue = DFMath.lerp( Constants.AudioSpectrumSmoothValue, 0.0, 0.025 )
        
        -- beat detection! Unfortunately different audio tracks will need different clip values here.
        local kCLIP_AVG = 0.30
        if sampleVal > kCLIP_AVG and AudioManager.mMusicTimer > 0.5 then    
            Constants.AudioSpectrumSmoothValue = sampleVal
        end
    end
end

function AudioManager:AdvanceFullMixTrack()
    --update the section index of the full mix track
    AudioManager.mFullMixIndex = AudioManager.mFullMixIndex + 1
    
    --if the section index is beyond the number of tracks for this section..
    if AudioManager.mFullMixIndex > #TrackDefs.FullMixSets[AudioManager.mFullMixTrack][AudioManager.mFullMixSection] then
        AudioManager.mFullMixIndex = 1
        
        --increment the section
        AudioManager.mFullMixSection = AudioManager.mFullMixSection + 1
        
        --if we're done with the final section..
        if AudioManager.mFullMixSection > 6 then
            --pick a new track
            AudioManager.mFullMixSection = 1
            AudioManager:PickFullMixTrack()
        end        
    end
    
    local tNextCue = TrackDefs.FullMixSets[AudioManager.mFullMixTrack][AudioManager.mFullMixSection][AudioManager.mFullMixIndex]
    AudioManager:PlayMusicTrack( tNextCue )
    
    if AudioManager.mFullMixSection == 1 then
        --this isn't used in full mix mode, but we're setting it to trick the visuals
        AudioManager.mMusicSection = 1
        Constants.LevelStateSwitchTimer = AudioManager.mMusicTimer
        Constants.DesiredLevelState = Constants.LevelStateStart
        Constants.MusicState = Constants.MusicStateLoop
    elseif AudioManager.mFullMixSection == 2 then
        AudioManager.mMusicSection = 2
        AudioManager.mIntensityDelay = AudioManager.mMusicTimer
    elseif AudioManager.mFullMixSection == 3 then
        Constants.LevelStateSwitchTimer = AudioManager.mMusicTimer        
        Constants.DesiredLevelState = Constants.LevelStateBuild
        Constants.MusicState = Constants.MusicStateBuilding
        AudioManager.mBuildStateTotalDuration = tNextCue[2]
    elseif AudioManager.mFullMixSection == 4 then
        AudioManager.mIntensityDelay = AudioManager.mMusicTimer        
        Constants.LevelStateSwitchTimer = AudioManager.mMusicTimer
        Constants.DesiredLevelState = Constants.LevelStateBonus        
        Constants.MusicState = Constants.MusicStateBonus
    elseif AudioManager.mFullMixSection == 5 then
        Constants.LevelStateSwitchTimer = AudioManager.mMusicTimer        
        Constants.DesiredLevelState = Constants.LevelStateBuild  
        Constants.MusicState = Constants.MusicStateBuilding  
        AudioManager.mBuildStateTotalDuration = tNextCue[2]
    elseif AudioManager.mFullMixSection == 6 then
        AudioManager.mIntensityDelay = AudioManager.mMusicTimer        
        Constants.LevelStateSwitchTimer = AudioManager.mMusicTimer
        Constants.DesiredLevelState = Constants.LevelStateBonus    
        Constants.MusicState = Constants.MusicStateBonus
    end
end

function AudioManager:AdvanceMusicTrack()    
    if Constants.MusicState == Constants.MusicStateLoop then
        local iAudioSetIndex = AudioManager:GetCurrentMusicSet(0)
        
        if AudioManager.mShouldEnterBuildState == true then
            --[[
            --debug block for testing game end
            Constants.LevelStateSwitchTimer = AudioManager.mMusicTimer
            Constants.DesiredLevelState = Constants.LevelStateBonus            
            Constants.MusicState = Constants.MusicStateBonus
            
            AudioManager:SetMusicSection(4, iAudioSetIndex)
            AudioManager:PlayMusicTrack( TrackDefs.MusicSets[iAudioSetIndex][4][1] )    
            --]]
            -- [[
            --delay our level state switch until the next track starts
            Constants.LevelStateSwitchTimer = AudioManager.mMusicTimer        
            Constants.DesiredLevelState = Constants.LevelStateBuild
            Constants.MusicState = Constants.MusicStateBuilding
            AudioManager.mShouldEnterBuildState = false
            
            AudioManager:CalculateBuildStateTotalDuration()
            AudioManager:SetMusicSection(3, iAudioSetIndex)
            
            AudioManager:PlayMusicTrack( TrackDefs.MusicSets[iAudioSetIndex][3][AudioManager.mMusicSectionIndex] )        
            --]]
        else
            --if we are in section 1, play through the tracks in order
            if AudioManager.mMusicSection == 1 then
                AudioManager.mMusicSectionIndex = AudioManager.mMusicSectionIndex + 1
                
                --if we're done playing through our tracks, switch to section 2
                if AudioManager.mMusicSectionIndex > AudioManager.mMusicSectionMaxIndex then
                    AudioManager:SetMusicSection(2, iAudioSetIndex)
                    AudioManager.mMusicSectionIndex = 1
                    AudioManager.mIntensityDelay = AudioManager.mMusicTimer        
                end            
            elseif AudioManager.mMusicSection == 2 then
                --if we're in section 2, pick the tracks randomly, without repeating
                local iRandTrack = math.random(1, AudioManager.mMusicSectionMaxIndex)
                while iRandTrack == AudioManager.mMusicSectionIndex and AudioManager.mMusicSectionMaxIndex > 1 do
                    iRandTrack = math.random(1, AudioManager.mMusicSectionMaxIndex)
                end        
                
                AudioManager.mMusicSectionIndex = iRandTrack
            end
            
            AudioManager:PlayMusicTrack( TrackDefs.MusicSets[iAudioSetIndex][AudioManager.mMusicSection][AudioManager.mMusicSectionIndex] )            
        end
    elseif Constants.MusicState == Constants.MusicStateBuilding then     
        local iAudioSetIndex = AudioManager:GetCurrentMusicSet(0)
             
        --if we're on our last build track, switch to bonus!        
        if AudioManager.mMusicSectionIndex == AudioManager.mMusicSectionMaxIndex then      
            --save off our music timer so we can update the level intensity after we leave this music state
            AudioManager.mIntensityDelay = AudioManager.mMusicTimer        
        
            --delay our level state switch until the next track starts
            Constants.LevelStateSwitchTimer = AudioManager.mMusicTimer
            Constants.DesiredLevelState = Constants.LevelStateBonus            
            Constants.MusicState = Constants.MusicStateBonus
            
            AudioManager:SetMusicSection(4, iAudioSetIndex)
            AudioManager:PlayMusicTrack( TrackDefs.MusicSets[iAudioSetIndex][4][1] )    
        else
            --progress through the building tracks
            AudioManager.mMusicSectionIndex = AudioManager.mMusicSectionIndex + 1
            AudioManager:PlayMusicTrack( TrackDefs.MusicSets[iAudioSetIndex][3][AudioManager.mMusicSectionIndex] )        
        end
    elseif Constants.MusicState == Constants.MusicStateBonus then
        if AudioManager.mMusicSectionIndex == AudioManager.mMusicSectionMaxIndex then            
            --delay our level state switch until the next track starts
            Constants.LevelStateSwitchTimer = AudioManager.mMusicTimer
            Constants.DesiredLevelState = Constants.LevelStateHealth
            Constants.MusicState = Constants.MusicStateLoop
            
            --once we exit the AudioManager:Update, we're going to advance the game level, so we need to compensate for that
            --  hence, the offset of 1            
            local iAudioSetIndex = AudioManager:GetCurrentMusicSet(1) 
            if iAudioSetIndex == -1 then
                --game complete!
                AudioManager:PlayMusicTrack( AudioManager.GameCompleteTrack )
                Constants.MusicState = Constants.MusicStateComplete
            else
                AudioManager:SetMusicSection(1, iAudioSetIndex)        
                AudioManager:PlayMusicTrack( TrackDefs.MusicSets[iAudioSetIndex][1][AudioManager.mMusicSectionIndex] )  
            end
        else
            --progress through the chorus tracks
            local iAudioSetIndex = AudioManager:GetCurrentMusicSet(0) 
            AudioManager.mMusicSectionIndex = AudioManager.mMusicSectionIndex + 1
            AudioManager:PlayMusicTrack( TrackDefs.MusicSets[iAudioSetIndex][4][AudioManager.mMusicSectionIndex] )           
        end
    elseif Constants.MusicState == Constants.MusicStateIntro then
        AudioManager:PlayMusicTrack( AudioManager.IntroTrack )
    end
end
         
function AudioManager:SetMusicSection(iIndex, iAudioSetIndex)
    AudioManager.mMusicSection = iIndex
    AudioManager.mMusicSectionIndex = 1
    AudioManager.mMusicSectionMaxIndex = #TrackDefs.MusicSets[iAudioSetIndex][iIndex] 
    
    --wipe out all buffered tracks because we're switching sets
    AudioManager:PrepMusicBuffer()
end

function AudioManager:SetEnterBuildStateFlag(bShouldEnter)
    AudioManager.mShouldEnterBuildState = bShouldEnter
end
     
function AudioManager:GetCurrentMusicSet(iLevelOffset)
    if AudioManager.OverrideSet ~= -1 then
        return AudioManager.OverrideSet
    end
        
    --index is determined by game level
    local iIndex = Constants.CurrentGameLevel + iLevelOffset
    
    --if we're advancing past the last level, return an identifier for "we're done"
    if iIndex > Constants.NumGameLevels then
        return -1
    end
    
    local iNumMusicSets = #TrackDefs.MusicSets
    while iIndex > iNumMusicSets do
        iIndex = iIndex - iNumMusicSets
    end
    return iIndex
end

--This is an internal function only!  Don't use it!  -ph
function AudioManager:PlayMusicTrack(track)
    if Constants.UseAudio == true then
        if not MOAIFmodEventMgr then
            return 
        end
                
        -- queue track
        sTrackFile = "Data/Audio/" .. track[1] .. ".fsb"
        print("adding track to queue: " .. sTrackFile)
        
        MOAIFmodEventMgr.addToSeamlessQueue(sTrackFile)
        AudioManager.mMusicSoundDuration = track[2]
    end
end

function AudioManager:FlushMusicBuffer()
    MOAIFmodEventMgr.resetSeamlessQueue()
end

function AudioManager:PrepMusicBuffer()
    MOAIFmodEventMgr.removeNonPlayingSoundsFromSeamlessQueue()
end

function AudioManager:PlaySound(eventString)
    if Constants.UseSFX == false then
        return nil
    end
    if not MOAIFmodEventMgr then
        return nil
    end
        
    return MOAIFmodEventMgr.playEvent2D( eventString )     
end

function AudioManager:StopSound(rSoundEvent, bHardStop)
    if rSoundEvent == nil then
        return
    end
    if not MOAIFmodEventMgr then
        return 
    end    
    
    rSoundEvent:stop( bHardStop )     
end

function AudioManager:StartAmbientTrack()
    Constants.MusicState = Constants.MusicStateIntro
end

function AudioManager:StartLoopTrack()    
    if Constants.UseAudio then
        --make sure whatever music we're currently playing is stopped
        AudioManager:StopAmbientTrack()
        AudioManager:FlushMusicBuffer()
            
        if Constants.FullMixMode then
            AudioManager:PlayMusicTrack( TrackDefs.FullMixSets[AudioManager.mFullMixTrack][1][1] )            
        else
            local iAudioSetIndex = AudioManager:GetCurrentMusicSet(0)
            AudioManager:SetMusicSection(1, iAudioSetIndex)

            AudioManager:PlayMusicTrack( TrackDefs.MusicSets[iAudioSetIndex][1][1] )
        end

        Constants.MusicState = Constants.MusicStateLoop
    end
end

function AudioManager:PlayGameEndAudio()
    if Constants.UseAudio then
        AudioManager:FlushMusicBuffer()
        AudioManager.mAmbientCue = AudioManager:PlaySound( AudioManager.GameOverTrack )
        Constants.MusicState = Constants.MusicStateEnd 
    end
end

function AudioManager:StopMusicForPostGame()
    AudioManager:FlushMusicBuffer()
    Constants.MusicState = Constants.MusicStateEnd
end

function AudioManager:StopAmbientTrack()
    AudioManager:StopSound( AudioManager.mAmbientCue, true )
    AudioManager.mAmbientCue = nil
end

function AudioManager:PauseMusic()
    MOAIFmodEventMgr.pauseSeamlessQueue( Constants.mPaused )
end

return AudioManager
