local Constants = require('Constants')
local DFMath = require('DFCommon.Math')
local AudioManager = require('AudioManager')
local InputTutorial = require('InputTutorial')
local RadEffects = require('RadEffects')
local Achievement = require('DFCommon.Achievement')
local ChargeManager = {}

-- Touch Position Object
function ChargeManager:NewTouchPosition()
    local TouchPosition = {}
  
    function TouchPosition:Init() 
        TouchPosition.mPos = {0, 0}
        TouchPosition.mUsed = false
    end
  
    TouchPosition:Init()
    return TouchPosition
end

-- Charge Object
function ChargeManager:NewCharge()
    local Charge = {}
    
    function Charge:Init()
        Charge.mState = Constants.ChargeStateDisabled
        Charge.mIndex = 0
        Charge.mPrevPos = {0, 0}
        Charge.mPos = {0, 0}
        Charge.mArcPos = {0, 0}    
        Charge.mZ = 1.0
    
        Charge.mChargeDisabledDuration = Constants.GameIntroChargeActivationDuration
        Charge.mChargeDisabledTimer = Charge.mChargeDisabledDuration
        Charge.mAllChargesEnabled = false
        Charge.mAltGameMode = false
        Charge.mBeamActiveRatio = 0.0
        Charge.mIntroTutorialValidAmount = 0.0
        
        Charge.mNumTouchPositions = 6
        Charge.mTouchPositions = {}
        
        local i = 0
        for i = 1, Charge.mNumTouchPositions do
            Charge.mTouchPositions[i] = ChargeManager:NewTouchPosition()
        end        
                  
        Charge.mNumThrowPositions = 10
        Charge.mPrevThrowPositions = {}
        for i = 1, Charge.mNumThrowPositions do
            Charge.mPrevThrowPositions[i] = {0, 0}
        end
            
        Charge.mPrevRotation = 0.0
        Charge.mRotation = 0.0
        Charge.mThrownRotationSpeed = 0.0    
        Charge.mThrowAmount = 0.0
                
        Charge.mThrowTimer = 0.0 --amount of time after a charge has been thrown
        
        Charge:DetermineSize()   
    end

    function Charge:DetermineSize()
        Charge.mArcRadius = Constants.GameRingRadius * 0.9 
        Charge.mRadius = Constants.ValidDrawHeight * 0.04  

        Charge.mDrawRadius = Charge.mRadius
        Charge.mDrawRadiusDecay = 10.0
        Charge.mDrawScale = 1.0
    end
    
    function Charge:CreateProp( tRenderer, iPriority )
        local fHalfPenWidth = math.max( Constants.BasePenWidth * 0.5, 2.0 )
        
        Charge.mBorderProp = ShapeProp.new()
        Charge.mBorderProp:setDeck( tRenderer.mShapeDeck )
        Charge.mBorderProp:SetNumSides( 32 )
        Charge.mBorderProp:SetRadii( 0.0, Charge.mRadius + fHalfPenWidth )  
        Charge.mBorderProp:setVisible( false )
        Charge.mBorderProp:setLoc( 0.0, 0.0, Charge.mZ )
        
        tRenderer.mGameLayer:insertProp( Charge.mBorderProp )        
        
        Charge.mInsideProp = ShapeProp.new()
        Charge.mInsideProp:setDeck( tRenderer.mShapeDeck )
        Charge.mInsideProp:SetNumSides( 32 )
        Charge.mInsideProp:SetRadii( 0.0, Charge.mRadius - fHalfPenWidth )  
        Charge.mInsideProp:setVisible( false )
        Charge.mInsideProp:setLoc( 0.0, 0.0, Charge.mZ + 0.01 )
        
        tRenderer.mGameLayer:insertProp( Charge.mInsideProp )        
    end
    
    function Charge:Reset()
        Charge:SetState( Constants.ChargeStateDisabled )
        Charge:ClearPreviousPositions()
        Charge:ResetThrowPositions()    
        Charge.mBorderProp:setVisible( false )
        Charge.mInsideProp:setVisible( false )  
        Charge.mIntroTutorialValidAmount = 0.0       
        Charge.mThrowAmount = 0.0
    end
    
    function Charge:UpdateSize()
        Charge:DetermineSize()
        
        local fHalfPenWidth = math.max( Constants.BasePenWidth * 0.5, 2.0 )
        Charge.mBorderProp:SetRadii( 0.0, Charge.mRadius + fHalfPenWidth )   
        Charge.mInsideProp:SetRadii( 0.0, Charge.mRadius - fHalfPenWidth )
        
        --update all positions
        Charge.mPrevPos[1], Charge.mPrevPos[2] = Charge:ProjectInput(Charge.mPrevPos[1], Charge.mPrevPos[2], Constants.GameRingRadius) 
        Charge.mPos[1], Charge.mPos[2] = Charge:ProjectInput(Charge.mPos[1], Charge.mPos[2], Constants.GameRingRadius) 
        Charge.mArcPos[1], Charge.mArcPos[2] = Charge:ProjectInput(Charge.mArcPos[1], Charge.mArcPos[2], Charge.mArcRadius)

        local i = 1
        for i = 1, Charge.mNumThrowPositions do
            Charge.mPrevThrowPositions[i][1], Charge.mPrevThrowPositions[i][2] = Charge:ProjectInput(Charge.mPrevThrowPositions[i][1], Charge.mPrevThrowPositions[i][2], Constants.GameRingRadius)
        end       
       
        Charge.mBorderProp:setLoc( Charge.mPos[1], Charge.mPos[2], Charge.mZ )
        Charge.mInsideProp:setLoc( Charge.mPos[1], Charge.mPos[2], Charge.mZ + 0.01 )       
    end
    
    function Charge:IsEnabled()
        return Charge.mChargeDisabledTimer ~= Charge.mChargeDisabledDuration
    end

    function Charge:IsActive()
        return Charge.mState == Constants.ChargeStateActive
    end
    
    function Charge:IsUncontrolled()
        return Charge.mState == Constants.ChargeStateThrown or Charge.mState == Constants.ChargeStateNoInput
    end
  
    function Charge:HasInput()
        return Charge.mState ~= Constants.ChargeStateNoInput and Charge.mState ~= Constants.ChargeStateDisabled
    end
    
    function Charge:IsThrown()
        return Charge.mState == Constants.ChargeStateThrown 
    end
        
    function Charge:SetState( iState )
        --exit state logic
        if Charge.mState == Constants.ChargeStateDisabled then
            if iState == Constants.ChargeStateActive then
                --if we're going from disabled to active, turn us on
                Charge.mBorderProp:setVisible( true )
                Charge.mInsideProp:setVisible( true )
                
                Charge:BurstEffect()
            end
        elseif Charge.mState == Constants.ChargeStateThrown then
            --if we used to be thrown, reset our rotation speed
            Charge.mThrownRotationSpeed = 0.0
        end

        Charge.mState = iState
        
        --enter state logic
        if Charge.mState == Constants.ChargeStateDisabled then
            --reset our disabled timer if we're disabled
            Charge.mChargeDisabledTimer = Charge.mChargeDisabledDuration
            Charge.mBorderProp:setVisible( false )
            Charge.mInsideProp:setVisible( false )  
        end
    end
    
    function Charge:SetInAltGameMode( bAltGame )
        Charge.mAltGameMode = bAltGame
        Charge.mBorderProp:setVisible( not Charge.mAltGameMode )
        Charge.mInsideProp:setVisible( not Charge.mAltGameMode )            
        if Charge.mAltGameMode then
            Charge:BurstEffect()
        end
    end
    
    function Charge:BurstEffect()
        RadEffects:SetEffectLayerToGameLayer()
        RadEffects:SetEffectTransform( Charge.mPos[1], Charge.mPos[2], 0.0, 0.0, 0.0, 0.0 )
        RadEffects:CreateEffect( "Effects/Particles/Chord/ChargeCreation.particles", nil, nil, nil, Constants.ParticlePriorityHigh )
        RadEffects:CreateEffect( "Effects/Particles/Chord/ChargeFlash.particles", nil, nil, nil, Constants.ParticlePriorityHigh ) 
        RadEffects:CreateEffect( "Effects/Particles/Chord/ChargeEmbers.particles", nil, nil, nil, Constants.ParticlePriorityHigh )    
    end
    
    function Charge:UpdatePaused( tGameInputs )
        Charge.mThrownRotationSpeed = 0.0
        if tGameInputs[Charge.mIndex].IsTouched() then
            Charge.mState = Constants.ChargeStateActive
            Charge:SetTouchPosition( tGameInputs[Charge.mIndex].mX, tGameInputs[Charge.mIndex].mY )       
        else
            Charge.mState = Constants.ChargeStateNoInput
        end    
    end
    
    function Charge:Update( tGameInputs )
        Charge.mThrowTimer = Charge.mThrowTimer + Constants.FrameStep  
                
        --record previous position for collision checking
        Charge.mPrevPos[1] = Charge.mPos[1]
        Charge.mPrevPos[2] = Charge.mPos[2]
                
        local bWasUncontrolled = Charge:IsUncontrolled()
        
        if tGameInputs[Charge.mIndex].IsTouched() then  
            local bAllowChargeActive = true
            
            --disabled state logic
            if Charge.mState == Constants.ChargeStateDisabled then
                --keep position up to date
                Charge:SetPosition( tGameInputs[Charge.mIndex].mX, tGameInputs[Charge.mIndex].mY ) 
                
                --tick down our disabled timer
                if Charge.mIntroTutorialValidAmount > 0.0 then
                    Charge.mChargeDisabledTimer = Charge.mChargeDisabledTimer - (Constants.FrameStep * Charge.mIntroTutorialValidAmount)
                    bAllowChargeActive = (Charge.mChargeDisabledTimer <= 0.0)
                else
                    Charge.mChargeDisabledTimer = Charge.mChargeDisabledDuration
                    bAllowChargeActive = false
                end               
            end
            
            if bAllowChargeActive == true then
                Charge:SetState( Constants.ChargeStateActive )
                Charge:SetPosition( tGameInputs[Charge.mIndex].mX, tGameInputs[Charge.mIndex].mY ) 
                
                --if we used to be uncontrolled and now we are, set our previous position to our current to avoid bad collision checks
                if bWasUncontrolled then
                    Charge.mPrevPos[1] = Charge.mPos[1]
                    Charge.mPrevPos[2] = Charge.mPos[2]           
                    Charge:ClearPreviousPositions()
                end
            end
        else            
            local bDisable = true
                                  
            --if we're thrown, update our position
            if Charge.mState == Constants.ChargeStateThrown then
                local bStillMoving = Charge:UpdateThrow()
                if bStillMoving then
                    bDisable = false
                end      
            elseif Charge.mAllChargesEnabled and Charge.mState == Constants.ChargeStateActive and Constants.ThrowUnlocked and not Charge.mAltGameMode then         
                --if we've got no input coming in but we're active, see if we should be thrown
                Charge:CheckForThrow()
                if Charge.mState == Constants.ChargeStateThrown then
                    bDisable = false
                end
            end
            
            --no input coming in, and not thrown, so shut us down 
            if bDisable then
                if Constants.GameState == Constants.GameStateIntro then
                    Charge:SetState( Constants.ChargeStateDisabled )
                else 
                    Charge:SetState( Constants.ChargeStateNoInput )
                    Charge:ClearPreviousPositions()
                end
            end
        end
        
        Charge:UpdateScale( tGameInputs )
        Charge:UpdateProps()
    end
    
    function Charge:UpdateScale( tGameInputs )
        Charge.mDrawScale = 0.25
        local fMaxDistance = Constants.ValidDrawHalfHeight
        local fDistanceToTouch = fMaxDistance
        if tGameInputs[Charge.mIndex].IsTouched() then
            fDistanceToTouch = DFMath.distance2D( Charge.mPos[1], Charge.mPos[2], tGameInputs[Charge.mIndex].mX, tGameInputs[Charge.mIndex].mY )
            local fDistanceRatio = math.min(fDistanceToTouch, fMaxDistance) / fMaxDistance
            Charge.mDrawScale = 0.25 + ((1.0 - fDistanceRatio) * 0.75)
        end       

        Charge.mDrawRadius = (Charge.mRadius * Charge.mDrawScale)
    end
    
    function Charge:UpdateProps() 
        --set drawing state
        local fHalfPenWidth = math.max( Constants.BasePenWidth * 0.5, 1.0 ) * Charge.mDrawScale
        Charge.mBorderProp:setLoc( Charge.mPos[1], Charge.mPos[2] )
        Charge.mInsideProp:setLoc( Charge.mPos[1], Charge.mPos[2] )
            
        if Charge.mState == Constants.ChargeStateActive then
            Charge.mBorderProp:SetDrawMode( ShapeProp.DrawModeCircle )
            Charge.mBorderProp:SetColor( Constants.Color1[1], Constants.Color1[2], Constants.Color1[3], 1.0 )
            Charge.mBorderProp:SetRadii( 0.0, Charge.mDrawRadius + fHalfPenWidth )    
        
            Charge.mInsideProp:SetDrawMode( ShapeProp.DrawModeCircle )
            Charge.mInsideProp:SetColor( Constants.Color2[1], Constants.Color2[2], Constants.Color2[3], 1.0 )
            Charge.mInsideProp:SetRadii( 0.0, Charge.mDrawRadius - fHalfPenWidth )        
        elseif Charge.mState == Constants.ChargeStateThrown then
            local fHalfRadius = Charge.mRadius * 0.4
            local fHalfArcWidth = 0.2
            Charge.mBorderProp:SetDrawMode( ShapeProp.DrawModeArc )
            Charge.mBorderProp:SetRadii( Charge.mArcRadius - fHalfRadius, Charge.mArcRadius + fHalfRadius )
            Charge.mBorderProp:SetColor( Constants.Color1[1], Constants.Color1[2], Constants.Color1[3], 1.0 )
            Charge.mBorderProp:SetArcRadians( Charge.mRotation - fHalfArcWidth, Charge.mRotation + fHalfArcWidth )    
            Charge.mBorderProp:setLoc( 0.0, 0.0 )        

            local fBorder = fHalfRadius * 0.2
            local fArcBorder = fHalfArcWidth * 0.1
            Charge.mInsideProp:SetDrawMode( ShapeProp.DrawModeArc )
            Charge.mInsideProp:SetRadii( Charge.mArcRadius - fHalfRadius + fBorder, Charge.mArcRadius + fHalfRadius - fBorder )
            Charge.mInsideProp:SetColor( Constants.Color2[1], Constants.Color2[2], Constants.Color2[3], 1.0 )
            Charge.mInsideProp:SetArcRadians( Charge.mRotation - fHalfArcWidth + fArcBorder, Charge.mRotation + fHalfArcWidth - fArcBorder )    
            Charge.mInsideProp:setLoc( 0.0, 0.0 ) 
        elseif Charge.mState == Constants.ChargeStateNoInput then
            local fHalfPenWidth = DFMath.lerp( 2.0, Constants.BasePenWidth * 0.5, Charge.mBeamActiveRatio )

            Charge.mBorderProp:SetDrawMode( ShapeProp.DrawModeCircle )
            Charge.mBorderProp:SetColor( Constants.Color1[1], Constants.Color1[2], Constants.Color1[3], 1.0 )            
            Charge.mBorderProp:SetRadii( 0.0, Charge.mDrawRadius + fHalfPenWidth )    

            Charge.mInsideProp:SetDrawMode( ShapeProp.DrawModeCircle )
            Charge.mInsideProp:SetColor( Constants.Color2[1], Constants.Color2[2], Constants.Color2[3], 1.0 )
            Charge.mInsideProp:SetRadii( 0.0, Charge.mDrawRadius - fHalfPenWidth )            
        elseif Charge.mState == Constants.ChargeStateDisabled then
            Charge.mBorderProp:setLoc( Charge.mPos[1], Charge.mPos[2], -9999.0 )
            Charge.mInsideProp:setLoc( Charge.mPos[1], Charge.mPos[2], -9999.0 )        
        end   
    end
      
    function Charge:UpdateThrow()
        local fDecayStep = Constants.ThrowRotationDecay * Constants.FrameStep
        if Charge.mAllChargesEnabled == false or math.abs(Charge.mThrownRotationSpeed) < fDecayStep or Charge.mAltGameMode then
            --done spinning, go back to active mode
            Charge.mThrownRotationSpeed = 0.0
            return false
        else
            --decay speed and update position
            if Charge.mThrownRotationSpeed > 0 then 
                Charge.mThrownRotationSpeed = Charge.mThrownRotationSpeed - fDecayStep
            else
                Charge.mThrownRotationSpeed = Charge.mThrownRotationSpeed + fDecayStep
            end

            --to calculate new position, use the angle we were at and update it by our speed
            local fAngleStep = Charge.mThrownRotationSpeed * Constants.FrameStep
            Charge.mThrowAmount = Charge.mThrowAmount + fAngleStep
            if Charge.mThrowAmount > DFMath.sixPi or Charge.mThrowAmount < -DFMath.sixPi then
                Achievement.updateAchievement( "DA_BeamSpin", 100 )
            end
            
            Charge.mPrevRotation = Charge.mRotation
            Charge.mRotation = Charge.mRotation + fAngleStep
            while Charge.mRotation < -math.pi do
                Charge.mRotation = Charge.mRotation + (math.pi * 2.0) 
                Charge.mPrevRotation = Charge.mPrevRotation + (math.pi * 2.0)
            end
            
            while Charge.mRotation > math.pi do 
                Charge.mRotation = Charge.mRotation - (math.pi * 2.0) 
                Charge.mPrevRotation = Charge.mPrevRotation - (math.pi * 2.0) 
            end
                        
            local fRotationCos = math.cos(Charge.mRotation)
            local fRotationSin = math.sin(Charge.mRotation)
                    
            local fNewX = Constants.ForwardVector[1]*fRotationCos - Constants.ForwardVector[2]*fRotationSin
            local fNewY = Constants.ForwardVector[1]*fRotationSin + Constants.ForwardVector[2]*fRotationCos
            Charge:SetPosition( fNewX, fNewY )
                    
            --scoot all throw positions down a couple indexes
            local i = Charge.mNumThrowPositions
            while i > 1 do
                Charge.mPrevThrowPositions[i][1] = Charge.mPrevThrowPositions[i-1][1]  
                Charge.mPrevThrowPositions[i][2] = Charge.mPrevThrowPositions[i-1][2]
                i = i - 1
            end
            
            --set previous throw position from our current arc position
            Charge.mPrevThrowPositions[1][1] = Charge.mArcPos[1]
            Charge.mPrevThrowPositions[1][2] = Charge.mArcPos[2]
        end              
        
        return true
    end
    
    function Charge:CheckForThrow(fSpeedThreshold)
        --how much did we swipe?
        local nNumPrevPositionsUsed = 0
        local nPrevTouchPositionSum = {0, 0}
        
        for i = 1, Charge.mNumTouchPositions - 1 do
            if Charge.mTouchPositions[i].mUsed and Charge.mTouchPositions[i+1].mUsed then
                nPrevTouchPositionSum[1] = nPrevTouchPositionSum[1] + (Charge.mTouchPositions[i].mPos[1] - Charge.mTouchPositions[i+1].mPos[1])
                nPrevTouchPositionSum[2] = nPrevTouchPositionSum[2] + (Charge.mTouchPositions[i].mPos[2] - Charge.mTouchPositions[i+1].mPos[2])
                nNumPrevPositionsUsed = nNumPrevPositionsUsed + 1                    
            end
        end
        
        --we can't swipe if we don't have a history
        if nNumPrevPositionsUsed > 1 then
            nPrevTouchPositionSum[1] = nPrevTouchPositionSum[1] / nNumPrevPositionsUsed 
            nPrevTouchPositionSum[2] = nPrevTouchPositionSum[2] / nNumPrevPositionsUsed 
            
            local fSwipeAmount = DFMath.length( nPrevTouchPositionSum[1], nPrevTouchPositionSum[2] ) 
            local fUndoFraction = (0.01666 / Constants.FrameStep)
            fSwipeAmount = fSwipeAmount * Constants.FrameStep * fUndoFraction       
            
            if fSwipeAmount > Constants.ThrowSpeedThreshold then                 
                --compute angle
                local nPosNormalized = {Charge.mPos[1], Charge.mPos[2]}
                nPosNormalized[1], nPosNormalized[2] = DFMath.normalize( nPosNormalized[1], nPosNormalized[2] )
                
                local fDot = DFMath.dot( Constants.ForwardVector[1], Constants.ForwardVector[2], nPosNormalized[1], nPosNormalized[2] )
                Charge.mRotation = math.acos( fDot )
                if Charge.mPos[2] < 0.0 then
                    Charge.mRotation = Charge.mRotation * -1.0
                end
                Charge.mPrevRotation = Charge.mRotation

                --clamp speed
                local fScaledSwipe = fSwipeAmount * Constants.SwipeSpeedScalar
                fScaledSwipe = math.min( fScaledSwipe, Constants.SwipeSpeedMax )
                Charge.mThrownRotationSpeed = fScaledSwipe

                --figure out which way we're spinning
                local nCrossProdHelp1 = {0, 0}
                nCrossProdHelp1[1] = Charge.mTouchPositions[1].mPos[1] + nPrevTouchPositionSum[1] 
                nCrossProdHelp1[2] = Charge.mTouchPositions[1].mPos[2] + nPrevTouchPositionSum[2]
                
                local nCrossProdHelp2 = {0, 0}
                nCrossProdHelp2[1] = Charge.mTouchPositions[1].mPos[1] 
                nCrossProdHelp2[2] = Charge.mTouchPositions[1].mPos[2]
                
                local tCrossResult = {}
                tCrossResult[1], tCrossResult[2], tCrossResult[3] = DFMath.cross( nCrossProdHelp1[1], nCrossProdHelp1[2], 0.0, nCrossProdHelp2[1], nCrossProdHelp2[2], 0.0 )
                if tCrossResult[3] > 0 then
                    Charge.mThrownRotationSpeed = Charge.mThrownRotationSpeed * -1.0
                end

                Charge.mThrowTimer = 0.0
                Charge.mThrowAmount = 0.0
                Charge:SetState( Constants.ChargeStateThrown )
                Charge:ClearPreviousPositions()
                Charge:ResetThrowPositions()
            end
        end     
    end
    
    function Charge:ClearPreviousPositions()
        local i = 1
        for i = 1, Charge.mNumTouchPositions do   
            Charge.mTouchPositions[i].mUsed = false
        end 
    end
  
    function Charge:ResetThrowPositions()
        local i = 1
        for i = 1, Charge.mNumThrowPositions do
            Charge.mPrevThrowPositions[i][1] = Charge.mPos[1]
            Charge.mPrevThrowPositions[i][2] = Charge.mPos[2]            
        end
    end
    
    function Charge:SetPosition(fInputX, fInputY)   
        Charge:SetTouchPosition(fInputX, fInputY)
        
        --compute line to center      
        Charge.mPos[1], Charge.mPos[2] = Charge:ProjectInput(fInputX, fInputY, Constants.GameRingRadius) 
        Charge.mArcPos[1], Charge.mArcPos[2] = Charge:ProjectInput(fInputX, fInputY, Charge.mArcRadius)
    end
    
    function Charge:SetTouchPosition(fInputX, fInputY)   
        --scoot all touch positions down an index 
        local i = Charge.mNumTouchPositions
        while i > 1 do
            Charge.mTouchPositions[i].mPos[1] = Charge.mTouchPositions[i-1].mPos[1]  
            Charge.mTouchPositions[i].mPos[2] = Charge.mTouchPositions[i-1].mPos[2]
            Charge.mTouchPositions[i].mUsed = Charge.mTouchPositions[i-1].mUsed
            i = i - 1
        end
          
        Charge.mTouchPositions[1].mPos[1] = fInputX
        Charge.mTouchPositions[1].mPos[2] = fInputY
        Charge.mTouchPositions[1].mUsed = true    
    end
  
    function Charge:ProjectInput(x, y, fProjectLength)
        local nFromCenter = {0, 0}
        nFromCenter[1] = x 
        nFromCenter[2] = y 
        if math.abs(nFromCenter[1]) > 0.0 or math.abs(nFromCenter[2]) > 0.0 then
            nFromCenter[1], nFromCenter[2] = DFMath.normalize( nFromCenter[1], nFromCenter[2] )
            nFromCenter[1] = nFromCenter[1] * fProjectLength
            nFromCenter[2] = nFromCenter[2] * fProjectLength
        else
            nFromCenter[1] = 0.0 
            nFromCenter[2] = 0.0 
        end
        
        return nFromCenter[1], nFromCenter[2]
    end
    
    function Charge:DrawConnectionLine()
        if not Charge.mAltGameMode then    
            --draw connection lines between charges and input
            if Charge.mState == Constants.ChargeStateDisabled then
                local t = Charge.mChargeDisabledTimer / Charge.mChargeDisabledDuration
                local x = DFMath.lerp( Charge.mPos[1], Charge.mTouchPositions[1].mPos[1], t )
                local y = DFMath.lerp( Charge.mPos[2], Charge.mTouchPositions[1].mPos[2], t )
                MOAIDraw.drawLine( Charge.mTouchPositions[1].mPos[1], Charge.mTouchPositions[1].mPos[2], x, y )        
            elseif Charge.mState == Constants.ChargeStateActive then
                MOAIDraw.drawLine( Charge.mPos[1], Charge.mPos[2], Charge.mTouchPositions[1].mPos[1], Charge.mTouchPositions[1].mPos[2] )
            end    
        end
    end
    
    Charge:Init()
    return Charge    
end

-- Charge Manager
function ChargeManager:Init( tRenderer )
    ChargeManager.mNumCharges = 2  --this really, really shouldn't be changed.
    ChargeManager.mCharges = {}
    ChargeManager.mCharges[1] = ChargeManager:NewCharge()
    ChargeManager.mCharges[1]:CreateProp( tRenderer )
    ChargeManager.mCharges[1].mIndex = 1
    ChargeManager.mCharges[2] = ChargeManager:NewCharge()
    ChargeManager.mCharges[2]:CreateProp( tRenderer )
    ChargeManager.mCharges[2].mIndex = 2
            
    ChargeManager.mBeamEnabled = false
    ChargeManager.mBeamHidden = false
    
    ChargeManager.mBeamDisableDuration = 1.0
    ChargeManager.mBeamDisableFadePoint = .5
    ChargeManager.mBeamDisableTimer = ChargeManager.mBeamDisableDuration       
    ChargeManager.mBeamDisableSound = nil
    ChargeManager.mBeamHiddenDisableScale = 0.5
    
    ChargeManager.mBeamSpinSound = nil
    
    ChargeManager.mBeamInvulnerableTimer = 0.0
    ChargeManager.mBeamInvulnerableDuration = 0.5
    
    ChargeManager.mBeamFlashTimer = 0.0
    ChargeManager.mBeamFlashDecay = 4.0
    
    ChargeManager.mBeamEffectA = nil
    ChargeManager.mBeamEffectA2 = nil
    ChargeManager.mBeamEffectB = nil    
    ChargeManager.mBeamEffectB2 = nil
end

function ChargeManager:Reset()
    local i = 1
    for i = 1, ChargeManager.mNumCharges do
        ChargeManager.mCharges[i]:Reset()
    end        
    
    ChargeManager.mBeamEnabled = false
    ChargeManager.mBeamDisableDuration = Constants.GameIntroBeamConnectionDuration
    ChargeManager.mBeamDisableTimer = 0.0   
end

function ChargeManager:UpdateSize()
    local i = 1
    for i = 1, ChargeManager.mNumCharges do
        ChargeManager.mCharges[i]:UpdateSize()
    end        
end

function ChargeManager:SetDisabledDurationForPause()
    ChargeManager.mBeamDisableDuration = 2.0
    ChargeManager.mBeamDisableTimer = 0.0
    ChargeManager.mBeamEnabled = false
end

function ChargeManager:SetDisabledDurationForGame()
    ChargeManager.mBeamDisableDuration = 1.5
    ChargeManager.mBeamDisableTimer = ChargeManager.mBeamDisableDuration
    ChargeManager.mBeamEnabled = true
end

function ChargeManager:IsBeamHidden()
    return ChargeManager.mBeamHidden
end

function ChargeManager:SetToAltGameMode(bAltGameMode)
    local i = 1
    if bAltGameMode then
        --hide charges
        RadEffects:SetEffectLayerToParticleAccumLayer()
            
        for i = 1, ChargeManager.mNumCharges do
            ChargeManager.mCharges[i]:SetInAltGameMode(true) 
        end  

        --hide beam
        ChargeManager.mBeamHidden = true
    else
        for i = 1, ChargeManager.mNumCharges do
            ChargeManager.mCharges[i]:SetInAltGameMode(false)
        end  
        
        ChargeManager.mBeamHidden = false
    end
end

function ChargeManager:GetBeamState()
    return ChargeManager.mBeamEnabled
end

function ChargeManager:MakeInvulnerable()
    ChargeManager.mBeamInvulnerableTimer = ChargeManager.mBeamInvulnerableDuration
end

function ChargeManager:IsInvulnerable()
    return ChargeManager.mBeamInvulnerableTimer > 0.0
end

function ChargeManager:FlashBeam()
    ChargeManager.mBeamFlashTimer = 1.0
end

function ChargeManager:AreAnyChargesThrown()
    local i = 1
    for i = 1, ChargeManager.mNumCharges do
        if ChargeManager.mCharges[i]:IsThrown() then
            return true
        end
    end    
    return false
end

function ChargeManager:AreBothChargesActive()
    local i = 1
    for i = 1, ChargeManager.mNumCharges do
        if ChargeManager.mCharges[i].mState ~= Constants.ChargeStateActive then
            return false
        end
    end  
    return true
end

function ChargeManager:AreBothChargesEnabled()
    local i = 1
    for i = 1, ChargeManager.mNumCharges do
        if ChargeManager.mCharges[i]:IsEnabled() == false then
            return false
        end
    end  
    return true
end

function ChargeManager:AreBothChargesShrunk()
    local i = 1
    for i = 1, ChargeManager.mNumCharges do
        if ChargeManager.mCharges[i].mDrawRadius > 0.0 then
            return false
        end
    end    
    return true
end

function ChargeManager:SetChargesAtExitPosition()
    ChargeManager.mCharges[1]:SetPosition( Constants.GameRingRadius * -0.9, Constants.GameRingRadius )
    ChargeManager.mCharges[2]:SetPosition( Constants.GameRingRadius * 0.9, Constants.GameRingRadius )
end

function ChargeManager:GetChargeTutorialSource()
    if ChargeManager.mCharges[1].mPos[1] > ChargeManager.mCharges[2].mPos[1] then
        return ChargeManager.mCharges[1].mPos[1], ChargeManager.mCharges[1].mPos[2]
    end
    
    return ChargeManager.mCharges[2].mPos[1], ChargeManager.mCharges[2].mPos[2]
end
  
function ChargeManager:GetBeamActiveRatio()
    local fBeamFadeRange = ChargeManager.mBeamDisableDuration - ChargeManager.mBeamDisableFadePoint
    return math.min( ChargeManager.mBeamDisableTimer / fBeamFadeRange, 1.0 )  
end

function ChargeManager:UpdateBeamSound()
    --create beam sound if it doesn't exist
    if ChargeManager.mBeamDisableSound == nil then
        ChargeManager.mBeamDisableSound = AudioManager:PlaySound( 'Radius/SFX/HoldandPlay' )
    end
    
    if ChargeManager.mBeamDisableSound ~= nil then
        --mess with beam parameter
        ChargeManager.mBeamDisableSound:setParameter( 'holdandplay', ChargeManager.mBeamDisableTimer/ChargeManager.mBeamDisableDuration )      
    end
end

function ChargeManager:UpdateBeamSpinSound( fParamValue )
    if fParamValue > 0.0 then
        if ChargeManager.mBeamSpinSound == nil then
            ChargeManager.mBeamSpinSound = AudioManager:PlaySound( 'Radius/SFX/Tremelo' )
        end
        if ChargeManager.mBeamSpinSound ~= nil then
            ChargeManager.mBeamSpinSound:setParameter( 'tremelosfx', fParamValue )      
        end
    else
        AudioManager:StopSound( ChargeManager.mBeamSpinSound, true )
        ChargeManager.mBeamSpinSound = nil
    end
end

function ChargeManager:UpdateBeamEffect()
    RadEffects:SetEffectLayerToGameLayer()
    
    local nDisabledDrawVector = {0.0, 0.0}
    nDisabledDrawVector[1] = ChargeManager.mCharges[2].mPos[1] - ChargeManager.mCharges[1].mPos[1]
    nDisabledDrawVector[2] = ChargeManager.mCharges[2].mPos[2] - ChargeManager.mCharges[1].mPos[2]
    
    local fConnectionRatio = ChargeManager.mBeamDisableTimer / ChargeManager.mBeamDisableDuration
    local fConnectionPercent = .5 * fConnectionRatio
    
    --figure out where the effect should be
    nDisabledDrawVector[1] = nDisabledDrawVector[1] * fConnectionPercent 
    nDisabledDrawVector[2] = nDisabledDrawVector[2] * fConnectionPercent        
    nDisabledDrawVector[1] = nDisabledDrawVector[1] + ChargeManager.mCharges[1].mPos[1]
    nDisabledDrawVector[2] = nDisabledDrawVector[2] + ChargeManager.mCharges[1].mPos[2]         
    
    if Constants.UsePostEffects then
        if fConnectionRatio > 0.0 then
            --create effects if they don't exist
            if ChargeManager.mBeamEffectA == nil then
                ChargeManager.mBeamEffectA = RadEffects:CreateEffect( "Effects/Particles/Chord/CreationBuild.particles", nil, {Constants.Color1[1], Constants.Color1[2], Constants.Color1[3], 1.0}, nil, Constants.ParticlePriorityHigh )    
            end
            if ChargeManager.mBeamEffectA2 == nil then
                ChargeManager.mBeamEffectA2 = RadEffects:CreateEffect( "Effects/Particles/Chord/CreationGlow.particles", nil, {Constants.Color1[1], Constants.Color1[2], Constants.Color1[3], 1.0}, nil, Constants.ParticlePriorityHigh )
            end
    
            local rEffectProp = ChargeManager.mBeamEffectA:getProp()
            rEffectProp:setLoc( nDisabledDrawVector[1], nDisabledDrawVector[2], 1.0 )
            local rMaterial = rEffectProp:getMaterial()
            if rMaterial then
               rMaterial:setShaderValue( "g_vColor", MOAIMaterial.VALUETYPE_VEC4, {Constants.Color1[1], Constants.Color1[2], Constants.Color1[3], 1.0} )
            end
            ChargeManager.mBeamEffectA.rParticleSystem:start()
            
            local rEffectProp2 = ChargeManager.mBeamEffectA2:getProp()
            rEffectProp2:setLoc( nDisabledDrawVector[1], nDisabledDrawVector[2], 1.0 )
            local rMaterial2 = rEffectProp2:getMaterial()
            if rMaterial2 then
                rMaterial2:setShaderValue( "g_vColor", MOAIMaterial.VALUETYPE_VEC4, {Constants.Color1[1], Constants.Color1[2], Constants.Color1[3], 1.0} )
            end
            ChargeManager.mBeamEffectA2.rParticleSystem:start()
        else
            --null out beam effect if our particle system has been cleaned up
            if ChargeManager.mBeamEffectA ~= nil then        
                if ChargeManager.mBeamEffectA.rParticleSystem ~= nil then
                    ChargeManager.mBeamEffectA.rParticleSystem:stop()
                end
                ChargeManager.mBeamEffectA = nil
            end
            
            if ChargeManager.mBeamEffectA2 ~= nil then
                if ChargeManager.mBeamEffectA2.rParticleSystem ~= nil then
                    ChargeManager.mBeamEffectA2.rParticleSystem:stop()
                end                
                ChargeManager.mBeamEffectA2 = nil
            end
        end
    end
    
    --figure out where the second effect should be
    nDisabledDrawVector[1] = nDisabledDrawVector[1] - ChargeManager.mCharges[1].mPos[1]
    nDisabledDrawVector[2] = nDisabledDrawVector[2] - ChargeManager.mCharges[1].mPos[2]     
    nDisabledDrawVector[1] = nDisabledDrawVector[1] * -1.0
    nDisabledDrawVector[2] = nDisabledDrawVector[2] * -1.0
    nDisabledDrawVector[1] = nDisabledDrawVector[1] + ChargeManager.mCharges[2].mPos[1]
    nDisabledDrawVector[2] = nDisabledDrawVector[2] + ChargeManager.mCharges[2].mPos[2]
    
    if Constants.UsePostEffects then
        if fConnectionRatio > 0.0 then
            --create effects if they don't exist
            if ChargeManager.mBeamEffectB == nil then
                ChargeManager.mBeamEffectB = RadEffects:CreateEffect( "Effects/Particles/Chord/CreationBuild.particles", nil, {Constants.Color1[1], Constants.Color1[2], Constants.Color1[3], 1.0}, nil, Constants.ParticlePriorityHigh )
            end
            if ChargeManager.mBeamEffectB2 == nil then
                ChargeManager.mBeamEffectB2 = RadEffects:CreateEffect( "Effects/Particles/Chord/CreationGlow.particles", nil, {Constants.Color1[1], Constants.Color1[2], Constants.Color1[3], 1.0}, nil, Constants.ParticlePriorityHigh ) 
            end
            
            local rEffectProp = ChargeManager.mBeamEffectB:getProp()            
            rEffectProp:setLoc( nDisabledDrawVector[1], nDisabledDrawVector[2], 1.0 )
            local rMaterial = rEffectProp:getMaterial()
            if rMaterial then
               rMaterial:setShaderValue( "g_vColor", MOAIMaterial.VALUETYPE_VEC4, {Constants.Color1[1], Constants.Color1[2], Constants.Color1[3], 1.0} )
            end
            ChargeManager.mBeamEffectB.rParticleSystem:start()
            
            local rEffectProp2 = ChargeManager.mBeamEffectB2:getProp()            
            rEffectProp2:setLoc( nDisabledDrawVector[1], nDisabledDrawVector[2], 1.0 )
            local rMaterial2 = rEffectProp2:getMaterial()
            if rMaterial2 then
                rMaterial2:setShaderValue( "g_vColor", MOAIMaterial.VALUETYPE_VEC4, {Constants.Color1[1], Constants.Color1[2], Constants.Color1[3], 1.0} )
            end
            ChargeManager.mBeamEffectB2.rParticleSystem:start()
        else
            --null out beam effect if our particle system has been cleaned up
            if ChargeManager.mBeamEffectB ~= nil then
                if ChargeManager.mBeamEffectB.rParticleSystem ~= nil then
                    ChargeManager.mBeamEffectB.rParticleSystem:stop()
                end
                ChargeManager.mBeamEffectB = nil
            end
            
            if ChargeManager.mBeamEffectB2 ~= nil then
                if ChargeManager.mBeamEffectB2.rParticleSystem ~= nil then
                    ChargeManager.mBeamEffectB2.rParticleSystem:stop()
                end
                ChargeManager.mBeamEffectB2 = nil
            end
        end
    end
end

function ChargeManager:UpdateIntro( tGameInputs )
    ChargeManager:UpdateBeamSound()
    ChargeManager:UpdateBeamEffect()
    
    --update input tutorial
    InputTutorial:SetBeamCompleteAmount( ChargeManager.mBeamDisableTimer/ChargeManager.mBeamDisableDuration )
    
    local i = 1
    if ChargeManager:AreBothChargesActive() then
        ChargeManager.mBeamDisableTimer = ChargeManager.mBeamDisableTimer + Constants.FrameStep
        if ChargeManager.mBeamDisableTimer >= ChargeManager.mBeamDisableDuration then
            --transition from game intro to menu!(?)
            if Constants.SkipMainMenu then
                Constants.DesiredGameState = Constants.GameStateInGame
            else
                Constants.DesiredGameState = Constants.GameStateMenu    
            end
            
            --beam is fully active
            ChargeManager.mBeamDisableTimer = 1.0
            ChargeManager.mBeamEnabled = true
            
            --stop sound
            AudioManager:StopSound( ChargeManager.mBeamDisableSound, true )
            ChargeManager.mBeamDisableSound = nil
            
            --trigger beam connect sound
            AudioManager:PlaySound( 'Radius/SFX/HoldandPlayMeet' )

            --trigger beam connect effect
            local fBeamCenterX = 0.0
            local fBeamCenterY = 0.0
            for i = 1, 2 do
                fBeamCenterX = fBeamCenterX + ChargeManager.mCharges[i].mPos[1]
                fBeamCenterY = fBeamCenterY + ChargeManager.mCharges[i].mPos[2]
            end       
            fBeamCenterX = fBeamCenterX * 0.5
            fBeamCenterY = fBeamCenterY * 0.5
            
            RadEffects:SetEffectLayerToGameLayer()
            RadEffects:SetEffectTransform( fBeamCenterX, fBeamCenterY, 0.0, 0.0, 0.0, 0.0 )
            RadEffects:CreateEffect( "Effects/Particles/Chord/CreationComplete.particles", nil, nil, nil, Constants.ParticlePriorityHigh )       
            RadEffects:CreateEffect( "Effects/Particles/Chord/CreationFlash.particles", nil, nil, nil, Constants.ParticlePriorityHigh )
            RadEffects:CreateEffect( "Effects/Particles/Chord/CreationSparkles.particles", nil, nil, nil, Constants.ParticlePriorityHigh )
            RadEffects:CreateEffect( "Effects/Particles/Chord/CreationEmbers.particles", nil, nil, nil, Constants.ParticlePriorityHigh )          
            
            --kill infinite effects
            if ChargeManager.mBeamEffectA ~= nil then
                ChargeManager.mBeamEffectA.rParticleSystem:stop() 
                ChargeManager.mBeamEffectA2.rParticleSystem:stop()              
            end
            
            if ChargeManager.mBeamEffectB ~= nil then
                ChargeManager.mBeamEffectB.rParticleSystem:stop()
                ChargeManager.mBeamEffectB2.rParticleSystem:stop()  
            end
        end
    else
        local fDoubleTime = ChargeManager.mBeamDisableTimer - (Constants.FrameStep * 2.0)
        ChargeManager.mBeamDisableTimer = math.max( fDoubleTime, 0.0 )
    end
    
    --update charge state against distance to tutorial props
    for i = 1, 2 do
        local fChargePosX = 0.0
        local fChargePosY = 0.0
        if tGameInputs[i]:IsTouched() then
            fChargePosX = tGameInputs[i].mX
            fChargePosY = tGameInputs[i].mY
        end
        
        ChargeManager.mCharges[i].mIntroTutorialValidAmount = InputTutorial:UpdateChargeNearInputProp(fChargePosX, fChargePosY, i)
        if ChargeManager.mCharges[i].mState ~= Constants.ChargeStateDisabled then
            ChargeManager.mCharges[i].mIntroTutorialValidAmount = 1.0
        end
        
        if ChargeManager.mCharges[i].mIntroTutorialValidAmount <= 0.0 then
            ChargeManager.mCharges[i]:SetState( Constants.ChargeStateDisabled )
            ChargeManager.mCharges[i].mChargeDisabledTimer = ChargeManager.mCharges[i].mChargeDisabledDuration
        end     
    end
end

function ChargeManager:UpdateMenu( tGameInputs )    
    --look for charge swaps and other shenanigans
    ChargeManager:UpdateChargePositions( tGameInputs )
    
    --run through the charges and see if they're all enabled
    local bAllEnabled = true 
    local i = 1
    for i = 1, ChargeManager.mNumCharges do
        if ChargeManager.mCharges[i].mState == Constants.ChargeStateDisabled then
            bAllEnabled = false
        end
    end  
    
    --let them know what we figured out and update them
    local fBeamActiveRatio = ChargeManager:GetBeamActiveRatio()    
    for i = 1, ChargeManager.mNumCharges do
        ChargeManager.mCharges[i].mAllChargesEnabled = bAllEnabled
        ChargeManager.mCharges[i].mBeamActiveRatio = fBeamActiveRatio
        ChargeManager.mCharges[i]:Update( tGameInputs )
    end  

    --keep lowpass in check 
    if MOAIFmodEventMgr then
        MOAIFmodEventMgr.setSeamlessQueueEventParam( "lowpass", 0.0 )
    end    
end

function ChargeManager:Update( tGameInputs, bCanPause )    
    --update invulnerable timer
    ChargeManager.mBeamInvulnerableTimer = math.max( ChargeManager.mBeamInvulnerableTimer - Constants.FrameStep, 0.0 )

    --update beam flash timer
    local fStep = ChargeManager.mBeamFlashDecay * Constants.FrameStep
    ChargeManager.mBeamFlashTimer = math.max( ChargeManager.mBeamFlashTimer - fStep, 0.0 )
    
    --look for charge swaps and other shenanigans
    ChargeManager:UpdateChargePositions( tGameInputs )
    
    --run through the charges and see if they're all enabled
    local bAllEnabled = true 
    local i = 1
    for i = 1, ChargeManager.mNumCharges do
        if ChargeManager.mCharges[i].mState == Constants.ChargeStateDisabled then
            bAllEnabled = false
        end
    end  
    
    --let them know what we figured out and update them
    local fBeamActiveRatio = ChargeManager:GetBeamActiveRatio()    
    for i = 1, ChargeManager.mNumCharges do
        ChargeManager.mCharges[i].mAllChargesEnabled = bAllEnabled
        ChargeManager.mCharges[i].mBeamActiveRatio = fBeamActiveRatio
        ChargeManager.mCharges[i]:Update( tGameInputs )
    end    
    
    --add treh-moe-low 
    if MOAIFmodEventMgr then
        local fThrow1Speed = math.abs( ChargeManager.mCharges[1].mThrownRotationSpeed )
        local fThrow2Speed = math.abs( ChargeManager.mCharges[2].mThrownRotationSpeed )
        if fThrow1Speed > Constants.SwipeSpeedMin or fThrow2Speed > Constants.SwipeSpeedMin then
            local fHighestSpeed = math.max( fThrow1Speed, fThrow2Speed )
            local fTremoloAmount = (fHighestSpeed - Constants.SwipeSpeedMin) / (Constants.SwipeSpeedMax - Constants.SwipeSpeedMin)
            
            MOAIFmodEventMgr.setSeamlessQueueEventParam( "volume", fTremoloAmount )
            ChargeManager:UpdateBeamSpinSound( fTremoloAmount )
        else
            MOAIFmodEventMgr.setSeamlessQueueEventParam( "volume", 0.0 )
            ChargeManager:UpdateBeamSpinSound( 0.0 )
        end
    end   
    
    if bCanPause then
        --if the beam is enabled, see if it's become disabled
        local fPrevBeamDisable = ChargeManager.mBeamDisableTimer
        if ChargeManager.mBeamEnabled then
            local bBeamIsControlled = ChargeManager.mCharges[1]:HasInput() or ChargeManager.mCharges[2]:HasInput()
            if bBeamIsControlled == false and Constants.AllowPause then
                local fDisableScale = 1.0
                if ChargeManager.mBeamHidden then
                    fDisableScale = ChargeManager.mBeamHiddenDisableScale
                end
                
                local fDisableStep = Constants.FrameStep * fDisableScale 
                ChargeManager.mBeamDisableTimer = math.max( ChargeManager.mBeamDisableTimer - fDisableStep, 0.0 )
                if ChargeManager.mBeamDisableTimer == 0.0 then
                    ChargeManager.mBeamEnabled = false
                end
            else
                ChargeManager.mBeamDisableTimer = ChargeManager.mBeamDisableDuration
            end
        end
        
        --update lowpass filter for beam disabled
        if Constants.GameState == Constants.GameStateInGame and fPrevBeamDisable ~= ChargeManager.mBeamDisableDuration then
            local fBeamRatio = ChargeManager.mBeamDisableTimer / ChargeManager.mBeamDisableDuration
            if MOAIFmodEventMgr then
                MOAIFmodEventMgr.setSeamlessQueueEventParam( "lowpass", 1.0 - fBeamRatio )
            end
        end
    else 
        ChargeManager.mBeamDisableTimer = ChargeManager.mBeamDisableDuration        
    end
end

function ChargeManager:UpdateComplete( tGameInputs, tInputVisuals )
    ChargeManager:UpdateChargePositions( tGameInputs )    
    
    --don't play the beam spin sound in the complete state
    ChargeManager:UpdateBeamSpinSound(0.0)
    
    --run through the charges and see if they're all enabled
    local bAllEnabled = true 
    local i = 1
    for i = 1, ChargeManager.mNumCharges do
        if ChargeManager.mCharges[i].mState == Constants.ChargeStateDisabled then
            bAllEnabled = false
        end
    end  
    
    --let them know what we figured out and update them
    local fBeamActiveRatio = ChargeManager:GetBeamActiveRatio()    
    for i = 1, ChargeManager.mNumCharges do
        ChargeManager.mCharges[i].mAllChargesEnabled = bAllEnabled
        ChargeManager.mCharges[i].mBeamActiveRatio = fBeamActiveRatio
        ChargeManager.mCharges[i]:Update( tGameInputs )
    end     
    
    local fDisableStep = Constants.FrameStep * 0.25
    local fPrevBeamTimer = ChargeManager.mBeamDisableTimer
    ChargeManager.mBeamDisableTimer = math.max( ChargeManager.mBeamDisableTimer - fDisableStep, 0.0 )
    
    if fPrevBeamTimer > 0.0 and ChargeManager.mBeamDisableTimer <= 0.0 then
        for i = 1, ChargeManager.mNumCharges do
            ChargeManager.mCharges[i]:SetInAltGameMode(true)            
        end
        
        tInputVisuals:SetGameComplete( true )
    end
end

function ChargeManager:UpdatePaused( tGameInputs )
    ChargeManager:UpdateBeamSound()
    
    --make sure tremolo is cleaned up during pause
    MOAIFmodEventMgr.setSeamlessQueueEventParam( "volume", 0.0 )   
    ChargeManager:UpdateBeamSpinSound( 0.0 )
            
    if not ChargeManager.mBeamHidden then
        ChargeManager:UpdateBeamEffect()
    end

    --look for charge swaps and other shenanigans
    ChargeManager:UpdateChargePositions( tGameInputs )
    
    --for beam connection drawing
    local i = 1
    for i = 1, 2 do
        ChargeManager.mCharges[i]:UpdatePaused( tGameInputs )
    end
    
    --when the beam is hidden, we only care about a single touch to unpause
    local bTouchUnpausing = tGameInputs[1].IsTouched()
    if not ChargeManager.mBeamHidden then
        bTouchUnpausing = bTouchUnpausing and tGameInputs[2].IsTouched()
    end
    
    if bTouchUnpausing then
        --if the beam is hidden, unpause immediately..
        local bUnPause = false
        if ChargeManager.mBeamHidden then
            --..if we're in the right spot
            if tGameInputs[1].mY < Constants.ValidDrawHalfHeight * 0.5 then
                bUnPause = true
            end
        else
            --otherwise, wait until the beam is complete
            ChargeManager.mBeamDisableTimer = math.min( ChargeManager.mBeamDisableTimer + Constants.FrameStep, ChargeManager.mBeamDisableDuration )
            bUnPause = (ChargeManager.mBeamDisableTimer == ChargeManager.mBeamDisableDuration)
        end
        
        if bUnPause then
            ChargeManager.mBeamDisableTimer = ChargeManager.mBeamDisableDuration
            ChargeManager:ExitPauseState()
    
            --trigger beam connect sound
            AudioManager:PlaySound( 'Radius/SFX/HoldandPlayMeet' )                      
        end
    else
        ChargeManager.mBeamDisableTimer = math.max( ChargeManager.mBeamDisableTimer - Constants.FrameStep, 0.0 )
    end    
    
    return ChargeManager.mBeamDisableTimer / ChargeManager.mBeamDisableDuration
end

function ChargeManager:ExitPauseState()
    ChargeManager.mBeamEnabled = true
    
    --stop sounds
    AudioManager:StopSound( ChargeManager.mBeamDisableSound, true )
    ChargeManager.mBeamDisableSound = nil
    ChargeManager:UpdateBeamSpinSound(0.0)
    
    --kill infinite effects
    if ChargeManager.mBeamEffectA ~= nil then
        if ChargeManager.mBeamEffectA.rParticleSystem ~= nil then
            ChargeManager.mBeamEffectA.rParticleSystem:stop()
        end
        ChargeManager.mBeamEffectA = nil
    end
    if ChargeManager.mBeamEffectA2 ~= nil then
        if ChargeManager.mBeamEffectA2.rParticleSystem ~= nil then
            ChargeManager.mBeamEffectA2.rParticleSystem:stop()
        end
        ChargeManager.mBeamEffectA2 = nil
    end
    
    if ChargeManager.mBeamEffectB ~= nil then
        if ChargeManager.mBeamEffectB.rParticleSystem ~= nil then
            ChargeManager.mBeamEffectB.rParticleSystem:stop()
        end
        ChargeManager.mBeamEffectB = nil
    end
    if ChargeManager.mBeamEffectB2 ~= nil then
        if ChargeManager.mBeamEffectB2.rParticleSystem ~= nil then
            ChargeManager.mBeamEffectB2.rParticleSystem:stop()
        end
        ChargeManager.mBeamEffectB2 = nil
    end
            
    --make sure our lowpass is reset for when we unpause
    if MOAIFmodEventMgr then
        MOAIFmodEventMgr.setSeamlessQueueEventParam( "lowpass", 0.0 )
    end    
end

function ChargeManager:UpdateChargePositions( tGameInputs )
    local fRingRadius = Constants.GameRingRadius
  
    --if a charge is out to lunch and there's new input near it, snap to it
    if ChargeManager.mCharges[1]:IsUncontrolled() and ChargeManager.mCharges[2]:IsUncontrolled() then
        local bMake2LookLike1 = false
        local bMake1LookLike2 = false
        
        --touch event 1 is down-- this is going to activate a charge later this frame, so see if we're near either of the charges
        if tGameInputs[1].IsTouched() then            
            local nProjectedTouchPos = {}
            nProjectedTouchPos[1], nProjectedTouchPos[2] = ChargeManager.mCharges[1]:ProjectInput( tGameInputs[1].mX, tGameInputs[1].mY, fRingRadius )
          
            local nChargePos1 = {}
            nChargePos1[1] = ChargeManager.mCharges[1].mPos[1]
            nChargePos1[2] = ChargeManager.mCharges[1].mPos[2]            
            local fTouchToPos1 = DFMath.distance2DSquared( nProjectedTouchPos[1], nProjectedTouchPos[2], nChargePos1[1], nChargePos1[2] )
            
            local nChargePos2 = {}
            nChargePos2[1] = ChargeManager.mCharges[2].mPos[1]
            nChargePos2[2] = ChargeManager.mCharges[2].mPos[2]                        
            local fTouchToPos2 = DFMath.distance2DSquared( nProjectedTouchPos[1], nProjectedTouchPos[2], nChargePos2[1], nChargePos2[2] )

            --closer to charge 2     
            if fTouchToPos1 > fTouchToPos2 then
                --touch 1 lines up with charge 2, but this event is going to position charge 1, so we need charge 2 to look like charge 1
                bMake2LookLike1 = true        
            end
        end
    
        --same for touch event 2
        if tGameInputs[2].IsTouched() then
            local nProjectedTouchPos = {}
            nProjectedTouchPos[1], nProjectedTouchPos[2] = ChargeManager.mCharges[1]:ProjectInput( tGameInputs[2].mX, tGameInputs[2].mY, fRingRadius )
            
            local nChargePos1 = {}
            nChargePos1[1] = ChargeManager.mCharges[1].mPos[1]
            nChargePos1[2] = ChargeManager.mCharges[1].mPos[2]            
            local fTouchToPos1 = DFMath.distance2DSquared( nProjectedTouchPos[1], nProjectedTouchPos[2], nChargePos1[1], nChargePos1[2] )
            
            local nChargePos2 = {}
            nChargePos2[1] = ChargeManager.mCharges[2].mPos[1]
            nChargePos2[2] = ChargeManager.mCharges[2].mPos[2]                        
            local fTouchToPos2 = DFMath.distance2DSquared( nProjectedTouchPos[1], nProjectedTouchPos[2], nChargePos2[1], nChargePos2[2] )
                      
            --closer to charge 1      
            if fTouchToPos2 > fTouchToPos1 then
                --touch 2 lines up with charge 1, but this event is going to position charge 2, so we need charge 1 to look like charge 2
                bMake1LookLike2 = true                   
            end
        end
    
        --swap charges if we caught the "other" one
        if bMake1LookLike2 == false and bMake2LookLike1 == true then
            ChargeManager:SwapCharges()
        elseif bMake1LookLike2 == true and bMake2LookLike1 == false then
            ChargeManager:SwapCharges()
        end 
    end
end

function ChargeManager:SwapCharges()       
    local fBackupPosX = ChargeManager.mCharges[2].mPos[1]
    local fBackupPosY = ChargeManager.mCharges[2].mPos[2]
    ChargeManager.mCharges[2].mPos[1] = ChargeManager.mCharges[1].mPos[1]
    ChargeManager.mCharges[2].mPos[2] = ChargeManager.mCharges[1].mPos[2]
    ChargeManager.mCharges[1].mPos[1] = fBackupPosX
    ChargeManager.mCharges[1].mPos[2] = fBackupPosY
  
    local i = 1
    local iNumTouchPositions = ChargeManager.mCharges[1].mNumTouchPositions
    for i = 1, iNumTouchPositions do
        local fBackupX = ChargeManager.mCharges[2].mTouchPositions[i].mPos[1]
        local fBackupY = ChargeManager.mCharges[2].mTouchPositions[i].mPos[2]
        local bBackupUsed = ChargeManager.mCharges[2].mTouchPositions[i].mUsed
        
        ChargeManager.mCharges[2].mTouchPositions[i].mPos[1] = ChargeManager.mCharges[1].mTouchPositions[i].mPos[1]
        ChargeManager.mCharges[2].mTouchPositions[i].mPos[2] = ChargeManager.mCharges[1].mTouchPositions[i].mPos[2]        
        ChargeManager.mCharges[2].mTouchPositions[i].mUsed = ChargeManager.mCharges[1].mTouchPositions[i].mUsed
        
        ChargeManager.mCharges[1].mTouchPositions[i].mPos[1] = fBackupX
        ChargeManager.mCharges[1].mTouchPositions[i].mPos[2] = fBackupY 
        ChargeManager.mCharges[1].mTouchPositions[i].mUsed = bBackupUsed
    end

    local iNumThrowPositions = ChargeManager.mCharges[1].mNumThrowPositions  
    for i = 1, iNumThrowPositions do
        local fBackupX = ChargeManager.mCharges[2].mPrevThrowPositions[i][1]
        local fBackupY = ChargeManager.mCharges[2].mPrevThrowPositions[i][2]
        ChargeManager.mCharges[2].mPrevThrowPositions[i][1] = ChargeManager.mCharges[1].mPrevThrowPositions[i][1]
        ChargeManager.mCharges[2].mPrevThrowPositions[i][2] = ChargeManager.mCharges[1].mPrevThrowPositions[i][2]        
        ChargeManager.mCharges[1].mPrevThrowPositions[i][1] = fBackupX
        ChargeManager.mCharges[1].mPrevThrowPositions[i][2] = fBackupY 
    end
  
    local fBackupPrevRot = ChargeManager.mCharges[2].mPrevRotation
    ChargeManager.mCharges[2].mPrevRotation = ChargeManager.mCharges[1].mPrevRotation
    ChargeManager.mCharges[1].mPrevRotation = fBackupPrevRot

    local fBackupRot = ChargeManager.mCharges[2].mRotation
    ChargeManager.mCharges[2].mRotation = ChargeManager.mCharges[1].mRotation
    ChargeManager.mCharges[1].mRotation = fBackupRot

    local fBackupSpeed = ChargeManager.mCharges[2].mThrownRotationSpeed
    ChargeManager.mCharges[2].mThrownRotationSpeed = ChargeManager.mCharges[1].mThrownRotationSpeed 
    ChargeManager.mCharges[1].mThrownRotationSpeed = fBackupSpeed

    local iBackupState = ChargeManager.mCharges[2].mState
    ChargeManager.mCharges[2].mState = ChargeManager.mCharges[1].mState 
    ChargeManager.mCharges[1].mState = iBackupState      
end

function ChargeManager:DrawExtra()
    local fPenR = (Constants.Color1[1] + Constants.BGColor[1]) * 0.5 
    local fPenG = (Constants.Color1[2] + Constants.BGColor[2]) * 0.5 
    local fPenB = (Constants.Color1[3] + Constants.BGColor[3]) * 0.5 
    
    MOAIGfxDevice.setPenWidth( 2.0 )
    MOAIGfxDevice.setPenColor( fPenR, fPenG, fPenB )
        
    for i = 1, ChargeManager.mNumCharges do    
        --draw connection lines between charges and input
        ChargeManager.mCharges[i]:DrawConnectionLine()
    end
end

function ChargeManager:DrawBeamDisabled()
    MOAIGfxDevice.setPenWidth( 6.0*Constants.GlobalScale )
    MOAIGfxDevice.setPenColor( Constants.Color1[1], Constants.Color1[2], Constants.Color1[3] )

    local nDisabledDrawVector = {0.0, 0.0}
    nDisabledDrawVector[1] = ChargeManager.mCharges[2].mPos[1] - ChargeManager.mCharges[1].mPos[1]
    nDisabledDrawVector[2] = ChargeManager.mCharges[2].mPos[2] - ChargeManager.mCharges[1].mPos[2]
    
    local fConnectionRatio = ChargeManager.mBeamDisableTimer / ChargeManager.mBeamDisableDuration
    local fConnectionPercent = .505 * fConnectionRatio
    
    nDisabledDrawVector[1] = nDisabledDrawVector[1] * fConnectionPercent 
    nDisabledDrawVector[2] = nDisabledDrawVector[2] * fConnectionPercent        
    nDisabledDrawVector[1] = nDisabledDrawVector[1] + ChargeManager.mCharges[1].mPos[1]
    nDisabledDrawVector[2] = nDisabledDrawVector[2] + ChargeManager.mCharges[1].mPos[2]        
    MOAIDraw.drawLine( ChargeManager.mCharges[1].mPos[1], ChargeManager.mCharges[1].mPos[2], nDisabledDrawVector[1], nDisabledDrawVector[2] )
   
    nDisabledDrawVector[1] = nDisabledDrawVector[1] - ChargeManager.mCharges[1].mPos[1]
    nDisabledDrawVector[2] = nDisabledDrawVector[2] - ChargeManager.mCharges[1].mPos[2]     
    nDisabledDrawVector[1] = nDisabledDrawVector[1] * -1.0
    nDisabledDrawVector[2] = nDisabledDrawVector[2] * -1.0
    nDisabledDrawVector[1] = nDisabledDrawVector[1] + ChargeManager.mCharges[2].mPos[1]
    nDisabledDrawVector[2] = nDisabledDrawVector[2] + ChargeManager.mCharges[2].mPos[2]     
    MOAIDraw.drawLine( ChargeManager.mCharges[2].mPos[1], ChargeManager.mCharges[2].mPos[2], nDisabledDrawVector[1], nDisabledDrawVector[2] )          
end

function ChargeManager:DrawBeam()
    if ChargeManager.mBeamHidden then
        return
    elseif not ChargeManager.mBeamEnabled then
        ChargeManager:DrawBeamDisabled()
        return
    end
    
    local fActiveRatio = ChargeManager:GetBeamActiveRatio() 
    local fStrokeWeight = DFMath.lerp( 2.0, Constants.BasePenWidth*Constants.GlobalScale, fActiveRatio )
    MOAIGfxDevice.setPenWidth( fStrokeWeight )

    local fBeamR = DFMath.lerp( Constants.Color1[1], Constants.SpecialColor[1], ChargeManager.mBeamFlashTimer )
    local fBeamG = DFMath.lerp( Constants.Color1[2], Constants.SpecialColor[2], ChargeManager.mBeamFlashTimer )
    local fBeamB = DFMath.lerp( Constants.Color1[3], Constants.SpecialColor[3], ChargeManager.mBeamFlashTimer )
    
    --see if any of the charges are thrown
    local iNumThrowPositions = ChargeManager.mCharges[1].mNumThrowPositions
    
    if ChargeManager.mCharges[1]:IsThrown() and ChargeManager.mCharges[2]:IsThrown() then
        --walk back through previous positions and draw ghost lines
        local i = iNumThrowPositions
        while i >= 1 do
            local fT = i / iNumThrowPositions
            local fBeamR = DFMath.lerp( fBeamR, Constants.BGColor[1], fT )
            local fBeamG = DFMath.lerp( fBeamG, Constants.BGColor[2], fT )
            local fBeamB = DFMath.lerp( fBeamB, Constants.BGColor[3], fT )
            MOAIGfxDevice.setPenColor( fBeamR, fBeamG, fBeamB )
            
            local fCharge1ThrowPosX = ChargeManager.mCharges[1].mPrevThrowPositions[i][1]
            local fCharge1ThrowPosY = ChargeManager.mCharges[1].mPrevThrowPositions[i][2]
            local fCharge2ThrowPosX = ChargeManager.mCharges[2].mPrevThrowPositions[i][1]
            local fCharge2ThrowPosY = ChargeManager.mCharges[2].mPrevThrowPositions[i][2]                
            MOAIDraw.drawLine( fCharge1ThrowPosX, fCharge1ThrowPosY, fCharge2ThrowPosX, fCharge2ThrowPosY )
            i = i - 1
        end
    elseif ChargeManager.mCharges[1]:IsThrown() then
        local i = iNumThrowPositions
        while i >= 1 do
            local fT = i / iNumThrowPositions
            local fBeamR = DFMath.lerp( fBeamR, Constants.BGColor[1], fT )
            local fBeamG = DFMath.lerp( fBeamG, Constants.BGColor[2], fT )
            local fBeamB = DFMath.lerp( fBeamB, Constants.BGColor[3], fT )
            MOAIGfxDevice.setPenColor( fBeamR, fBeamG, fBeamB )

            local fCharge1ThrowPosX = ChargeManager.mCharges[1].mPrevThrowPositions[i][1]
            local fCharge1ThrowPosY = ChargeManager.mCharges[1].mPrevThrowPositions[i][2]                
            MOAIDraw.drawLine( fCharge1ThrowPosX, fCharge1ThrowPosY, ChargeManager.mCharges[2].mPos[1], ChargeManager.mCharges[2].mPos[2] )
            i = i - 1
        end  
    elseif ChargeManager.mCharges[2]:IsThrown() then
        local i = iNumThrowPositions
        while i >= 1 do
            local fT = i / iNumThrowPositions
            local fBeamR = DFMath.lerp( fBeamR, Constants.BGColor[1], fT )
            local fBeamG = DFMath.lerp( fBeamG, Constants.BGColor[2], fT )
            local fBeamB = DFMath.lerp( fBeamB, Constants.BGColor[3], fT )
            MOAIGfxDevice.setPenColor( fBeamR, fBeamG, fBeamB )

            local fCharge2ThrowPosX = ChargeManager.mCharges[2].mPrevThrowPositions[i][1]
            local fCharge2ThrowPosY = ChargeManager.mCharges[2].mPrevThrowPositions[i][2] 
            MOAIDraw.drawLine( fCharge2ThrowPosX, fCharge2ThrowPosY, ChargeManager.mCharges[1].mPos[1], ChargeManager.mCharges[1].mPos[2] )
            i = i - 1
        end  
    else
        MOAIGfxDevice.setPenColor( fBeamR, fBeamG, fBeamB, fActiveRatio )
        MOAIDraw.drawLine( ChargeManager.mCharges[1].mPos[1], ChargeManager.mCharges[1].mPos[2], ChargeManager.mCharges[2].mPos[1], ChargeManager.mCharges[2].mPos[2] )
    end
end

function ChargeManager:DebugDrawBeam()
    MOAIGfxDevice.setPenColor( 1.0, 0.0, 0.0 )
    
    local i = 0
    for i = 0, 9 do
        local ax, ay, bx, by = DFRad:DebugGetLine(i)
        MOAIDraw.drawLine( ax, ay, bx, by )
    end
end

return ChargeManager