local Constants = require('Constants')
local DFMath = require('DFCommon.Math')
local DFUtil = require('DFCommon.Util')
local GameRing = require('GameRing')
local AudioManager = require('AudioManager')
local UI = require('UI')
local ChargeManager = require('ChargeManager')
local ColorManager = require('ColorManager')
local PatternManager = require('PatternManager')
local Score = require('Score')
local ChargeField = require('ChargeField')
local Grid = require('Grid')
local Wind = require('Wind')
local InputVisuals = require('InputVisuals')
local Renderer = require('Renderer')
local DFGraphics = require('DFCommon.Graphics')
local Post = require('postfx.Post')
local PostManager = require('postfx.PostManager')
local PostDebug = require('postfx.PostDebug')
local PostLastLevel = require('postfx.PostState_Level10')
local Input = require('Input')
local Credits = require('Credits')
local HowToPlay = require('HowToPlay')
local Leaderboard = require('Leaderboard')
local LeaderboardEntry = require('LeaderboardEntryTouch')
local TextManager = require('TextManager')
local PauseScreen = require('PauseScreen')
local RadEffects = require('RadEffects')
local DebugDisplay = require('DebugDisplay')
local Letterbox = require('Letterbox')
local TitleScreen = require('TitleScreen')
local HUD = require('HUD')
local InputTutorial = require('InputTutorial')
local AudioBars = require('AudioBars')
local AudioRing = require('AudioRing')
local PerformanceTest = require('PerformanceTest')
local OpponentScore = require('OpponentScore')
local Achievement = require('DFCommon.Achievement')

local GameRules = {} 

-- Game Input Object
function GameRules:NewGameInput()
    local GameInput = {}
    
    function GameInput:Init()
        GameInput.mX = 0.0
        GameInput.mY = 0.0
        GameInput.mZ = 0.0
        GameInput.mDirty = false
        GameInput.mTouchIndex = -1
        GameInput.mTouched = false
        GameInput.mWasTouched = false
        GameInput.mFirstTouch = false
    end
  
    function GameInput:Reset()
        GameInput.mTouchIndex = -1
        GameInput.mWasTouched = GameInput.mTouched
        GameInput.mTouched = false 
        GameInput.mFirstTouch = false
    end
    
    function GameInput:WasTouched()
        return GameInput.mWasTouched 
    end
    
    function GameInput:IsTouched()
        return GameInput.mTouched 
    end
    
    function GameInput:Touch( index )
        GameInput.mFirstTouch = not GameInput.mTouched
        GameInput.mTouchIndex = index
        GameInput.mTouched = true
    end
    
    GameInput:Init()
    return GameInput
end

-- Camera Tilt Object
function GameRules:NewCameraTiltInput()
    local CameraTiltInput = {}
    
    function CameraTiltInput:Init()
        CameraTiltInput.mX = 0.0
        CameraTiltInput.mY = 0.0
        CameraTiltInput.mZ = 0.0
        CameraTiltInput.mActiveAmount = 0.0
        CameraTiltInput.mActiveSpeed = 2.0
    end
  
    function CameraTiltInput:NoInput()
        local fStep = CameraTiltInput.mActiveSpeed * Constants.FrameStep
        CameraTiltInput.mActiveAmount = math.max( CameraTiltInput.mActiveAmount - fStep, 0.0 )
    end

    function CameraTiltInput:Update(x, y, z) 
        CameraTiltInput.mX = x
        CameraTiltInput.mY = y
        CameraTiltInput.mZ = z
        local fStep = CameraTiltInput.mActiveSpeed * Constants.FrameStep
        CameraTiltInput.mActiveAmount = math.min( CameraTiltInput.mActiveAmount + fStep, 1.0 )
    end
    
    CameraTiltInput:Init()
    return CameraTiltInput
end

function GameRules:Init()
    local i = 1
    GameRules.mNumGameInputs = Constants.NumGameInputs
    GameRules.mGameInputs = {}
    for i = 1, GameRules.mNumGameInputs do
        GameRules.mGameInputs[i] = GameRules:NewGameInput()
    end
    
    GameRules.mLastElapsedTime = 0.0
    GameRules.mGameStateElapsedTime = 0.0
    GameRules.mLevelIntensity = 0.0
    GameRules.mEatInputTimer = 0.0
    GameRules.mEatEscInput = true
    GameRules.mResetGameAfterLeaderboards = false
    GameRules.mPreIntroDurations = {1.0, 4.0}
    GameRules.mChargeDeckImpressionSize = 1.0
    GameRules.mGridImpressionSize = 1.0
    
    --for camera tilt
    GameRules.mCurrentCameraOffset = {}
    GameRules.mCurrentCameraOffset[1] = 0.0
    GameRules.mCurrentCameraOffset[2] = 0.0
    GameRules.mCurrentCameraOffset[3] = 0.0
    GameRules.mDesiredCameraOffset = {}
    GameRules.mDesiredCameraOffset[1] = 0.0
    GameRules.mDesiredCameraOffset[2] = 0.0
    GameRules.mDesiredCameraOffset[3] = 0.0
    GameRules.mCameraOffsetSpeed = 1000.0
    GameRules.mCameraTiltMaxHoriz = 0.25
    GameRules.mCameraTiltMaxVert = 0.125
    GameRules.mCameraTiltInputs = {}
    for i = 1, 2 do
        GameRules.mCameraTiltInputs[i] = GameRules:NewCameraTiltInput()
    end
    
    --for camera zoom
    GameRules.mDesiredCameraZ = 0.0
    GameRules.mDesiredCameraSpeed = 2.0
    GameRules.mDesiredCameraPercents = {0.0, 0.0, 1.0}
        
    --for tutorials
    GameRules.mTutorialType = Constants.mTutorialTypeNone
    GameRules.mTutorialState = Constants.mTutorialStateNone 
    GameRules.mTutorialStateTimer = 0.0
    GameRules.mTutorialStateTransitionDuration = 0.125
    GameRules.mTutorialStateWaitDuration = 2.5    
    
    --for menus
    GameRules.mMenuDrawTimer = 0.0
    GameRules.mMenuDrawDuration = 0.125
    GameRules.mMenuDrawState = Constants.mMenuDrawStateNone
    GameRules.mMenuDrawType = Constants.mMenuSelectionNone
    GameRules.mMenuSourceX = 0.0
    GameRules.mMenuSourceY = 0.0
    GameRules.mMenuHUDIndex = -1
    GameRules.mMenuDepth = 0
    GameRules.mOverrideMenuDepth = -1
    
    --for health state
    GameRules.mHealthState = Constants.mHealthStateNone
    GameRules.mHealthStateSleepDuration = 0.5
    GameRules.mHealthStateBreakDuration = 1.0
    GameRules.mHealthStateTimer = 0.0
    GameRules.mHealthStateTotalDuration = 7.0
    GameRules.mHealthStateTotalTime = 0.0
    GameRules.mHealthStatePatternCount = 0
    GameRules.mHealthStatePerfectCount = 0
    GameRules.mHealthStatePercent = 0.0
    GameRules.mHealthAwarded = 0.0
    
    --for dsp filters
    GameRules.mCurrentHighpassValue = 0.0
    GameRules.mDesiredHighpassValue = 0.0
    GameRules.mHighpassChange = 4.0
    GameRules.mCurrentLowpassValue = 0.0
    GameRules.mDesiredLowpassValue = 0.0
    GameRules.mLowpassChange = 4.0
    GameRules.mCurrentFlangeValue = 0.0
    GameRules.mDesiredFlangeValue = 0.0
    GameRules.mFlangeChange = 4.0
    GameRules.mCurrentPanValue = 0.0
    GameRules.mDesiredPanValue = 0.0
    GameRules.mPanChange = 2.0
    
    --for delaying pause
    GameRules.mPauseDelayTimer = 0.0
    
    Input.Init()
    
	GameRules.mUpdateThread = MOAIThread.new()
	GameRules.mUpdateThread:run(GameRules.Update, GameRules)
    
    Constants.DesiredGameState = Constants.GameStateTitle--Constants.GameStatePerformanceTest
end

-- called when all other systems exist. Use this to set up glue-y stuff that
--  depends on multiple systems.
function GameRules:PostSystemInit()
    GameRules:VisualQualityChange()
    GameRules.mLetterbox = Letterbox:new(Renderer.mLetterboxLayer)
    if Constants.ShowDebugInfo then
        g_DebugDisplay = DebugDisplay:new(Renderer.mUIViewport, Renderer.mGameCamera)
    end
    GameRules.mTitleScreen = TitleScreen:new(Renderer.mGameLayer)
end

function GameRules:OnScreenResize(width, height)
    if Constants.ShowDebugInfo then
        g_DebugDisplay:OnScreenResize(width, height)
    end
    GameRules.mLetterbox:OnScreenResize(width, height)
    GameRules.mTitleScreen:OnScreenResize(width, height)
    
    GameRules.mDesiredCameraZ = Renderer.mGameCameraZDistance * GameRules.mDesiredCameraPercents[3]
    Renderer.mGameCamera:setLoc( 0.0, 0.0, GameRules.mDesiredCameraZ )    
    
    GameRules.mChargeDeckImpressionSize = Constants.ScreenBufferMaxWidth_High * 0.06
    GameRules.mGridImpressionSize = Constants.ScreenBufferMaxWidth_High * 0.03
    
    if g_DebugDisplay then
        g_DebugDisplay:SetValue("RES", width .. "x" .. height)
    end 
end

function GameRules:Update()
    while true do
        --set tick time
        local fCurrentElapsedTime = MOAISim.getElapsedTime()
        Constants.FrameStep = fCurrentElapsedTime - GameRules.mLastElapsedTime 
        GameRules.mLastElapsedTime = fCurrentElapsedTime
        GameRules.mGameStateElapsedTime = GameRules.mGameStateElapsedTime + Constants.FrameStep
        
        --update input
        GameRules:UpdateInput()    
        
        --audio drives game state and should update first
        MOAISim.profileEnterScope("LUA AudioManager update")
        AudioManager:Update()
        MOAISim.profileLeaveScope("LUA AudioManager update")
    
        --look for game state change
        if Constants.GameState ~= Constants.DesiredGameState then
            GameRules:GameStateChange()
        end
        
        MOAISim.profileEnterScope("LUA game state update")
        --update game state
        if Constants.GameState == Constants.GameStateTitle                      then GameRules:UpdateGameStateTitle()
        elseif Constants.GameState == Constants.GameStatePreIntro               then GameRules:UpdateGameStatePreIntro()
        elseif Constants.GameState == Constants.GameStateIntro                  then GameRules:UpdateGameStateIntro()    
        elseif Constants.GameState == Constants.GameStateMenu                   then GameRules:UpdateGameStateMenu()    
        elseif Constants.GameState == Constants.GameStateInGame                 then GameRules:UpdateGameStateInGame()  
        elseif Constants.GameState == Constants.GameStateComplete               then GameRules:UpdateGameStateComplete()
        elseif Constants.GameState == Constants.GameStatePost                   then GameRules:UpdateGameStatePost()    
        elseif Constants.GameState == Constants.GameStateLeaderboardEntry       then GameRules:UpdateGameStateLeaderboardEntry()    
        elseif Constants.GameState == Constants.GameStateLeaderboardShowcase    then GameRules:UpdateGameStateLeaderboardShowcase()   
        elseif Constants.GameState == Constants.GameStateCredits                then GameRules:UpdateGameStateCredits()     
        elseif Constants.GameState == Constants.GameStateHowToPlay              then GameRules:UpdateGameStateHowToPlay()    
        elseif Constants.GameState == Constants.GameStatePerformanceTest        then PerformanceTest.Update() 
        end
        MOAISim.profileLeaveScope("LUA game state update")
        
        --classes to always update
        local i = 1
        for i = 1, GameRules.mNumGameInputs do
            if GameRules.mGameInputs[i].IsTouched() then
                ChargeField:AddImpression( GameRules.mGameInputs[i].mX, GameRules.mGameInputs[i].mY, GameRules.mChargeDeckImpressionSize )
                Grid:AddImpression( GameRules.mGameInputs[i].mX, GameRules.mGameInputs[i].mY, GameRules.mGridImpressionSize )
            end
        end        
         
        MOAISim.profileEnterScope("LUA GameRing - Update")
        GameRing:Update()
        MOAISim.profileLeaveScope("LUA GameRing - Update")
        
        ChargeField:Update()
        AudioBars:Update()
        AudioRing:Update()
        Grid:Update()
        Wind:Update()
        
        MOAISim.profileEnterScope("LUA Post - Update")
        if Constants.UsePostEffects then Post:Update() end
        MOAISim.profileLeaveScope("LUA Post - Update")
        
        GameRules:PostDebugInput()
        
        GameRules:UpdateScreenTilt()
        GameRules:UpdateDSPValues()
        
        HUD:Update( TextManager, GameRules.mGameInputs )
        
        --sleep for game step
        coroutine.yield()
    end  
end

function GameRules:UpdateLevelState()
    --update level state switch timer and see if we need to change states
    Constants.LevelStateSwitchTimer = Constants.LevelStateSwitchTimer - Constants.FrameStep
    if Constants.LevelState ~= Constants.DesiredLevelState then
        if Constants.LevelStateSwitchTimer <= 0.0 then
            GameRules:LevelStateChange()
        end
    end
        
    --update level intensity
    GameRules.mLevelIntensity = AudioManager:GetLevelIntensityFromMusic()
    
    if Constants.LevelState == Constants.LevelStateStart then
        --tell the audio manager if we've hit our quota of bits hit
        local bQuotaFilled = Score:ShouldEnterBuildState()
        AudioManager:SetEnterBuildStateFlag( bQuotaFilled )
        PatternManager:SetGameSpeed( 0.0 )
        GameRing:UpdateIntensityRings( 0.0 )        
    elseif Constants.LevelState == Constants.LevelStateBuild then
        PatternManager:SetGameSpeed( GameRules.mLevelIntensity )
        GameRing:UpdateIntensityRings( GameRules.mLevelIntensity )
    elseif Constants.LevelState == Constants.LevelStateBonus then
        local fTrackPercent = AudioManager:GetTrackPercent()
        Score:SetNumBitsHitFromPercentage( 1.0 - fTrackPercent )
        PatternManager:SetGameSpeed( 0.0 )
        GameRing:UpdateIntensityRings( 0.0 )    
    elseif Constants.LevelState == Constants.LevelStateHealth then      
        GameRules:UpdateLevelStateHealth()
    end    
end

function GameRules:UpdateLevelStateHealth()    
    GameRules.mHealthStateTimer = GameRules.mHealthStateTimer - Constants.FrameStep
    if GameRules.mHealthStateTimer <= 0.0 then    
        --advance state
        GameRules.mHealthState = GameRules.mHealthState + 1
        if GameRules.mHealthState > Constants.mHealthStateHealthUp then
            GameRules.mHealthState = Constants.mHealthStateNone
        end    
    
        if GameRules.mHealthState == Constants.mHealthStateNone then
            --we wrapped states, time to get out
            Constants.DesiredLevelState = Constants.LevelStateStart
    
        elseif GameRules.mHealthState == Constants.mHealthStateTrackComplete then
            --determine if we're going to give the player health            
            GameRules.mHealthStatePercent = GameRules.mHealthStatePerfectCount / GameRules.mHealthStatePatternCount
            if GameRules.mHealthStatePercent > 0.99 then
                GameRules.mHealthAwarded = 1.0
            elseif GameRules.mHealthStatePercent > 0.8 then
                GameRules.mHealthAwarded = 0.6
            elseif GameRules.mHealthStatePercent > 0.6 then
                GameRules.mHealthAwarded = 0.3
            else
                GameRules.mHealthAwarded = 0.0
            end
            
            GameRules.mHealthStateTimer = GameRules.mHealthStateBreakDuration          
            HUD:ShowTrackCompleteText( TextManager, GameRules.mHealthStateBreakDuration )
            
        elseif GameRules.mHealthState == Constants.mHealthStatePatterns then
            GameRules.mHealthStateTimer = GameRules.mHealthStateSleepDuration
            HUD:ShowPatternCountText( TextManager, GameRules.mHealthStateTotalTime )
            
        elseif GameRules.mHealthState == Constants.mHealthStateNumPatterns then
            GameRules.mHealthStateTimer = GameRules.mHealthStateBreakDuration
            HUD:ShowPatternCount( TextManager, GameRules.mHealthStatePatternCount, GameRules.mHealthStateTotalTime ) 
            
        elseif GameRules.mHealthState == Constants.mHealthStatePerfect then
            GameRules.mHealthStateTimer = GameRules.mHealthStateSleepDuration
            HUD:ShowPerfectCountText( TextManager, GameRules.mHealthStateTotalTime )
            
        elseif GameRules.mHealthState == Constants.mHealthStateNumPerfect then
            GameRules.mHealthStateTimer = GameRules.mHealthStateBreakDuration
            HUD:ShowPerfectCount( TextManager, GameRules.mHealthStatePerfectCount, GameRules.mHealthStateTotalTime )
            
        elseif GameRules.mHealthState == Constants.mHealthStatePercent then
            GameRules.mHealthStateTimer = GameRules.mHealthStateBreakDuration
            HUD:ShowPercent( TextManager, GameRules.mHealthStatePercent, GameRules.mHealthStateTotalTime )
            HUD:ShowStarsBad()
            
        elseif GameRules.mHealthState == Constants.mHealthStateHealthUp then
            GameRules.mHealthStateTimer = GameRules.mHealthStateBreakDuration * 2.0
            if GameRules.mHealthAwarded > 0.8 then
                HUD:QueueHealthUpText( 3 )
                Score:AddHealthForLevelComplete( GameRules.mHealthAwarded )                        
            elseif GameRules.mHealthAwarded > 0.5 then
                HUD:QueueHealthUpText( 2 )
                Score:AddHealthForLevelComplete( GameRules.mHealthAwarded )                        
            elseif GameRules.mHealthAwarded > 0.2 then
                HUD:QueueHealthUpText( 1 )
                Score:AddHealthForLevelComplete( GameRules.mHealthAwarded )            
            else
                HUD:ShowNoBonusText( TextManager, GameRules.mHealthStateTotalTime )
            end
        end
                
        --update total timer
        GameRules.mHealthStateTotalTime = GameRules.mHealthStateTotalTime - GameRules.mHealthStateTimer
    end
end

function GameRules:LevelStateChange()    
    --exit state logic
    if Constants.LevelState == Constants.LevelStateHealth then
        HUD:HideStars()
    end
    
    --custom logic for advancing levels
    local bAdvanceToNextLevel = false
    if Constants.DesiredLevelState == Constants.LevelStateStart and Constants.LevelState ~= Constants.LevelStateHealth then
        bAdvanceToNextLevel = true
    elseif Constants.DesiredLevelState == Constants.LevelStateHealth then
        bAdvanceToNextLevel = true
    end
    
    --full mix mode only changes visual types and gets harder, it doesn't do pattern state switching
    if Constants.FullMixMode then
        if bAdvanceToNextLevel then
            PostManager:RandomizeFullMixIndex()
            PostManager:SetLevelStateStart()
        end
        
        if Constants.DesiredLevelState == Constants.LevelStateBuild then
            PostManager:SetLevelStateBuild()
        elseif Constants.DesiredLevelState == Constants.LevelStateBonus then
            PostManager:SetLevelStateBonus()
        end        
    else    
        if bAdvanceToNextLevel then            
            --level complete!!!
            Constants.MaxCompletedLevel = math.max( Constants.MaxCompletedLevel, Constants.CurrentGameLevel )      
            Leaderboard:SaveLeaders()
            
            GameRules:IncrementCurrentLevel()   
          
            --check for game complete
            if Constants.CurrentGameLevel > Constants.NumGameLevels then
                --switch to complete state
                Constants.DesiredGameState = Constants.GameStateComplete
                bRestartPostManager = false
            else 
                PostManager:SetLevelStateStart()
            end          
        end
                
        --exit current state logic
        if Constants.LevelState == Constants.LevelStateBonus then
            GameRules.mHealthStatePatternCount = PatternManager.mNumPatternsAttempted
            GameRules.mHealthStatePerfectCount = PatternManager.mNumPatternsPerfect      
            PatternManager:PrimeForNewLevel()
        end
        
        --enter desired state logic 
        if Constants.DesiredLevelState == Constants.LevelStateStart then         
            --trigger tutorials 
            if Constants.CurrentGameLevel == Constants.ThrowTutorialLevel then
                UI:ShowNewAbility( Constants.UINewAbilityStateThrow, TextManager )
            end        
            Score:SetNumBitsToBuildState()
            PatternManager:CleanPattern()
        elseif Constants.DesiredLevelState == Constants.LevelStateBuild then
            PostManager:SetLevelStateBuild()
            HUD:ShowBuildModeText( TextManager )
            PatternManager:SetPatternType( Constants.PatternTypeStandard )        
            PatternManager:CleanPattern()       
        elseif Constants.DesiredLevelState == Constants.LevelStateBonus then
            PatternManager:SetBonusPattern()
            PostManager:SetLevelStateBonus()
            HUD:ShowBonusText( TextManager )
        elseif Constants.DesiredLevelState == Constants.LevelStateHealth then  
            --if we're going in to the complete state, jack this timer up so nothing happens for a bit..
            if Constants.DesiredGameState == Constants.GameStateComplete then
                GameRules.mHealthStateTimer = 999.0    
            else
                GameRules.mHealthStateTimer = 0.0    
            end
            
            GameRules.mHealthStateTotalTime = GameRules.mHealthStateTotalDuration
            Score:SetNumBitsToBuildState()        
            AudioManager:SetEnterBuildStateFlag(false)
            PatternManager:DelayPattern(GameRules.mHealthStateTotalTime)
        end
    end
    
    Constants.LevelState = Constants.DesiredLevelState    
    
    Score:UpdateVerseTextForGameState()
end

function GameRules:UpdateGameStateTitle()
    GameRules.mTitleScreen:Update()
    TextManager:Update() 
    
    if Input:CheckForExcessiveMovement(GameRules.mGameInputs) then
        GameRules.mTitleScreen:RequestClear()
        Input:ResetExcessiveMovement()
    end    
end

function GameRules:UpdateGameStatePreIntro()
    UI:UpdatePreIntro( GameRules.mGameInputs )
    ColorManager:Update()
    GameRing:UpdatePreIntro()
end

function GameRules:UpdateGameStateIntro()
    local iNumChargesPrevEnabled = 0
    local iNumChargesEnabled = 0
    
    if ChargeManager.mCharges[1]:IsActive() then iNumChargesPrevEnabled = iNumChargesPrevEnabled + 1 end
    if ChargeManager.mCharges[2]:IsActive() then iNumChargesPrevEnabled = iNumChargesPrevEnabled + 1 end
    
    ChargeManager:UpdateIntro( GameRules.mGameInputs )
    ChargeManager:Update( GameRules.mGameInputs, true )
    
    if ChargeManager.mCharges[1]:IsActive() then iNumChargesEnabled = iNumChargesEnabled + 1 end
    if ChargeManager.mCharges[2]:IsActive() then iNumChargesEnabled = iNumChargesEnabled + 1 end
    
    if iNumChargesPrevEnabled == 0 and iNumChargesEnabled ~= 0 then
        AudioManager:PlaySound( 'Radius/SFX/RadiusActivationLow' )
    elseif iNumChargesPrevEnabled == 1 and iNumChargesEnabled > 1 then
        AudioManager:PlaySound( 'Radius/SFX/RadiusActivationHigh' )
    end
    
    UI:UpdateIntro( ChargeManager, TextManager )
    ColorManager:Update()
    InputVisuals:UpdateIntro( GameRules.mGameInputs )
    TextManager:Update()    
    InputTutorial:Update( GameRules.mGameInputs, ChargeManager, TextManager )
end

function GameRules:UpdateGameStateMenu()
    --game center button specific
    local bUpdateChargeManager = true
    if GameRules.mMenuDepth == Constants.mMenuDepthOptions and GameRules.mGameCenterProp ~= nil then
        local fXBounds = -Constants.ValidDrawHalfWidth + (Constants.ValidDrawHalfWidth * .3)
        local fYBounds = -Constants.ValidDrawHalfHeight + (Constants.ValidDrawHalfHeight * .4)
        for i = 1, 2 do
            if GameRules.mGameInputs[i].mX < fXBounds and GameRules.mGameInputs[i].mY < fYBounds then
                if GameRules.mGameInputs[i].WasTouched() then 
                    Achievement.showAchievements()
                    bUpdateChargeManager = false
                end            
                if GameRules.mGameInputs[i].IsTouched() then
                    bUpdateChargeManager = false
                end
            end
        end
    end
    
    if bUpdateChargeManager then
        ChargeManager:UpdateMenu( GameRules.mGameInputs )
    end
    
    --update bits and bit queue
    local iRequestedDepth = PatternManager:UpdateMenu( GameRules.mMenuDepth, Renderer, Leaderboard )
    if GameRules.mOverrideMenuDepth ~= -1 then
        iRequestedDepth = GameRules.mOverrideMenuDepth
        GameRules.mOverrideMenuDepth = -1
    end
    
    --update collision
    PatternManager:CheckCollision( ChargeManager, Renderer, GameRules.mGameInputs )

    if GameRules.mMenuDrawState == Constants.mMenuDrawStateOut then
        --continue exiting
        GameRules.mMenuDrawTimer = math.max( GameRules.mMenuDrawTimer - Constants.FrameStep, 0.0 )
        local fStateRatio = GameRules.mMenuDrawTimer / GameRules.mMenuDrawDuration
        HUD:UpdateTutorialDrawRatio( GameRules.mMenuHUDIndex, fStateRatio, GameRules.mMenuSourceX, GameRules.mMenuSourceY )
        
        if GameRules.mMenuDrawTimer <= 0.0 then
            GameRules.mMenuDrawState = Constants.mMenuDrawStateNone
        end
    else
        --look for some text to show
        local iSelectedItem, fSourceX, fSourceY = PatternManager:CheckMenuSelection()   
    
        --what we want to show is not what we're showing..
        if iSelectedItem ~= GameRules.mMenuDrawType then
            --reset!
            TextManager:ClearAll()
                
            local bFadeOut = false
            if iSelectedItem == -1 then
                --nothing selected, fade out 
                bFadeOut = true
            else
                --fade in!
                GameRules.mMenuDrawState = Constants.mMenuDrawStateIn
                GameRules.mMenuSourceX = fSourceX
                GameRules.mMenuSourceY = fSourceY
                HUD:UpdateTutorialDrawRatio( GameRules.mMenuHUDIndex, GameRules.mMenuDrawTimer / GameRules.mMenuDrawDuration, GameRules.mMenuSourceX, GameRules.mMenuSourceY )
                
                if GameRules.mMenuDepth == Constants.mMenuDepthMenu then
                    if iSelectedItem == Constants.mMenuSelectionPlay then
                        HUD:ClearTutorialText()
                        GameRules.mMenuHUDIndex = HUD:InitTutorialText( "play", 4, TextManager, Constants.mHUDPositionPlay, Constants.Color1 )
                    elseif iSelectedItem == Constants.mMenuSelectionFreePlay then
                        HUD:ClearTutorialText()
                        GameRules.mMenuHUDIndex = HUD:InitTutorialText( "full.mix", 8, TextManager, Constants.mHUDPositionFullMix, Constants.SpecialColor )
                    elseif iSelectedItem == Constants.mMenuSelectionLeaders then
                        HUD:ClearTutorialText()
                        GameRules.mMenuHUDIndex = HUD:InitTutorialText( "leaders", 7, TextManager, Constants.mHUDPositionLeaders, Constants.SpecialColor )
                    elseif iSelectedItem == Constants.mMenuSelectionOptions then
                        HUD:ClearTutorialText()
                        GameRules.mMenuHUDIndex = HUD:InitTutorialText( "options", 7, TextManager, Constants.mHUDPositionOptions, Constants.SpecialColor )
                    end
                elseif GameRules.mMenuDepth == Constants.mMenuDepthOptions then
                    if iSelectedItem == Constants.mMenuSelectionHowToPlay then
                        HUD:ClearTutorialText()
                        GameRules.mMenuHUDIndex = HUD:InitTutorialText( "how.to.play", 11, TextManager, Constants.mHUDPositionHowToPlay, Constants.Color1 )                                        
                    elseif iSelectedItem == Constants.mMenuSelectionVisuals then
                        HUD:ClearTutorialText()
                        GameRules.mMenuHUDIndex = HUD:InitTutorialText( "visual.quality", 14, TextManager, Constants.mHUDPositionVQ, Constants.Color1 )                                                        
                    elseif iSelectedItem == Constants.mMenuSelectionCredits then
                        HUD:ClearTutorialText()
                        GameRules.mMenuHUDIndex = HUD:InitTutorialText( "credits", 7, TextManager, Constants.mHUDPositionCredits, Constants.Color1 )                                    
                    elseif iSelectedItem == Constants.mMenuSelectionMenu then
                        HUD:ClearTutorialText()
                        GameRules.mMenuHUDIndex = HUD:InitTutorialText( "back", 4, TextManager, Constants.mHUDPositionBack, Constants.Color1 )                                    
                    end         
                elseif GameRules.mMenuDepth == Constants.mMenuDepthVisuals then
                    if iSelectedItem == Constants.mMenuSelectionLow then
                        HUD:ClearTutorialText()
                        GameRules.mMenuHUDIndex = HUD:InitTutorialText( "standard", 8, TextManager, Constants.mHUDPositionLow, Constants.Color1 )                                        
                    elseif iSelectedItem == Constants.mMenuSelectionHigh then
                        HUD:ClearTutorialText()
                        GameRules.mMenuHUDIndex = HUD:InitTutorialText( "high", 4, TextManager, Constants.mHUDPositionHigh, Constants.Color1 )                                    
                    elseif iSelectedItem == Constants.mMenuSelectionUltra then
                        HUD:ClearTutorialText()
                        GameRules.mMenuHUDIndex = HUD:InitTutorialText( "the.future", 10, TextManager, Constants.mHUDPositionUltra, Constants.Color1 )                                    
                    elseif iSelectedItem == Constants.mMenuSelectionBack then
                        HUD:ClearTutorialText()
                        GameRules.mMenuHUDIndex = HUD:InitTutorialText( "back", 4, TextManager, Constants.mHUDPositionBack, Constants.Color1 )                                    
                    end          
                end
            end 
            
            if bFadeOut then
                GameRules.mMenuDrawState = Constants.mMenuDrawStateOut          
            end
           
            GameRules.mMenuDrawType = iSelectedItem
        else
            --keep showing what we're showing
            if GameRules.mMenuDrawState == Constants.mMenuDrawStateIn then
                GameRules.mMenuDrawTimer = math.min( GameRules.mMenuDrawTimer + Constants.FrameStep, GameRules.mMenuDrawDuration )
                local fStateRatio = GameRules.mMenuDrawTimer / GameRules.mMenuDrawDuration
                HUD:UpdateTutorialDrawRatio( GameRules.mMenuHUDIndex, fStateRatio, GameRules.mMenuSourceX, GameRules.mMenuSourceY )  
                
                if GameRules.mMenuDrawTimer >= GameRules.mMenuDrawDuration then
                    GameRules.mMenuDrawState = Constants.mMenuDrawStateShow
                    HUD:ShowTutorialText( GameRules.mMenuHUDIndex, 999.0, TextManager )
                end   
            elseif GameRules.mMenuDrawState == Constants.mMenuDrawStateShow then
                HUD:UpdateTutorialDrawRatio( GameRules.mMenuHUDIndex, 1.0, GameRules.mMenuSourceX, GameRules.mMenuSourceY )
            elseif GameRules.mMenuDrawState == Constants.mMenuDrawStateNone then
                HUD:ClearTutorialText()
            end         
        end        
    end
    
    local fPercentSelected = PatternManager:GetMenuSelectionPercentComplete()
    UI:UpdateMenu( 1.0 - fPercentSelected )
    ColorManager:Update()
    InputVisuals:UpdateIntro( GameRules.mGameInputs )
    TextManager:Update()     
   
    --if we request a different depth than what we're at
    if iRequestedDepth ~= GameRules.mMenuDepth then        
        if GameRules.mMenuDepth == Constants.mMenuDepthVisuals then
            --if we're exiting the visual quality menu, jiggle the handle on the visual style
            GameRules:VisualQualityChange()
            PostManager:SetStateIntro()
            
            --save our visual style if we're exiting out of the visual quality menu
            if iRequestedDepth == Constants.mMenuDepthOptions then
                Leaderboard:SaveLeaders()
            end
        end
        
        --visuals reset is just an identifier to jiggle the handle
        if iRequestedDepth == Constants.mMenuDepthVisualsReset then
            iRequestedDepth = Constants.mMenuDepthVisuals
        end
        
        GameRules.mMenuDepth = iRequestedDepth
        
        if GameRules.mMenuDepth == Constants.mMenuDepthMenu then
            UI:CreateMenuText( Renderer, "menu" )
        elseif GameRules.mMenuDepth == Constants.mMenuDepthOptions then
            UI:CreateMenuText( Renderer, "options" )
        elseif GameRules.mMenuDepth == Constants.mMenuDepthVisuals then
            UI:CreateMenuText( Renderer, "visuals" )
        end
        
        PatternManager:MenuTransition()
        PatternManager:Reset()
        PatternManager:InitForMenu(GameRules.mMenuDepth)
        
        --create or destroy game center icon
        if GameRules.mMenuDepth == Constants.mMenuDepthOptions then
            if GameRules.mGameCenterProp == nil and MOAIGameCenterIOS then
                local gameCenterDeck = DFGraphics.loadSpriteSheet( 'Images/misc', true )
                local fDesiredImageSize = Constants.ValidDrawHeight / 8.0
                local fGameCenterScale = fDesiredImageSize / 128.0   
                local iDeckIndex = gameCenterDeck.names[ "gamecenter" ]
                
                GameRules.mGameCenterProp = MOAIProp.new()
                GameRules.mGameCenterProp:setDeck( gameCenterDeck )
                GameRules.mGameCenterProp:setLoc( -Constants.ValidDrawHalfWidth + fDesiredImageSize, -Constants.ValidDrawHalfHeight + fDesiredImageSize, 0.0 )  
                GameRules.mGameCenterProp:setScl( fGameCenterScale, -fGameCenterScale ) 
                GameRules.mGameCenterProp:setIndex( iDeckIndex )                
                
                Renderer.mGameLayer:insertProp( GameRules.mGameCenterProp )
            end
        else
            GameRules:DestroyGameCenterProp()
        end
        
        TextManager:ClearAll() 
        HUD:ClearTutorialText()
        AudioManager:PlaySound( 'Radius/SFX/HoldandPlayMeet' )
        
        GameRules.mMenuDrawState = Constants.mMenuDrawStateNone
        GameRules.mMenuDrawTimer = 0.0   
    end
end

function GameRules:UpdateGameStateInGame()
    MOAISim.profileEnterScope("LUA levelstate update")
    GameRules:UpdateLevelState()
    MOAISim.profileLeaveScope("LUA levelstate update")
    
    --check for game pause
    GameRules:Pause( not ChargeManager:GetBeamState() )
    
    --update pause timer
    GameRules.mPauseDelayTimer = GameRules.mPauseDelayTimer - Constants.FrameStep
            
    --update game
    if Constants.mPaused == true then
        local fPauseRatio = ChargeManager:UpdatePaused( GameRules.mGameInputs )
        local iReturn = PauseScreen:UpdatePauseScreen( GameRules.mGameInputs )
        if iReturn == Constants.mPauseScreenReturnQuit then
            GameRules:Pause( false )
            Constants.DesiredGameState = Constants.GameStatePost
        elseif iReturn == Constants.mPauseScreenReturnNext then
            if not Constants.FullMixMode then
                GameRules:SkipToNextLevel()        
            end
        end        
    else    
        --if we're in a tutorial state, run through the sequence
        if GameRules.mTutorialState ~= Constants.mTutorialStateNone then
            GameRules:UpdateGameStateInGameTutorial()
        else
            --standard gameplay
            local bTimerPreventPause = GameRules.mPauseDelayTimer > 0.0
            local bCanPause = not UI:IsInAbilityTutorial() and Constants.LevelState ~= Constants.LevelStateHealth and not bTimerPreventPause
            
            ChargeManager:Update( GameRules.mGameInputs, bCanPause )     
            
            --MOAISim.profileEnterScope("LUA pattern manager update")
            PatternManager:CheckCollision( ChargeManager, Renderer, GameRules.mGameInputs )
            PatternManager:Update( TextManager )
            if PatternManager:CheckForPatternTypeChange( ChargeManager ) then
                GameRules.mPauseDelayTimer = 2.0
            end
            --MOAISim.profileLeaveScope("LUA pattern manager update")
            
            --MOAISim.profileEnterScope("LUA UI/color update")
            UI:UpdateInGame( GameRules.mGameInputs, ChargeManager, TextManager )
            --MOAISim.profileLeaveScope("LUA UI/color update")
            
            --HUD responses
            --MOAISim.profileEnterScope("LUA hudresponse update")
            if PatternManager.mShowTutorial == true then
                GameRules:InitInGameTutorial( PatternManager.mTutorialType )
                PatternManager.mShowTutorial = false                
            end
            --MOAISim.profileLeaveScope("LUA hudresponse update")
            
            --MOAISim.profileEnterScope("LUA score update")
            Score:Update( ChargeManager, GameRules.mLevelIntensity )
            --MOAISim.profileLeaveScope("LUA score update")
        end
    end
    
    MOAISim.profileEnterScope("LUA xtra ui update")
    GameRing:UpdateRotation()   
    InputVisuals:Update( GameRules.mGameInputs )
    TextManager:Update()    
    ColorManager:Update()
    OpponentScore:Update( Score.mScore )
    MOAISim.profileLeaveScope("LUA xtra ui update")
end

function GameRules:UpdateGameStateInGameTutorial()
    GameRules.mTutorialStateTimer = math.max( GameRules.mTutorialStateTimer - Constants.FrameStep, 0.0 )

    --update tutorial source
    local fSourceX, fSourceY = GameRules:GetInGameTutorialSource()
    local fLowpassLimit = 200.0
    
    if GameRules.mTutorialState == Constants.mTutorialStateEnterTutorial then
        local fStateRatio = 1.0 - (GameRules.mTutorialStateTimer / GameRules.mTutorialStateTransitionDuration)
        HUD:UpdateTutorialDrawRatio( 1, fStateRatio, fSourceX, fSourceY )
        GameRules:SetCameraOffsetForTutorial( fStateRatio )
            
        if GameRules.mTutorialStateTimer <= 0.0 then
            GameRules.mTutorialStateTimer = GameRules.mTutorialStateWaitDuration
            GameRules.mTutorialState = Constants.mTutorialStateWaitTutorial
            
            --show text for tutorial
            local fTextDuration = GameRules.mTutorialStateWaitDuration
            if GameRules.mTutorialType == Constants.mTutorialTypeThrow then
                fTextDuration = 999999.0
            end
            
            HUD:ShowTutorialText( 1, fTextDuration, TextManager )
        end
    elseif GameRules.mTutorialState == Constants.mTutorialStateWaitTutorial then
        HUD:UpdateTutorialDrawRatio( 1, 1.0, fSourceX, fSourceY )
        if GameRules.mTutorialType ~= Constants.mTutorialTypeThrow then                
            if GameRules.mTutorialStateTimer <= 0.0 then
                GameRules.mTutorialStateTimer = GameRules.mTutorialStateTransitionDuration
                GameRules.mTutorialState = Constants.mTutorialStateExitTutorial                
            end
        elseif ChargeManager:AreAnyChargesThrown() then
            TextManager:ClearAll()
            GameRules.mTutorialStateTimer = GameRules.mTutorialStateTransitionDuration
            GameRules.mTutorialState = Constants.mTutorialStateExitTutorial                        
        end
    elseif GameRules.mTutorialState == Constants.mTutorialStateExitTutorial then
        local fStateRatio = GameRules.mTutorialStateTimer / GameRules.mTutorialStateTransitionDuration
        HUD:UpdateTutorialDrawRatio( 1, fStateRatio, fSourceX, fSourceY )
        GameRules:SetCameraOffsetForTutorial( fStateRatio )
                
        if GameRules.mTutorialStateTimer <= 0.0 then
            Score:SetScoreVisible(true)
            HUD:ClearTutorialText()
            GameRules:SetCameraOffsetForTutorial( 0.0 )
            GameRules.mTutorialStateTimer = 0.0
            GameRules.mTutorialState = Constants.mTutorialStateNone 
            GameRules.mTutorialType = Constants.mTutorialTypeNone
        end
    end
    
    --allow player to keep moving
    ChargeManager:Update( GameRules.mGameInputs, false )
end

function GameRules:UpdateGameStateComplete()
    PatternManager:Update( TextManager )
    ChargeManager:UpdateComplete( GameRules.mGameInputs, InputVisuals )
    ColorManager:Update()
    GameRules.mLevelIntensity = UI:UpdateComplete( Post, PostLastLevel )
    
    InputVisuals:UpdateGameComplete( GameRules.mGameInputs )
    Score:UpdateComplete( GameRules.mLevelIntensity )
end

function GameRules:UpdateGameStatePost()
    UI:UpdatePost( GameRules.mGameInputs )
    ColorManager:Update()
end

function GameRules:UpdateGameStateLeaderboardEntry()
    LeaderboardEntry:Update( GameRules.mGameInputs )
    ColorManager:Update()
    InputVisuals:Update( GameRules.mGameInputs )
    UI:UpdateLeaderboardEntry()
end

function GameRules:UpdateGameStateLeaderboardShowcase()
    ColorManager:Update()
    TextManager:Update()    
    local bDone = Leaderboard:Update( GameRules.mGameInputs )
    InputVisuals:Update( GameRules.mGameInputs )
    
    if bDone then
        Leaderboard:HideLeaders( Renderer )
    end
    
    if not Leaderboard:AreLeadersShowing() then
        if GameRules.mResetGameAfterLeaderboards == true then
            Constants.DesiredGameState = Constants.GameStatePreIntro
        else
            Constants.DesiredGameState = Constants.GameStateMenu  
        end
    end    
end

function GameRules:UpdateGameStateCredits()
    ColorManager:Update()
    TextManager:Update()
    local bDone = Credits:Update( GameRules.mGameInputs )
    InputVisuals:Update( GameRules.mGameInputs )
    
    if bDone then
        Credits:EndCredits( Renderer )
    end
    
    if not Credits:AreCreditsShowing() then
        Constants.DesiredGameState = Constants.GameStateMenu  
    end
end

function GameRules:UpdateGameStateHowToPlay()
    ColorManager:Update()
    TextManager:Update()
    HowToPlay:Update( GameRules.mGameInputs )
    InputVisuals:Update( GameRules.mGameInputs )
        
    if not HowToPlay:IsHowToPlayShowing() then
        Constants.DesiredGameState = Constants.GameStateMenu  
    end
end

function GameRules:VisualQualityChange()
    Constants.VisualQuality = Constants.DesiredVisualQuality
    
    Renderer:VisualQualityChange()
    Post:VisualQualityChange()
    Grid:VisualQualityChange()
    Wind:VisualQualityChange()
    ChargeField:VisualQualityChange()
end

function GameRules:GameStateChange()    
    --exit state logic
    if Constants.GameState == Constants.GameStatePreIntro then
        MOAISim.forceGarbageCollection()
        UI:InitIntro()
        GameRing:ExitPreIntro()
    elseif Constants.GameState == Constants.GameStatePerformanceTest then
        PerformanceTest.Stop()
    elseif Constants.GameState == Constants.GameStateIntro then
        UI:SetTitleVisible(false)
        InputTutorial:Disable(Renderer)
    elseif Constants.GameState == Constants.GameStateMenu then
        PatternManager:Reset()
        TextManager:ClearAll() 
        HUD:ClearTutorialText()
        UI:ExitMenu()
        UI:DestroyMenuText( Renderer )
        if Constants.DesiredGameState ~= Constants.GameStateInGame then
            AudioManager:PlaySound( 'Radius/SFX/MenuSelect' )
        end
        GameRules.mMenuDrawState = Constants.mMenuDrawStateNone
        GameRules.mMenuDrawTimer = 0.0        
        GameRules:DestroyGameCenterProp()
    elseif Constants.GameState == Constants.GameStateComplete then
        AudioManager:SetAdvanceGameFlag(false)
        AudioManager:SetAdvanceGameScalar(1.0)
    elseif Constants.GameState == Constants.GameStatePost then
        ColorManager:SetLineModeIntro()
        ChargeManager:Reset()
        ChargeManager:SetToAltGameMode(false)
        if Constants.DesiredGameState == Constants.GameStateLeaderboardShowcase then
            UI:InitPreIntro( GameRules.mPreIntroDurations[1], GameRules.mPreIntroDurations[2] )
        end
        AudioManager:SetFullMixMode( false )
    elseif Constants.GameState == Constants.GameStateCredits then
        Credits:EndCredits( Renderer )
        TextManager:ClearAll()
        GameRules:SetZoomPercents(0.0, 0.0, 1.0) 
    elseif Constants.GameState == Constants.GameStateLeaderboardShowcase then
        PatternManager:Reset()
        Leaderboard:HideLeaders( Renderer )
        TextManager:ClearAll()
        GameRules:SetZoomPercents(0.0, 0.0, 1.0) 
    elseif Constants.GameState == Constants.GameStateLeaderboardEntry then  
        PatternManager:Reset()
        LeaderboardEntry:Disable()
        UI:InitPreIntro( GameRules.mPreIntroDurations[1], GameRules.mPreIntroDurations[2] )
        InputVisuals:Reset()
        GameRules:SetZoomPercents(0.0, 0.0, 1.0)
    elseif Constants.GameState == Constants.GameStateHowToPlay then
        HowToPlay:EndHowToPlay( Renderer )
        TextManager:ClearAll()
        GameRules:SetZoomPercents(0.0, 0.0, 1.0) 
    end
    
    --enter state logic
    if Constants.DesiredGameState == Constants.GameStateTitle then
        PostManager:SetStateIntro()
        UI:InitPreIntro( GameRules.mPreIntroDurations[1], GameRules.mPreIntroDurations[2] )
        ChargeManager:Reset()
        GameRules.mTitleScreen:Show()
        GameRing:InitTitle()
    elseif Constants.DesiredGameState == Constants.GameStatePerformanceTest then
        PerformanceTest.Start()
    elseif Constants.DesiredGameState == Constants.GameStatePreIntro then  
        GameRing:InitPreIntro( GameRules.mPreIntroDurations[2] )
        ColorManager:SetLineModeIntro()
        PatternManager:Reset()
        Score:InitForIntro()    
        UI:InitPreIntro( GameRules.mPreIntroDurations[1], GameRules.mPreIntroDurations[2] )
        InputVisuals:Reset()
        ChargeManager:Reset()
        if Constants.GameState ~= Constants.GameStateTitle then
            AudioManager:StopAmbientTrack()
            PostManager:SetStateIntro()
        end
        AudioManager:StartAmbientTrack()
        GameRules:SetZoomPercents(0.0, 0.0, 1.0)
        AudioManager:PlaySound( 'Radius/SFX/DiscordLogo' )
        Constants.DesiredLevelState = Constants.LevelStateStart
        Constants.CurrentGameLevel = 1
    elseif Constants.DesiredGameState == Constants.GameStateIntro then   
        InputTutorial:InitForIntro(Renderer)
        GameRing:MakeRingVisible()
    elseif Constants.DesiredGameState == Constants.GameStateMenu then
        GameRules.mMenuDepth = 0
        PatternManager:InitForMenu( GameRules.mMenuDepth )
        UI:InitMenu()
        UI:CreateMenuText( Renderer, "menu" )
        if Constants.GameState ~= Constants.GameStateIntro then
            PostManager:SetStateIntro()
        end
        GameRules.mResetGameAfterLeaderboards = false
    elseif Constants.DesiredGameState == Constants.GameStateInGame then
        GameRules.mResetGameAfterLeaderboards = true
        Constants.LevelState = Constants.LevelStateStart
        Constants.DesiredLevelState = Constants.LevelState        
        Constants.ThrowUnlocked = false
        Constants.LeaderboardIndex = -1
        Constants.PreviousGameScore = 0
        Constants.PreviousGameInitials = "!!!"
        Constants.FirstTimePlay = false
        AudioManager:StartLoopTrack()
        ColorManager:SetLineModeInGame()
        PatternManager:InitForGame()
        UI:FlashWhite(3.0)
        UI:InitInGame()
        Score:InitForGame()
        ChargeManager:SetDisabledDurationForGame()
        GameRing:MakeRingTimerVisible(true)
        PostManager:RandomizeFullMixIndex()
        PostManager:SetLevelStateStart()
        OpponentScore:InitForGame()
    elseif Constants.DesiredGameState == Constants.GameStateComplete then
        Score:InitForComplete()
        PatternManager:InitForComplete()
        AudioManager:SetAdvanceGameFlag(true)
        InputVisuals:SetFingerEffects("Effects/Particles/Fingers/BlobDotsFade.particles", false)
        Achievement.updateAchievement( "DA_FullSetComplete", 100 )
    elseif Constants.DesiredGameState == Constants.GameStatePost then
        AudioManager:StopMusicForPostGame()
        UI:InitPost( Score.mScore )
        Score:InitForPost()
        if Constants.GameState ~= Constants.GameStateComplete then
            PostManager:SetStateGameOver() 
        end
        Leaderboard:DetermineScoreRank( Score.mScore )
        TextManager:ClearAll()     
        GameRing:MakeRingTimerVisible(false)
        ChargeManager:ExitPauseState()
        InputVisuals:SetAltGameMode(false)
        GameRing:InitPost()      
        OpponentScore:ClearScores()
        Constants.GameCameraTiltDesiredMode = Constants.GameCameraTiltModeNone
    elseif Constants.DesiredGameState == Constants.GameStateCredits then
        GameRules:SetZoomPercents(0.0, 0.0, 0.65)
        PostManager:SetStateCredits() 
        PostManager:SetStateLeaderboardShowcase()  
        Credits:BeginCredits( Renderer )
        Input:ResetExcessiveMovement()
    elseif Constants.DesiredGameState == Constants.GameStateLeaderboardEntry then
        UI:InitLeaderboardEntry()
        LeaderboardEntry:Enable()
        GameRules:SetZoomPercents(0.0, 0.0, 0.75)
    elseif Constants.DesiredGameState == Constants.GameStateLeaderboardShowcase then  
        GameRules:SetZoomPercents(0.0, 0.0, 0.65)
        PostManager:SetStateLeaderboardShowcase()  
        Leaderboard:ShowHighScores( Renderer ) 
        Input:ResetExcessiveMovement()
    elseif Constants.DesiredGameState == Constants.GameStateHowToPlay then  
        GameRules:SetZoomPercents(0.0, 0.0, 0.65)
        PostManager:SetStateLeaderboardShowcase() 
        HowToPlay:BeginHowToPlay( Renderer )
        Input:ResetExcessiveMovement()
    end
    
    GameRules.mGameStateElapsedTime = 0.0
    Constants.GameState = Constants.DesiredGameState
end

function GameRules:IncrementCurrentLevel()
    Constants.CurrentGameLevel = Constants.CurrentGameLevel + 1
    
    --make sure the throw is unlocked if we're skipping levels
    if Constants.CurrentGameLevel > Constants.ThrowTutorialLevel then
        Constants.ThrowUnlocked = true
    end
end

function GameRules:Pause( bPause )
    local bCanPause = not UI:IsInAbilityTutorial() and Constants.LevelState ~= Constants.LevelStateHealth and GameRules.mTutorialType == Constants.mTutorialTypeNone
    if Constants.GameState == Constants.GameStateInGame and bPause ~= Constants.mPaused and Constants.AllowPause and bCanPause then
        Constants.mPaused = bPause
        
        if Constants.mPaused == true then
            ChargeManager:SetDisabledDurationForPause()    
            PauseScreen:ShowPauseScreen()
        else
            ChargeManager:SetDisabledDurationForGame()    
            PauseScreen:HidePauseScreen()
        end
        
        AudioManager:PauseMusic()
    end
end

function GameRules:InitInGameTutorial( iType )
    GameRules.mTutorialType = iType
    GameRules.mTutorialState = Constants.mTutorialStateEnterTutorial
    GameRules.mTutorialStateTimer = GameRules.mTutorialStateTransitionDuration    
    
    HUD:ClearTutorialText()
    if GameRules.mTutorialType == Constants.mTutorialTypeBad then
        HUD:InitTutorialText( "avoid", 5, TextManager, Constants.mHUDPositionBit, Constants.GeneralRed )
    elseif GameRules.mTutorialType == Constants.mTutorialTypePoint then
        HUD:InitTutorialText( "tap", 3, TextManager, Constants.mHUDPositionBit, Constants.Color1 )   
    elseif GameRules.mTutorialType == Constants.mTutorialTypeHold then
        HUD:InitTutorialText( "hold", 4, TextManager, Constants.mHUDPositionBit, Constants.Color1 )          
    elseif GameRules.mTutorialType == Constants.mTutorialTypeThrow then
        HUD:InitTutorialText( "swipe", 5, TextManager, Constants.mHUDPositionBit, Constants.Color1 )
        Constants.ThrowUnlocked = true
    end
    
    Score:SetScoreVisible(false)
end

function GameRules:GetInGameTutorialSource()
    if GameRules.mTutorialType == Constants.mTutorialTypeBad then
        return PatternManager:GetBadTutorialSource()
    elseif GameRules.mTutorialType == Constants.mTutorialTypePoint then
        return PatternManager:GetPointTutorialSource()
    elseif GameRules.mTutorialType == Constants.mTutorialTypeHold then
        return PatternManager:GetHoldTutorialSource()        
    elseif GameRules.mTutorialType == Constants.mTutorialTypeThrow then
        return ChargeManager:GetChargeTutorialSource()
    end
end

function GameRules:SetCameraOffsetForTutorial( fRatio )
    local x, y, z = Renderer.mGameCamera:getLoc()
    local fOffset = DFMath.lerp( 0.0, Constants.ValidDrawHalfWidth*0.1, fRatio )
    
    Renderer.mGameCamera:setLoc( fOffset, y, z )
end

function GameRules:DestroyGameCenterProp()
    if GameRules.mGameCenterProp ~= nil then
        Renderer.mGameLayer:removeProp( GameRules.mGameCenterProp )
        GameRules.mGameCenterProp = nil
    end
end
            
function GameRules:SetZoomPercents(x, y, z)
    GameRules.mDesiredCameraPercents[1] = x
    GameRules.mDesiredCameraPercents[2] = y
    GameRules.mDesiredCameraPercents[3] = z
    
    GameRules.mDesiredCameraZ = Renderer.mGameCameraZDistance * GameRules.mDesiredCameraPercents[3]
    Constants.GameCameraTiltDesiredMode = Constants.GameCameraTiltModeNone
end

function GameRules:UpdateScreenTilt()   
    --state change
    Constants.GameCameraTiltMode = Constants.GameCameraTiltDesiredMode
    
    local fTiltMax = 0.0  

    local fDesiredZ = Renderer.mGameCameraZDistance
    
    GameRules.mDesiredCameraOffset[1] = 0.0
    GameRules.mDesiredCameraOffset[2] = 0.0
    GameRules.mDesiredCameraOffset[3] = fDesiredZ
    
    local i = 1
    for i = 1, 2 do
        if GameRules.mGameInputs[i].IsTouched() then
            GameRules.mCameraTiltInputs[i]:Update(GameRules.mGameInputs[i].mX, GameRules.mGameInputs[i].mY, GameRules.mGameInputs[i].mZ) 
        else
            GameRules.mCameraTiltInputs[i]:NoInput()
        end
    end

    local fTotalActive = GameRules.mCameraTiltInputs[1].mActiveAmount + GameRules.mCameraTiltInputs[2].mActiveAmount
    
    --if we're not in tilt mode, update our camera to the desired zoom level
    if Constants.GameCameraTiltMode == Constants.GameCameraTiltModeNone then
        local x, y, z = Renderer.mGameCamera:getLoc()
        local fMovementStep = GameRules.mDesiredCameraSpeed * Renderer.mGameCameraZDistance * Constants.FrameStep
        if z < GameRules.mDesiredCameraZ - fMovementStep then
            fDesiredZ = z + fMovementStep
        elseif z > GameRules.mDesiredCameraZ + fMovementStep then
            fDesiredZ = z - fMovementStep 
        else
            fDesiredZ = GameRules.mDesiredCameraZ 
        end
    elseif Constants.GameCameraTiltMode == Constants.GameCameraTiltModeTiltHoriz then
        --for horizontal tilt, move the camera left/right according to x and z positions
        fTiltMax = (Renderer.mGameCameraZDistance * GameRules.mCameraTiltMaxHoriz)
        for i = 1, 2 do
            if GameRules.mCameraTiltInputs[i].mActiveAmount > 0.0 then
                local fXRatio = GameRules.mCameraTiltInputs[i].mX / Constants.ValidDrawHalfWidth
                local fXAmount = DFMath.pin( fXRatio * GameRules.mCameraTiltInputs[i].mActiveAmount, -1.0, 1.0 )
                GameRules.mDesiredCameraOffset[1] = GameRules.mDesiredCameraOffset[1] + (-fXAmount * fTiltMax)
            end
        end 
    elseif Constants.GameCameraTiltMode == Constants.GameCameraTiltModeTiltVert then
        --for horizontal tilt, move the camera up/down according to y and z positions
        fTiltMax = (Renderer.mGameCameraZDistance * GameRules.mCameraTiltMaxVert)
        for i = 1, 2 do
            if GameRules.mCameraTiltInputs[i].mActiveAmount > 0.0 then
                local fYRatio = GameRules.mCameraTiltInputs[i].mY / Constants.ValidDrawHalfHeight
                local fYAmount = DFMath.pin( fYRatio * GameRules.mCameraTiltInputs[i].mActiveAmount, -1.0, 1.0 )
                GameRules.mDesiredCameraOffset[2] = GameRules.mDesiredCameraOffset[2] + (-fYAmount * fTiltMax)
            end
        end 
    elseif Constants.GameCameraTiltMode == Constants.GameCameraTiltModeTiltBoth or Constants.GameCameraTiltMode == Constants.GameCameraTiltModeTiltBothNoAudio then
        --for both tilt, move the camera all around according to x, y, and z positions
        fTiltMax = (Renderer.mGameCameraZDistance * GameRules.mCameraTiltMaxHoriz)
        local fHalfDrawHeightSquared = Constants.ValidDrawHalfHeight * Constants.ValidDrawHalfHeight
        
        for i = 1, 2 do
            if GameRules.mCameraTiltInputs[i].mActiveAmount > 0.0 then
                local fNormPos = { GameRules.mCameraTiltInputs[i].mX, GameRules.mCameraTiltInputs[i].mY }
                local fLengthSquared = DFMath.lengthSquared( fNormPos[1], fNormPos[2] )
                if fLengthSquared > fHalfDrawHeightSquared then
                    fNormPos[1], fNormPos[2] = DFMath.normalize( fNormPos[1], fNormPos[2] )
                    fNormPos = DFMath.scaleValueTable( fNormPos, Constants.ValidDrawHalfHeight )
                end
                
                local fXRatio = fNormPos[1] / Constants.ValidDrawHalfHeight
                local fXAmount = DFMath.pin( fXRatio * GameRules.mCameraTiltInputs[i].mActiveAmount, -1.0, 1.0 )
                
                local fYRatio = fNormPos[2] / Constants.ValidDrawHalfHeight
                local fYAmount = DFMath.pin( fYRatio * GameRules.mCameraTiltInputs[i].mActiveAmount, -1.0, 1.0 )
                                
                GameRules.mDesiredCameraOffset[1] = GameRules.mDesiredCameraOffset[1] + (-fXAmount * fTiltMax)
                GameRules.mDesiredCameraOffset[2] = GameRules.mDesiredCameraOffset[2] + (-fYAmount * fTiltMax)
            end
        end 
    end

    --only update the camera if we're not in the tutorial state (which also updates it)
    if GameRules.mTutorialState == Constants.mTutorialStateNone then
        local fDistToDesired = 0.0
        local fCameraDiff = {}
        fCameraDiff[1] = GameRules.mDesiredCameraOffset[1] - GameRules.mCurrentCameraOffset[1]
        fCameraDiff[2] = GameRules.mDesiredCameraOffset[2] - GameRules.mCurrentCameraOffset[2]        
        fCameraDiff[3] = GameRules.mDesiredCameraOffset[3] - GameRules.mCurrentCameraOffset[3]        
        fCameraDiff[1], fCameraDiff[2], fCameraDiff[3], fDistToDesired = DFMath.normalize3DGetLength( fCameraDiff[1], fCameraDiff[2], fCameraDiff[3] )
        local fMovementStep = GameRules.mCameraOffsetSpeed * Constants.FrameStep
        fCameraDiff = DFMath.scaleValueTable( fCameraDiff, math.min(fMovementStep, fDistToDesired) )
        
        GameRules.mCurrentCameraOffset[1] = GameRules.mCurrentCameraOffset[1] + fCameraDiff[1]
        GameRules.mCurrentCameraOffset[2] = GameRules.mCurrentCameraOffset[2] + fCameraDiff[2]
        GameRules.mCurrentCameraOffset[3] = GameRules.mCurrentCameraOffset[3] + fCameraDiff[3]

        DFRad:RotateCameraTowardPoint( Renderer.mGameCamera, GameRules.mCurrentCameraOffset[1], GameRules.mCurrentCameraOffset[2], GameRules.mCurrentCameraOffset[3], fDesiredZ )
        
        if MOAIFmodEventMgr then
            if Constants.GameCameraTiltMode == Constants.GameCameraTiltModeTiltHoriz or Constants.GameCameraTiltMode == Constants.GameCameraTiltModeTiltBoth then
                local fTiltRatio = GameRules.mCurrentCameraOffset[1] / fTiltMax
                local fModifiedRatio = DFMath.pin( fTiltRatio * 1.5, -1.0, 1.0 )
                GameRules.mDesiredPanValue = fTiltRatio*-1.0
                GameRules.mDesiredHighpassValue = math.abs(fTiltRatio) 
            end
            
            if Constants.GameCameraTiltMode == Constants.GameCameraTiltModeTiltVert or Constants.GameCameraTiltMode == Constants.GameCameraTiltModeTiltBoth then
                local fTiltBuffer = 20.0
                if GameRules.mCurrentCameraOffset[2] > fTiltBuffer then
                    local fTiltRange = fTiltMax - fTiltBuffer
                    local fValue = (GameRules.mCurrentCameraOffset[2] - fTiltBuffer) / fTiltRange
                    GameRules.mDesiredHighpassValue = math.max( fValue, 0.0 )
                    GameRules.mDesiredLowpassValue = 0.0
                elseif GameRules.mCurrentCameraOffset[2] < -fTiltBuffer then
                    local fTiltRange = fTiltMax - fTiltBuffer
                    local fValue = (math.abs(GameRules.mCurrentCameraOffset[2]) - fTiltBuffer) / fTiltRange                
                    GameRules.mDesiredLowpassValue = math.max( fValue, 0.0 )
                    GameRules.mDesiredHighpassValue = 0.0                
                else
                    GameRules.mDesiredLowpassValue = 0.0
                    GameRules.mDesiredHighpassValue = 0.0
                end
            end
            
            if Constants.GameCameraTiltMode == Constants.GameCameraTiltModeNone or Constants.GameCameraTiltMode == Constants.GameCameraTiltModeTiltBothNoAudio then
                GameRules.mDesiredHighpassValue = 0.0
                GameRules.mDesiredLowpassValue = 0.0
                GameRules.mDesiredPanValue = 0.0
                GameRules.mDesiredFlangeValue = 0.0
            end
        end            
    end
end

function GameRules:UpdateDSPValues()    
    --update highpass
    local fHighpassStep = GameRules.mHighpassChange * Constants.FrameStep
    local fHighpassDiff = math.abs(GameRules.mDesiredHighpassValue - GameRules.mCurrentHighpassValue)
    if fHighpassDiff > 0.0 then
        if fHighpassDiff < fHighpassStep then
            GameRules.mCurrentHighpassValue = GameRules.mDesiredHighpassValue
        elseif GameRules.mCurrentHighpassValue < GameRules.mDesiredHighpassValue then
            GameRules.mCurrentHighpassValue = GameRules.mCurrentHighpassValue + fHighpassStep
        else
            GameRules.mCurrentHighpassValue = GameRules.mCurrentHighpassValue - fHighpassStep
        end
        
        MOAIFmodEventMgr.setSeamlessQueueEventParam( "highpass", GameRules.mCurrentHighpassValue )
    end
    
    --update lowpass
    local fLowpassStep = GameRules.mLowpassChange * Constants.FrameStep
    local fLowpassDiff = math.abs(GameRules.mDesiredLowpassValue - GameRules.mCurrentLowpassValue)
    if fLowpassDiff > 0.0 then
        if fLowpassDiff < fLowpassStep then
            GameRules.mCurrentLowpassValue = GameRules.mDesiredLowpassValue
        elseif GameRules.mCurrentLowpassValue < GameRules.mDesiredLowpassValue then
            GameRules.mCurrentLowpassValue = GameRules.mCurrentLowpassValue + fLowpassStep
        else
            GameRules.mCurrentLowpassValue = GameRules.mCurrentLowpassValue - fLowpassStep
        end
        
        MOAIFmodEventMgr.setSeamlessQueueEventParam( "lowpass", GameRules.mCurrentLowpassValue )
    end
    
    --update flange
    local fFlangeStep = GameRules.mFlangeChange * Constants.FrameStep
    local fFlangeDiff = math.abs(GameRules.mDesiredFlangeValue - GameRules.mCurrentFlangeValue)
    if fFlangeDiff > 0.0 then
        if fFlangeDiff < fFlangeStep then
            GameRules.mCurrentFlangeValue = GameRules.mDesiredFlangeValue
        elseif GameRules.mCurrentFlangeValue < GameRules.mDesiredFlangeValue then
            GameRules.mCurrentFlangeValue = GameRules.mCurrentFlangeValue + fFlangeStep
        else
            GameRules.mCurrentFlangeValue = GameRules.mCurrentFlangeValue - fFlangeStep
        end
        
        MOAIFmodEventMgr.setSeamlessQueueEventParam( "flange", GameRules.mCurrentFlangeValue )
    end
    
    --update pan
    local fPanStep = GameRules.mPanChange * Constants.FrameStep
    local fPanDiff = math.abs(GameRules.mDesiredPanValue - GameRules.mCurrentPanValue)
    if fPanDiff > 0.0 then
        if fPanDiff < fPanStep then
            GameRules.mCurrentPanValue = GameRules.mDesiredPanValue
        elseif GameRules.mCurrentPanValue < GameRules.mDesiredPanValue then
            GameRules.mCurrentPanValue = GameRules.mCurrentPanValue + fPanStep
        else
            GameRules.mCurrentPanValue = GameRules.mCurrentPanValue - fPanStep
        end
        
        MOAIFmodEventMgr.setSeamlessQueuePan( GameRules.mCurrentPanValue )
    end 
end

function GameRules:SkipToNextLevel()
    local bCanSkipLevel = Constants.MaxCompletedLevel >= Constants.CurrentGameLevel and Constants.CurrentGameLevel ~= Constants.NumGameLevels
    if (bCanSkipLevel or Constants.AllowSkipAllLevels) and not Constants.FullMixMode then 
        GameRules:Pause( false )
         
        ChargeManager:ExitPauseState()
        
        GameRules:IncrementCurrentLevel()
        Constants.LevelState = Constants.LevelStateStart
        Constants.DesiredLevelState = Constants.LevelState     
        Score:SetNumBitsToBuildState()
        Score:UpdateVerseTextForGameState()
        
        PatternManager:PrimeForNewLevel()
        PatternManager:CleanPattern()
        
        if Constants.CurrentGameLevel == Constants.ThrowTutorialLevel then
            UI:ShowNewAbility( Constants.UINewAbilityStateThrow, TextManager )       
        end

        PostManager:SetLevelStateStart()
        AudioManager:StartLoopTrack()
        UI:FlashWhite(3.0)
    end
end

function GameRules:EatInput( fDuration ) 
    local i = 1
    for i = 1, GameRules.mNumGameInputs do
        GameRules.mGameInputs[i]:Reset()
    end
    
    GameRules.mEatInputTimer = fDuration
end

function GameRules:UpdateInput()
    local i = 1
    local j = 1
    
    --used for transitioning between states
    if GameRules.mEatInputTimer > 0.0 then
        GameRules.mEatInputTimer = GameRules.mEatInputTimer - Constants.FrameStep
        
        for i = 1, GameRules.mNumGameInputs do
            GameRules.mGameInputs[i]:Reset()
        end            
        return
    end

    if Input.HasMultitouch() and not Constants.TouchBotEnabled then
        local iInputIndex = 1
        
        --reset dirty flags in case we need them
        Input.ResetDirtyFlags()
        for i = 1, GameRules.mNumGameInputs do
            GameRules.mGameInputs[i].mDirty = false
        end
        
        --first, run through our inputs and see if their touch index is still valid
        for i = 1, GameRules.mNumGameInputs do
            local iInputTouchIndex = GameRules.mGameInputs[i].mTouchIndex
            if iInputTouchIndex ~= -1 then
                local touch = Input.GetTouch( iInputTouchIndex )
                if touch.bIsDown and not touch.bDirty then
                    --our input was valid last frame and it's down this frame, hooray!
                    local fMouseX, fMouseY
                    fMouseX, fMouseY = GameRules:CenterInput( touch.x, touch.y )                
                    GameRules.mGameInputs[i].mX = fMouseX
                    GameRules.mGameInputs[i].mY = fMouseY                    
                    GameRules.mGameInputs[i]:Touch( iInputTouchIndex )
                    GameRules.mGameInputs[i].mDirty = true
                    
                    --flag this input as dirty
                    Input.SetDirtyFlag( iInputTouchIndex )
                end
            end
        end
        
        --now, run back through the touches and assign any clean values to game inputs        
        for i = 1, GameRules.mNumGameInputs do
            local touch = Input.GetTouch(i)
            if touch.bIsDown and not touch.bDirty then                
                -- center the touch coordinates
                local fMouseX, fMouseY
                fMouseX, fMouseY = GameRules:CenterInput( touch.x, touch.y )
            
                --find a non-dirty game input to assign it to
                for j = 1, GameRules.mNumGameInputs do
                    if not GameRules.mGameInputs[j].mDirty then
                        GameRules.mGameInputs[j].mX = fMouseX
                        GameRules.mGameInputs[j].mY = fMouseY
                        GameRules.mGameInputs[j]:Touch( i )        
                        GameRules.mGameInputs[j].mDirty = true
                        break
                    end
                end
            end
        end
        
        --reset all other inputs
        for i = 1, GameRules.mNumGameInputs do
            if not GameRules.mGameInputs[i].mDirty then
                GameRules.mGameInputs[i]:Reset()
            end
        end
    else
        --get mouse position and normalize to zero-in-the-center space
        local touch = Input.GetMouse()
        
        -- BEGIN TOUCHBOT
        if Constants.TouchBotEnabled then
            if Constants.GameState == Constants.GameStateIntro then
                Constants.DesiredGameState = Constants.GameStateInGame
            elseif Constants.GameState ~= Constants.GameStateLeaderboardEntry then
                timeBetweenBotTouch = 2.0
                
                if not botTouchTimeRemaining then
                    botTouchTimeRemaining = timeBetweenBotTouch
                end
            
                if not botTouchTargetX then
                    botTouchTargetX = 0.0
                    botTouchTargetY = 0.0
                end
            
                if not botTouchX then
                    botTouchX = botTouchTargetX
                    botTouchY = botTouchTargetY
                end
            
                botTouchTimeRemaining = botTouchTimeRemaining - Constants.FrameStep
            
                if botTouchTimeRemaining < 0.0 then
                    botTouchTimeRemaining = timeBetweenBotTouch
                    botTouchTargetX = math.random() * Constants.ValidDrawWidth
                    botTouchTargetY = math.random() * Constants.ValidDrawHeight
                    
                    -- wow touchbot you're really good at this game
                    Score:GoodHit(5)
                end
                
                botTouchX = DFMath.lerp(botTouchX, botTouchTargetX, 0.05)
                botTouchY = DFMath.lerp(botTouchY, botTouchTargetY, 0.05)
                
                touch = {x = botTouchX, y = botTouchY, bIsDown = true}
            end
        end
        -- END TOUCHBOT
        
        if touch.bIsDown and Constants.UseMouseInput then
            --print("debug mouse x, y: " .. fMouseX .. ", " .. fMouseY)    
            
            -- center the mouse coordinates
            local fMouseX, fMouseY
            fMouseX, fMouseY = GameRules:CenterInput( touch.x, touch.y )
           
            GameRules.mGameInputs[1].mX = fMouseX
            GameRules.mGameInputs[1].mY = fMouseY   
            GameRules.mGameInputs[1]:Touch(1)

            --cast across the circle for other touch            
            GameRules.mGameInputs[2].mX = -fMouseX
            GameRules.mGameInputs[2].mY = -fMouseY
            GameRules.mGameInputs[2]:Touch(2)
        else  
            for i = 1, GameRules.mNumGameInputs do
                GameRules.mGameInputs[i]:Reset()
            end    
        end
    end
    
    --keyboard controls
    local keyboard = MOAIInputMgr.device.keyboard
    if keyboard then             
        if Constants.ShowFullProfilingInfo then
            if keyboard:keyDown( "?" ) then
                g_DebugDisplay:ShowLuaDebugMetrics()
            end
        end
            
        --ESC is ASCII 27
        if keyboard:keyDown( 27 ) then
            GameRules:HandleEsc()
        elseif keyboard:keyDown( "q" ) then
            if Constants.mPaused then
                GameRules:Pause( false )
                Constants.DesiredGameState = Constants.GameStatePost
            end
        elseif keyboard:keyDown( 32 ) then
            if Constants.mPaused and not Constants.FullMixMode then
                GameRules:SkipToNextLevel()
                GameRules:EatInput(0.5)
            end            
        end
    end
end

function GameRules:CenterInput( fInputX, fInputY )
    local fCenteredX = fInputX - Constants.ValidDrawHalfWidth
    local fCenteredY = fInputY - Constants.ValidDrawHalfHeight
    
    local fZoomRatio = GameRules.mDesiredCameraZ / Renderer.mGameCameraZDistance
    fCenteredX = fCenteredX * fZoomRatio
    fCenteredY = fCenteredY * fZoomRatio
    
    return fCenteredX, fCenteredY
end

--Android
function GameRules:HandleBack()
    if Constants.GameState == Constants.GameStateInGame then 
        --don't allow pausing when we're in a tutorial
        if not Constants.mPaused and GameRules.mTutorialState == Constants.mTutorialStateNone and Constants.LevelState ~= Constants.LevelStateHealth then
            GameRules:Pause( true )
            GameRules:EatInput(1.0)
            return true
        end    
    elseif Constants.GameState == Constants.GameStateComplete then
        return true
    elseif Constants.GameState == Constants.GameStatePost then 
        UI:EscPressedDuringPost()    
        return true
    elseif Constants.GameState == Constants.GameStateLeaderboardEntry then 
        LeaderboardEntry:HandleEsc() 
        return true
    elseif Constants.GameState == Constants.GameStateLeaderboardShowcase then 
        Leaderboard:HideLeaders( Renderer )
        return true
    elseif Constants.GameState == Constants.GameStateCredits then 
        Credits:EndCredits( Renderer )
        return true
    elseif Constants.GameState == Constants.GameStateHowToPlay then 
        HowToPlay:EndHowToPlay( Renderer )
        return true
    end      

    --save and give the app some time..
    Leaderboard:SaveLeaders()
    DFUtil.sleep(0.5)
    
    --false is identifier for, "let android OS exit app"
    return false
end

function GameRules:HandleEsc()
    if Constants.GameState == Constants.GameStateTitle then
        GameRules.mTitleScreen:RequestClear()
    elseif Constants.GameState == Constants.GameStatePreIntro then
        if UI:IsAbleToSkipPreIntro() then
            Constants.DesiredGameState = Constants.GameStateIntro
        end
    elseif Constants.GameState == Constants.GameStateIntro then
        Constants.DesiredGameState = Constants.GameStateMenu
    elseif Constants.GameState == Constants.GameStateMenu then
        if GameRules.mMenuDepth == Constants.mMenuDepthMenu then
            ChargeManager:SetChargesAtExitPosition()
        else
            GameRules.mOverrideMenuDepth = GameRules.mMenuDepth - 1
        end
    elseif Constants.GameState == Constants.GameStateInGame then 
        --don't allow pausing when we're in a tutorial
        if not Constants.mPaused and GameRules.mTutorialState == Constants.mTutorialStateNone and Constants.LevelState ~= Constants.LevelStateHealth then
            GameRules:Pause( true )
            GameRules:EatInput(1.0)
        else
            GameRules:Pause( false )
        end        
    elseif Constants.GameState == Constants.GameStatePost then 
        UI:EscPressedDuringPost()    
    elseif Constants.GameState == Constants.GameStateLeaderboardEntry then 
        LeaderboardEntry:HandleEsc() 
    elseif Constants.GameState == Constants.GameStateLeaderboardShowcase then 
        Leaderboard:HideLeaders( Renderer )
    elseif Constants.GameState == Constants.GameStateCredits then 
        Credits:EndCredits( Renderer )
    elseif Constants.GameState == Constants.GameStateHowToPlay then 
        HowToPlay:EndHowToPlay( Renderer )
    end   
end

local halfBufferIndex = 1
local quarterBufferIndex = 1
local eighthBufferIndex = 1

function GameRules:PostDebugInput()
    if not Constants.UsePostEffectDebugInput then return end

    local keyboard = MOAIInputMgr.device.keyboard

    if keyboard then        
        local kBUFFERKEYS = "1234567"
        for i=1,#kBUFFERKEYS do
            local bufferChar = kBUFFERKEYS:sub(i, i + 1)
            if keyboard:keyDown(bufferChar) then
                Post:ShowBuffer(i)
            end
        end
        
        if (keyboard:keyDown("7")) then
            halfBufferIndex = (halfBufferIndex + 1) % kNUM_POST_HALFBUFFERS
            Post:ShowHalfBuffer(halfBufferIndex)
        end
            
         if (keyboard:keyDown("8")) then
            quarterBufferIndex = (quarterBufferIndex + 1) % kNUM_POST_QUARTERBUFFERS
            Post:ShowQuarterBuffer(quarterBufferIndex)
        end
    
        if (keyboard:keyDown("9")) then
            quarterBufferIndex = (quarterBufferIndex + 1) % kNUM_POST_EIGHTHBUFFERS
            Post:ShowEighthBuffer(quarterBufferIndex)
        end
    
      
        --------------------------------
        -- DEBUG FILTERS
        --------------------------------    
        if (keyboard:keyDown("q")) then
            PostDebug:BasicComp()
        end

        if (keyboard:keyDown("w")) then
            PostDebug:ShowGame()
        end
    
        if (keyboard:keyDown("e")) then
            PostDebug:ShowUI()
        end
        
        if (keyboard:keyDown("r")) then
            PostDebug:ShowBackground()
        end

        if (keyboard:keyDown("t")) then 
            PostDebug:ShowDirectionalBlur()
        end
                
        if (keyboard:keyDown("y")) then
            PostDebug:ShowEdges()
        end

         if (keyboard:keyDown("u")) then
            PostDebug:ShowEnergyRipple()
        end

        if (keyboard:keyDown("i")) then
            PostDebug:ShowTVNoise()
        end
        
        if (keyboard:keyDown("o")) then
            PostDebug:ShowAccumWithEdges(true)
        end
        
        if (keyboard:keyDown("p")) then
            PostDebug:ShowAccumWithEdges(false)
        end
        
        if (keyboard:keyDown("[")) then
            PostDebug:ShowWideAccumBlur()
        end
        
        if (keyboard:keyDown("a")) then
            PostDebug:ShowAccumFlow()
        end
        
        if (keyboard:keyDown("s")) then
            PostDebug:ShowAccumFlowWithEdges(true)
        end
        
        if (keyboard:keyDown("d")) then
            PostDebug:ShowAccumFlowWithEdges(false)
        end        
        
    --    if (keyboard:keyDown("f")) then
      --      PostDebug:ShowBlurredRemap(0, 0, false)
     --   end        
        
        if (keyboard:keyDown("F")) then
            PostDebug:ShowBlurredRemap(0, 0, true)
        end       
        
        if (keyboard:keyDown("g")) then
            PostDebug:ShowBlurredRemap(.3, 1, false)
        end        
        
        if (keyboard:keyDown("G")) then
            PostDebug:ShowBlurredRemap(.3, 1, true)
        end           
        if (keyboard:keyDown("h")) then
            PostDebug:ShowBlurredRemap(0, 2, false)
        end         
        if (keyboard:keyDown("H")) then
            PostDebug:ShowBlurredRemap(0, 2, true)
        end               
        if (keyboard:keyDown("j")) then
            PostDebug:ShowCrossBlur(0, 2, true)
        end       
        
        --------------------------------
        -- Actual Game State
        --------------------------------    
        
        if (keyboard:keyDown("v")) then
            PostManager:SetStateLeaderboardShowcase()
        end
        if (keyboard:keyDown("b")) then
            PostManager:SetStateGameOver()
        end
        if (keyboard:keyDown("n")) then
            PostManager:SetStateIntro()
        end
        if (keyboard:keyDown(",")) then
            PostManager:SetLevelStateStart()
        end
        if (keyboard:keyDown(".")) then
            PostManager:SetLevelStateBuild()
        end
        if (keyboard:keyDown("/")) then
            PostManager:SetLevelStateBonus()
        end
        
        --------------------------------
        -- RANDOM SHIT
        --------------------------------
        
        if (keyboard:keyDown("z")) then
            Grid:Enable(not Grid.mEnabled)
        end
        if (keyboard:keyDown("x")) then
            ChargeField:Enable(not ChargeField.mEnabled)
        end
        
        -- Cycle level 
        if (keyboard:keyDown("c")) then
            GameRules:IncrementCurrentLevel()
            if Constants.CurrentGameLevel > Constants.NumGameLevels then
                Constants.CurrentGameLevel = 1
            end
            print("Constants.CurrentGameLevel: " .. Constants.CurrentGameLevel)
        end
        
        if (keyboard:keyDown("K")) then
            print ("Killing Game")
            Constants.DesiredGameState = Constants.GameStatePost
        end
        
        if (keyboard:keyDown("V")) then
            print("Forcing Build State")
            Score:ForceBuildState()
        end

    end
            
    -- local testFps = MOAISim.getPerformance()
    -- local memUsage = MOAISim.getMemoryUsage()
    -- print("FPS: " .. tostring(testFps) .. " MEM: " .. memUsage.total)
end

function GameRules:DebugJumpGameState()
    GameRing:InitPreIntro( GameRules.mPreIntroDurations[2] )
    GameRing:MakeRingVisible()
    GameRing:ClearFlavorRingRotations()
    AudioManager:StartAmbientTrack()
    ColorManager:SetLineModeIntro()
    PatternManager:Reset()
    Score:InitForIntro()
    UI:InitPreIntro( GameRules.mPreIntroDurations[1], GameRules.mPreIntroDurations[2] )
    InputVisuals:Reset()
    ChargeManager:Reset()
    PostManager:SetStateIntro()
    
    GameRules:SetZoomPercents(0.0, 0.0, 1.0)
    Constants.DesiredLevelState = Constants.LevelStateStart
end

return GameRules
