local Constants = require('Constants')
local DFMath = require('DFCommon.Math')
local DFGraphics = require("DFCommon.Graphics")
local RadFont = require('RadFont')
local Score = require('Score')
local AudioManager = require('AudioManager')
local HUD = {}

function HUD:NewTutorialBox()
    local TutorialBox = {}
    
    function TutorialBox:Init()
        TutorialBox.mUsed = false
        
        TutorialBox.mText = nil
        TutorialBox.mTextX = 0.0
        TutorialBox.mTextY = 0.0
        TutorialBox.mTextWidth = 0.0
        TutorialBox.mTextHeight = 0.0
        TutorialBox.mTextBuffer = 0.0  
        TutorialBox.mTextBufferDraw = 0.0  
        TutorialBox.mTextBufferDrawMax = 0.0  
        TutorialBox.mTextSize = 0.0   
        TutorialBox.mTextScale = 0.0   
     
        TutorialBox.mDrawRatio = 0.0
        TutorialBox.mDecay = 1.0
        TutorialBox.mSourceX = 0.0
        TutorialBox.mSourceY = 0.0     
    
        TutorialBox.mColor = {}
    end
    
    function TutorialBox:Enable(sText, x, y, fWidth, fHeight, fBuffer, fSize, fScale, tColor)
        TutorialBox.mUsed = true
        TutorialBox.mText = sText
        TutorialBox.mTextX = x
        TutorialBox.mTextY = y
        TutorialBox.mTextWidth = fWidth
        TutorialBox.mTextHeight = fHeight     
        TutorialBox.mTextBuffer = fBuffer   
        TutorialBox.mTextBufferDraw = fBuffer
        TutorialBox.mTextBufferDrawMax = Constants.GameRingRadius * 0.02
        TutorialBox.mTextSize = fSize   
        TutorialBox.mTextScale = fScale
        TutorialBox.mColor = tColor
        TutorialBox.mDrawRatio = 0.0
    end

    function TutorialBox:Disable()
        TutorialBox.mUsed = false
        TutorialBox.mText = nil
    end

    function TutorialBox:ShowTutorialText( fDuration, tTextManager )
        tTextManager:SetTextAttributes( TutorialBox.mTextSize, TutorialBox.mTextScale, TutorialBox.mColor )
        tTextManager:SetFadeAttributes( 0.0, 0.0 )    
        tTextManager:SetZoomAttributes( 0.0, 0.0, 0.0 ) 
        tTextManager:AddLine( TutorialBox.mText, TutorialBox.mTextX, TutorialBox.mTextY, fDuration )     
    end

    function TutorialBox:Update()
        TutorialBox.mTextBufferDraw = TutorialBox.mTextBuffer + (Constants.Heartbeat * TutorialBox.mTextBufferDrawMax)
    end
    
    function TutorialBox:Draw()
        local fBoxX = TutorialBox.mTextX - TutorialBox.mTextBufferDraw
        local fBoxY = TutorialBox.mTextY + TutorialBox.mTextBufferDraw
        local fWidth = TutorialBox.mTextWidth + (TutorialBox.mTextBufferDraw * 2.0)
        local fHeight = TutorialBox.mTextHeight + (TutorialBox.mTextBufferDraw * 2.0)
            
        local fTextBoxX = fBoxX + (TutorialBox.mTextWidth * 0.5)
        local fTextBoxY = fBoxY - (TutorialBox.mTextHeight * 0.5)
        
        --draw connection lines
        local fXDiff = fTextBoxX - TutorialBox.mSourceX
        local fYDiff = fTextBoxY - TutorialBox.mSourceY
        
        local fLineEndX = 0.0
        local fLineEndY = 0.0
        if fXDiff > 0.0 then                                            --on right side
            if math.abs(fXDiff) > math.abs(fYDiff) then                 --attach to side
                fLineEndX = fBoxX
                fLineEndY = fBoxY - (fHeight*0.5)
            else
                if fYDiff < 0.0 then                                    --attach to bottom
                    fLineEndX = fBoxX + (fWidth*0.5)
                    fLineEndY = fBoxY                
                else                                                    --attach to top
                    fLineEndX = fBoxX + (fWidth*0.5)
                    fLineEndY = fBoxY - fHeight                                   
                end
            end
        else                                                            --on left side
            if math.abs(fXDiff) > math.abs(fYDiff) then                 --attach to side
                fLineEndX = fBoxX + fWidth
                fLineEndY = fBoxY - (fHeight*0.5)                
            else
                if fYDiff < 0.0 then                                    --attach to bottom
                    fLineEndX = fBoxX + (fWidth*0.5)
                    fLineEndY = fBoxY                
                else                                                    --attach to top
                    fLineEndX = fBoxX + (fWidth*0.5)
                    fLineEndY = fBoxY - fHeight                                   
                end            
            end       
        end

        fLineEndX = DFMath.lerp( TutorialBox.mSourceX, fLineEndX, TutorialBox.mDrawRatio )
        fLineEndY = DFMath.lerp( TutorialBox.mSourceY, fLineEndY, TutorialBox.mDrawRatio )
        MOAIDraw.drawLine( TutorialBox.mSourceX, TutorialBox.mSourceY, fLineEndX, fLineEndY )
        
        --draw text bounding box
        if TutorialBox.mDrawRatio >= 1.0 then
            MOAIDraw.drawRect( fBoxX, fBoxY, fBoxX + fWidth, fBoxY - fHeight )
        end    
    end
    
    TutorialBox:Init()
    return TutorialBox
end

function HUD:Init( tRenderer )
    HUD:DetermineSize( tRenderer )
    
    local i = 1 
    local j = 1
    
    local sEscToExit = {"q","u","i","t"}
    HUD.mEscCount = 4
    HUD.mEscProps = {}
    for i = 1, HUD.mEscCount do
        HUD.mEscProps[i] = {}
        for j = 1, 2 do
            HUD.mEscProps[i][j] = MOAIProp.new()
            HUD.mEscProps[i][j]:setDeck( RadFont.mFontDeck )        
            HUD.mEscProps[i][j]:setScl( HUD.mImageScale, -HUD.mImageScale )
            HUD.mEscProps[i][j]:setIndex( RadFont.mFontDeck.names[ sEscToExit[i] ] )
            HUD.mEscProps[i][j]:setVisible(false)
            HUD.mEscProps[i][j]:setBlendMode( MOAIProp2D.GL_SRC_ALPHA, MOAIProp2D.GL_ONE_MINUS_SRC_ALPHA )
            
            tRenderer.mUILayer:insertProp( HUD.mEscProps[i][j] )        
        end
    end  

    local sNextTrack = {"n","e","x","t"}
    HUD.mNextCount = 4
    HUD.mNextProps = {}
    for i = 1, HUD.mNextCount do
        HUD.mNextProps[i] = {}
        for j = 1, 2 do    
            HUD.mNextProps[i][j] = MOAIProp.new()
            HUD.mNextProps[i][j]:setDeck( RadFont.mFontDeck )        
            HUD.mNextProps[i][j]:setScl( HUD.mImageScale, -HUD.mImageScale )
            HUD.mNextProps[i][j]:setIndex( RadFont.mFontDeck.names[ sNextTrack[i] ] )
            HUD.mNextProps[i][j]:setVisible(false)
            HUD.mNextProps[i][j]:setBlendMode( MOAIProp2D.GL_SRC_ALPHA, MOAIProp2D.GL_ONE_MINUS_SRC_ALPHA )
            
            tRenderer.mUILayer:insertProp( HUD.mNextProps[i][j] )        
        end
    end  
    
    HUD.mLetterboxDeck = MOAIGfxQuad2D.new()
    local whiteTexture = DFGraphics.loadTexture( "White", true )
    HUD.mLetterboxDeck:setTexture( whiteTexture )
    HUD.mLetterboxDeck:setRect( -1, -1, 1, 1 )
        
    HUD.mLetterboxCount = 2
    HUD.mLetterboxProps = {}
    for i = 1, HUD.mLetterboxCount do        
        HUD.mLetterboxProps[i] = MOAIProp2D.new()
        HUD.mLetterboxProps[i]:setDeck( HUD.mLetterboxDeck )
        HUD.mLetterboxProps[i]:setColor( Constants.LetterboxColor[1], Constants.LetterboxColor[2], Constants.LetterboxColor[3], Constants.LetterboxColor[4] )    
        HUD.mLetterboxProps[i]:setVisible( false )
        HUD.mLetterboxProps[i]:setLoc( 0.0, 0.0, 1.0 )
        HUD.mLetterboxProps[i]:setScl( 1.0, 1.0 )
        
        tRenderer.mLetterboxLayer:insertProp( HUD.mLetterboxProps[i] )
    end
        
    HUD.mStarPropCount = 3
    HUD.mStarsActivated = {}
    HUD.mStarProps = {}
    for i = 1, HUD.mStarPropCount do
        HUD.mStarProps[i] = ShapeProp.new()
        HUD.mStarProps[i]:setDeck( tRenderer.mShapeDeck )      
        HUD.mStarProps[i]:SetNumSides( 32 )  
        HUD.mStarProps[i]:SetRadii( HUD.mStarPropWidth * 0.5, HUD.mStarPropWidth )
        HUD.mStarProps[i]:SetDrawMode( ShapeProp.DrawModeStar )
        HUD.mStarProps[i]:SetColor( Constants.Color1[1], Constants.Color1[2], Constants.Color1[3], 1.0 )
        HUD.mStarProps[i]:setVisible( false )
        
        tRenderer.mUILayer:insertProp( HUD.mStarProps[i] )        
        
        HUD.mStarsActivated[i] = false
    end
    HUD:PositionStars()
        
    HUD.mEscapeToExitTimer = 0.0
    HUD.mEscapeToExitDuration = 0.5
    HUD.mEscapeToExitAlphaAmount = 0.0

    HUD.mNextTrackTimer = 0.0
    HUD.mNextTrackDuration = 0.5
    HUD.mNextTrackAlphaAmount = 0.0
    
    HUD.mTextAlphaTimer = math.pi + DFMath.halfPi
    HUD.mTextAlphaScale = 4.0
    
    HUD.mHealthUpCount = 0
    HUD.mHealthUpInterval = 0.25
    HUD.mHealthUpTimer = 0.0
    
    HUD.mTutorialColor = {}
    
    HUD.mNumTutorialBoxes = 4
    HUD.mTutorialBoxes = {}
    for i = 1, HUD.mNumTutorialBoxes do
        HUD.mTutorialBoxes[i] = HUD:NewTutorialBox()
    end
end
    
function HUD:DetermineSize( tRenderer )
    HUD.mPropZ = tRenderer.mGameCameraZDistance * 0.039
    HUD.mButtonZ = tRenderer.mGameCameraZDistance * 0.037
    
    local fDesiredLetterImageSize = Constants.ValidDrawHeight / 15.0
    HUD.mImageScale = fDesiredLetterImageSize / 128.0
    HUD.mLetterWidth = fDesiredLetterImageSize
    
    HUD.mButtonWidth = 32.0
    
    HUD.mLargeLetterSize = Constants.ValidDrawHeight / 10.0
    HUD.mLargeLetterScale = HUD.mLargeLetterSize / 128.0 
    
    HUD.mReviewLetterSize = Constants.ValidDrawHeight / 15.0
    HUD.mReviewLetterScale = HUD.mReviewLetterSize / 128.0    
    
    HUD.mStarPropWidth = HUD.mReviewLetterSize * 1.0
end

function HUD:UpdateSize( tRenderer )
    HUD:DetermineSize( tRenderer )
    
    local i = 1
    local j = 1
    for i = 1, HUD.mEscCount do
        for j = 1, 2 do
            HUD.mEscProps[i][j]:setScl( HUD.mImageScale, -HUD.mImageScale )
        end
    end         
    
    for i = 1, HUD.mNextCount do
        for j = 1, 2 do
            HUD.mNextProps[i][j]:setScl( HUD.mImageScale, -HUD.mImageScale )
        end
    end           
    
    for i = 1, HUD.mStarPropCount do
        HUD.mStarProps[i]:SetRadii( HUD.mStarPropWidth * 0.5, HUD.mStarPropWidth )
    end       
    HUD:PositionStars()
end

function HUD:Update( tTextManager, tGameInputs )
    local i = 1
    local j = 1 
    
    --if we're paused, ramp up the "q to quit" text
    if Constants.mPaused then
        HUD.mTextAlphaTimer = HUD.mTextAlphaTimer + (Constants.FrameStep * HUD.mTextAlphaScale)
        local fBaseAlpha = ((math.sin(HUD.mTextAlphaTimer) * 0.2) + 0.8)
        
        if HUD.mEscapeToExitTimer > 0.0 then
            HUD.mEscapeToExitTimer = HUD.mEscapeToExitTimer - Constants.FrameStep
        else
            HUD.mEscapeToExitAlphaAmount = math.min( HUD.mEscapeToExitAlphaAmount + Constants.FrameStep, 1.0 )
        
            local fAlpha = fBaseAlpha * HUD.mEscapeToExitAlphaAmount
            for i = 1, HUD.mEscCount do
                HUD.mEscProps[i][1]:setVisible( true )
                HUD.mEscProps[i][1]:setColor( Constants.SpecialColor[1], Constants.SpecialColor[2], Constants.SpecialColor[3], fAlpha )
            end          

            for i = 1, HUD.mEscCount do
                HUD.mEscProps[i][2]:setVisible( true )
                HUD.mEscProps[i][2]:setColor( 0.0, 0.0, 0.0, HUD.mEscapeToExitAlphaAmount*0.25 )
            end              
            
            HUD:PositionEsc( tGameInputs )
        end

        --if we've beaten this level before, allow skipping
        local bCanSkipLevel = Constants.MaxCompletedLevel >= Constants.CurrentGameLevel and Constants.CurrentGameLevel ~= Constants.NumGameLevels 
        if (bCanSkipLevel or Constants.AllowSkipAllLevels) and not Constants.FullMixMode then     
            if HUD.mNextTrackTimer > 0.0 then
                HUD.mNextTrackTimer = HUD.mNextTrackTimer - Constants.FrameStep
            else
                HUD.mNextTrackAlphaAmount = math.min(HUD.mNextTrackAlphaAmount + Constants.FrameStep, 1.0)
                
                local fAlpha = fBaseAlpha * HUD.mNextTrackAlphaAmount
                for i = 1, HUD.mNextCount do
                    HUD.mNextProps[i][1]:setVisible( true )
                    HUD.mNextProps[i][1]:setColor( Constants.SpecialColor[1], Constants.SpecialColor[2], Constants.SpecialColor[3], fAlpha )
                end 
                
                for i = 1, HUD.mNextCount do
                    HUD.mNextProps[i][2]:setVisible( true )
                    HUD.mNextProps[i][2]:setColor( 0.0, 0.0, 0.0, HUD.mNextTrackAlphaAmount*0.25 )
                end        
    
                HUD:PositionNext( tGameInputs )           
            end
        end
    else
        HUD.mEscapeToExitTimer = HUD.mEscapeToExitDuration
        HUD.mEscapeToExitAlphaAmount = 0.0   

        HUD.mNextTrackTimer = HUD.mNextTrackDuration 
        HUD.mNextTrackAlphaAmount = 0.0
        
        HUD.mTextAlphaTimer = math.pi + DFMath.halfPi 
    
        for i = 1, HUD.mEscCount do
            for j = 1, 2 do        
                HUD.mEscProps[i][j]:setVisible( false )
            end
        end          
        
        for i = 1, HUD.mNextCount do
            for j = 1, 2 do
                HUD.mNextProps[i][j]:setVisible( false )
            end
        end             
    end
    
    for i = 1, HUD.mNumTutorialBoxes do
        HUD.mTutorialBoxes[i]:Update()
    end
    
    --if we have health up messages queued, trigger them
    if HUD.mHealthUpCount > 0 then
        HUD.mHealthUpTimer = HUD.mHealthUpTimer - Constants.FrameStep
        if HUD.mHealthUpTimer < 0.0 then           
            local fYOffset = HUD.mHealthUpCount * (HUD.mLargeLetterSize * 0.25)
            local tColor = Constants.SpecialColor
            if HUD.mHealthUpCount % 2 == 0 then
                tColor = Constants.Color1
            end
            
            HUD:ShowHealthUpText( tTextManager, fYOffset, tColor )
            HUD:ActivateStar()
            
            HUD.mHealthUpCount = HUD.mHealthUpCount - 1
            HUD.mHealthUpTimer = HUD.mHealthUpInterval            
        end
    end
end

function HUD:UpdateTutorialDrawRatio( iIndex, fRatio, fSourceX, fSourceY )
    if iIndex > 0 then
        HUD.mTutorialBoxes[iIndex].mDrawRatio = fRatio
        HUD.mTutorialBoxes[iIndex].mSourceX = fSourceX
        HUD.mTutorialBoxes[iIndex].mSourceY = fSourceY

        local i = 1
        local bAnythingDrawing = false
        for i = 1, HUD.mNumTutorialBoxes do
            if HUD.mTutorialBoxes[i].mDrawRatio > 0.0 then
                bAnythingDrawing = true
                break
            end
        end
        
        if not bAnythingDrawing then
            local i = 1
            for i = 1, HUD.mLetterboxCount do        
                HUD.mLetterboxProps[i]:setVisible( false )
            end    
            HUD.mLetterboxProps[1]:setLoc( 0.0, -Constants.ValidDrawHeight - Constants.ValidDrawHalfHeight, 1.0 )
            HUD.mLetterboxProps[2]:setLoc( 0.0, Constants.ValidDrawHeight + Constants.ValidDrawHalfHeight, 1.0 )
        end
    end
end

function HUD:Draw()
    local i = 1
    local fHighestDrawRatio = 0.0
    
    MOAIGfxDevice.setPenWidth( 2.0 )
    MOAIGfxDevice.setPenColor( Constants.Color1[1], Constants.Color1[2], Constants.Color1[3] )    
    
    for i = 1, HUD.mNumTutorialBoxes do
        if HUD.mTutorialBoxes[i].mUsed then
            HUD.mTutorialBoxes[i]:Draw() 
            fHighestDrawRatio = HUD.mTutorialBoxes[i].mDrawRatio
        end
    end
               
    --position letterbox props if anything is drawing
    if fHighestDrawRatio > 0.0 then     
        local fEntranceOffset = Constants.ValidDrawHalfHeight * 0.2 * (1.0 - fHighestDrawRatio)
        local fBaseOffset = (Constants.ValidDrawHeight * 0.925)
        
        HUD.mLetterboxProps[1]:setVisible( true )
        HUD.mLetterboxProps[1]:setLoc( 0.0, -fBaseOffset - fEntranceOffset, 1.0 )
        HUD.mLetterboxProps[1]:setScl( Constants.ValidDrawHalfWidth, Constants.ValidDrawHalfHeight )        
        
        HUD.mLetterboxProps[2]:setVisible( true )
        HUD.mLetterboxProps[2]:setLoc( 0.0, fBaseOffset + fEntranceOffset, 1.0 )        
        HUD.mLetterboxProps[2]:setScl( Constants.ValidDrawHalfWidth, Constants.ValidDrawHalfHeight )        
    end
end

function HUD:InitTutorialText( sText, iNumChars, tTextManager, iPosID, tColor )
    --get an open tutorial box
    local i = 1
    local iIndex = -1
    for i = 1, HUD.mNumTutorialBoxes do
        if not HUD.mTutorialBoxes[i].mUsed then
            iIndex = i
            break 
        end
    end
    
    if iIndex ~= -1 then
        local fLetterSize = Constants.ValidDrawHeight / 20.0
        if iPosID == Constants.mHUDPositionPlay then
            fLetterSize = Constants.ValidDrawHeight / 17.5
        end
        
        local fLetterScale = fLetterSize / 128.0
           
        local fBuffer = fLetterSize * 0.15
        local fWidth = fLetterSize * iNumChars
        local fHeight = fLetterSize
        
        local x, y
        if iPosID == Constants.mHUDPositionBit then
            x = Constants.GameRingRadius * 1.2
            y = Constants.GameRingRadius * -0.6           
        elseif iPosID == Constants.mHUDPositionPlay then
            x = Constants.GameRingRadius * 1.2
            y = Constants.GameRingRadius * -1.0         
        elseif iPosID == Constants.mHUDPositionFullMix then        
            x = (Constants.GameRingRadius * -1.2) - fWidth
            y = Constants.GameRingRadius * 0.6          
        elseif iPosID == Constants.mHUDPositionLeaders then        
            x = Constants.GameRingRadius * 1.2
            y = Constants.GameRingRadius * -0.3             
        elseif iPosID == Constants.mHUDPositionHowToPlay then        
            x = (Constants.GameRingRadius * -0.8) - fWidth
            y = Constants.GameRingRadius * 0.9              
        elseif iPosID == Constants.mHUDPositionCredits then        
            x = Constants.GameRingRadius * 1.2
            y = Constants.GameRingRadius * -0.8
        elseif iPosID == Constants.mHUDPositionHold1 then        
            x = Constants.GameRingRadius * 1.4
            y = Constants.GameRingRadius * -0.6                
        elseif iPosID == Constants.mHUDPositionHold2 then        
            x = (Constants.GameRingRadius * -1.4) - fWidth
            y = Constants.GameRingRadius * -0.6            
        elseif iPosID == Constants.mHUDPositionOptions then        
            x = (Constants.GameRingRadius * -1.2) - fWidth
            y = Constants.GameRingRadius * -0.6            
        elseif iPosID == Constants.mHUDPositionVQ then        
            x = Constants.GameRingRadius * 0.4
            y = Constants.GameRingRadius * -1.0            
        elseif iPosID == Constants.mHUDPositionBack then        
            x = (Constants.GameRingRadius * -1.3) - fWidth
            y = Constants.GameRingRadius * -0.2            
        elseif iPosID == Constants.mHUDPositionLow then        
            x = (Constants.GameRingRadius * -1.0) - fWidth
            y = Constants.GameRingRadius * -0.8            
        elseif iPosID == Constants.mHUDPositionHigh then        
            x = (Constants.GameRingRadius * -1.1) - fWidth
            y = Constants.GameRingRadius * 0.8    
        elseif iPosID == Constants.mHUDPositionUltra then
            x = Constants.GameRingRadius * 0.8
            y = Constants.GameRingRadius * 1.0                
        elseif iPosID == Constants.mHUDPositionFullScreen then
            x = (Constants.GameRingRadius * -0.7) - fWidth
            y = Constants.GameRingRadius * -0.8               
        end
                
        HUD.mTutorialBoxes[iIndex]:Enable(sText, x, y, fWidth, fHeight, fBuffer, fLetterSize, fLetterScale, tColor)
    end
    
    return iIndex
end

function HUD:ShowTutorialText( iIndex, fDuration, tTextManager )
    if iIndex > 0 then
        HUD.mTutorialBoxes[iIndex]:ShowTutorialText( fDuration, tTextManager )
    end
end

function HUD:ClearTutorialText()
    local i = 1
    for i = 1, HUD.mLetterboxCount do        
        HUD.mLetterboxProps[i]:setVisible( false )
    end    
    HUD.mLetterboxProps[1]:setLoc( 0.0, -Constants.ValidDrawHeight - Constants.ValidDrawHalfHeight, 1.0 )
    HUD.mLetterboxProps[2]:setLoc( 0.0, Constants.ValidDrawHeight + Constants.ValidDrawHalfHeight, 1.0 )    
    
    for i = 1, HUD.mNumTutorialBoxes do
        if HUD.mTutorialBoxes[i].mUsed then
            HUD.mTutorialBoxes[i]:Disable() 
        end
    end    
end

function HUD:PositionEsc( tGameInputs )
    local fDepressedOffset = HUD.mLetterWidth * 0.15
    if tGameInputs[1]:IsTouched() and HUD:IsInEscBounds( tGameInputs[1].mX, tGameInputs[1].mY ) then
        fDepressedOffset = HUD.mLetterWidth * 0.075
    end
    
    local fEscTextPositionX = -Constants.ValidDrawHalfWidth + (HUD.mLetterWidth * 2.0)
    local fEscTextPositionY = Constants.ValidDrawHalfHeight - (HUD.mLetterWidth * 1.0)
        
    local i = 1
    for i = 1, HUD.mEscCount do
        HUD.mEscProps[i][1]:setLoc( fEscTextPositionX + fDepressedOffset, fEscTextPositionY - fDepressedOffset, HUD.mPropZ ) 
        HUD.mEscProps[i][2]:setLoc( fEscTextPositionX - fDepressedOffset, fEscTextPositionY + fDepressedOffset, HUD.mButtonZ ) 
        fEscTextPositionX = fEscTextPositionX + HUD.mLetterWidth
    end 
end

function HUD:PositionNext( tGameInputs )
    local fDepressedOffset = HUD.mLetterWidth * 0.15
    if tGameInputs[1]:IsTouched() and HUD:IsInNextBounds( tGameInputs[1].mX, tGameInputs[1].mY ) then
        fDepressedOffset = HUD.mLetterWidth * 0.075
    end
    
    local fNextTextPositionX = Constants.ValidDrawHalfWidth - (HUD.mLetterWidth * 6.0)
    local fNextTextPositionY = Constants.ValidDrawHalfHeight - (HUD.mLetterWidth * 1.0)
    
    local i = 1
    for i = 1, HUD.mNextCount do
        HUD.mNextProps[i][1]:setLoc( fNextTextPositionX - fDepressedOffset, fNextTextPositionY - fDepressedOffset, HUD.mPropZ ) 
        HUD.mNextProps[i][2]:setLoc( fNextTextPositionX + fDepressedOffset, fNextTextPositionY + fDepressedOffset, HUD.mButtonZ ) 
        fNextTextPositionX = fNextTextPositionX + HUD.mLetterWidth
    end  
end

function HUD:IsInEscBounds( x, y )
    local fMinX = -Constants.ValidDrawHalfWidth
    local fMaxX = fMinX + ((HUD.mEscCount + 4.0) * HUD.mLetterWidth)
    local fMaxY = Constants.ValidDrawHalfHeight
    local fMinY = fMaxY - (HUD.mLetterWidth * 3.0)
    
    return x > fMinX and x < fMaxX and y > fMinY and y < fMaxY
end

function HUD:IsInNextBounds( x, y )
    local fMinX = Constants.ValidDrawHalfWidth - (HUD.mLetterWidth * 8.0)
    local fMaxX = Constants.ValidDrawHalfWidth
    local fMaxY = Constants.ValidDrawHalfHeight
    local fMinY = fMaxY - (HUD.mLetterWidth * 3.0)
    
    return x > fMinX and x < fMaxX and y > fMinY and y < fMaxY    
end

function HUD:PositionStars()
    local fLineX = HUD.mStarPropWidth * -2.0    
    local fLineY = HUD.mReviewLetterSize * 0.75        
    
    local i = 1
    for i = 1, HUD.mStarPropCount do
        HUD.mStarProps[i]:setLoc( fLineX, fLineY, HUD.mPropZ )        
        fLineX = fLineX + HUD.mStarPropWidth + HUD.mStarPropWidth
    end    
end

function HUD:ShowPatternPerfect( tTextManager, fDuration )
    local iDivisibleBy10 = Score.mMultiplier % 10
    if Score.mMultiplier == 5 then
        tTextManager:SetTextAttributes( HUD.mLargeLetterSize, HUD.mLargeLetterScale, Constants.Color2 )
        tTextManager:SetFadeAttributes( 0.0, 0.25 )
        tTextManager:SetZoomAttributes( 250.0, 10.0, 0.25 )   
       
        local fLineX = HUD.mLargeLetterSize * -0.5
        local fLineY = HUD.mLargeLetterSize * -0.5    
        tTextManager:AddLine( "5", fLineX, fLineY, fDuration )  
        
        tTextManager:SetTextAttributes( HUD.mReviewLetterSize, HUD.mReviewLetterScale, Constants.Color1 )
        fLineX = HUD.mReviewLetterSize * -4.5
        fLineY = fLineY + HUD.mReviewLetterSize
        tTextManager:AddLine( "in.a.row!", fLineX, fLineY, fDuration )        
        
        --this is kind of hacky, but assures that later text won't zoom in
        tTextManager:SetZoomAttributes( 0.0, 0.0, 0.0 ) 
    elseif Score.mMultiplier == 99 then
        local fScaleSize = ( HUD.mLargeLetterSize * 2.0 )
        local fScaleScale = ( HUD.mLargeLetterScale * 2.0 )        
        tTextManager:SetTextAttributes( fScaleSize, fScaleScale, Constants.SpecialColor )
        tTextManager:SetFadeAttributes( 0.0, 0.25 )
        tTextManager:SetZoomAttributes( 250.0, 10.0, 0.25 )   
       
        local fLineX = fScaleSize * -1.5
        local fLineY = 0.0    
        tTextManager:AddLine( "pro", fLineX, fLineY, fDuration )  
                
        --this is kind of hacky, but assures that later text won't zoom in
        tTextManager:SetZoomAttributes( 0.0, 0.0, 0.0 )    
    elseif iDivisibleBy10 == 0 then
        local fTenCount = math.min(Score.mMultiplier / 10.0, 5)
        local fScaleSize = ( HUD.mLargeLetterSize * (1.0 + (fTenCount * 0.2)) )
        local fScaleScale = ( HUD.mLargeLetterScale * (1.0 + (fTenCount * 0.2)) )
        tTextManager:SetTextAttributes( fScaleSize, fScaleScale, Constants.SpecialColor )
        tTextManager:SetFadeAttributes( 0.0, 0.25 )
        tTextManager:SetZoomAttributes( 250.0, 10.0, 0.25 )   
       
        local fLineX = -fScaleSize 
        local fLineY = HUD.mLargeLetterSize * -0.5    
        tTextManager:AddLine( tostring(Score.mMultiplier), fLineX, fLineY, fDuration )  
        
        tTextManager:SetTextAttributes( HUD.mReviewLetterSize, HUD.mReviewLetterScale, Constants.Color1 )
        fLineX = HUD.mReviewLetterSize * -4.5
        fLineY = fLineY + HUD.mReviewLetterSize
        tTextManager:AddLine( "in.a.row!", fLineX, fLineY, fDuration )        
        
        --this is kind of hacky, but assures that later text won't zoom in
        tTextManager:SetZoomAttributes( 0.0, 0.0, 0.0 )        
    else
        Score:ShowGreatFanfare( fDuration )
    end
end

function HUD:ShowLostMultiplier( tTextManager, fDuration )
    tTextManager:SetTextAttributes( HUD.mReviewLetterSize, HUD.mReviewLetterScale, Constants.GeneralRed )
    tTextManager:SetFadeAttributes( 0.0, fDuration ) 

    local fLineX = HUD.mReviewLetterSize * -5.0
    local fLineY = HUD.mReviewLetterSize * -0.5    
    tTextManager:AddLine( "multiplier", fLineX, fLineY, fDuration*2.0 )  

    fLineX = HUD.mReviewLetterSize * -2.0
    fLineY = fLineY + HUD.mReviewLetterSize
    tTextManager:AddLine( "lost", fLineX, fLineY, fDuration*2.0 )      
end

function HUD:ShowBuildModeText( tTextManager )
    --don't show state text in full mix mode
    if not Constants.FullMixMode then   
        tTextManager:ClearAll()
        Score:HideFanfare()
        
        tTextManager:SetTextAttributes( HUD.mLargeLetterSize, HUD.mLargeLetterScale, Constants.Color1 )
        tTextManager:SetFadeAttributes( 0.0, 1.0 )
        tTextManager:SetZoomAttributes( 500.0, 10.0, 0.25 )
        
        local fLineX = HUD.mLargeLetterSize * -2.5
        local fLineY = HUD.mLargeLetterSize * -0.5    
        tTextManager:AddLine( "verse", fLineX, fLineY, 1.5 )  
        
        fLineX = HUD.mLargeLetterSize * -4.5
        fLineY = fLineY + HUD.mLargeLetterSize
        tTextManager:AddLine( "complete!", fLineX, fLineY, 1.5 )        
        
        --this is kind of hacky, but assures that later text won't zoom in
        tTextManager:SetZoomAttributes( 0.0, 0.0, 0.0 )        
        
        AudioManager:PlaySound( 'Radius/SFX/VerseComplete' )  
    end
end

function HUD:ShowTrackCompleteText( tTextManager, fDuration )
    tTextManager:ClearAll()
    Score:HideFanfare()
    
    tTextManager:SetTextAttributes( HUD.mLargeLetterSize, HUD.mLargeLetterScale, Constants.Color1 )
    tTextManager:SetFadeAttributes( 0.0, 1.0 )
    tTextManager:SetZoomAttributes( 500.0, 10.0, 0.25 )
    
    local fLineX = HUD.mLargeLetterSize * -2.5
    local fLineY = HUD.mLargeLetterSize * -1.0    
    tTextManager:AddLine( "track", fLineX, fLineY, fDuration ) 
   
    fLineX = HUD.mLargeLetterSize * -4.5
    fLineY = fLineY + HUD.mLargeLetterSize
    tTextManager:AddLine( "complete!", fLineX, fLineY, fDuration )     
    
    tTextManager:SetZoomAttributes( 0.0, 0.0, 0.0 )
    tTextManager:SetFadeAttributes( 0.0, 0.0 )
    
    AudioManager:PlaySound( 'Radius/SFX/TrackComplete' )
end

function HUD:ShowPatternCountText( tTextManager, fDuration )
    tTextManager:SetTextAttributes( HUD.mReviewLetterSize, HUD.mReviewLetterScale, Constants.Color1 )
    
    local fLineX = HUD.mReviewLetterSize * -6.0
    local fLineY = HUD.mReviewLetterSize * -3.0    
    tTextManager:AddLine( "patterns_", fLineX, fLineY, fDuration )  
    AudioManager:PlaySound( 'Radius/SFX/ShowStats' )    
end

function HUD:ShowPatternCount( tTextManager, iCount, fDuration )
    tTextManager:SetTextAttributes( HUD.mReviewLetterSize, HUD.mReviewLetterScale, Constants.Color2 )
    
    local iRealCount = math.floor( math.min(iCount, 99) )
    local sCountString = tostring( iRealCount )
                
    local fLineX = HUD.mReviewLetterSize * 4.0
    local fLineY = HUD.mReviewLetterSize * -3.0    
    tTextManager:AddLine( sCountString, fLineX, fLineY, fDuration )  
    AudioManager:PlaySound( 'Radius/SFX/ShowStats' )    
end

function HUD:ShowPerfectCountText( tTextManager, fDuration )
    tTextManager:SetTextAttributes( HUD.mReviewLetterSize, HUD.mReviewLetterScale, Constants.Color1 )
    
    local fLineX = HUD.mReviewLetterSize * -5.0
    local fLineY = HUD.mReviewLetterSize * -1.75    
    tTextManager:AddLine( "perfect_", fLineX, fLineY, fDuration )
    AudioManager:PlaySound( 'Radius/SFX/ShowStats' )
end

function HUD:ShowPerfectCount( tTextManager, iCount, fDuration )
    tTextManager:SetTextAttributes( HUD.mReviewLetterSize, HUD.mReviewLetterScale, Constants.Color2 )
    
    local iRealCount = math.floor( math.min(iCount, 99) )
    local sCountString = tostring( iRealCount )
                
    local fLineX = HUD.mReviewLetterSize * 4.0
    local fLineY = HUD.mReviewLetterSize * -1.75    
    tTextManager:AddLine( sCountString, fLineX, fLineY, fDuration )  
    AudioManager:PlaySound( 'Radius/SFX/ShowStats' )    
end

function HUD:ShowPercent( tTextManager, iPercent, fDuration )    
    local iRealPercent = math.floor( iPercent * 100.0 )
    local sPercentString = tostring( iRealPercent ) .. "("
                
    local fLineX = HUD.mReviewLetterSize * 4.0
    if iRealPercent == 100 then
        tTextManager:SetTextAttributes( HUD.mReviewLetterSize, HUD.mReviewLetterScale, Constants.SpecialColor )
        fLineX = HUD.mReviewLetterSize * 3.0
        AudioManager:PlaySound( 'Radius/SFX/Score_Pro' )
    elseif iRealPercent < 40 then
        tTextManager:SetTextAttributes( HUD.mReviewLetterSize, HUD.mReviewLetterScale, Constants.GeneralRed )
        AudioManager:PlaySound( 'Radius/SFX/ShowStats' )
    else
        tTextManager:SetTextAttributes( HUD.mReviewLetterSize, HUD.mReviewLetterScale, Constants.Color2 )
        AudioManager:PlaySound( 'Radius/SFX/Score_Leader' )
    end
    
    local fLineY = HUD.mReviewLetterSize * -0.25    
    tTextManager:AddLine( sPercentString, fLineX, fLineY, fDuration )  
end

function HUD:ShowStarsBad()
    local i = 1
    for i = 1, HUD.mStarPropCount do
        HUD.mStarProps[i]:SetColor( Constants.GeneralRed[1], Constants.GeneralRed[2], Constants.GeneralRed[3], 1.0 )        
        HUD.mStarProps[i]:setVisible( true )        
        HUD.mStarsActivated[i] = false
    end    
end
 
function HUD:HideStars()
    local i = 1
    for i = 1, HUD.mStarPropCount do
        HUD.mStarProps[i]:setVisible( false )    
        HUD.mStarsActivated[i] = false    
    end    
end

function HUD:ActivateStar()
    local i = 1
    for i = 1, HUD.mStarPropCount do
        if HUD.mStarsActivated[i] == false then
            HUD.mStarsActivated[i] = true
            HUD.mStarProps[i]:SetColor( Constants.SpecialColor[1], Constants.SpecialColor[2], Constants.SpecialColor[3], 1.0 )
            return
        end
    end   
end
   
function HUD:ShowNoBonusText( tTextManager, fDuration )
    local fLineY = HUD.mLargeLetterSize * 2.5    
    local fLineX = HUD.mLargeLetterSize * -4.0
    tTextManager:SetTextAttributes( HUD.mLargeLetterSize, HUD.mLargeLetterScale, Constants.GeneralRed )
    tTextManager:AddLine( "no.bonus", fLineX, fLineY, fDuration )    
    AudioManager:PlaySound( 'Radius/SFX/ShowStats' )      
end

function HUD:QueueHealthUpText( iCount )
    HUD.mHealthUpCount = iCount
    HUD.mHealthUpTimer = 0.0
end
        
function HUD:ShowHealthUpText( tTextManager, fYOffset, tColor )    
    local fLineX = HUD.mLargeLetterSize * -5.0
    local fLineY = HUD.mLargeLetterSize * 2.0    
    tTextManager:SetFadeAttributes( 0.0, 0.75 )
    tTextManager:SetZoomAttributes( 1000.0, 10.0, 0.25 )        
    tTextManager:SetTextAttributes( HUD.mLargeLetterSize, HUD.mLargeLetterScale, tColor )
    tTextManager:AddLine( "health.up!", fLineX, fLineY + fYOffset, 1.5 )  
    AudioManager:PlaySound( 'Radius/SFX/Health' )
end

function HUD:ShowBonusText( tTextManager )
    --don't show state text in full mix mode
    if not Constants.FullMixMode then    
        Score:HideFanfare()
        
        tTextManager:SetTextAttributes( HUD.mLargeLetterSize, HUD.mLargeLetterScale, Constants.Color1 )
        tTextManager:SetFadeAttributes( 0.0, 1.0 )
        tTextManager:SetZoomAttributes( 500.0, 10.0, 0.25 )
        
        --show bonus round text for first 6 levels
        if Constants.CurrentGameLevel < 7 then
            local fLineX = HUD.mLargeLetterSize * -3.0
            local fLineY = HUD.mLargeLetterSize * -0.5    
            tTextManager:AddLine( "bonus!", fLineX, fLineY, 2.0 )  
        --for every level except the last, pick a random text
        elseif Constants.CurrentGameLevel ~= Constants.NumGameLevels then
            local fLineX = 0.0
            local fLineY = HUD.mLargeLetterSize * -0.5
            
            local iRandomText = math.random( 1, 10 )
            if iRandomText < 7 then
                fLineX = HUD.mLargeLetterSize * -3.0
                tTextManager:AddLine( "bonus!", fLineX, fLineY, 2.0 )         
            elseif iRandomText == 7 then
                fLineX = HUD.mLargeLetterSize * -3.5
                tTextManager:AddLine( "frenzy!", fLineX, fLineY, 2.0 )         
            elseif iRandomText == 8 then
                fLineX = HUD.mLargeLetterSize * -3.0
                tTextManager:AddLine( "mania!", fLineX, fLineY, 2.0 )           
            elseif iRandomText == 9 then
                fLineX = HUD.mLargeLetterSize * -3.0
                tTextManager:AddLine( "fever!", fLineX, fLineY, 2.0 )            
            elseif iRandomText == 10 then
                fLineX = HUD.mLargeLetterSize * -4.5
                tTextManager:AddLine( "skidditz!", fLineX, fLineY, 2.0 )            
            end  
        end
        
        --this is kind of hacky, but assures that later text won't zoom in
        tTextManager:SetZoomAttributes( 0.0, 0.0, 0.0 )    
    end
end

return HUD