local Renderer = require('Renderer')
local Constants = require('Constants')

local Input = {}

function Input:NewInputTouch()
    local Ob = {}
    
    function Ob:Init()
        self:Reset()
    end
    
    function Ob:Reset()
        self.x = -1.0
        self.y = -1.0
        self.z = -1.0
        self.bIsDown = false
        self.nHandId = -1
        self.nId = -1      
        self.bDirty = false  
    end
    
    Ob:Init()
    return Ob
end

function Input:NewSwipeGesture()
    local Ob = {}
    
    function Ob:Init()
        self:Reset()
    end
    
    function Ob:Reset()
        self.nId = -1
        self.nPointableId = -1
        self.nState = Input.GESTURE_STATE_INVALID
        self.fSpeed = 0.0
        self.fDuration = 0.0
        self.fCurrentX = -1.0
        self.fCurrentY = -1.0
        self.fCurrentZ = -1.0
        self.fStartX = -1.0
        self.fStartY = -1.0
        self.fStartZ = -1.0
        self.fDirectionX = -1.0
        self.fDirectionY = -1.0
        self.fDirectionZ = -1.0
    end
    
    Ob:Init()
    
    return Ob
end

function Input:NewCircleGesture()
    local Ob = {}
    
    function Ob:Init()
        self:Reset()
    end
    
    function Ob:Reset()
        self.nId = -1
        self.nPointableId = -1
        self.nState = Input.GESTURE_STATE_INVALID
        self.fCenterX = -1.0
        self.fCenterY = -1.0
        self.fCenterZ = -1.0
        self.fProgress = 0.0
        self.fRadius = 0.0
        self.fDuration = 0.0
        self.fAngle = 0.0
        self.bIsClockwise = true
    end
    
    Ob:Init()
    
    return Ob
end

function Input:NewTapGesture()
    local Ob = {}
    
    function Ob:Init()
        self:Reset()
    end
    
    function Ob:Reset()
        self.nId = -1
        self.nPointableId = -1
        self.nState = Input.GESTURE_STATE_INVALID
        self.bIsKeyTap = true
        self.fProgress = 0.0
        self.fPositionX = -1.0
        self.fPositionY = -1.0
        self.fPositionZ = -1.0
        self.fDirectionX = -1.0
        self.fDirectionY = -1.0
        self.fDirectionZ = -1.0
    end
    
    Ob:Init()
    
    return Ob
end

function Input.Init(fnTouchCb)
    Input.kMAX_HANDS = 2

    Input.kMAX_CIRCLE_GESTURES = 5
    Input.kMAX_SWIPE_GESTURES = 5
    Input.kMAX_TAP_GESTURES = 10

    -- Just for laziness' sake
    Input.TOUCH_DOWN = MOAITouchSensor.TOUCH_DOWN
    Input.TOUCH_MOVE = MOAITouchSensor.TOUCH_MOVE
    Input.TOUCH_UP = MOAITouchSensor.TOUCH_UP

    Input.GESTURE_STATE_INVALID = -1
    Input.GESTURE_STATE_START = 1
    Input.GESTURE_STATE_UPDATE = 2
    Input.GESTURE_STATE_STOP = 3
    
    Input.fnTouchCb = fnTouchCb

    Input.kMAX_POINTABLES = Constants.NumGameInputs
    
    Input:ResetExcessiveMovement()
    
    Input.tTouches = {}
    Input.tSwipeGestures = {}
    Input.tCircleGestures = {}
    Input.tTapGestures = {}
    
    Input.tMouse = Input:NewInputTouch()
    
    -- support arbitrary numbers of these
    local i = 1
    for i=1,Input.kMAX_POINTABLES do
        local touch = Input:NewInputTouch()
        table.insert(Input.tTouches, touch)
    end
        
    Input.nNumCircleGestures = 0
    for i=1,Input.kMAX_CIRCLE_GESTURES do
        local circle = Input:NewCircleGesture()
        table.insert(Input.tCircleGestures, circle)
    end
    
    Input.nNumSwipeGestures = 0
    for i=1,Input.kMAX_SWIPE_GESTURES do
        local swipe = Input:NewSwipeGesture()
        table.insert(Input.tSwipeGestures, swipe)
    end
    
    Input.nNumTapGestures = 0
    for i=1,Input.kMAX_TAP_GESTURES do
        local tap = Input:NewTapGesture()
        table.insert(Input.tTapGestures, tap)
    end
        
    if Input.HasMouse() then
        -- called on button state _change_ (not continuously)
        local function onLeftMouse(bIsDown)
            Input.tMouse.bIsDown = bIsDown
            
            -- remap mouse up/down to touch states
            local evt = Input.TOUCH_UP
            if bIsDown then evt = Input.TOUCH_DOWN end
            
            Input.OnMouse(evt, Input.tMouse.x, Input.tMouse.y)
        end

        -- called on pointer position _change_ (not continuously)
        local function onMouseMove(x,y)
            Input.tMouse.x = x
            Input.tMouse.y = y
            if Input.tMouse.bIsDown then
                Input.OnMouse(TOUCH_MOVE, x, y)
            end
        end

        MOAIInputMgr.device.pointer:setCallback ( onMouseMove )
        MOAIInputMgr.device.mouseLeft:setCallback ( onLeftMouse )
    else   
        local function onTouch(eventType, index, x, y, z)
            if (eventType == Input.TOUCH_DOWN or
                eventType == Input.TOUCH_MOVE or
                eventType == Input.TOUCH_UP) then

                index = index + 1

                -- must offset based on letterboxing
                local barSize = (Constants.NativeScreenHeight - Constants.ValidDrawHeight) / 2

                y = y - barSize

                -- hack: for touch devices like iOS we have no idea of handId, just mod the index
                -- (assuming hands start at 1)
                local handId = (index % 2) + 1
                
                Input.OnTouch(eventType, x, y, z, index, handId, index)
            end
        end
        MOAIInputMgr.device.touch:setCallback(onTouch)
    end
    
    --[[
    -- use for finding keycodes
    if MOAIInputMgr.device.keyboard then
        local function onKey(keyCode, bIsDown)
            print("key: " .. tostring(keyCode) .. ", " .. tostring(bIsDown))
        end
    
        MOAIInputMgr.device.keyboard:setCallback(onKey)
    end
    ]]--
end

function Input.GetTouch(nIndex)
    if nIndex >= 1 and nIndex <= #Input.tTouches then
        return Input.tTouches[nIndex]
    end
    
    return nil
end

function Input.GetMouse()
    return Input.tMouse
end

function Input.GetTouchById(nId)
    for nIndex,rTouch in ipairs(Input.tTouches) do
        if rTouch.nId == nId then
            return rTouch
        end
    end
    
    return nil
end

function Input.GetTouchIndexForId(nId)    
    for nIndex,rTouch in ipairs(Input.tTouches) do
        if rTouch.nId == nId then
            return nIndex
        end
    end
    
    return -1
end

function Input.ResetDirtyFlags()
    for nIndex,rTouch in ipairs(Input.tTouches) do
        rTouch.bDirty = false
    end
end

function Input.SetDirtyFlag(nIndex)
    if nIndex >= 1 and nIndex <= #Input.tTouches then
        Input.tTouches[nIndex].bDirty = true
    end
end

function Input.HasMultitouch()
    return Constants.IsMobile
end

function Input.HasMouse()
    return nil ~= MOAIInputMgr.device.pointer
end

function Input.OnTouch(eventType, x, y, z, index, handId, touchId)
    if index > Input.kMAX_POINTABLES then return end
    
    if eventType == Input.TOUCH_DOWN then 
        Input.tTouches[index].bIsDown = true
    elseif eventType == Input.TOUCH_UP then
        Input.tTouches[index].bIsDown = false
    end
    
    Input.tTouches[index].x = x
    Input.tTouches[index].y = y
    Input.tTouches[index].z = z
    Input.tTouches[index].nHandId = handId
    Input.tTouches[index].nId = touchId
       
    if Input.fnTouchCb then 
        Input.fnTouchCb(eventType, x, y, z, touchId, handId)
    end
end

function Input.OnMouse(eventType, x, y)
    if eventType == Input.TOUCH_DOWN then 
        Input.tMouse.bIsDown = true
    elseif eventType == Input.TOUCH_UP then
        Input.tMouse.bIsDown = false
    end
    
    Input.tMouse.x = x
    Input.tMouse.y = y
    Input.tMouse.nHandId = 1
    Input.tMouse.nId = 1
       
    if Input.fnTouchCb then 
        Input.fnTouchCb(eventType, x, y, 1, 1)
    end
end

function Input:ResetExcessiveMovement()
    Input.mMovementAmount = 0.0
    Input.mMovementThreshold = 1000.0
    Input.mMovementDecay = 2000.0
    Input.mInput1Valid = false
    Input.mInput2Valid = false
    Input.mInput1 = {0.0, 0.0}
    Input.mInput2 = {0.0, 0.0}
end

function Input:CheckForExcessiveMovement( tGameInputs )
    --input 1
    if tGameInputs[1]:IsTouched() then
        if Input.mInput1Valid then
            Input.mMovementAmount = Input.mMovementAmount + math.abs( tGameInputs[1].mX - Input.mInput1[1] )
            Input.mMovementAmount = Input.mMovementAmount + math.abs( tGameInputs[1].mY - Input.mInput1[2] )
        end
        
        Input.mInput1[1] = tGameInputs[1].mX
        Input.mInput1[2] = tGameInputs[1].mY
        Input.mInput1Valid = true
    else
        Input.mInput1Valid = false
    end

    --input 2
    if tGameInputs[2]:IsTouched() then
        if Input.mInput2Valid then
            Input.mMovementAmount = Input.mMovementAmount + math.abs( tGameInputs[2].mX - Input.mInput2[1] )
            Input.mMovementAmount = Input.mMovementAmount + math.abs( tGameInputs[2].mY - Input.mInput2[2] )
        end
        
        Input.mInput2[1] = tGameInputs[2].mX
        Input.mInput2[2] = tGameInputs[2].mY
        Input.mInput2Valid = true
    else
        Input.mInput2Valid = false
    end
    
    --decay movement amount
    local fDecayAmount = Input.mMovementDecay * Constants.FrameStep
    Input.mMovementAmount = math.max( Input.mMovementAmount - fDecayAmount, 0.0 )
    
    return Input.mMovementAmount > Input.mMovementThreshold
end

return Input
