local Constants = require("Constants")

local ParticleSystemManager = {}

ParticleSystemManager.tActiveParticleSystems = {}
ParticleSystemManager.lastTimePrint = 0

-- PUBLIC FUNCTIONS --
function ParticleSystemManager.register( rParticleSystem )
    table.insert(ParticleSystemManager.tActiveParticleSystems, rParticleSystem)
end

function ParticleSystemManager.stopAll()

    local numActiveParticleSystems = #ParticleSystemManager.tActiveParticleSystems
    for i=1,numActiveParticleSystems do
        local rParticleSystem = ParticleSystemManager.tActiveParticleSystems[i]
        rParticleSystem:stop(true)
        rParticleSystem:unload()
    end
    
    ParticleSystemManager.update()
    ParticleSystemManager.update()
    
    DFEffects.stopAllParticleSystems()
end

function ParticleSystemManager.checkForLeaks()
    
    local numParticleSystemsAlive = DFEffects.printParticleSystems()
    if numParticleSystemsAlive > 0 then
        Trace(TT_Error, tostring(numParticleSystemsAlive) .. " particle systems leakd!")
    end
end

function ParticleSystemManager.update()

    MOAISim.profileEnterScope("LUA ParticleSystemManager update")

    -- Delete dead particle systems
    local numActiveParticleSystems = #ParticleSystemManager.tActiveParticleSystems
    
    local iParticleCountLow = Constants.MaxParticlesLow
    local iParticleCountHigh = Constants.MaxParticlesHigh
    
    for i=1,numActiveParticleSystems do
    
        local idx = numActiveParticleSystems - i + 1
        local rParticleSystem = ParticleSystemManager.tActiveParticleSystems[idx]
        
        local bCountExceeded = false
        if rParticleSystem ~= nil then
            if rParticleSystem.iPriority == Constants.ParticlePriorityLow then
                iParticleCountLow = iParticleCountLow - 1
                if iParticleCountLow <= 0 then
                    bCountExceeded = true
                end
            elseif rParticleSystem.iPriority == Constants.ParticlePriorityHigh then
                iParticleCountHigh = iParticleCountHigh - 1
                if iParticleCountHigh <= 0 then
                    bCountExceeded = true
                end
            end
        end
        
        if bCountExceeded then
            rParticleSystem:stop( true )
            rParticleSystem:unload()
            table.remove(ParticleSystemManager.tActiveParticleSystems, idx)        
        elseif rParticleSystem.bStarted == true then
        
            local bUnregister = true
            if rParticleSystem.rParticleSystem ~= nil then
            
                bUnregister = false
                
                local bIsDead = rParticleSystem.rParticleSystem:isDead()    
                if bIsDead == true then
                
                    if rParticleSystem.mCompleteCallback ~= nil then
                        rParticleSystem:mCompleteCallback( rParticleSystem )
                    end                
                    rParticleSystem:stop()
                
                    if rParticleSystem.rEntity ~= nil then
                        rParticleSystem:removeFromEntity()
                    elseif rParticleSystem.rSceneLayer ~= nil then
                        rParticleSystem:removeFromSceneLayer()
                    else
                        assert(0)
                    end
                    
                    rParticleSystem:unload()
                    rParticleSystem = nil
                    
                    bUnregister = true
                end
            end
            
            if bUnregister == true then
                table.remove(ParticleSystemManager.tActiveParticleSystems, idx)
            end
            
        elseif rParticleSystem.bUnloaded == true then
        
            rParticleSystem:stop()
            
            table.remove(ParticleSystemManager.tActiveParticleSystems, idx)
        end
    end
    
    if ParticleSystemManager.debugDrawEnabled then
        -- Print some debug spam
        local curTime = MOAISim.getDeviceTime()
        local deltaTime = curTime - ParticleSystemManager.lastTimePrint
        if deltaTime > 2 then
        
            ParticleSystemManager.lastTimePrint = curTime
        
            numActiveParticleSystems = #ParticleSystemManager.tActiveParticleSystems
            Trace("Active particles systems: " .. tostring(numActiveParticleSystems))
        end
    end
    
    MOAISim.profileLeaveScope("LUA ParticleSystemManager update")
end

-- DEBUG --
function ParticleSystemManager._debugDraw()

    local numActiveParticleSystems = #ParticleSystemManager.tActiveParticleSystems
    for i=1,numActiveParticleSystems do
        local rParticleSystem = ParticleSystemManager.tActiveParticleSystems[i]
        rParticleSystem:_debugDraw()
    end	
end

function ParticleSystemManager.enableDebugDraw(bEnable)
	if ParticleSystemManager.debugDrawEnabled and not bEnable then
		ParticleSystemManager.debugDrawEnabled = false
        Renderer.dDebugRenderWorldSpace:unregister(ParticleSystemManager._debugDraw)
	elseif not ParticleSystemManager.debugDrawEnabled and bEnable then
		ParticleSystemManager.debugDrawEnabled = true
        Renderer.dDebugRenderWorldSpace:register(ParticleSystemManager._debugDraw)
	end
end

function ParticleSystemManager.toggleDebugDraw()
	if ParticleSystemManager.debugDrawEnabled == nil then
		ParticleSystemManager.debugDrawEnabled = false
	end
	ParticleSystemManager.enableDebugDraw(not ParticleSystemManager.debugDrawEnabled)
end

function ParticleSystemManager._debugList()

    numActiveParticleSystems = #ParticleSystemManager.tActiveParticleSystems
    if numActiveParticleSystems <= 0 then
        return
    end
            
    local xCursor = 100
    local yCursor = 100
    local yOffset = 15
    
    local rFont = Renderer.getGlobalFont("debug")
    MOAIGfxDevice.setPenColor( 1, 1, 1, 1 )
    
    MOAIDraw.drawText(rFont, 30, tostring(numActiveParticleSystems) .. " active particle systems:", xCursor, yCursor, 1, 2, 2)
    xCursor = xCursor + 20
    yCursor = yCursor + yOffset

    for i=1,numActiveParticleSystems do
        local rParticleSystem = ParticleSystemManager.tActiveParticleSystems[i]
        if rParticleSystem ~= nil then
            
            local cx = math.floor( xCursor )
            local cy = math.floor( yCursor )
        
            MOAIDraw.drawText(rFont, 30, rParticleSystem.sResource, cx, cy, 1, 1, 1)
            yCursor = yCursor + yOffset
        end
    end
end

function ParticleSystemManager.enableDebugList(bEnable)
	if ParticleSystemManager.debugListEnabled and not bEnable then
		ParticleSystemManager.debugListEnabled = false
        Renderer.dDebugRenderWorldSpace:unregister(ParticleSystemManager._debugList)
	elseif not ParticleSystemManager.debugListEnabled and bEnable then
		ParticleSystemManager.debugListEnabled = true
        Renderer.dDebugRenderWorldSpace:register(ParticleSystemManager._debugList)
	end
end

function ParticleSystemManager.toggleDebugList()
	if ParticleSystemManager.debugListEnabled == nil then
		ParticleSystemManager.debugListEnabled = false
	end
	ParticleSystemManager.enableDebugList(not ParticleSystemManager.debugListEnabled)
end

function ParticleSystemManager.toggleDisable()
    if ParticleSystemManager.bDisabled == nil then
        ParticleSystemManager.bDisabled = false
    end
    ParticleSystemManager.bDisabled = not ParticleSystemManager.bDisabled
    DFEffects.debugDisableParticles( ParticleSystemManager.bDisabled )
end

return ParticleSystemManager
