local Constants = require('Constants')
local DFMath = require('DFCommon.Math')
local DFFile = require('DFCommon.File')
local AudioManager = require('AudioManager')
local BitFactory = require('BitFactory')
local SliceFactory = require('SliceFactory')
local Score = require('Score')
local UI = require('UI')
local ColorManager = require('ColorManager')
local Grid = require('Grid')
local RadEffects = require('RadEffects')
local GameRing = require('GameRing')
local InputVisuals = require('InputVisuals')
local HUD = require('HUD')
local Achievement = require('DFCommon.Achievement')
local PatternManager = {}

function PatternManager:Init( tRenderer )      
    local i = 0
    PatternManager.mNumBits = 15
    PatternManager.mBits = {}
    for i = 1, PatternManager.mNumBits do
        PatternManager.mBits[i] = BitFactory:NewBit(i)
        PatternManager.mBits[i]:CreateProp( tRenderer, i )
    end
    
    PatternManager.mNumSlices = 16
    PatternManager.mSlices = {}
    for i = 1, PatternManager.mNumSlices do
        PatternManager.mSlices[i] = SliceFactory:NewSlice()
    end
        
    PatternManager.mNumBitQueueSlots = 15
    PatternManager.mBitQueue = {}
    for i = 1, PatternManager.mNumBitQueueSlots do
        PatternManager.mBitQueue[i] = BitFactory:NewBitQueueSlot()
    end
    
    PatternManager.mPatternSpeed = 1.0
    PatternManager.mNewBitTimer = 0.0
    PatternManager.mNewBitDuration = 1.0
    PatternManager.mBitPatternCount = 0
    PatternManager.mBitPatternHelpIndex = -1
    PatternManager.mPatternType = Constants.PatternTypeIntro
    PatternManager.mPatternTypePrev = Constants.PatternTypeIntro
    PatternManager.mPatternTypeDirty = false
    PatternManager.mFanfareTimer = 0.0
    PatternManager.mPatternState = Constants.PatternStateStandard
    
    PatternManager.mPatternFlawless = false
    PatternManager.mHadMultiplier = false
        
    PatternManager.mBasePatternStep = 0.75
    PatternManager.mPatternStep = 1.0
    PatternManager.mLastDestroyedBit = {}
        
    PatternManager.mTutorialType = Constants.mTutorialTypeNone
    PatternManager.mPrepForTutorial = false
    PatternManager.mPrepTutorialTimer = 0.0
    PatternManager.mShowTutorial = false
    
    PatternManager.mNumPatternsAttempted = 0
    PatternManager.mNumPatternsPerfect = 0
    
    PatternManager.mFullMixTimer = 0.0
    
    PatternManager.mLoadPatternDataBuffer = MOAIDataBuffer.new()
end
  
function PatternManager:UpdateSize()
    local i = 1
    for i = 1, PatternManager.mNumBits do
        PatternManager.mBits[i]:UpdateSize()
    end
    
    for i = 1, PatternManager.mNumSlices do
        PatternManager.mSlices[i]:UpdateSize()
    end
end

function PatternManager:ClearBitTimer()
    PatternManager.mNewBitTimer = 0.0
    PatternManager.mNewBitDuration = 1.0
end

function PatternManager:Reset()
    PatternManager:ResetBitsAndSlices()
    PatternManager:ResetQueue()
        
    PatternManager.mBitPatternCount = 0
    PatternManager:SetPatternType( Constants.PatternTypeIntro )
    PatternManager.mFanfareTimer = 0.0
    PatternManager.mPatternState = Constants.PatternStateStandard
    
    PatternManager.mTutorialCleared = false
    PatternManager.mTutorialType = Constants.mTutorialTypeNone
    PatternManager.mPrepForTutorial = false
    PatternManager.mShowTutorial = false
    
    PatternManager.mPatternFlawless = false
    PatternManager.mHadMultiplier = false
    
    PatternManager.mNumPatternsAttempted = -1
    PatternManager.mNumPatternsPerfect = 0    
end
    
function PatternManager:ResetQueue()
    local i = 1
    for i = 1, PatternManager.mNumBitQueueSlots do
        PatternManager.mBitQueue[i].mEnabled = false
    end
end

function PatternManager:ResetBitsAndSlices()
    local i = 1
    for i = 1, PatternManager.mNumBits do
        PatternManager.mBits[i]:Disable()
    end
    
    for i = 1, PatternManager.mNumSlices do
        PatternManager.mSlices[i].mEnabled = false
    end    
end

function PatternManager:SetPatternType( iType )
    PatternManager.mPatternTypeDirty = true
    PatternManager.mPatternTypePrev = PatternManager.mPatternType
    PatternManager.mPatternType = iType
end

function PatternManager:MenuTransition()
    RadEffects:SetEffectLayerToGameLayer()
    
    local i = 1
    for i = 1, PatternManager.mNumBits do
        if PatternManager.mBits[i].mEnabled then            
            RadEffects:SetEffectTransform( PatternManager.mBits[i].mStartPos[1], PatternManager.mBits[i].mStartPos[2], 0.0, 0.0, 0.0, 0.0 )
            RadEffects:CreateEffect( "Effects/Particles/Chord/ChargeCreation.particles", nil, nil, nil, Constants.ParticlePriorityHigh )
            RadEffects:CreateEffect( "Effects/Particles/Chord/ChargeFlash.particles", nil, nil, nil, Constants.ParticlePriorityHigh ) 
            RadEffects:CreateEffect( "Effects/Particles/Chord/ChargeEmbers.particles", nil, nil, nil, Constants.ParticlePriorityHigh )
        end
    end
end

function PatternManager:InitForMenu( iDepth )
    --create menu pattern
    if iDepth == Constants.mMenuDepthMenu then
        PatternManager:LoadPattern( "Patterns/menu.txt" ) 
    elseif iDepth == Constants.mMenuDepthOptions then
        PatternManager:LoadPattern( "Patterns/menu2.txt" ) 
    elseif iDepth == Constants.mMenuDepthVisuals then
        PatternManager:LoadPattern( "Patterns/menu3.txt" )         
    end
    
    PatternManager:MakePatternLong()    
end

function PatternManager:InitForGame()
    PatternManager.mFullMixTimer = 0.0
    PatternManager:SetGameSpeed(0.0)
    PatternManager:PrimeForNewLevel()
    PatternManager:ClearBitTimer()
    if Constants.FullMixMode then
        PatternManager:SetPatternType( Constants.PatternTypeFullMix )
    else
        PatternManager:SetPatternType( Constants.PatternTypeIntro )    
    end
    PatternManager:CleanPattern()
end

function PatternManager:InitForComplete()
    PatternManager:SetPatternType( Constants.PatternTypeComplete )
end

function PatternManager:GetBadTutorialSource()
    --run through active bits and return the location of the first bad
    local i = 1
    for i = 1, PatternManager.mNumBits do
        if PatternManager.mBits[i].mEnabled and PatternManager.mBits[i].mType == Constants.BitTypeBad then
            return PatternManager.mBits[i].mStartPos[1], PatternManager.mBits[i].mStartPos[2]
        end
    end      
    return 0.0, 0.0
end

function PatternManager:GetPointTutorialSource()
    --run through active bits and return the location of the first
    local i = 1
    for i = 1, PatternManager.mNumBits do
        if PatternManager.mBits[i].mEnabled then
            return PatternManager.mBits[i].mStartPos[1], PatternManager.mBits[i].mStartPos[2]
        end
    end      
    return 0.0, 0.0    
end

function PatternManager:GetHoldTutorialSource()
    --run through active bits and return the location of the first hold bit
    local i = 1
    for i = 1, PatternManager.mNumBits do
        if PatternManager.mBits[i].mEnabled and PatternManager.mBits[i].mType == Constants.BitTypeHold then
            return PatternManager.mBits[i].mStartPos[1], PatternManager.mBits[i].mStartPos[2]
        end
    end      
    return 0.0, 0.0    
end

function PatternManager:UpdateMenu( iDepth, tRenderer, tLeaderboard )
    local i = 1
    local j = 1
    for i = 1, PatternManager.mNumBits do
        if PatternManager.mBits[i].mEnabled then
            PatternManager.mBits[i]:UpdateMenu(i, iDepth)
            if PatternManager.mBits[i].mType == Constants.BitTypeFreeze then
                if iDepth == Constants.mMenuDepthMenu then
                    --a bit died!  switch game states according to index
                    if i == Constants.mMenuSelectionPlay then
                        Constants.DesiredGameState = Constants.GameStateInGame
                        AudioManager:SetFullMixMode( false )
                    elseif i == Constants.mMenuSelectionFreePlay then
                        Constants.DesiredGameState = Constants.GameStateInGame
                        AudioManager:SetFullMixMode( true )
                    elseif i == Constants.mMenuSelectionLeaders then
                        Constants.DesiredGameState = Constants.GameStateLeaderboardShowcase
                    elseif i == Constants.mMenuSelectionOptions then
                        iDepth = Constants.mMenuDepthOptions
                    end
                elseif iDepth == Constants.mMenuDepthOptions then
                    --a bit died!  switch game states according to index
                    if i == Constants.mMenuSelectionHowToPlay then
                        Constants.DesiredGameState = Constants.GameStateHowToPlay
                    elseif i == Constants.mMenuSelectionVisuals then
                        iDepth = Constants.mMenuDepthVisuals                        
                    elseif i == Constants.mMenuSelectionCredits then
                        Constants.DesiredGameState = Constants.GameStateCredits
                    elseif i == Constants.mMenuSelectionMenu then
                        iDepth = Constants.mMenuDepthMenu
                    end                                
                elseif iDepth == Constants.mMenuDepthVisuals then
                    if i == Constants.mMenuSelectionUltra then
                        Constants.DesiredVisualQuality = Constants.VisualQualityLevelUltra
                        iDepth = Constants.mMenuDepthVisualsReset
                    elseif i == Constants.mMenuSelectionHigh then
                        Constants.DesiredVisualQuality = Constants.VisualQualityLevelHigh
                        iDepth = Constants.mMenuDepthVisualsReset
                    elseif i == Constants.mMenuSelectionLow then
                        Constants.DesiredVisualQuality = Constants.VisualQualityLevelLow
                        iDepth = Constants.mMenuDepthVisualsReset
                    elseif i == Constants.mMenuSelectionBack then
                        iDepth = Constants.mMenuDepthOptions
                    end                                
                end
                
                return iDepth
            else
                --artificially bump the timer so it never ends
                PatternManager.mBits[i].mLifetime = PatternManager.mBits[i].mLifetime + Constants.FrameStep     
            end
        end
    end    
    
    --update queue
    for i = 1, PatternManager.mNumBitQueueSlots do  
        if PatternManager.mBitQueue[i].mEnabled == true then        
            --tick down timer and activate a bit if we hit zero
            PatternManager.mBitQueue[i].mBitDelay = PatternManager.mBitQueue[i].mBitDelay - Constants.FrameStep
                        
            if PatternManager.mBitQueue[i].mBitDelay <= 0.0 then
                local bShowBit = true
                if Constants.FirstTimePlay and iDepth == Constants.mMenuDepthMenu then
                    if i == Constants.mMenuSelectionFreePlay or i == Constants.mMenuSelectionLeaders then
                        bShowBit = false
                    end                
                end
                
                --activate a bit with the queue slot's characteristics
                for j = 1, PatternManager.mNumBits do
                    if PatternManager.mBits[j].mEnabled == false then
                        local iType = PatternManager.mBitQueue[i].mBitType
                        local fStartX = PatternManager.mBitQueue[i].mStartX
                        local fStartY = PatternManager.mBitQueue[i].mStartY
                        
                        --haha.. don't show bits by putting them way the f off screen
                        if not bShowBit then
                            fStartY = fStartY + (Constants.ValidDrawHeight * 10.0)
                        end
                        PatternManager.mBits[j]:EnableForMenu(iType, fStartX, fStartY, fStartX, fStartY, 999999.0, 1.0)   
    
                        --override hold duration for menu
                        PatternManager.mBits[j].mHoldDuration = Constants.BeatDuration * 4.0
                        PatternManager.mBits[j].mHoldTimer = PatternManager.mBits[j].mHoldDuration
                        break
                    end
                end
                
                --queue slot is done
                PatternManager.mBitQueue[i].mEnabled = false
            end
        end
    end    
    
    return iDepth
end
    
function PatternManager:CheckMenuSelection()
    --run through the bits and see if any were held this frame
    local i = 1
    local j = 1
    for i = 1, PatternManager.mNumBits do
        if PatternManager.mBits[i].mEnabled and PatternManager.mBits[i].mHoldActiveFrames > 0 then        
            --we've found one we're going to send back.. make sure every other one didn't get excited to go
            for j = i + 1, PatternManager.mNumBits do
                if PatternManager.mBits[j].mEnabled and PatternManager.mBits[j].mHoldActiveFrames > 0 then
                    PatternManager.mBits[j].mHoldTimer = PatternManager.mBits[j].mHoldDuration
                end
            end
        
            --default to first bit highlighted (cause "play" is #1)
            return i, PatternManager.mBits[i].mStartPos[1], PatternManager.mBits[i].mStartPos[2]
        end
    end    
    
    return -1, 0, 0
end
    
function PatternManager:GetMenuSelectionPercentComplete()
    local i = 1
    for i = 1, PatternManager.mNumBits do
        if PatternManager.mBits[i].mEnabled and PatternManager.mBits[i].mHoldActiveFrames > 0 then        
            return 1.0 - (PatternManager.mBits[i].mHoldTimer / PatternManager.mBits[i].mHoldDuration)
        end
    end    
    
    return 0.0
end

function PatternManager:Update( tTextManager )     
    if PatternManager.mPatternType < Constants.PatternTypePointHoriz then
        PatternManager:UpdateStandard(tTextManager)
    elseif PatternManager.mPatternType ~= Constants.PatternTypeComplete then
        PatternManager:UpdatePoint(tTextManager)
    end
    
    PatternManager.mFullMixTimer = PatternManager.mFullMixTimer + Constants.FrameStep
    
    --full mix timer achievement
    if PatternManager.mFullMixTimer > 300 then
        Achievement.updateAchievement( "DA_FullMixMode", 100 ) 
    end
        
    --update tutorial timer
    if PatternManager.mPrepTutorialTimer > 0.0 then
        PatternManager.mPrepTutorialTimer = PatternManager.mPrepTutorialTimer - Constants.FrameStep
        if PatternManager.mPrepTutorialTimer < 0.0 then
            PatternManager.mShowTutorial = true 
        end
    end  
end

function PatternManager:UpdateStandard( tTextManager )
    --reset counters
    local i = 1
    local j = 1
    
    --update slices 
    for i = 1, PatternManager.mNumSlices do
        if PatternManager.mSlices[i].mEnabled then                        
            PatternManager.mSlices[i]:Update()
            if PatternManager.mSlices[i].mEnabled == false and PatternManager.mSlices[i].mActivated == false then
                --slice died by its own will-- that's not flawless
                PatternManager.mPatternFlawless = false          
            end
        end
    end
    
    --update bits
    for i = 1, PatternManager.mNumBits do
        if PatternManager.mBits[i].mEnabled then
            local bStillAlive = PatternManager.mBits[i]:Update()
            local iType = PatternManager.mBits[i].mType
            
            if not bStillAlive then
                --one of the good bits died by its own will-- that's not flawless
                if iType == Constants.BitTypeGood or iType == Constants.BitTypeHold then
                    PatternManager.mPatternFlawless = false
                elseif iType == Constants.BitTypeBad then
                    --if a bad just disappeared, we may want to finish the pattern
                    PatternManager:CheckForQueueAdvancement()
                end
            end
        end
    end
    
    --update queue
    for i = 1, PatternManager.mNumBitQueueSlots do  
        if PatternManager.mBitQueue[i].mEnabled == true then        
            --tick down timer and activate a bit if we hit zero
            PatternManager.mBitQueue[i].mBitDelay = PatternManager.mBitQueue[i].mBitDelay - Constants.FrameStep
                        
            if PatternManager.mBitQueue[i].mBitDelay <= 0.0 then
                --activate a slice with the queue slot's characteristics
                if PatternManager.mBitQueue[i].mBitType == Constants.BitTypeSlice then
                    for j = 1, PatternManager.mNumSlices do
                        if PatternManager.mSlices[j].mEnabled == false then
                            PatternManager.mSlices[j]:Enable(PatternManager.mBitQueue[i].mOrientation, PatternManager.mBitQueue[i].mLifetime)
                            break
                        end
                    end
                else
                    --activate a bit with the queue slot's characteristics
                    for j = 1, PatternManager.mNumBits do
                        if PatternManager.mBits[j].mEnabled == false then
                            local iType = PatternManager.mBitQueue[i].mBitType
                            local fStartX = PatternManager.mBitQueue[i].mStartX
                            local fStartY = PatternManager.mBitQueue[i].mStartY
                            local fDestX = PatternManager.mBitQueue[i].mDestX
                            local fDestY = PatternManager.mBitQueue[i].mDestY       
                            local iLife = PatternManager.mBitQueue[i].mLifetime
                            local fEnterTimer = PatternManager.mBitQueue[i].mEnterTimer
                            local fEnterDuration = PatternManager.mBitQueue[i].mEnterDuration
                            PatternManager.mBits[j]:Enable(iType, fStartX, fStartY, fDestX, fDestY, iLife, fEnterTimer, fEnterDuration)
                            
                            local iType = PatternManager.mBits[j].mType
                            if iType == Constants.BitTypeBad then
                                --if we're waiting to show the bad tutorial, throw our flag now
                                if PatternManager.mPrepForTutorial and PatternManager.mTutorialType == Constants.mTutorialTypeBad then
                                    PatternManager.mPrepForTutorial = false
                                    PatternManager.mPrepTutorialTimer = fEnterTimer * 0.35
                                end                                  
                            elseif iType == Constants.BitTypeHold then
                                --if we're waiting to show the hold tutorial, throw our flag now
                                if PatternManager.mPrepForTutorial and PatternManager.mTutorialType == Constants.mTutorialTypeHold then
                                    PatternManager.mPrepForTutorial = false 
                                    PatternManager.mPrepTutorialTimer = fEnterTimer * 0.5
                                end   
                            end
                            
                            break
                        end
                    end
                end
                
                --queue slot is done
                PatternManager.mBitQueue[i].mEnabled = false
            end
        end
    end
    
    --update timer
    PatternManager.mNewBitTimer = PatternManager.mNewBitTimer - Constants.FrameStep
    if Constants.LevelState ~= Constants.LevelStateHealth then
        GameRing:UpdateTimer( PatternManager.mNewBitTimer / PatternManager.mNewBitDuration, Constants.PatternStateStandard )
    else
        GameRing:UpdateTimer( 1.0, Constants.PatternStateFanfare )
    end
    
    if PatternManager.mNewBitTimer <= 0.0 then   
        -- [[
        --if we're currently showing fanfare, tick down until that's gone and then set the next pattern
        if PatternManager.mPatternState == Constants.PatternStateStandard then
            PatternManager:CheckPatternLeftoversForPerfect()
            PatternManager:ResetQueue()
            PatternManager.mFanfareTimer = 0.5
            
            --we attempted a pattern, so increment that
            PatternManager.mNumPatternsAttempted = PatternManager.mNumPatternsAttempted + 1
            
            --was the pattern flawless?
            if PatternManager.mPatternFlawless then  
                PatternManager:ClearBadBits()
                AudioManager:PlaySound( 'Radius/SFX/PatternComplete' )
                Score:IncMultiplier()  
                PatternManager:EnableBitFanfare()  
            
                PatternManager.mPatternState = Constants.PatternStateFanfare
                HUD:ShowPatternPerfect( tTextManager, PatternManager.mFanfareTimer )   
            
                --increment perfect patterns!
                PatternManager.mNumPatternsPerfect = PatternManager.mNumPatternsPerfect + 1
            else
                --boo!
                PatternManager.mPatternState = Constants.PatternStateFail
                Score:ClearMultiplier()
                if PatternManager.mHadMultiplier then
                    HUD:ShowLostMultiplier( tTextManager, PatternManager.mFanfareTimer ) 
                end
            end   
        else 
            GameRing:UpdateTimer( 1.0, PatternManager.mPatternState )
            PatternManager.mFanfareTimer = PatternManager.mFanfareTimer - Constants.FrameStep
            if PatternManager.mFanfareTimer <= 0.0 then
                PatternManager:ClearFrozenBits()
                PatternManager:SetNewBitPattern()      
                PatternManager.mPatternState = Constants.PatternStateStandard                
            end         
        end
        --]]
    end
end

function PatternManager:UpdatePoint( tTextManager )
    local i = 1
    local j = 1
    
    --update bits
    for i = 1, PatternManager.mNumBits do
        if PatternManager.mBits[i].mEnabled then
            local bStillAlive = PatternManager.mBits[i]:UpdatePointType()  
            if not bStillAlive then
                PatternManager.mPatternFlawless = false
            end            
        end
    end
    
    --update queue
    for i = 1, PatternManager.mNumBitQueueSlots do  
        if PatternManager.mBitQueue[i].mEnabled == true then        
            --tick down timer and activate a bit if we hit zero
            PatternManager.mBitQueue[i].mBitDelay = PatternManager.mBitQueue[i].mBitDelay - Constants.FrameStep
                        
            if PatternManager.mBitQueue[i].mBitDelay <= 0.0 then
                --activate a bit with the queue slot's characteristics
                for j = 1, PatternManager.mNumBits do
                    if PatternManager.mBits[j].mEnabled == false then
                        local iCount = PatternManager.mBitQueue[i].mOrientation
                        local fStartX = PatternManager.mBitQueue[i].mStartX
                        local fStartY = PatternManager.mBitQueue[i].mStartY
                        local fDestX = PatternManager.mBitQueue[i].mDestX
                        local fDestY = PatternManager.mBitQueue[i].mDestY       
                        local iLife = PatternManager.mBitQueue[i].mLifetime
                        local fEnterTimer = PatternManager.mBitQueue[i].mEnterTimer
                        local fEnterDuration = PatternManager.mBitQueue[i].mEnterDuration
                        PatternManager.mBits[j]:EnableForPoint(fStartX, fStartY, fDestX, fDestY, iLife, fEnterTimer, fEnterDuration, iCount)
                        
                        if PatternManager.mPrepForTutorial then
                            PatternManager.mPrepForTutorial = false
                            PatternManager.mPrepTutorialTimer = fEnterTimer + Constants.FrameStep
                        end   
                                
                        break
                    end
                end
                
                --queue slot is done
                PatternManager.mBitQueue[i].mEnabled = false
            end
        end
    end
        
    --update timer
    PatternManager.mNewBitTimer = PatternManager.mNewBitTimer - Constants.FrameStep
    if Constants.LevelState ~= Constants.LevelStateHealth then
        GameRing:UpdateTimer( PatternManager.mNewBitTimer / PatternManager.mNewBitDuration, Constants.PatternStateStandard )
    else
        GameRing:UpdateTimer( 1.0, Constants.PatternStateFanfare )
    end
    
    if PatternManager.mNewBitTimer <= 0.0 then     
        --if we're currently showing fanfare, tick down until that's gone and then set the next pattern
        if PatternManager.mPatternState == Constants.PatternStateStandard then
            PatternManager:CheckPatternLeftoversForPerfect()
            PatternManager:ResetQueue()
            PatternManager.mFanfareTimer = .5

            --we attempted a pattern, so increment that
            PatternManager.mNumPatternsAttempted = PatternManager.mNumPatternsAttempted + 1
            
            --was the pattern flawless?
            if PatternManager.mPatternFlawless then  
                AudioManager:PlaySound( 'Radius/SFX/PatternComplete' )
                Score:IncMultiplier()     
                PatternManager:EnableBitFanfare()  
                
                HUD:ShowPatternPerfect( tTextManager, PatternManager.mFanfareTimer )             
                PatternManager.mPatternState = Constants.PatternStateFanfare                
                
                --increment perfect patterns!
                PatternManager.mNumPatternsPerfect = PatternManager.mNumPatternsPerfect + 1      
            else
                --boo!
                PatternManager.mPatternState = Constants.PatternStateFail
                Score:ClearMultiplier()
                if PatternManager.mHadMultiplier then
                    HUD:ShowLostMultiplier( tTextManager, PatternManager.mFanfareTimer ) 
                end                
            end   
        else 
            GameRing:UpdateTimer( 1.0, PatternManager.mPatternState )
            PatternManager.mFanfareTimer = PatternManager.mFanfareTimer - Constants.FrameStep
            if PatternManager.mFanfareTimer <= 0.0 then
                PatternManager:ClearFrozenBits()
                PatternManager:SetNewBitPattern()     
                PatternManager.mPatternState = Constants.PatternStateStandard
            end         
        end
    end
end
   
function PatternManager:CheckForPatternTypeChange( tChargeManager )
    --if current pattern doesn't match previous, we just switched states
    if PatternManager.mPatternTypeDirty then
        if PatternManager.mPatternType == Constants.PatternTypePointHoriz then
            GameRing:SetDesiredRotation( 0.0, -90.0, 0.0, 180.0 )
            GameRing:SetVisibilityTimer( false, 0.5 )
            Score:SetScoreRotation( 110.0, 220.0 )      
            tChargeManager:SetToAltGameMode( true )   
            InputVisuals:SetAltGameMode( true )
            UI:SetClockAlphaDesired( 1.0 )
            Constants.GameCameraTiltDesiredMode = Constants.GameCameraTiltModeTiltHoriz
        elseif PatternManager.mPatternType == Constants.PatternTypePointVert then
            GameRing:SetDesiredRotation( 90.0, 0.0, 0.0, 180.0 )
            GameRing:SetVisibilityTimer( false, 0.5 )
            Score:SetScoreRotation( 110.0, 220.0 )      
            tChargeManager:SetToAltGameMode( true )   
            InputVisuals:SetAltGameMode( true )
            UI:SetClockAlphaDesired( 1.0 )
            Constants.GameCameraTiltDesiredMode = Constants.GameCameraTiltModeTiltVert
        elseif PatternManager.mPatternType == Constants.PatternTypePointBoth then
            GameRing:SetDesiredRotation( 90.0, 90.0, 0.0, 180.0 )
            GameRing:SetVisibilityTimer( false, 0.5 )
            Score:SetScoreRotation( 110.0, 220.0 )      
            tChargeManager:SetToAltGameMode( true )   
            InputVisuals:SetAltGameMode( true )
            UI:SetClockAlphaDesired( 1.0 )
            Constants.GameCameraTiltDesiredMode = Constants.GameCameraTiltModeTiltBoth
        else
            --standard mode.. reset everything if we're coming from an alt mode
            if PatternManager.mPatternTypePrev == Constants.PatternTypePointHoriz or PatternManager.mPatternTypePrev == Constants.PatternTypePointVert or PatternManager.mPatternTypePrev == Constants.PatternTypePointBoth then
                GameRing:SetDesiredRotation( 0.0, 0.0, 0.0, 90.0 )
                GameRing:SetVisibilityTimer( true, 0.01 )
                Score:SetScoreRotation( 0.0, 220.0 )     
                tChargeManager:SetToAltGameMode( false )
                InputVisuals:SetAltGameMode( false )
                PatternManager:DelayPattern( 1.0 ) 
                Constants.GameCameraTiltDesiredMode = Constants.GameCameraTiltModeNone
                UI:SetClockAlphaDesired( 0.0 )
            end
        end
        
        PatternManager.mPatternTypePrev = PatternManager.mPatternType
        PatternManager.mPatternTypeDirty = false
        return true
    end
    return false
end
     
function PatternManager:Draw()
    --draw active bits movement lines
    MOAIGfxDevice.setPenWidth( 4.0 )
    
    local fForwardPrediction = Constants.BeatDuration * 4.0
    local i = 1
    for i = 1, PatternManager.mNumBits do
        local iType = PatternManager.mBits[i].mType
        if PatternManager.mBits[i].mEnabled and iType ~= Constants.BitTypeFreeze and iType ~= Constants.BitTypeFanFare and iType ~= Constants.BitTypePoint then
            local fPosX, fPosY = PatternManager.mBits[i]:GetCurrentPos()
            
            if iType == Constants.BitTypeBad then
                MOAIGfxDevice.setPenColor( Constants.BadEffectColor[1], Constants.BadEffectColor[2], Constants.BadEffectColor[3] )
            else
                MOAIGfxDevice.setPenColor( Constants.Color1[1], Constants.Color1[2], Constants.Color1[3] )
            end
            
            local fDestX, fDestY
            if PatternManager.mBits[i].mEntering then
                --if we're entering, trace the line out as we come in
                fDestX, fDestY = PatternManager.mBits[i]:GetEnteringDest()
            else
                fDestX = PatternManager.mBits[i].mDestination[1]
                fDestY = PatternManager.mBits[i].mDestination[2]
            end
            
            MOAIDraw.drawLine( fPosX, fPosY, fDestX, fDestY )
        end
    end
end

function PatternManager:DrawSlices()
    MOAIGfxDevice.setPenWidth( 4.0 )
    
    local i = 1
    for i = 1, PatternManager.mNumSlices do
        if PatternManager.mSlices[i].mEnabled == true then    
            PatternManager.mSlices[i]:Draw()
        end
    end    
end
  
function PatternManager:SetNewBitPattern()    
    --make sure slices are cleaned up
    local i = 1
    for i = 1, PatternManager.mNumSlices do
        PatternManager.mSlices[i].mEnabled = false
    end
    
    if PatternManager.mPatternType == Constants.PatternTypeIntro then
        PatternManager:PickNewIntroSet()
    elseif PatternManager.mPatternType == Constants.PatternTypeStandard then
        PatternManager:PickNewStandardSet()
    elseif PatternManager.mPatternType == Constants.PatternTypeBonus then
        PatternManager:PickBonusSet()
    elseif PatternManager.mPatternType == Constants.PatternTypeFullMix then
        PatternManager:PickFullMixSet()
    elseif PatternManager.mPatternType == Constants.PatternTypeLearnThrow then
        PatternManager:PickLearnThrowSet()    
    elseif PatternManager.mPatternType == Constants.PatternTypeLearnHold then
        PatternManager:PickLearnHoldSet()
    elseif PatternManager.mPatternType == Constants.PatternTypePointHoriz or PatternManager.mPatternType == Constants.PatternTypePointVert or PatternManager.mPatternType == Constants.PatternTypePointBoth then
        PatternManager:PickNewPointSet()
    end
    
    PatternManager.mPatternFlawless = true
    PatternManager.mHadMultiplier = Score:HasMultiplier()
end
  
function PatternManager:PickNewIntroSet()
    local iLastIndex = PatternManager.mBitPatternHelpIndex

    --Constants.ThrowUnlocked = true
    --PatternManager:LoadPattern( "Patterns/leftspin_3.txt" )
        
    -- [[ 
    --show, swipe left, up/down, swipe right
    if PatternManager.mBitPatternCount == 0 then
        PatternManager:LoadPattern( "Patterns/intro_start.txt" ) 
        PatternManager:MakePatternLong()
    elseif PatternManager.mBitPatternCount == 1 then 
        PatternManager:LoadPattern( "Patterns/intro0.txt" )
        PatternManager:MakePatternLong()     
    elseif PatternManager.mBitPatternCount == 2 then 
        PatternManager:LoadPattern( "Patterns/intro1.txt" )
        PatternManager:MakePatternLong()
    elseif PatternManager.mBitPatternCount == 3 then 
        PatternManager:LoadPattern( "Patterns/intro2.txt" )
        PatternManager:MakePatternLong()
    elseif PatternManager.mBitPatternCount == 4 then    
        --introduce bad! --bad on top, good on bottom (or vice versa)
        local fRandom = math.random(0, 1)
        if fRandom == 0 then
            PatternManager:LoadPattern( "Patterns/intro3a.txt" )
        else
            PatternManager:LoadPattern( "Patterns/intro3b.txt" )
        end
        
        if not Constants.FullMixMode then
            PatternManager.mPrepForTutorial = true 
            PatternManager.mTutorialType = Constants.mTutorialTypeBad
        end
    elseif PatternManager.mBitPatternCount >= 5 and PatternManager.mBitPatternCount <= 8 then   
        --bad with good        
        while PatternManager.mBitPatternHelpIndex == iLastIndex do
            PatternManager.mBitPatternHelpIndex = math.random(0, 2)
        end
        
        if PatternManager.mBitPatternHelpIndex == 0 then PatternManager:LoadPattern( "Patterns/intro4a.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 1 then PatternManager:LoadPattern( "Patterns/intro4b.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 2 then PatternManager:LoadPattern( "Patterns/intro4c.txt" )
        end
    elseif PatternManager.mBitPatternCount == 9 then
        --introduce moving bits!
        PatternManager:LoadPattern( "Patterns/intro5.txt" )
    elseif PatternManager.mBitPatternCount > 9 then
        --various easy patterns
        while PatternManager.mBitPatternHelpIndex == iLastIndex do
            PatternManager.mBitPatternHelpIndex = math.random(0, 1)   
        end

        if PatternManager.mBitPatternHelpIndex == 0 then PatternManager:LoadPattern( "Patterns/intro6a.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 1 then PatternManager:LoadPattern( "Patterns/intro6b.txt" )
        end        
    end
    
    PatternManager.mBitPatternCount = PatternManager.mBitPatternCount + 1
    if PatternManager.mBitPatternCount > 10 then
        PatternManager.mBitPatternCount = 0
        PatternManager:SetPatternType( Constants.PatternTypeStandard )
    end 
    --]]
end
  
function PatternManager:PickLearnThrowSet()    
    if PatternManager.mBitPatternCount == 0 then 
        PatternManager:LoadPattern( "Patterns/throw0.txt" )
        PatternManager:DelayPattern( 3.0 ) 
        if not Constants.TouchBotEnabled then
            PatternManager.mPrepTutorialTimer = 2.5
            PatternManager.mTutorialType = Constants.mTutorialTypeThrow
        end
    elseif PatternManager.mBitPatternCount == 1 then PatternManager:LoadPattern( "Patterns/throw1.txt" )
    elseif PatternManager.mBitPatternCount == 2 then PatternManager:LoadPattern( "Patterns/throw2.txt" )
    elseif PatternManager.mBitPatternCount == 3 then PatternManager:LoadPattern( "Patterns/throw3.txt" )
    end    
    
    --allow the player lots of time to learn
    PatternManager:MakePatternLong()
    
    PatternManager.mBitPatternCount = PatternManager.mBitPatternCount + 1
    if PatternManager.mBitPatternCount > 3 then
        PatternManager.mBitPatternCount = 0
        PatternManager:SetPatternType( Constants.PatternTypeStandard )
    end   
end

function PatternManager:PickLearnHoldSet()    
    if PatternManager.mBitPatternCount == 0 then 
        PatternManager:LoadPattern( "Patterns/hold0.txt" )
        --PatternManager.mPrepForTutorial = true 
        --PatternManager.mTutorialType = Constants.mTutorialTypeHold
        PatternManager:MakePatternLong()        
    elseif PatternManager.mBitPatternCount == 1 then PatternManager:LoadPattern( "Patterns/hold1.txt" )
    elseif PatternManager.mBitPatternCount == 2 then PatternManager:LoadPattern( "Patterns/hold2.txt" )
    elseif PatternManager.mBitPatternCount == 3 then PatternManager:LoadPattern( "Patterns/hold3.txt" )
    end    
    
    PatternManager.mBitPatternCount = PatternManager.mBitPatternCount + 1
    if PatternManager.mBitPatternCount > 3 then
        PatternManager.mBitPatternCount = 0
        PatternManager:SetPatternType( Constants.PatternTypeStandard )
    end   
end
   
function PatternManager:PickNewStandardSet()   
    --pick random pattern that isn't the same as the last one we picked
    local iLastIndex = PatternManager.mBitPatternHelpIndex   
    
    --[[
        _one bad patterns      
        triangle_low
        triangle_high
        bigslash_ur
        bigslash_ul
        five_top
        five_bottom
        cross_up
        cross_down
        long_slash_r
        long_slash_l
        vert_slash_r
        vert_slash_l
        
        _two bad patterns
        keyturn_ccw
        keyturn_cw
        bs_to_fs
        fs_to_bs
        criss_cross_left
        criss_cross_right
        side_to_side_left_e
        side_to_side_right_e
        pin_bl
        pin_tr
        
        _advanced patterns
        steady_up
        steady_down
        bs_rotate_cw
        fs_rotate_ccw
        up_down_up
        slash_left
        slash_right
        back_forth_bottom
        back_forth_top
        
        _hard patterns
        diamond_bs
        diamond_fs
        eggslicer_right_up
        eggslicer_right_down
        eggslicer_left_up
        eggslicer_left_down
        
        _throw patterns
        wait_left
        wait_right
        leftspin_3                   
        leftspin_down_e
        leftspin_up_e    
        rightspin_3
        rightspin_down_e
        rightspin_up_e
        top_bottom_spin
        
        _throw advanced patterns
        leftspin_down
        leftspin_up
        rightspin_down
        rightspin_up
        out_n_in_fs
        out_n_in_bs
        spin_top_left
        spin_top_right
        
        _hold patterns
        all_fall
        all_rise
        hold_down_right
        hold_up_left
        hold_box_cw
        hold_box_ccw
        hold_pinch_tr
        hold_pinch_bl
        hold_bad_bs
        hold_bad_fs
        
        _advanced hold patterns
        hold_center_fs        
        hold_center_bs       
        hold_steady_up
        hold_steady_down
    --]]
    
    --first level, use only one-bad patterns
    if Constants.CurrentGameLevel == 1 then
        while PatternManager.mBitPatternHelpIndex == iLastIndex do
            PatternManager.mBitPatternHelpIndex = math.random(0, 13) 
        end    
        
        if PatternManager.mBitPatternHelpIndex == 0 then PatternManager:LoadPattern( "Patterns/intro6c.txt" ) 
        elseif PatternManager.mBitPatternHelpIndex == 1 then PatternManager:LoadPattern( "Patterns/intro6d.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 2 then PatternManager:LoadPattern( "Patterns/five_bottom.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 3 then PatternManager:LoadPattern( "Patterns/five_top.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 4 then PatternManager:LoadPattern( "Patterns/bigslash_ul.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 5 then PatternManager:LoadPattern( "Patterns/bigslash_ur.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 6 then PatternManager:LoadPattern( "Patterns/triangle_low.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 7 then PatternManager:LoadPattern( "Patterns/triangle_high.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 8 then PatternManager:LoadPattern( "Patterns/cross_up.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 9 then PatternManager:LoadPattern( "Patterns/cross_down.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 10 then PatternManager:LoadPattern( "Patterns/long_slash_r.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 11 then PatternManager:LoadPattern( "Patterns/long_slash_l.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 12 then PatternManager:LoadPattern( "Patterns/vert_slash_r.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 13 then PatternManager:LoadPattern( "Patterns/vert_slash_l.txt" ) 
        end       
    --only the build section, do previous level difficulty
    elseif Constants.CurrentGameLevel == Constants.PointHorizLevel then
        while PatternManager.mBitPatternHelpIndex == iLastIndex do
            PatternManager.mBitPatternHelpIndex = math.random(0, 9) 
        end    
        
        if PatternManager.mBitPatternHelpIndex == 0 then PatternManager:LoadPattern( "Patterns/five_bottom.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 1 then PatternManager:LoadPattern( "Patterns/five_top.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 2 then PatternManager:LoadPattern( "Patterns/bigslash_ul.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 3 then PatternManager:LoadPattern( "Patterns/bigslash_ur.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 4 then PatternManager:LoadPattern( "Patterns/triangle_low.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 5 then PatternManager:LoadPattern( "Patterns/triangle_high.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 6 then PatternManager:LoadPattern( "Patterns/long_slash_r.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 7 then PatternManager:LoadPattern( "Patterns/long_slash_l.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 8 then PatternManager:LoadPattern( "Patterns/vert_slash_r.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 9 then PatternManager:LoadPattern( "Patterns/vert_slash_l.txt" ) 
        end   
    --hold level, use only easy hold patterns, with a couple one-bad patterns
    elseif Constants.CurrentGameLevel == Constants.HoldTutorialLevel then
        while PatternManager.mBitPatternHelpIndex == iLastIndex do
            PatternManager.mBitPatternHelpIndex = math.random(0, 11) 
        end
        
        if PatternManager.mBitPatternHelpIndex == 0 then PatternManager:LoadPattern( "Patterns/five_top.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 1 then PatternManager:LoadPattern( "Patterns/five_bottom.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 2 then PatternManager:LoadPattern( "Patterns/all_fall.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 3 then PatternManager:LoadPattern( "Patterns/all_rise.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 4 then PatternManager:LoadPattern( "Patterns/hold_down_right.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 5 then PatternManager:LoadPattern( "Patterns/hold_up_left.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 6 then PatternManager:LoadPattern( "Patterns/hold_box_cw.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 7 then PatternManager:LoadPattern( "Patterns/hold_box_ccw.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 8 then PatternManager:LoadPattern( "Patterns/hold_pinch_bl.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 9 then PatternManager:LoadPattern( "Patterns/hold_pinch_tr.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 10 then PatternManager:LoadPattern( "Patterns/hold_bad_bs.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 11 then PatternManager:LoadPattern( "Patterns/hold_bad_fs.txt" )      
        end   
    --interim level, use harder hold and bring in two-bad patterns
    elseif Constants.CurrentGameLevel == Constants.HardHoldLevel then
        while PatternManager.mBitPatternHelpIndex == iLastIndex do
            PatternManager.mBitPatternHelpIndex = math.random(0, 11) 
        end
                
        if PatternManager.mBitPatternHelpIndex == 0 then PatternManager:LoadPattern( "Patterns/keyturn_ccw.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 1 then PatternManager:LoadPattern( "Patterns/keyturn_cw.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 2 then PatternManager:LoadPattern( "Patterns/bs_to_fs.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 3 then PatternManager:LoadPattern( "Patterns/fs_to_bs.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 4 then PatternManager:LoadPattern( "Patterns/hold_steady_up.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 5 then PatternManager:LoadPattern( "Patterns/hold_steady_down.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 6 then PatternManager:LoadPattern( "Patterns/side_to_side_left_e.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 7 then PatternManager:LoadPattern( "Patterns/side_to_side_right_e.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 8 then PatternManager:LoadPattern( "Patterns/pin_bl.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 9 then PatternManager:LoadPattern( "Patterns/pin_tr.txt" )      
        elseif PatternManager.mBitPatternHelpIndex == 10 then PatternManager:LoadPattern( "Patterns/hold_center_bs.txt" )      
        elseif PatternManager.mBitPatternHelpIndex == 11 then PatternManager:LoadPattern( "Patterns/hold_center_fs.txt" )      
        end        
    --only the build section, do previous level difficulty
    elseif Constants.CurrentGameLevel == Constants.PointVertLevel then
        while PatternManager.mBitPatternHelpIndex == iLastIndex do
            PatternManager.mBitPatternHelpIndex = math.random(0, 7) 
        end    
        
        if PatternManager.mBitPatternHelpIndex == 0 then PatternManager:LoadPattern( "Patterns/criss_cross_right.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 1 then PatternManager:LoadPattern( "Patterns/criss_cross_left.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 2 then PatternManager:LoadPattern( "Patterns/hold_center_bs.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 3 then PatternManager:LoadPattern( "Patterns/hold_center_fs.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 4 then PatternManager:LoadPattern( "Patterns/keyturn_ccw.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 5 then PatternManager:LoadPattern( "Patterns/keyturn_cw.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 6 then PatternManager:LoadPattern( "Patterns/hold_steady_up.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 7 then PatternManager:LoadPattern( "Patterns/hold_steady_down.txt" )
        end           
    --throw level, use only easy throw patterns, with a couple two-bad patterns
    elseif Constants.CurrentGameLevel == Constants.ThrowTutorialLevel then
        while PatternManager.mBitPatternHelpIndex == iLastIndex do
            PatternManager.mBitPatternHelpIndex = math.random(0, 14) 
        end
        
        if PatternManager.mBitPatternHelpIndex == 0 then PatternManager:LoadPattern( "Patterns/steady_up.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 1 then PatternManager:LoadPattern( "Patterns/steady_down.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 2 then PatternManager:LoadPattern( "Patterns/criss_cross_left.txt" )          
        elseif PatternManager.mBitPatternHelpIndex == 3 then PatternManager:LoadPattern( "Patterns/criss_cross_right.txt" )              
        elseif PatternManager.mBitPatternHelpIndex == 4 then PatternManager:LoadPattern( "Patterns/wait_left.txt" )                    
        elseif PatternManager.mBitPatternHelpIndex == 5 then PatternManager:LoadPattern( "Patterns/wait_right.txt" )    
        elseif PatternManager.mBitPatternHelpIndex == 6 then PatternManager:LoadPattern( "Patterns/leftspin_3.txt" )                    
        elseif PatternManager.mBitPatternHelpIndex == 7 then PatternManager:LoadPattern( "Patterns/leftspin_down_e.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 8 then PatternManager:LoadPattern( "Patterns/leftspin_up_e.txt" )     
        elseif PatternManager.mBitPatternHelpIndex == 9 then PatternManager:LoadPattern( "Patterns/rightspin_3.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 10 then PatternManager:LoadPattern( "Patterns/rightspin_down_e.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 11 then PatternManager:LoadPattern( "Patterns/rightspin_up_e.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 12 then PatternManager:LoadPattern( "Patterns/top_bottom_spin.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 13 then PatternManager:LoadPattern( "Patterns/side_to_side_l.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 14 then PatternManager:LoadPattern( "Patterns/side_to_side_r.txt" )
        end 
    --interim level, use harder throw and bring in advanced patterns
    elseif Constants.CurrentGameLevel == Constants.HardThrowLevel then
        while PatternManager.mBitPatternHelpIndex == iLastIndex do
            PatternManager.mBitPatternHelpIndex = math.random(0, 12) 
        end
        
        if PatternManager.mBitPatternHelpIndex == 0 then PatternManager:LoadPattern( "Patterns/up_down_up.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 1 then PatternManager:LoadPattern( "Patterns/leftspin_down.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 2 then PatternManager:LoadPattern( "Patterns/leftspin_up.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 3 then PatternManager:LoadPattern( "Patterns/rightspin_down.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 4 then PatternManager:LoadPattern( "Patterns/rightspin_up.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 5 then PatternManager:LoadPattern( "Patterns/out_n_in_fs.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 6 then PatternManager:LoadPattern( "Patterns/out_n_in_bs.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 7 then PatternManager:LoadPattern( "Patterns/spin_top_left.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 8 then PatternManager:LoadPattern( "Patterns/spin_top_right.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 9 then PatternManager:LoadPattern( "Patterns/fs_rotate_ccw.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 10 then PatternManager:LoadPattern( "Patterns/bs_rotate_cw.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 11 then PatternManager:LoadPattern( "Patterns/back_forth_bottom.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 12 then PatternManager:LoadPattern( "Patterns/back_forth_top.txt" )
        end        
    --only the build section, do previous level difficulty
    elseif Constants.CurrentGameLevel == Constants.PointBothLevel then
        while PatternManager.mBitPatternHelpIndex == iLastIndex do
            PatternManager.mBitPatternHelpIndex = math.random(0, 11) 
        end
        
        if PatternManager.mBitPatternHelpIndex == 0 then PatternManager:LoadPattern( "Patterns/up_down_up.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 1 then PatternManager:LoadPattern( "Patterns/leftspin_down.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 2 then PatternManager:LoadPattern( "Patterns/rightspin_up.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 3 then PatternManager:LoadPattern( "Patterns/out_n_in_fs.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 4 then PatternManager:LoadPattern( "Patterns/spin_top_left.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 5 then PatternManager:LoadPattern( "Patterns/spin_top_right.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 6 then PatternManager:LoadPattern( "Patterns/fs_rotate_ccw.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 7 then PatternManager:LoadPattern( "Patterns/bs_rotate_cw.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 8 then PatternManager:LoadPattern( "Patterns/back_forth_bottom.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 9 then PatternManager:LoadPattern( "Patterns/back_forth_top.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 10 then PatternManager:LoadPattern( "Patterns/slash_left.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 11 then PatternManager:LoadPattern( "Patterns/slash_right.txt" )
        end 
    --final levels, use all advanced patterns
    elseif Constants.CurrentGameLevel == Constants.AllLevel or Constants.CurrentGameLevel == Constants.NumGameLevels then
        while PatternManager.mBitPatternHelpIndex == iLastIndex do
            PatternManager.mBitPatternHelpIndex = math.random(0, 21) 
        end
        
        if PatternManager.mBitPatternHelpIndex == 0 then PatternManager:LoadPattern( "Patterns/diamond_bs.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 1 then PatternManager:LoadPattern( "Patterns/diamond_fs.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 2 then PatternManager:LoadPattern( "Patterns/eggslicer_right_up.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 3 then PatternManager:LoadPattern( "Patterns/eggslicer_right_down.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 4 then PatternManager:LoadPattern( "Patterns/eggslicer_left_down.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 5 then PatternManager:LoadPattern( "Patterns/eggslicer_left_down.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 6 then PatternManager:LoadPattern( "Patterns/slash_left.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 7 then PatternManager:LoadPattern( "Patterns/slash_right.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 8 then PatternManager:LoadPattern( "Patterns/leftspin_down.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 9 then PatternManager:LoadPattern( "Patterns/leftspin_up.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 10 then PatternManager:LoadPattern( "Patterns/rightspin_down.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 11 then PatternManager:LoadPattern( "Patterns/rightspin_up.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 12 then PatternManager:LoadPattern( "Patterns/out_n_in_fs.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 13 then PatternManager:LoadPattern( "Patterns/out_n_in_bs.txt" )        
        elseif PatternManager.mBitPatternHelpIndex == 14 then PatternManager:LoadPattern( "Patterns/hold_split_ccw.txt" )        
        elseif PatternManager.mBitPatternHelpIndex == 15 then PatternManager:LoadPattern( "Patterns/hold_split_cw.txt" )      
        elseif PatternManager.mBitPatternHelpIndex == 16 then PatternManager:LoadPattern( "Patterns/hold_horiz_slash_ccw.txt" )        
        elseif PatternManager.mBitPatternHelpIndex == 17 then PatternManager:LoadPattern( "Patterns/hold_horiz_slash_cw.txt" )        
        elseif PatternManager.mBitPatternHelpIndex == 18 then PatternManager:LoadPattern( "Patterns/hold_one_two_ccw.txt" )        
        elseif PatternManager.mBitPatternHelpIndex == 19 then PatternManager:LoadPattern( "Patterns/hold_one_two_cw.txt" )        
        elseif PatternManager.mBitPatternHelpIndex == 20 then PatternManager:LoadPattern( "Patterns/all_down_up.txt" )        
        elseif PatternManager.mBitPatternHelpIndex == 21 then PatternManager:LoadPattern( "Patterns/all_up_down.txt" )        
        end     
    end   
end
  
function PatternManager:PickBonusSet()    
    PatternManager.mBitPatternHelpIndex = PatternManager.mBitPatternHelpIndex + 1
    
    if Constants.CurrentGameLevel == 1 then
        PatternManager:LoadPattern( "Patterns/star.txt" )

    elseif Constants.CurrentGameLevel == Constants.PointHorizLevel then
        if PatternManager.mBitPatternHelpIndex == 2 then PatternManager.mBitPatternHelpIndex = 0 end    
        if PatternManager.mBitPatternHelpIndex == 0 then PatternManager:LoadPattern( "Patterns/circle_cw.txt" ) 
        elseif PatternManager.mBitPatternHelpIndex == 1 then PatternManager:LoadPattern( "Patterns/circle_ccw.txt" )
        end
        
    elseif Constants.CurrentGameLevel == Constants.HoldTutorialLevel then
        if PatternManager.mBitPatternHelpIndex == 2 then PatternManager.mBitPatternHelpIndex = 0 end    
        if PatternManager.mBitPatternHelpIndex == 0 then PatternManager:LoadPattern( "Patterns/hold_left_bonus.txt" ) 
        elseif PatternManager.mBitPatternHelpIndex == 1 then PatternManager:LoadPattern( "Patterns/hold_right_bonus.txt" )
        end
         
    elseif Constants.CurrentGameLevel == Constants.HardHoldLevel then
        if PatternManager.mBitPatternHelpIndex == 2 then PatternManager.mBitPatternHelpIndex = 0 end    
        if PatternManager.mBitPatternHelpIndex == 0 then PatternManager:LoadPattern( "Patterns/hold_bonus_bs.txt" ) 
        elseif PatternManager.mBitPatternHelpIndex == 1 then PatternManager:LoadPattern( "Patterns/hold_bonus_fs.txt" )
        end
                 
    elseif Constants.CurrentGameLevel == Constants.PointVertLevel then
        if PatternManager.mBitPatternHelpIndex == 2 then PatternManager.mBitPatternHelpIndex = 0 end    
        if PatternManager.mBitPatternHelpIndex == 0 then PatternManager:LoadPattern( "Patterns/bonus_hidden_a.txt" ) 
        elseif PatternManager.mBitPatternHelpIndex == 1 then PatternManager:LoadPattern( "Patterns/bonus_hidden_b.txt" )
        end
                 
    elseif Constants.CurrentGameLevel == Constants.ThrowTutorialLevel then
        if PatternManager.mBitPatternHelpIndex == 2 then PatternManager.mBitPatternHelpIndex = 0 end 
        if PatternManager.mBitPatternHelpIndex == 0 then PatternManager:LoadPattern( "Patterns/circle_slice_cw.txt" ) 
        elseif PatternManager.mBitPatternHelpIndex == 1 then PatternManager:LoadPattern( "Patterns/circle_slice_ccw.txt" )
        end  
        
    elseif Constants.CurrentGameLevel == Constants.HardThrowLevel then
        if PatternManager.mBitPatternHelpIndex == 2 then PatternManager.mBitPatternHelpIndex = 0 end 
        if PatternManager.mBitPatternHelpIndex == 0 then PatternManager:LoadPattern( "Patterns/bonus_fan_top.txt" ) 
        elseif PatternManager.mBitPatternHelpIndex == 1 then PatternManager:LoadPattern( "Patterns/bonus_fan_bottom.txt" )
        end 
    
    elseif Constants.CurrentGameLevel == Constants.PointBothLevel then
        if PatternManager.mBitPatternHelpIndex == 2 then PatternManager.mBitPatternHelpIndex = 0 end 
        if PatternManager.mBitPatternHelpIndex == 0 then PatternManager:LoadPattern( "Patterns/bonus_hold_down.txt" ) 
        elseif PatternManager.mBitPatternHelpIndex == 1 then PatternManager:LoadPattern( "Patterns/bonus_hold_up.txt" )
        end 

    elseif Constants.CurrentGameLevel == Constants.AllLevel then
        if PatternManager.mBitPatternHelpIndex == 2 then PatternManager.mBitPatternHelpIndex = 0 end 
        if PatternManager.mBitPatternHelpIndex == 0 then PatternManager:LoadPattern( "Patterns/bonus_bad_top.txt" ) 
        elseif PatternManager.mBitPatternHelpIndex == 1 then PatternManager:LoadPattern( "Patterns/bonus_bad_bottom.txt" )
        end  

    elseif Constants.CurrentGameLevel == Constants.NumGameLevels then
        if PatternManager.mBitPatternHelpIndex == 9 then PatternManager.mBitPatternHelpIndex = 0 end 
        if PatternManager.mBitPatternHelpIndex == 0 then PatternManager:LoadPattern( "Patterns/bonus_d.txt" ) 
        elseif PatternManager.mBitPatternHelpIndex == 1 then PatternManager:LoadPattern( "Patterns/bonus_r.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 2 then PatternManager:LoadPattern( "Patterns/bonus_o.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 3 then PatternManager:LoadPattern( "Patterns/bonus_p.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 4 then PatternManager:LoadPattern( "Patterns/bonus_c.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 5 then PatternManager:LoadPattern( "Patterns/bonus_h.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 6 then PatternManager:LoadPattern( "Patterns/bonus_o.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 7 then PatternManager:LoadPattern( "Patterns/bonus_r.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 8 then PatternManager:LoadPattern( "Patterns/bonus_d.txt" )
        end 
    end
end
  
function PatternManager:PickFullMixSet()
    --pick range according to time
    local iLastIndex = PatternManager.mBitPatternHelpIndex
    if PatternManager.mFullMixTimer < 30.0 then    
        while PatternManager.mBitPatternHelpIndex == iLastIndex do PatternManager.mBitPatternHelpIndex = math.random(0, 13) end    
        
        if PatternManager.mBitPatternHelpIndex == 0 then PatternManager:LoadPattern( "Patterns/intro6c.txt" ) 
        elseif PatternManager.mBitPatternHelpIndex == 1 then PatternManager:LoadPattern( "Patterns/intro6d.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 2 then PatternManager:LoadPattern( "Patterns/five_bottom.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 3 then PatternManager:LoadPattern( "Patterns/five_top.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 4 then PatternManager:LoadPattern( "Patterns/bigslash_ul.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 5 then PatternManager:LoadPattern( "Patterns/bigslash_ur.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 6 then PatternManager:LoadPattern( "Patterns/triangle_low.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 7 then PatternManager:LoadPattern( "Patterns/triangle_high.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 8 then PatternManager:LoadPattern( "Patterns/cross_up.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 9 then PatternManager:LoadPattern( "Patterns/cross_down.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 10 then PatternManager:LoadPattern( "Patterns/long_slash_r.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 11 then PatternManager:LoadPattern( "Patterns/long_slash_l.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 12 then PatternManager:LoadPattern( "Patterns/vert_slash_r.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 13 then PatternManager:LoadPattern( "Patterns/vert_slash_l.txt" ) 
        end    
    elseif PatternManager.mFullMixTimer < 60.0 then    
        while PatternManager.mBitPatternHelpIndex == iLastIndex do PatternManager.mBitPatternHelpIndex = math.random(0, 11) end    
        
        if PatternManager.mBitPatternHelpIndex == 0 then PatternManager:LoadPattern( "Patterns/hold_pinch_bl.txt" ) 
        elseif PatternManager.mBitPatternHelpIndex == 1 then PatternManager:LoadPattern( "Patterns/hold_pinch_tr.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 2 then PatternManager:LoadPattern( "Patterns/five_bottom.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 3 then PatternManager:LoadPattern( "Patterns/five_top.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 4 then PatternManager:LoadPattern( "Patterns/bigslash_ul.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 5 then PatternManager:LoadPattern( "Patterns/bigslash_ur.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 6 then PatternManager:LoadPattern( "Patterns/all_fall.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 7 then PatternManager:LoadPattern( "Patterns/all_rise.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 8 then PatternManager:LoadPattern( "Patterns/cross_up.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 9 then PatternManager:LoadPattern( "Patterns/cross_down.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 10 then PatternManager:LoadPattern( "Patterns/long_slash_r.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 11 then PatternManager:LoadPattern( "Patterns/long_slash_l.txt" )
        end       
    elseif PatternManager.mFullMixTimer < 120.0 then
        while PatternManager.mBitPatternHelpIndex == iLastIndex do PatternManager.mBitPatternHelpIndex = math.random(0, 11) end    
        
        if PatternManager.mBitPatternHelpIndex == 0 then PatternManager:LoadPattern( "Patterns/five_top.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 1 then PatternManager:LoadPattern( "Patterns/five_bottom.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 2 then PatternManager:LoadPattern( "Patterns/all_fall.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 3 then PatternManager:LoadPattern( "Patterns/all_rise.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 4 then PatternManager:LoadPattern( "Patterns/hold_down_right.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 5 then PatternManager:LoadPattern( "Patterns/hold_up_left.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 6 then PatternManager:LoadPattern( "Patterns/hold_box_cw.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 7 then PatternManager:LoadPattern( "Patterns/hold_box_ccw.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 8 then PatternManager:LoadPattern( "Patterns/hold_pinch_bl.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 9 then PatternManager:LoadPattern( "Patterns/hold_pinch_tr.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 10 then PatternManager:LoadPattern( "Patterns/hold_bad_bs.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 11 then PatternManager:LoadPattern( "Patterns/hold_bad_fs.txt" )      
        end   
    elseif PatternManager.mFullMixTimer < 150.0 then
        while PatternManager.mBitPatternHelpIndex == iLastIndex do PatternManager.mBitPatternHelpIndex = math.random(0, 11) end    
        
        if PatternManager.mBitPatternHelpIndex == 0 then PatternManager:LoadPattern( "Patterns/keyturn_ccw.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 1 then PatternManager:LoadPattern( "Patterns/keyturn_cw.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 2 then PatternManager:LoadPattern( "Patterns/hold_center_bs.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 3 then PatternManager:LoadPattern( "Patterns/hold_center_fs.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 4 then PatternManager:LoadPattern( "Patterns/hold_down_right.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 5 then PatternManager:LoadPattern( "Patterns/hold_up_left.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 6 then PatternManager:LoadPattern( "Patterns/hold_box_cw.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 7 then PatternManager:LoadPattern( "Patterns/hold_box_ccw.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 8 then PatternManager:LoadPattern( "Patterns/side_to_side_left_e.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 9 then PatternManager:LoadPattern( "Patterns/side_to_side_right_e.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 10 then PatternManager:LoadPattern( "Patterns/hold_bad_bs.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 11 then PatternManager:LoadPattern( "Patterns/hold_bad_fs.txt" )      
        end         
    elseif PatternManager.mFullMixTimer < 180.0 then
        while PatternManager.mBitPatternHelpIndex == iLastIndex do PatternManager.mBitPatternHelpIndex = math.random(0, 11) end
                
        if PatternManager.mBitPatternHelpIndex == 0 then PatternManager:LoadPattern( "Patterns/keyturn_ccw.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 1 then PatternManager:LoadPattern( "Patterns/keyturn_cw.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 2 then PatternManager:LoadPattern( "Patterns/bs_to_fs.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 3 then PatternManager:LoadPattern( "Patterns/fs_to_bs.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 4 then PatternManager:LoadPattern( "Patterns/hold_steady_up.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 5 then PatternManager:LoadPattern( "Patterns/hold_steady_down.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 6 then PatternManager:LoadPattern( "Patterns/side_to_side_left_e.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 7 then PatternManager:LoadPattern( "Patterns/side_to_side_right_e.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 8 then PatternManager:LoadPattern( "Patterns/pin_bl.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 9 then PatternManager:LoadPattern( "Patterns/pin_tr.txt" )      
        elseif PatternManager.mBitPatternHelpIndex == 10 then PatternManager:LoadPattern( "Patterns/hold_center_bs.txt" )      
        elseif PatternManager.mBitPatternHelpIndex == 11 then PatternManager:LoadPattern( "Patterns/hold_center_fs.txt" )      
        end 
    elseif PatternManager.mFullMixTimer < 210.0 then
        --first pattern with slices; make sure throw is unlocked!
        Constants.ThrowUnlocked = true
        
        while PatternManager.mBitPatternHelpIndex == iLastIndex do PatternManager.mBitPatternHelpIndex = math.random(0, 11) end

        if PatternManager.mBitPatternHelpIndex == 0 then PatternManager:LoadPattern( "Patterns/leftspin_3.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 1 then PatternManager:LoadPattern( "Patterns/rightspin_3.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 2 then PatternManager:LoadPattern( "Patterns/bs_to_fs.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 3 then PatternManager:LoadPattern( "Patterns/fs_to_bs.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 4 then PatternManager:LoadPattern( "Patterns/hold_steady_up.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 5 then PatternManager:LoadPattern( "Patterns/hold_steady_down.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 6 then PatternManager:LoadPattern( "Patterns/rightspin_down_e.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 7 then PatternManager:LoadPattern( "Patterns/rightspin_up_e.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 8 then PatternManager:LoadPattern( "Patterns/pin_bl.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 9 then PatternManager:LoadPattern( "Patterns/pin_tr.txt" )      
        elseif PatternManager.mBitPatternHelpIndex == 10 then PatternManager:LoadPattern( "Patterns/hold_center_bs.txt" )      
        elseif PatternManager.mBitPatternHelpIndex == 11 then PatternManager:LoadPattern( "Patterns/hold_center_fs.txt" )      
        end        
    elseif PatternManager.mFullMixTimer < 240.0 then
        while PatternManager.mBitPatternHelpIndex == iLastIndex do PatternManager.mBitPatternHelpIndex = math.random(0, 12) end
        
        if PatternManager.mBitPatternHelpIndex == 0 then PatternManager:LoadPattern( "Patterns/steady_up.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 1 then PatternManager:LoadPattern( "Patterns/steady_down.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 2 then PatternManager:LoadPattern( "Patterns/criss_cross_left.txt" )          
        elseif PatternManager.mBitPatternHelpIndex == 3 then PatternManager:LoadPattern( "Patterns/criss_cross_right.txt" )              
        elseif PatternManager.mBitPatternHelpIndex == 4 then PatternManager:LoadPattern( "Patterns/wait_left.txt" )                    
        elseif PatternManager.mBitPatternHelpIndex == 5 then PatternManager:LoadPattern( "Patterns/wait_right.txt" )    
        elseif PatternManager.mBitPatternHelpIndex == 6 then PatternManager:LoadPattern( "Patterns/leftspin_3.txt" )                    
        elseif PatternManager.mBitPatternHelpIndex == 7 then PatternManager:LoadPattern( "Patterns/leftspin_down_e.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 8 then PatternManager:LoadPattern( "Patterns/leftspin_up_e.txt" )     
        elseif PatternManager.mBitPatternHelpIndex == 9 then PatternManager:LoadPattern( "Patterns/rightspin_3.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 10 then PatternManager:LoadPattern( "Patterns/rightspin_down_e.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 11 then PatternManager:LoadPattern( "Patterns/rightspin_up_e.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 12 then PatternManager:LoadPattern( "Patterns/top_bottom_spin.txt" )
        end     
    elseif PatternManager.mFullMixTimer < 270.0 then
        while PatternManager.mBitPatternHelpIndex == iLastIndex do PatternManager.mBitPatternHelpIndex = math.random(0, 14) end
        
        if PatternManager.mBitPatternHelpIndex == 0 then PatternManager:LoadPattern( "Patterns/steady_up.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 1 then PatternManager:LoadPattern( "Patterns/steady_down.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 2 then PatternManager:LoadPattern( "Patterns/back_forth_bottom.txt" )          
        elseif PatternManager.mBitPatternHelpIndex == 3 then PatternManager:LoadPattern( "Patterns/back_forth_top.txt" )              
        elseif PatternManager.mBitPatternHelpIndex == 4 then PatternManager:LoadPattern( "Patterns/wait_left.txt" )                    
        elseif PatternManager.mBitPatternHelpIndex == 5 then PatternManager:LoadPattern( "Patterns/wait_right.txt" )    
        elseif PatternManager.mBitPatternHelpIndex == 6 then PatternManager:LoadPattern( "Patterns/leftspin_3.txt" )                    
        elseif PatternManager.mBitPatternHelpIndex == 7 then PatternManager:LoadPattern( "Patterns/leftspin_down.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 8 then PatternManager:LoadPattern( "Patterns/leftspin_up_e.txt" )     
        elseif PatternManager.mBitPatternHelpIndex == 9 then PatternManager:LoadPattern( "Patterns/rightspin_3.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 10 then PatternManager:LoadPattern( "Patterns/rightspin_down.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 11 then PatternManager:LoadPattern( "Patterns/rightspin_up_e.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 12 then PatternManager:LoadPattern( "Patterns/up_down_up.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 13 then PatternManager:LoadPattern( "Patterns/side_to_side_l.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 14 then PatternManager:LoadPattern( "Patterns/side_to_side_r.txt" )        
        end        
    elseif PatternManager.mFullMixTimer < 300.0 then
        while PatternManager.mBitPatternHelpIndex == iLastIndex do PatternManager.mBitPatternHelpIndex = math.random(0, 12) end
        
        if PatternManager.mBitPatternHelpIndex == 0 then PatternManager:LoadPattern( "Patterns/up_down_up.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 1 then PatternManager:LoadPattern( "Patterns/leftspin_down.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 2 then PatternManager:LoadPattern( "Patterns/leftspin_up.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 3 then PatternManager:LoadPattern( "Patterns/rightspin_down.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 4 then PatternManager:LoadPattern( "Patterns/rightspin_up.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 5 then PatternManager:LoadPattern( "Patterns/out_n_in_fs.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 6 then PatternManager:LoadPattern( "Patterns/out_n_in_bs.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 7 then PatternManager:LoadPattern( "Patterns/spin_top_left.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 8 then PatternManager:LoadPattern( "Patterns/spin_top_right.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 9 then PatternManager:LoadPattern( "Patterns/fs_rotate_ccw.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 10 then PatternManager:LoadPattern( "Patterns/bs_rotate_cw.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 11 then PatternManager:LoadPattern( "Patterns/back_forth_bottom.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 12 then PatternManager:LoadPattern( "Patterns/back_forth_top.txt" )
        end     
    else
        while PatternManager.mBitPatternHelpIndex == iLastIndex do PatternManager.mBitPatternHelpIndex = math.random(0, 21) end
        
        if PatternManager.mBitPatternHelpIndex == 0 then PatternManager:LoadPattern( "Patterns/diamond_bs.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 1 then PatternManager:LoadPattern( "Patterns/diamond_fs.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 2 then PatternManager:LoadPattern( "Patterns/eggslicer_right_up.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 3 then PatternManager:LoadPattern( "Patterns/eggslicer_right_down.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 4 then PatternManager:LoadPattern( "Patterns/eggslicer_left_down.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 5 then PatternManager:LoadPattern( "Patterns/eggslicer_left_down.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 6 then PatternManager:LoadPattern( "Patterns/slash_left.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 7 then PatternManager:LoadPattern( "Patterns/slash_right.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 8 then PatternManager:LoadPattern( "Patterns/leftspin_down.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 9 then PatternManager:LoadPattern( "Patterns/leftspin_up.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 10 then PatternManager:LoadPattern( "Patterns/rightspin_down.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 11 then PatternManager:LoadPattern( "Patterns/rightspin_up.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 12 then PatternManager:LoadPattern( "Patterns/out_n_in_fs.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 13 then PatternManager:LoadPattern( "Patterns/out_n_in_bs.txt" )        
        elseif PatternManager.mBitPatternHelpIndex == 14 then PatternManager:LoadPattern( "Patterns/hold_split_ccw.txt" )        
        elseif PatternManager.mBitPatternHelpIndex == 15 then PatternManager:LoadPattern( "Patterns/hold_split_cw.txt" )      
        elseif PatternManager.mBitPatternHelpIndex == 16 then PatternManager:LoadPattern( "Patterns/hold_horiz_slash_ccw.txt" )        
        elseif PatternManager.mBitPatternHelpIndex == 17 then PatternManager:LoadPattern( "Patterns/hold_horiz_slash_cw.txt" )        
        elseif PatternManager.mBitPatternHelpIndex == 18 then PatternManager:LoadPattern( "Patterns/hold_one_two_ccw.txt" )        
        elseif PatternManager.mBitPatternHelpIndex == 19 then PatternManager:LoadPattern( "Patterns/hold_one_two_cw.txt" )        
        elseif PatternManager.mBitPatternHelpIndex == 20 then PatternManager:LoadPattern( "Patterns/all_down_up.txt" )        
        elseif PatternManager.mBitPatternHelpIndex == 21 then PatternManager:LoadPattern( "Patterns/all_up_down.txt" )        
        end          
    end
end

function PatternManager:PickNewPointSet()
    --[[
    PatternManager:LoadPattern( "Patterns/pattern.txt" )
    PatternManager:ConfigPatternForPoint( 3 )
    --]]
    
    -- [[
    if PatternManager.mPatternType == Constants.PatternTypePointHoriz then    
        if PatternManager.mBitPatternCount == 0 then
            PatternManager:LoadPattern( "Patterns/point0.txt" ) 
            PatternManager:MakePatternLong()
            PatternManager:ConfigPatternForPoint( 1 )
            PatternManager.mPrepForTutorial = true 
            PatternManager.mTutorialType = Constants.mTutorialTypePoint
        elseif PatternManager.mBitPatternCount == 1 then
            PatternManager:LoadPattern( "Patterns/point1.txt" ) 
            PatternManager:MakePatternLong()
            PatternManager:ConfigPatternForPoint( 2 )   
        elseif PatternManager.mBitPatternCount == 2 then
            PatternManager:LoadPattern( "Patterns/point2.txt" ) 
            PatternManager:MakePatternLong()
            PatternManager:ConfigPatternForPoint( 3 )            
        else               
            local iLastIndex = PatternManager.mBitPatternHelpIndex    
            while PatternManager.mBitPatternHelpIndex == iLastIndex do
                PatternManager.mBitPatternHelpIndex = math.random(0, 8)
            end
            
            if PatternManager.mBitPatternHelpIndex == 0 then PatternManager:LoadPattern( "Patterns/point_still0.txt" ) 
            elseif PatternManager.mBitPatternHelpIndex == 1 then PatternManager:LoadPattern( "Patterns/point_still1.txt" )
            elseif PatternManager.mBitPatternHelpIndex == 2 then PatternManager:LoadPattern( "Patterns/point_still2.txt" )
            elseif PatternManager.mBitPatternHelpIndex == 3 then PatternManager:LoadPattern( "Patterns/point_still3.txt" )
            elseif PatternManager.mBitPatternHelpIndex == 4 then PatternManager:LoadPattern( "Patterns/point_still4.txt" )
            elseif PatternManager.mBitPatternHelpIndex == 5 then PatternManager:LoadPattern( "Patterns/point_still5.txt" )           
            elseif PatternManager.mBitPatternHelpIndex == 6 then PatternManager:LoadPattern( "Patterns/point_still6.txt" )           
            elseif PatternManager.mBitPatternHelpIndex == 7 then PatternManager:LoadPattern( "Patterns/point_still7.txt" )           
            elseif PatternManager.mBitPatternHelpIndex == 8 then PatternManager:LoadPattern( "Patterns/point_still8.txt" )           
            end
            
            PatternManager:ConfigPatternForPoint( 3 )
        end
    else
        local iLastIndex = PatternManager.mBitPatternHelpIndex
        while PatternManager.mBitPatternHelpIndex == iLastIndex do
            PatternManager.mBitPatternHelpIndex = math.random(0, 8) 
        end        
        
        if PatternManager.mBitPatternHelpIndex == 0 then PatternManager:LoadPattern( "Patterns/point_move0.txt" ) 
        elseif PatternManager.mBitPatternHelpIndex == 1 then PatternManager:LoadPattern( "Patterns/point_move1.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 2 then PatternManager:LoadPattern( "Patterns/point_move2.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 3 then PatternManager:LoadPattern( "Patterns/point_move3.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 4 then PatternManager:LoadPattern( "Patterns/point_move4.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 5 then PatternManager:LoadPattern( "Patterns/point_move5.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 6 then PatternManager:LoadPattern( "Patterns/point_short0.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 7 then PatternManager:LoadPattern( "Patterns/point_short1.txt" )
        elseif PatternManager.mBitPatternHelpIndex == 8 then PatternManager:LoadPattern( "Patterns/point_short2.txt" )
        end 
 
        PatternManager:ConfigPatternForPoint( math.random(2, 6) )
    end
    --]]  
    
    PatternManager.mBitPatternCount = PatternManager.mBitPatternCount + 1
    PatternManager.mPatternFlawless = true    
    PatternManager.mHadMultiplier = Score:HasMultiplier()
end

function PatternManager:SetBonusPattern()
    PatternManager:SetPatternType( Constants.PatternTypeBonus )
    PatternManager.mBitPatternHelpIndex = -1
    PatternManager:CleanPattern()
end
  
function PatternManager:PrimeForNewLevel()
    --clear the tutorial type
    PatternManager.mTutorialType = Constants.mTutorialTypeNone
    PatternManager.mPrepForTutorial = false
    PatternManager.mPrepTutorialTimer = 0.0
    PatternManager.mShowTutorial = false
    
    --set pattern type according to level
    if Constants.CurrentGameLevel == Constants.ThrowTutorialLevel then
        PatternManager:SetPatternType( Constants.PatternTypeLearnThrow )
    elseif Constants.CurrentGameLevel == Constants.HoldTutorialLevel then
        PatternManager:SetPatternType( Constants.PatternTypeLearnHold )
    elseif Constants.CurrentGameLevel == Constants.PointHorizLevel then
        PatternManager:SetPatternType( Constants.PatternTypePointHoriz )
    elseif Constants.CurrentGameLevel == Constants.PointVertLevel then
        PatternManager:SetPatternType( Constants.PatternTypePointVert )    
    elseif Constants.CurrentGameLevel == Constants.PointBothLevel then
        PatternManager:SetPatternType( Constants.PatternTypePointBoth ) 
    else   
        PatternManager:SetPatternType( Constants.PatternTypeStandard )
    end
    
    --if we're switching to the last level, allow super tilt (without audio distortion)
    if Constants.CurrentGameLevel == Constants.NumGameLevels then
        Constants.GameCameraTiltDesiredMode = Constants.GameCameraTiltModeTiltBothNoAudio
    end
    
    PatternManager:ResetBitsAndSlices()
    PatternManager:ResetQueue()
    PatternManager.mPatternFlawless = false
    PatternManager.mHadMultiplier = false
    PatternManager.mBitPatternCount = 0
    PatternManager.mNewBitDuration = 1.0
    PatternManager.mNewBitTimer = PatternManager.mNewBitDuration
    
    PatternManager.mNumPatternsAttempted = 0
    PatternManager.mNumPatternsPerfect = 0
end
  
function PatternManager:CleanPattern()
    PatternManager:ResetBitsAndSlices()        
    PatternManager:ResetQueue()
    PatternManager.mFanfareTimer = 0.0
    PatternManager.mPatternState = Constants.PatternStateStandard       
    PatternManager:SetNewBitPattern()  
    PatternManager:DelayPattern(1.0)    
end
  
function PatternManager:SetGameSpeed( fIntensity )
    --game speed is time based in full mix mode
    if Constants.FullMixMode then
        --soft timer to make game as hard as last level by minute 6
        local fHardTime = 360.0
        local fRatioToHard = PatternManager.mFullMixTimer / fHardTime
        
        PatternManager.mBasePatternStep = DFMath.lerp( 0.75, 0.5, fRatioToHard )
        PatternManager.mPatternSpeed = DFMath.lerp( 1.0, 2.5, fRatioToHard )
    else
        --levels 1, 2
        if Constants.CurrentGameLevel < 3 then
            PatternManager.mBasePatternStep = 0.75
            PatternManager.mPatternSpeed = DFMath.lerp( 1.0, 1.5, fIntensity )
        --levels 3, 4
        elseif Constants.CurrentGameLevel < 5 then
            PatternManager.mBasePatternStep = 0.65
            PatternManager.mPatternSpeed = DFMath.lerp( 1.2, 1.7, fIntensity )
        --levels 5, 6
        elseif Constants.CurrentGameLevel < 7 then
            PatternManager.mBasePatternStep = 0.6
            PatternManager.mPatternSpeed = DFMath.lerp( 1.4, 1.9, fIntensity )
        --levels 7, 8
        elseif Constants.CurrentGameLevel < 9 then
            PatternManager.mBasePatternStep = 0.55
            PatternManager.mPatternSpeed = DFMath.lerp( 1.6, 2.1, fIntensity )
        --levels 9, 10
        else
            PatternManager.mBasePatternStep = 0.5
            PatternManager.mPatternSpeed = DFMath.lerp( 1.8, 2.5, fIntensity )
        end
    end
end
  
function PatternManager:CheckCollision( tChargeManager, tRenderer, tGameInputs )
    if PatternManager.mPatternType < Constants.PatternTypePointHoriz then
        PatternManager:CheckCollisionWithBeam( tChargeManager, tRenderer )
    elseif PatternManager.mPatternType ~= Constants.PatternTypeComplete then
        PatternManager:CheckCollisionWithInputs( tGameInputs, tRenderer )
    end
end

function PatternManager:CheckCollisionWithInputs( tGameInputs, tRenderer )
    --check inputs versus *assumed* point bits
    local i = 1
    local j = 1
    for i = 1, PatternManager.mNumBits do
        if PatternManager.mBits[i]:CanBeDestroyed() then    
            local fBitX, fBitY = PatternManager.mBits[i]:GetCurrentPos()
            local fBitRadius = PatternManager.mBits[i].mWidth
            local fBitRadSquared = fBitRadius * fBitRadius
            
            --run through all the inputs and see if they're inside the bit
            local bAtLeastOneJustWentDown = false
            local iNumCollisions = 0
            for j = 1, Constants.NumGameInputs do
                if tGameInputs[j]:IsTouched() then                    
                    if DFMath.distance2DSquared( fBitX, fBitY, tGameInputs[j].mX, tGameInputs[j].mY ) < fBitRadSquared then
                        if tGameInputs[j].mFirstTouch then
                            bAtLeastOneJustWentDown = true
                        end                    
                    
                        iNumCollisions = iNumCollisions + 1
                    end
                end
            end
            
            --if the correct number are colliding, flag it!
            if bAtLeastOneJustWentDown and iNumCollisions == PatternManager.mBits[i].mPointCount then
                PatternManager.mBits[i]:GetHoldRatio(2)    
                PatternManager:ExplodeBit(i, tRenderer)                
            end
        end
    end
end

function PatternManager:CheckCollisionWithBeam( tChargeManager, tRenderer )
    --check beam versus all bits
    local i = 1
    local f1PosX = tChargeManager.mCharges[1].mPos[1]
    local f1PosY = tChargeManager.mCharges[1].mPos[2]   
    local f2PosX = tChargeManager.mCharges[2].mPos[1]
    local f2PosY = tChargeManager.mCharges[2].mPos[2] 
    local f1PrvPosX = tChargeManager.mCharges[1].mPrevPos[1]
    local f1PrvPosY = tChargeManager.mCharges[1].mPrevPos[2]   
    local f2PrvPosX = tChargeManager.mCharges[2].mPrevPos[1]
    local f2PrvPosY = tChargeManager.mCharges[2].mPrevPos[2] 
                
    DFRad:InitCircleForIntersectionChecks(0.0, 0.0, Constants.GameRingRadius)            
    DFRad:InitBeamForIntersectionChecks(f1PosX, f1PosY, f2PosX, f2PosY, f1PrvPosX, f1PrvPosY, f2PrvPosX, f2PrvPosY)
                
    for i = 1, PatternManager.mNumBits do
        if PatternManager.mBits[i]:CanBeDestroyed() then
            local fBitX, fBitY = PatternManager.mBits[i]:GetCurrentPos()
            local fBitRadius = PatternManager.mBits[i].mHalfWidth
            local iBitType = PatternManager.mBits[i].mType
            
            --if we're invulnerable and this is a bad, don't bother checking collision
            if iBitType ~= Constants.BitTypeBad or tChargeManager:IsInvulnerable() == false then
                local bIntersecting = DFRad:BeamBitIntersectionCheck(fBitX, fBitY, fBitRadius)
                if bIntersecting == true then
                    local fBitHoldRatio = PatternManager.mBits[i]:GetHoldRatio(1)                    
                    if fBitHoldRatio <= 0.0 then
                        PatternManager:ExplodeBit(i, tRenderer)
                        
                        --if we just hit a bad, go invulnerable for a bit
                        if iBitType == Constants.BitTypeBad then
                            tChargeManager:MakeInvulnerable()
                        else
                            tChargeManager:FlashBeam()
                        end
                    end
                end
            end
        end
    end 
    
    --check thrown charges against slices
    if tChargeManager.mCharges[1]:IsThrown() then
        PatternManager:CheckSliceCollision( tChargeManager.mCharges[1] )
    end
    
    if tChargeManager.mCharges[2]:IsThrown() then
        PatternManager:CheckSliceCollision( tChargeManager.mCharges[2] )
    end      
end
  
function PatternManager:CheckSliceCollision( tCharge )
    local i = 1
    for i = 1, PatternManager.mNumSlices do
        if PatternManager.mSlices[i].mEnabled and PatternManager.mSlices[i].mActivated == false then
            --check the angle of the charge against the angle of the slice
            local fBaseRotation = tCharge.mRotation
            local fPrevRotation = tCharge.mPrevRotation      
            local fBaseY = tCharge.mPos[2]
            local fPrevY = tCharge.mPrevPos[2]
            
            local j = 1
            local iNumSteps = 10
            local fStep = 1 / iNumSteps
            local fT = fStep
            for j = 1, iNumSteps do
                local fRot = DFMath.lerp( fPrevRotation, fBaseRotation, fT )
                local fY = DFMath.lerp( fPrevY, fBaseY, fT )
                
                --if the charge is above the center line, we need to mess with the rotation angle to get it in to the [0 : 2*PI] space
                if fY < 0.0 then
                    fRot = math.pi + (math.pi - math.abs(fRot)) 
                end
                PatternManager.mSlices[i]:CheckForActivation(fRot)                
            
                fT = fT + fStep
            end

            --slice activated!
            if PatternManager.mSlices[i].mActivated then
                if Constants.LevelState == Constants.LevelStateBonus then
                    AudioManager:PlaySound( 'Radius/SFX/SliceChorus' )
                else
                    AudioManager:PlaySound( 'Radius/SFX/Slice' )
                end
            
                Score:GoodHit(1)
                PatternManager:CheckForQueueAdvancement()
            end
        end
    end   
end
  
function PatternManager:ClearBadBits()
    RadEffects:SetEffectLayerToGameLayer()
    
    local i = 1
    for i = 1, PatternManager.mNumBits do
        if PatternManager.mBits[i].mEnabled then
            if PatternManager.mBits[i].mType == Constants.BitTypeBad then
                local x, y, z = PatternManager.mBits[i].GetCurrentPos()
                RadEffects:SetEffectTransform( x, y, z, 0.0, 0.0, 0.0 )
                RadEffects:CreateEffect( "Effects/Particles/Bursts/ScratchDeath.particles", nil, {Constants.BadEffectColor[1],Constants.BadEffectColor[2],Constants.BadEffectColor[3],1}, nil, Constants.ParticlePriorityLow )
                RadEffects:CreateEffect( "Effects/Particles/Bursts/ScratchDeathHorizontalFlare.particles", nil, {Constants.BadEffectColor[1],Constants.BadEffectColor[2],Constants.BadEffectColor[3],1}, nil, Constants.ParticlePriorityLow )
                RadEffects:CreateEffect( "Effects/Particles/Bursts/ScratchShards.particles", nil, {Constants.BadEffectColor[1],Constants.BadEffectColor[2],Constants.BadEffectColor[3],1}, nil, Constants.ParticlePriorityLow )
                
                PatternManager.mBits[i]:Disable()
            end
        end
    end          
end
  
function PatternManager:ClearFrozenBits()
    local i = 1
    for i = 1, PatternManager.mNumBits do
        if PatternManager.mBits[i].mEnabled then
            if PatternManager.mBits[i].mType == Constants.BitTypeFanFare or PatternManager.mBits[i].mType == Constants.BitTypeFreeze then
                PatternManager.mBits[i]:Disable()
            end
        end
    end          
end

function PatternManager:CheckPatternLeftoversForPerfect()
    local i = 1
    for i = 1, PatternManager.mNumBits do
        if PatternManager.mBits[i].mEnabled then
            local iType = PatternManager.mBits[i].mType
            if iType == Constants.BitTypeGood or iType == Constants.BitTypeHold or iType == Constants.BitTypePoint then
                PatternManager.mPatternFlawless = false
                return
            end
        end
    end    
        
    for i = 1, PatternManager.mNumSlices do
        if PatternManager.mSlices[i].mEnabled and not PatternManager.mSlices[i].mActivated then
            PatternManager.mPatternFlawless = false
            return
        end
    end    
end
  
function PatternManager:EnableBitFanfare()
    local i = 1
    for i = 1, PatternManager.mNumBits do
        if PatternManager.mBits[i].mEnabled and PatternManager.mBits[i].mType == Constants.BitTypeFreeze then
            PatternManager.mBits[i]:SetFanfare()
        end
    end    
    
    for i = 1, PatternManager.mNumSlices do
        if PatternManager.mSlices[i].mEnabled then
            PatternManager.mSlices[i].mFanfare = true
        end
    end    
end
  
function PatternManager:ExplodeBit(iIndex, tRenderer)   
    local fBitX, fBitY = PatternManager.mBits[iIndex]:GetCurrentPos()
    
    PatternManager.mLastDestroyedBit[1] = fBitX
    PatternManager.mLastDestroyedBit[2] = fBitY
    
    local iType = PatternManager.mBits[iIndex].mType
        
    --response to type
    if iType == Constants.BitTypeBad then
        ColorManager:Interrupt()    
        Score:ClearMultiplier()
        Score:BadHit()
        AudioManager:PlaySound( 'Radius/SFX/ScratchHit' )
        PatternManager.mPatternFlawless = false
        PatternManager.mBits[iIndex]:Disable()
        PatternManager:ClearFrozenBits()      
        RadEffects:SetEffectLayerToGameLayer()
        RadEffects:SetEffectTransform( fBitX, fBitY,0.0, 0.0, 0.0, 0.0 )
        RadEffects:CreateEffect( "Effects/Particles/Bursts/ScratchHit.particles", nil, {Constants.BadEffectColor[1],Constants.BadEffectColor[2],Constants.BadEffectColor[3],1}, nil, Constants.ParticlePriorityLow )
        RadEffects:CreateEffect( "Effects/Particles/Bursts/ScratchHitRing.particles", nil, {Constants.BadEffectColor[1],Constants.BadEffectColor[2],Constants.BadEffectColor[3],1}, nil, Constants.ParticlePriorityLow )                     
    elseif iType == Constants.BitTypeGood or iType == Constants.BitTypeHold then
        Score:GoodHit(1)
        if Constants.LevelState == Constants.LevelStateBonus then
            AudioManager:PlaySound( 'Radius/SFX/NoteHitChorus' )
        elseif Constants.GameState ~= Constants.GameStateMenu then
            AudioManager:PlaySound( 'Radius/SFX/NoteHit' )
        end        
        PatternManager.mBits[iIndex]:Freeze()
        PatternManager:CheckForQueueAdvancement()
        Grid:AddImpression( fBitX, fBitY, Constants.NativeScreenWidth*0.04 )
        RadEffects.NewGoodExplosionFunction( fBitX, fBitY, 0.0, Score:GetBitHitTargetProp() )
    elseif iType == Constants.BitTypePoint then
        Score:GoodHit(PatternManager.mBits[iIndex].mPointCount)
        AudioManager:PlaySound( 'Radius/SFX/CirclePop' )
        PatternManager.mBits[iIndex]:FreezePoint()
        PatternManager:CheckForQueueAdvancement()
        RadEffects:SetEffectTransform( fBitX, fBitY, 0.0, 0.0, 0.0, 0.0 )        
        RadEffects:SetEffectLayerToParticleAccumLayer()
        RadEffects:CreateEffect( "Effects/Particles/Bursts/HealthRing.particles", nil, {Constants.Color1[1],Constants.Color1[2],Constants.Color1[3],1}, nil, Constants.ParticlePriorityLow )        
        RadEffects:CreateEffect( "Effects/Particles/Bursts/HealthBits.particles", nil, {Constants.Color1[1],Constants.Color1[2],Constants.Color1[3],1}, nil, Constants.ParticlePriorityLow ) 
        RadEffects:CreateEffect( "Effects/Particles/Bursts/HealthTracers.particles", nil, {Constants.Color1[1],Constants.Color1[2],Constants.Color1[3],1}, nil, Constants.ParticlePriorityLow )        
    end
end
  
function PatternManager:CheckForQueueAdvancement()
    --see if there are any good or slices left on screen
    local i = 1
    for i = 1, PatternManager.mNumBits do        
        if PatternManager.mBits[i]:IsPreventingQueueAdvancement() then
           --good (or bonus) on screen, don't do anything
           return
        end
    end 
    
    for i = 1, PatternManager.mNumSlices do
        if PatternManager.mSlices[i].mEnabled and PatternManager.mSlices[i].mActivated == false then
            --unactivated slices on screen, get out
            return
        end
    end
    
    --if there aren't, see if there are any good or slices in the queue
    local bAnyQueuedUp = false
    local fAdvanceTime = 999.0
    
    for i = 1, PatternManager.mNumBitQueueSlots do
        if PatternManager.mBitQueue[i].CanAdvanceQueue() then
            --take the nearest time
            local fTimeToActive = PatternManager.mBitQueue[i].mBitDelay + PatternManager.mBitQueue[i].mEnterTimer
            fAdvanceTime = math.min( fAdvanceTime, fTimeToActive )
            bAnyQueuedUp = true
        end
    end
    
    --and see if there are any on screen in the entering state
    for i = 1, PatternManager.mNumBits do
        if PatternManager.mBits[i]:CanAdvanceQueue() then
            --take the nearest time
            fAdvanceTime = math.min( fAdvanceTime, PatternManager.mBits[i].mEnterTimer )
            bAnyQueuedUp = true
        end
    end 
        
    if bAnyQueuedUp == true then
        --decrease the bit delay while increasing the lifetime on all good so they show up faster but remain present as long as they were scripted
        for i = 1, PatternManager.mNumBitQueueSlots do
            local iType = PatternManager.mBitQueue[i].mBitType
            if PatternManager.mBitQueue[i]:CanAdvanceQueue() then  
                --remove time from bit delay
                local fTimeLeftToAdvance = fAdvanceTime
                local fBitDelayTime = PatternManager.mBitQueue[i].mBitDelay
                PatternManager.mBitQueue[i].mBitDelay = fBitDelayTime - fTimeLeftToAdvance
                
                --if there's time left over after bit delay, subtract from the enter timer
                fTimeLeftToAdvance = fTimeLeftToAdvance - fBitDelayTime
                if fTimeLeftToAdvance > 0.0 then
                    PatternManager.mBitQueue[i].mEnterTimer = PatternManager.mBitQueue[i].mEnterTimer - fTimeLeftToAdvance
                end
                
                --add total time to lifetime
                PatternManager.mBitQueue[i].mLifetime = PatternManager.mBitQueue[i].mLifetime + fAdvanceTime
            end
        end
        
        for i = 1, PatternManager.mNumBits do
            if PatternManager.mBits[i]:CanAdvanceQueue() then
                PatternManager.mBits[i].mEnterTimer = PatternManager.mBits[i].mEnterTimer - fAdvanceTime
                PatternManager.mBits[i].mLifetime = PatternManager.mBits[i].mLifetime + fAdvanceTime
                PatternManager.mBits[i].mLifeDuration = PatternManager.mBits[i].mLifeDuration + fAdvanceTime
            end
        end         
    else
        --if there's nothing in the queue and nothing good on screen, end this sucka if we've done it perfect
        if PatternManager.mPatternFlawless then
            PatternManager:ClearBitTimer()
        end
    end
end
  
function PatternManager:BitIsNear(fX, fY)
    local fNearAmount = 10.0
    local i = 1
    for i = 1, PatternManager.mNumBits do
        if PatternManager.mBits[i].mEnabled then
            local xPos, yPos = PatternManager.mBits[i]:GetCurrentPos()
            local xDiff = math.abs(xPos - fX)
            local yDiff = math.abs(yPos - fY)
            if xDiff < fNearAmount or yDiff < fNearAmount then
                return true
            end
        end
    end
    
    for i = 1, PatternManager.mNumBitQueueSlots do
        if PatternManager.mBitQueue[i].mEnabled then
            local xDiff = math.abs(PatternManager.mBitQueue[i].mStartX - fX)
            local yDiff = math.abs(PatternManager.mBitQueue[i].mStartY - fY)        
            if xDiff < fNearAmount or yDiff < fNearAmount then
                return true
            end
        end
    end
    
    return false
end
  
function PatternManager:LoadPattern(sFilename)
    PatternManager.mLoadPatternDataBuffer:load( DFFile.getDataPath(sFilename) )
    local patternContents = PatternManager.mLoadPatternDataBuffer:getString()
        
    local iQueueIndex = 1
  
    PatternManager.mPatternStep = (1.0 / PatternManager.mPatternSpeed) * PatternManager.mBasePatternStep
    
    local patternContentsWalker = string.gmatch(patternContents, "[^%s]+")
    PatternManager.mNewBitTimer = patternContentsWalker() / PatternManager.mPatternSpeed
    
    --add a step to the bit timer because the step is tacked on to bit entry
    PatternManager.mNewBitTimer = PatternManager.mNewBitTimer + PatternManager.mPatternStep
    PatternManager.mNewBitDuration = PatternManager.mNewBitTimer
    
    while true do
        local nextPieceOfContent = patternContentsWalker()
        if nextPieceOfContent == nil then break end
        local iBitType = nextPieceOfContent * 1.0   --really hate you, lua.  Multiplying by 1.0 to force it to be a number for table use.
        
        nextPieceOfContent = patternContentsWalker()
        if nextPieceOfContent == nil then break end
        local iOrientation = nextPieceOfContent * 1.0
        
        nextPieceOfContent = patternContentsWalker()
        if nextPieceOfContent == nil then break end
        local fBitDelay = nextPieceOfContent / PatternManager.mPatternSpeed
       
        nextPieceOfContent = patternContentsWalker()
        if nextPieceOfContent == nil then break end
        local fStartX = nextPieceOfContent * 1.0
        
        nextPieceOfContent = patternContentsWalker()
        if nextPieceOfContent == nil then break end
        local fStartY = nextPieceOfContent * 1.0
        
        nextPieceOfContent = patternContentsWalker()
        if nextPieceOfContent == nil then break end
        local fDestX = nextPieceOfContent * 1.0
        
        nextPieceOfContent = patternContentsWalker()
        if nextPieceOfContent == nil then break end
        local fDestY = nextPieceOfContent * 1.0
        
        nextPieceOfContent = patternContentsWalker()
        if nextPieceOfContent == nil then break end
        local fSpeed = nextPieceOfContent * 1.0  --speed was deprecated..
        
        nextPieceOfContent = patternContentsWalker()
        if nextPieceOfContent == nil then break end
        local fLifetime = nextPieceOfContent / PatternManager.mPatternSpeed  
      
        if iBitType == Constants.BitTypeSlice then
            PatternManager.mBitQueue[iQueueIndex]:SetSlice( iOrientation, fBitDelay, fLifetime )
        elseif fStartX ~= fDestX or fStartY ~= fDestY then
            local fDistanceToDest = DFMath.distance2D( fStartX, fStartY, fDestX, fDestY )
            fSpeed = fDistanceToDest / fLifetime
            PatternManager.mBitQueue[iQueueIndex]:SetMoving( iBitType, fBitDelay, fStartX, fStartY, fDestX, fDestY, fLifetime, PatternManager.mPatternStep )
        else           
            PatternManager.mBitQueue[iQueueIndex]:Set( iBitType, fBitDelay, fStartX, fStartY, fLifetime, PatternManager.mPatternStep )
        end
                
        iQueueIndex = iQueueIndex + 1
    end
end

function PatternManager:ConfigPatternForPoint(iNumInputs)
    --count the number of active bits
    local iNumBits = 0
    local i = 1
    for i = 1, PatternManager.mNumBitQueueSlots do
        if PatternManager.mBitQueue[i].mEnabled then
            iNumBits = iNumBits + 1
        end
    end    
    
    --make sure we've got at least 1 input per bit
    if iNumInputs < iNumBits then
        iNumInputs = iNumBits
    end
    
    --run back through the bits, setting the number of inputs they require
    for i = 1, PatternManager.mNumBitQueueSlots do
        if PatternManager.mBitQueue[i].mEnabled then
            if iNumBits == 1 then
                PatternManager.mBitQueue[i].mOrientation = math.min( iNumInputs, 2 )
            else
                --make sure we leave at least 1 input for each of the remaining bits
                local iMaxInputs = iNumInputs - (iNumBits-1)
                
                --don't let a single bit have more than 2 inputs
                iMaxInputs = math.min( iMaxInputs, 2 )
                
                local iRandInputs = math.random(1, iMaxInputs)
                PatternManager.mBitQueue[i].mOrientation = iRandInputs
                iNumInputs = iNumInputs - iRandInputs
            end
            
            iNumBits = iNumBits - 1
        end
    end    
end

function PatternManager:DelayPattern(fDuration)    
    PatternManager.mNewBitTimer = PatternManager.mNewBitTimer + fDuration
    PatternManager.mNewBitDuration = PatternManager.mNewBitDuration + fDuration
    
    local i = 1
    for i = 1, PatternManager.mNumBitQueueSlots do
        if PatternManager.mBitQueue[i].mEnabled then
            PatternManager.mBitQueue[i].mBitDelay = PatternManager.mBitQueue[i].mBitDelay + fDuration
        end
    end
end
  
function PatternManager:MakePatternLong()
    local fLong = 999.0   
    PatternManager.mNewBitTimer = PatternManager.mNewBitTimer + fLong
    PatternManager.mNewBitDuration = PatternManager.mNewBitDuration + fLong
    
    local i = 1
    for i = 1, PatternManager.mNumBitQueueSlots do
        if PatternManager.mBitQueue[i].mEnabled then
            PatternManager.mBitQueue[i].mLifetime = PatternManager.mBitQueue[i].mLifetime + fLong
        end
    end
end

return PatternManager
