local Constants = require('Constants')
local DFMath = require('DFCommon.Math')
local AudioManager = require('AudioManager')
local SliceFactory = {}

-- Slice Object
function SliceFactory:NewSlice()
    local Slice = {}
      
    function Slice:Init()
        Slice.mEnabled = false
        Slice.mOrientation = 0
        
        Slice.mAngle = 0.0
        Slice.mAngleMin = 0.0
        Slice.mAngleMax = 0.0
        Slice.mArcDiameter = 1.0
        Slice.mSegmentMinStart = {}
        Slice.mSegmentMaxStart = {}        
        Slice.mSegmentMinEnd = {}
        Slice.mSegmentMaxEnd = {}
      
        Slice.mActivated = false
        Slice.mActivatedTimer = 0.0
        Slice.mActivatedTimerScalar = 3.0
        Slice.mFanfare = false
        Slice.mLifetime = 0.0
        Slice.mSpecial = false
    end
  
    function Slice:Enable(iOrientation, fLifetime)
        Slice.mEnabled = true
        Slice.mActivated = false
        Slice.mActivatedTimer = 0.0
        Slice.mFanfare = false
        Slice.mOrientation = iOrientation
        
        Slice:UpdateSize()
        
        Slice.mLifetime = fLifetime
    end

    function Slice:UpdateSize()
        --total of 16 orientations
        local fInterval = (2.0 * math.pi) / 16.0
        local fRotationExtent = math.pi * .045
        local fArcLengthOutside = Constants.ValidDrawHeight * 0.65 * 0.5
        local fArcLengthInside = Constants.ValidDrawHeight * 0.65 * 0.1
        
        Slice.mAngle = (fInterval * Slice.mOrientation)
        Slice.mAngleMin = Slice.mAngle - fRotationExtent
        Slice.mAngleMax = Slice.mAngle + fRotationExtent
        Slice.mSpecial = false
        if Slice.mAngleMin < 0.0 then
            Slice.mAngleMin = Slice.mAngleMin + (2.0 * math.pi)  
            Slice.mSpecial = true
        end
        
        local fMinCos = math.cos(Slice.mAngleMin)
        local fMinSin = math.sin(Slice.mAngleMin)    
        Slice.mSegmentMinStart[1] = Constants.ForwardVector[1]*fMinCos - Constants.ForwardVector[2]*fMinSin
        Slice.mSegmentMinStart[2] = Constants.ForwardVector[1]*fMinSin + Constants.ForwardVector[2]*fMinCos
        Slice.mSegmentMinEnd[1] = Slice.mSegmentMinStart[1]
        Slice.mSegmentMinEnd[2] = Slice.mSegmentMinStart[2]
        
        Slice.mSegmentMinStart[1] = Slice.mSegmentMinStart[1] * fArcLengthInside
        Slice.mSegmentMinStart[2] = Slice.mSegmentMinStart[2] * fArcLengthInside
        Slice.mSegmentMinEnd[1] = Slice.mSegmentMinEnd[1] * fArcLengthOutside
        Slice.mSegmentMinEnd[2] = Slice.mSegmentMinEnd[2] * fArcLengthOutside     
        
        local fMaxCos = math.cos(Slice.mAngleMax)
        local fMaxSin = math.sin(Slice.mAngleMax)
        Slice.mSegmentMaxStart[1] = Constants.ForwardVector[1]*fMaxCos - Constants.ForwardVector[2]*fMaxSin
        Slice.mSegmentMaxStart[2] = Constants.ForwardVector[1]*fMaxSin + Constants.ForwardVector[2]*fMaxCos    
        Slice.mSegmentMaxEnd[1] = Slice.mSegmentMaxStart[1]
        Slice.mSegmentMaxEnd[2] = Slice.mSegmentMaxStart[2]
        
        Slice.mSegmentMaxStart[1] = Slice.mSegmentMaxStart[1] * fArcLengthInside
        Slice.mSegmentMaxStart[2] = Slice.mSegmentMaxStart[2] * fArcLengthInside 
        Slice.mSegmentMaxEnd[1] = Slice.mSegmentMaxEnd[1] * fArcLengthOutside
        Slice.mSegmentMaxEnd[2] = Slice.mSegmentMaxEnd[2] * fArcLengthOutside   
    end
    
    function Slice:CheckForActivation( fAngleCurrent )
        if Slice.mSpecial then
            if (fAngleCurrent >= Slice.mAngleMin) or (fAngleCurrent <= Slice.mAngleMax) then
                Slice:Activate()
            end
        else
            --is this frame within bounds?
            if (fAngleCurrent >= Slice.mAngleMin) and (fAngleCurrent <= Slice.mAngleMax) then
                Slice:Activate()
            end       
        end
    end
  
    function Slice:Activate()
        Slice.mActivatedTimer = 0.0
        Slice.mActivated = true
    end
    
    function Slice:Update()
        Slice.mLifetime = Slice.mLifetime - Constants.FrameStep
        Slice.mEnabled = Slice.mLifetime > 0.0 or Slice.mActivated
        Slice.mActivatedTimer = math.min( Slice.mActivatedTimer + (Constants.FrameStep*Slice.mActivatedTimerScalar), 1.0 )
    end
  
    function Slice:Draw()
        if Slice.mActivated then
            MOAIGfxDevice.setPenColor( Constants.Color2[1], Constants.Color2[2], Constants.Color2[3] )    
            local fActivatedMinStartX = DFMath.lerp( Slice.mSegmentMinEnd[1], Slice.mSegmentMinStart[1], Slice.mActivatedTimer )
            local fActivatedMinStartY = DFMath.lerp( Slice.mSegmentMinEnd[2], Slice.mSegmentMinStart[2], Slice.mActivatedTimer )
            local fActivatedMaxStartX = DFMath.lerp( Slice.mSegmentMaxEnd[1], Slice.mSegmentMaxStart[1], Slice.mActivatedTimer )
            local fActivatedMaxStartY = DFMath.lerp( Slice.mSegmentMaxEnd[2], Slice.mSegmentMaxStart[2], Slice.mActivatedTimer )                        
            MOAIDraw.fillFan( fActivatedMinStartX, fActivatedMinStartY, Slice.mSegmentMinEnd[1], Slice.mSegmentMinEnd[2], Slice.mSegmentMaxEnd[1], Slice.mSegmentMaxEnd[2] )
            MOAIDraw.fillFan( fActivatedMinStartX, fActivatedMinStartY, fActivatedMaxStartX, fActivatedMaxStartY, Slice.mSegmentMaxEnd[1], Slice.mSegmentMaxEnd[2] )        
        end

        if Slice.mFanfare then
            MOAIGfxDevice.setPenColor( Constants.SpecialColor[1], Constants.SpecialColor[2], Constants.SpecialColor[3] )
        else
            MOAIGfxDevice.setPenColor( Constants.Color1[1], Constants.Color1[2], Constants.Color1[3] )
        end
        MOAIDraw.drawLine( Slice.mSegmentMinStart[1], Slice.mSegmentMinStart[2], Slice.mSegmentMinEnd[1], Slice.mSegmentMinEnd[2] )
        MOAIDraw.drawLine( Slice.mSegmentMaxStart[1], Slice.mSegmentMaxStart[2], Slice.mSegmentMaxEnd[1], Slice.mSegmentMaxEnd[2] )    
        MOAIDraw.drawLine( Slice.mSegmentMinStart[1], Slice.mSegmentMinStart[2], Slice.mSegmentMaxStart[1], Slice.mSegmentMaxStart[2] )  
        MOAIDraw.drawLine( Slice.mSegmentMinEnd[1], Slice.mSegmentMinEnd[2], Slice.mSegmentMaxEnd[1], Slice.mSegmentMaxEnd[2] )
    end
    
    Slice:Init()
    return Slice    
end

return SliceFactory
