local Constants = require('Constants')
local RadFont = require('RadFont')
local Renderer = require('Renderer')
local TextManager = {}

-- Text Line Object
function TextManager:NewTextLine()
    local TextLine = {}
    
    function TextLine:Init() 
        TextLine.mEnabled = false
        TextLine.mPos = {}
        TextLine.mVelocity = {}
        TextLine.mNumCharactersInLine = 0
        TextLine.mLetterWidth = 0.0
        TextLine.mLetterScale = 1.0
        TextLine.mLetterProps = {}
        TextLine.mColor = {}
        TextLine.mFadeInDuration = 0.0
        TextLine.mFadeOutDuration = 0.0
        TextLine.mLifetime = 0.0
        TextLine.mMaxLife = 0.0
        TextLine.mZPosStart = 0.0
        TextLine.mZPosEnd = 0.0
        TextLine.mZEnterDuration = 0.0
    end
  
    function TextLine:SetZAttributes( fStartZ, fEndZ, fDuration )
        TextLine.mZPosStart = fStartZ
        TextLine.mZPosEnd = fEndZ
        TextLine.mZEnterDuration = fDuration
    end
    
    function TextLine:SetVelocity( fVelX, fVelY )
        TextLine.mVelocity[1] = fVelX
        TextLine.mVelocity[2] = fVelY
    end
    
    function TextLine:SetText( sText, fPosX, fPosY, fLetterWidth, fLetterScale )
        TextLine.mEnabled = true
        TextLine.mNumCharactersInLine = #sText
        TextLine.mPos[1] = fPosX
        TextLine.mPos[2] = fPosY
        TextLine.mLetterWidth = fLetterWidth
        TextLine.mLetterScale = fLetterScale
        
        local textWalker = string.gmatch(sText, ".")
        
        local fAdditionalX = 0.0
        local i = 1
        for i = 1, TextLine.mNumCharactersInLine do
            TextLine.mLetterProps[i] = MOAIProp.new()
            TextLine.mLetterProps[i].__typename = "TextLineLetterProp"
            TextLine.mLetterProps[i]:setDeck( RadFont.mFontDeck )        
            TextLine.mLetterProps[i]:setScl( TextLine.mLetterScale, -TextLine.mLetterScale )
            TextLine.mLetterProps[i]:setLoc( fPosX + fAdditionalX, fPosY, TextLine.mZPosStart )
            TextLine.mLetterProps[i]:setIndex( RadFont.mFontDeck.names[ textWalker() ] )
            TextLine.mLetterProps[i]:setBlendMode( MOAIProp2D.GL_SRC_ALPHA, MOAIProp2D.GL_ONE_MINUS_SRC_ALPHA )
        
            Renderer.mUILayer:insertProp( TextLine.mLetterProps[i] )
            
            fAdditionalX = fAdditionalX + TextLine.mLetterWidth
        end
    end
      
    function TextLine:SetBehavior(fLifetime, tColor, fFadeInDuration, fFadeOutDuration)
        TextLine.mMaxLife = fLifetime
        TextLine.mLifetime = fLifetime
        TextLine.mColor[1] = tColor[1]
        TextLine.mColor[2] = tColor[2]
        TextLine.mColor[3] = tColor[3]

        TextLine.mFadeInDuration = fFadeInDuration
        TextLine.mFadeOutDuration = fFadeOutDuration
        
        local fAlpha = 1.0
        if TextLine.mFadeInDuration > 0.0 then
            fAlpha = 0.0
        end
        
        local i = 1
        for i = 1, TextLine.mNumCharactersInLine do
            TextLine.mLetterProps[i]:setColor( TextLine.mColor[1], TextLine.mColor[2], TextLine.mColor[3], fAlpha )
        end
    end
    
    function TextLine:Update()
        local i = 1
        TextLine.mLifetime = TextLine.mLifetime - Constants.FrameStep        
        
        if TextLine.mLifetime <= 0.0 then  
            TextLine:Disable()
            return
        elseif TextLine.mFadeInDuration > 0.0 or TextLine.mFadeOutDuration > 0.0 then
            --update alpha if we've got a fade
            local fAlpha = 1.0
            local fFullyFadedInTime = TextLine.mMaxLife - TextLine.mFadeInDuration
            if TextLine.mLifetime > fFullyFadedInTime then
                --fading in
                fAlpha = 1.0 - ((TextLine.mLifetime-fFullyFadedInTime) / TextLine.mFadeInDuration)
            elseif TextLine.mLifetime < TextLine.mFadeOutDuration then
                --fading out
                fAlpha = TextLine.mLifetime / TextLine.mFadeOutDuration
            end
                        
            for i = 1, TextLine.mNumCharactersInLine do
                TextLine.mLetterProps[i]:setColor( TextLine.mColor[1], TextLine.mColor[2], TextLine.mColor[3], fAlpha )
            end        
        end
        
        --update z position if we've got an enter duration
        if TextLine.mZEnterDuration > 0.0 then
            local fFullyEnteredTime = TextLine.mMaxLife - TextLine.mZEnterDuration
            if TextLine.mLifetime > fFullyEnteredTime then
                local fZRange = TextLine.mZPosEnd - TextLine.mZPosStart
                local fFadeRatio = 1.0 - ((TextLine.mLifetime-fFullyEnteredTime) / TextLine.mZEnterDuration)
                local fZPos = (fZRange * fFadeRatio) + TextLine.mZPosStart
                
                for i = 1, TextLine.mNumCharactersInLine do
                    local x, y, z = TextLine.mLetterProps[i]:getLoc()
                    TextLine.mLetterProps[i]:setLoc( x, y, fZPos )
                end   
            else
                for i = 1, TextLine.mNumCharactersInLine do
                    local x, y, z = TextLine.mLetterProps[i]:getLoc()
                    TextLine.mLetterProps[i]:setLoc( x, y, TextLine.mZPosEnd )
                end                  
            end        
        end
        
        --update position
        if math.abs(TextLine.mVelocity[1]) > 0.0 or math.abs(TextLine.mVelocity[2]) > 0.0 then
            for i = 1, TextLine.mNumCharactersInLine do
                local x, y, z = TextLine.mLetterProps[i]:getLoc()
                x = x + (TextLine.mVelocity[1] * Constants.FrameStep) 
                y = y + (TextLine.mVelocity[2] * Constants.FrameStep) 
                TextLine.mLetterProps[i]:setLoc( x, y, z )
            end                    
        end
    end
      
    function TextLine:Disable()
        TextLine.mEnabled = false
        local i = 1
        for i = 1, TextLine.mNumCharactersInLine do
            Renderer.mUILayer:removeProp( TextLine.mLetterProps[i] )
        end
        TextLine.mLetterProps = {}
    end
    
    TextLine:Init()
    return TextLine
end

function TextManager:Init()
    TextManager.mNumTextLines = 25
    TextManager.mTextLines = {}
    local i = 1
    for i = 1, TextManager.mNumTextLines do
        TextManager.mTextLines[i] = TextManager:NewTextLine()
    end    
    
    TextManager.mTextSize = 1.0
    TextManager.mTextScale = 1.0
    TextManager.mTextColor = {}
    TextManager.mFadeInDuration = 0.0
    TextManager.mFadeOutDuration = 0.0
    TextManager.mZStart = 0.0
    TextManager.mZEnd = 0.0
    TextManager.mZEnterDuration = 0.0
    
    TextManager.mVelX = 0.0
    TextManager.mVelY = 0.0
end

function TextManager:Update()
    local i = 1
    for i = 1, TextManager.mNumTextLines do
        if TextManager.mTextLines[i].mEnabled then
            TextManager.mTextLines[i]:Update()
        end
    end 
end

function TextManager:SetTextAttributes( fSize, fScale, tColor )
    TextManager.mTextSize = fSize
    TextManager.mTextScale = fScale
    TextManager.mTextColor[1] = tColor[1]
    TextManager.mTextColor[2] = tColor[2]
    TextManager.mTextColor[3] = tColor[3]
end

function TextManager:SetVelocity(fVelX, fVelY)
    TextManager.mVelX = fVelX
    TextManager.mVelY = fVelY
end

function TextManager:SetFadeAttributes(fIn, fOut)
    TextManager.mFadeInDuration = fIn
    TextManager.mFadeOutDuration = fOut
end

function TextManager:SetZoomAttributes(fStart, fEnd, fDuration)
    TextManager.mZStart = fStart
    TextManager.mZEnd = fEnd
    TextManager.mZEnterDuration = fDuration
end

function TextManager:AddLine( sTest, fPosX, fPosY, fLifetime )
    local i = 1
    for i = 1, TextManager.mNumTextLines do
        if not TextManager.mTextLines[i].mEnabled then
            TextManager.mTextLines[i]:SetZAttributes( TextManager.mZStart, TextManager.mZEnd, TextManager.mZEnterDuration )
            TextManager.mTextLines[i]:SetVelocity( TextManager.mVelX, TextManager.mVelY )
            TextManager.mTextLines[i]:SetText( sTest, fPosX, fPosY, TextManager.mTextSize, TextManager.mTextScale )
            TextManager.mTextLines[i]:SetBehavior( fLifetime, TextManager.mTextColor, TextManager.mFadeInDuration, TextManager.mFadeOutDuration )
            return 
        end
    end 
end

function TextManager:ClearAll()
    local i = 1
    for i = 1, TextManager.mNumTextLines do
        if TextManager.mTextLines[i].mEnabled then
            TextManager.mTextLines[i]:Disable()
        end
    end 
    
    TextManager.mTextSize = 1.0
    TextManager.mTextScale = 1.0
    TextManager.mFadeInDuration = 0.0
    TextManager.mFadeOutDuration = 0.0
    TextManager.mZStart = 0.0
    TextManager.mZEnd = 0.0 
    TextManager.mZEnterDuration = 0.0 
    TextManager.mVelX = 0.0
    TextManager.mVelY = 0.0    
end

return TextManager