local Constants = require('Constants')
local GameRules = require('GameRules')
local AudioManager = require('AudioManager')
local RadFont = require('RadFont')
local GameRing = require('GameRing')
local Renderer = require('Renderer')
local TextManager = require('TextManager')
local UI = require('UI')
local ChargeManager = require('ChargeManager')
local ColorManager = require('ColorManager')
local PatternManager = require('PatternManager')
local InputVisuals = require('InputVisuals')
local Score = require('Score')
local ChargeField = require('ChargeField')
local Grid = require('Grid')
local Wind = require('Wind')
local Post = require('postfx.Post')
local PostManager = require('postfx.PostManager')
local Credits = require('Credits')
local HowToPlay = require('HowToPlay')
local Leaderboard = require('Leaderboard')
local LeaderboardEntry = require('LeaderboardEntryTouch')
local RadEffects = require('RadEffects')
local RadShapes = require('RadShapes')
local PauseScreen = require('PauseScreen')
local HUD = require('HUD')
local InputTutorial = require('InputTutorial')
local AudioBars = require('AudioBars')
local AudioRing = require('AudioRing')
local GlowRing = require('GlowRing')
local HUDGlow = require('HUDGlow')
local OpponentScore = require('OpponentScore')
local Achievement = require('DFCommon.Achievement')
local AndroidDevices = require('AndroidDevices')
local Input = require('Input')

debug.traceback = function()
    print("\nthis is a traceback! \n")
    -- I put this callback in here because the traceback function in MOAI seems to crash in android with luajit
end

local function OnScreenResize(width, height)
    --resize all classes
    GameRing:UpdateSize()
    ChargeManager:UpdateSize()
    InputVisuals:UpdateSize()
    PatternManager:UpdateSize()
    UI:UpdateSize()
    Score:UpdateSize()
    PauseScreen:UpdateSize()
    LeaderboardEntry:UpdateSize()
    Credits:UpdateSize( Renderer )
    HowToPlay:UpdateSize( Renderer )
    Leaderboard:UpdateSize( Renderer )
    HUD:UpdateSize( Renderer )
    Wind:UpdateSize()
    Grid:UpdateSize()
    ChargeField:UpdateSize()
    AudioBars:UpdateSize()
    InputTutorial:UpdateSize()
    OpponentScore:UpdateSize()
    GlowRing:UpdateSize()
    HUDGlow:UpdateSize()
    
    if Constants.UsePostEffects then
        Post:OnScreenResize(width, height)
    end
    
    GameRules:OnScreenResize(width, height)
end
        
local function OnBackButtonPressed()
    return GameRules:HandleBack()
end

local function DetectDeviceIOS()
    print("DetectDeviceIOS()")
    Constants.DefaultVisualQuality = Constants.VisualQualityLevelHigh
    Constants.GlobalScale = 1.0
    
    local sDeviceId = MOAIEnvironment.devModel
    print("sDeviceId: " .. tostring(sDeviceId))
    local bIsIpad = nil ~= string.find(sDeviceId, "iPad")
    
    local bIsRetina = (MOAIEnvironment.iosRetinaDisplay ~= false)

    local nativeW, nativeH = MOAIGfxDevice.getViewSize()

    if sDeviceId == "iPhone4,1" then -- iPhone 4S 
        Constants.DefaultVisualQuality = Constants.VisualQualityLevelHigh
        Constants.ScreenBufferMaxWidth_High = 512
        Constants.ScreenBufferMaxHeight_High = 512
    elseif sDeviceId == "iPad2,5" then -- iPad mini
        Constants.DefaultVisualQuality = Constants.VisualQualityLevelHigh
        Constants.ScreenBufferMaxWidth_High = 1024
        Constants.ScreenBufferMaxHeight_High = 512
    elseif sDeviceId == "iPod5,1" then -- ipod touch 5 has a worse GPU than iPhone 5
        Constants.DefaultVisualQuality = Constants.VisualQualityLevelHigh
        Constants.ScreenBufferMaxWidth_High = 640
        Constants.ScreenBufferMaxHeight_High = 512
    elseif nativeW > 960 and nativeH >= 640 and false == bIsIpad then -- high res phone, go crazy
        Constants.DefaultVisualQuality = Constants.VisualQualityLevelHigh
        Constants.ScreenBufferMaxWidth_High = 1024
        Constants.ScreenBufferMaxHeight_High = 512        
    elseif nativeW > 1024 and nativeH > 768 then -- big devices, most likely high end ipads
        Constants.DefaultVisualQuality = Constants.VisualQualityLevelHigh
        Constants.ScreenBufferMaxWidth_High = nativeW / 2
        Constants.ScreenBufferMaxHeight_High = nativeH / 2
    else
        Constants.GlobalScale = 0.5
        Constants.DefaultVisualQuality = Constants.VisualQualityLevelLow
    end
end

local function DetectDeviceAndroid()
    print("DetectDeviceAndroid()")
    
    -- unlike on iPhone it is generally impossible to cater to specific devices, so we have to use a more generic approach
    
    local nativeW, nativeH = MOAIGfxDevice.getViewSize()
    
    local minHighResX = 960
    local minHighResY = 640
    
    local nativePixels = nativeW * nativeH
    local minHighResPixels = minHighResX * minHighResY

    print(" DEVICE STATS\n------------------------------------------- ")
    print("cpuabi: " .. tostring(MOAIEnvironment.cpuabi))
    print("devBrand: " .. tostring(MOAIEnvironment.devBrand))
    print("devName: " .. tostring(MOAIEnvironment.devName))
    print("devManufacturer: " .. tostring(MOAIEnvironment.devManufacturer))
    print("devModel: " .. tostring(MOAIEnvironment.devModel))
    print("devPlatform: " .. tostring(MOAIEnvironment.devPlatform))
    print("devProduct: " .. tostring(MOAIEnvironment.devProduct))
    print("numProcessors: " .. tostring(MOAIEnvironment.numProcessors))
    print("screenDpi: " .. tostring(MOAIEnvironment.screenDpi))
    print("osVersion: " .. tostring(MOAIEnvironment.osVersion))
    print("gpuVendor: " .. tostring(MOAIEnvironment.gpuVendor))
    print("gpuRenderer: " .. tostring(MOAIEnvironment.gpuRenderer))

    print("native res: " .. tostring(nativeW) .. " x " .. tostring(nativeH) .. " - " .. tostring(nativePixels))
    
    -- TODO: we want to check GPU capabilities if we can
    if nativePixels <= minHighResPixels then
        Constants.DefaultVisualQuality = Constants.VisualQualityLevelLow
        Constants.ScreenBufferMaxWidth_High = 512
        Constants.ScreenBufferMaxHeight_High = 512
    else
        --Check against list of known android GPUs
        local sDeviceRenderer = tostring(MOAIEnvironment.gpuRenderer)
        
        Constants.DefaultVisualQuality = Constants.VisualQualityUninitialized
        
        local i = 1
        for i = 1, #AndroidDevices.DevicesByGPU do
            --look for a match to the primary
            if string.find( sDeviceRenderer, AndroidDevices.DevicesByGPU[i].Primary ) then
                --primary match, look for secondary
                local j = 1
                for j = 1, #AndroidDevices.DevicesByGPU[i].Secondary do
                    if string.find( sDeviceRenderer, AndroidDevices.DevicesByGPU[i].Secondary[j][1] ) then
                        --we found a secondary match, get out!
                        Constants.DefaultVisualQuality = AndroidDevices.DevicesByGPU[i].Secondary[j][2]
                        print( "Match: "..sDeviceRenderer.."="..AndroidDevices.DevicesByGPU[i].Primary.."+"..AndroidDevices.DevicesByGPU[i].Secondary[j][1] )
                        break
                    end
                end
            end
            
            if Constants.DefaultVisualQuality ~= Constants.VisualQualityUninitialized then
                break
            end
        end
        
        --if we didn't find a match in the GPU list, dig through the misc list in the hopes we get lucky
        if Constants.DefaultVisualQuality == Constants.VisualQualityUninitialized then
            for i = 1, #AndroidDevices.MiscDevices do
                if string.find( sDeviceRenderer, AndroidDevices.MiscDevices[i][1] ) then
                    Constants.DefaultVisualQuality = AndroidDevices.MiscDevices[i][2]
                    print( "Match: "..sDeviceRenderer.."="..AndroidDevices.MiscDevices[i][1] )
                    break
                end
            end
        end
        
        --if we STILL didn't find anything, default to low
        if Constants.DefaultVisualQuality == Constants.VisualQualityUninitialized then
            print( "No Match Found" )
            Constants.DefaultVisualQuality = Constants.VisualQualityLevelLow
        end
    end
    
    --look for devices that need low performance audio
    local sDevModel = tostring(MOAIEnvironment.devModel)
    local iDevModelStringCount = string.len( sDevModel )
    for i = 1, #AndroidDevices.LowPerformanceAudioDevices do
        if string.find( sDevModel, AndroidDevices.LowPerformanceAudioDevices[i] ) then
            local iDeviceStringCount = string.len( AndroidDevices.LowPerformanceAudioDevices[i] )
            if iDevModelStringCount == iDeviceStringCount then
                print( "Using Low Performance Audio" )
                Constants.UseLowPerformanceAudio = true
                break
            end
        end
    end
end

local function DetectDevice()
    if not Constants.IsMobile then return end
    
    --after initial game load, set the desired quality to whatever we had saved
    -- hack for now, just use whatever visual quality you set up in Constants.
    --Constants.DefaultVisualQuality = nil
    
    -- if there was no setting saved, try and find a reasonable default for this device
    if(MOAIEnvironment.osBrand == "iOS") then
        DetectDeviceIOS()
    elseif(MOAIEnvironment.osBrand == "Android") then
        DetectDeviceAndroid()
    end    
    
    -- load the saved setting if it's there
    if Constants.Leaderboards and Constants.Leaderboards[Constants.VisualQualityIndex] and #Constants.Leaderboards[Constants.VisualQualityIndex] > 0 then
        Constants.DefaultVisualQuality = Constants.Leaderboards[Constants.VisualQualityIndex][1]
    end
    
    Constants.DesiredVisualQuality = Constants.DefaultVisualQuality
end

local function LateInit()     
    print("initializing systems")
    AndroidDevices:Init()
    Leaderboard:Init()

    DetectDevice()

    AudioManager:Init()
    GameRules:Init()
    Renderer:Init(OnScreenResize)   
    TextManager:Init()
    ChargeField:Init( Renderer )
    Grid:Init( Renderer )   
    AudioBars:Init( Renderer )
    AudioRing:Init( Renderer )
    Wind:Init( Renderer ) 
    GameRing:Init( Renderer )
    HUDGlow:Init( Renderer )

    print("initializing systems - halfway!")
    
    RadEffects:Init( Renderer )
    RadShapes:Init()
    RadFont:Init()
    UI:Init( Renderer )
    ChargeManager:Init( Renderer )
    ColorManager:Init()
    PatternManager:Init( Renderer )
    InputVisuals:Init( Renderer )
    Score:Init( Renderer )
    Post:Init()
    PostManager:Init()
    Credits:Init()
    HowToPlay:Init()
    LeaderboardEntry:Init()
    HUD:Init( Renderer )
    PauseScreen:Init()
    InputTutorial:Init( Renderer )
    OpponentScore:Init()
    GlowRing:Init( Renderer )
        
    -- all done initializing subsystems, let gamerules do any glue-y stuff it wants between them.
    GameRules:PostSystemInit()
        
    local screenW, screenH = Renderer:GetScreenSize()
    OnScreenResize(screenW, screenH)

    print("initializing systems - Complete")
                
    Achievement.init( Constants.Achievements )
        
    --spin the randomizer a bit on startup
    local i = 1
    for i = 1, 20 do
        math.random()
    end
    
    -- Uncomment these lines to jump directly in to the game.  Or, modify this to jump in to a specific state.
    --Constants.DesiredGameState = Constants.GameStateMenu
    --GameRules:DebugJumpGameState() 
end

function DelayInit()

    print("delay init!")

    --[[
    local nFramesRemaining = 100
    
    while (nFramesRemaining > 0) do
        nFramesRemaining = nFramesRemaining - 1
        coroutine.yield()
    end
    ]]--
    
    LateInit()
end
        
function ResumeSession()
    print("LUA ResumeSession")
end

function SuspendSession()
    print("LUA SuspendSession")
    GameRules:Pause(true)
    GameRules:EatInput(0.5)
    if Input ~= nil and Input.Reset ~= nil then
    	Input.Reset()
    end    
    AudioManager:PauseMusic()
end
  
function TeardownGame()
    print("LUA TeardownGame")
end
        
        
local function InitGame()   
    Constants:Init()
 
    if Constants.ShowFullProfilingInfo then
        MOAISim.setHistogramEnabled(true);
    end
    
    if MOAIApp ~= nil then
        --Android back button
        MOAIApp.setListener(MOAIApp.BACK_BUTTON_PRESSED, OnBackButtonPressed)
        
        if(MOAIEnvironment.osBrand == "iOS") then
            MOAISim.setListener(MOAISim.EVENT_FINALIZE, TeardownGame)
            MOAIApp.setListener(MOAIApp.SESSION_START, ResumeSession)
            MOAIApp.setListener(MOAIApp.SESSION_END, SuspendSession)
        end
    end

    MOAISim.setGcFrameMaxDuration (2000) -- Maximum time (in micro second) Lua can spend for GC during one frame
    MOAISim.setGcConstantDuration (1000) -- Amount of time (in micro second) Lua will spend at the end of the frame for GC
    
    MOAISim.clearLoopFlags()
    MOAISim.setLoopFlags( MOAISim.SIM_LOOP_VARIABLE )   
    
    Renderer.PreInit()
    
    -- hack: for now, we delay initialization for a while on android because we were getting black screens. we need to fine
    --        the core problem, but for now we need some other systems work first.
    local rLateInitThread = MOAIThread.new()
    rLateInitThread:run(DelayInit)
end

InitGame()
