local Renderer = require('Renderer')
local DFGraphics = require('DFCommon.Graphics')
local DFFile = require('DFCommon.File')
local Constants = require('Constants')
local ChargeField= require('ChargeField')
local Grid = require('Grid')
local AudioBars = require('AudioBars')
local AudioRing = require('AudioRing')
local Wind = require('Wind')
local InputVisuals = require('InputVisuals')
local ColorManager= require('ColorManager')
local DFMath = require('DFCommon.Math')
local RadEffects = require('RadEffects')
local RadShapes = require('RadShapes')
local GlowRing = require('GlowRing')
local HUDGlow = require('HUDGlow')

local Post = {}
 
kBLEND_MULTIPLY = { MOAIProp.GL_DST_COLOR, MOAIProp.GL_ONE_MINUS_SRC_ALPHA }
kBLEND_ADD = { MOAIProp.GL_ONE, MOAIProp.GL_ONE }
kBLEND_ALPHA = { MOAIProp.GL_SRC_ALPHA, MOAIProp.GL_ONE_MINUS_SRC_ALPHA}
kBLEND_PREMULTIPLIED = { MOAIProp.GL_ONE, MOAIProp.GL_ONE_MINUS_SRC_ALPHA}

kNUM_POST_FILTERS = 30
kNUM_POST_BUFFERS = 8
kNUM_POST_HALFBUFFERS = 4
kNUM_POST_QUARTERBUFFERS = 4
kNUM_POST_EIGHTHBUFFERS = 4
kNUM_POST_SCRATCHLAYERS = 4
 
function Post.CreateScreenQuad()
    local Obj = {}
    
    function Obj:setInBuffer(rInputFrameBuffer)
        self:enableFullScreenEffects()
    
        self.quad:setTexture(rInputFrameBuffer)
            
        MOAISim.removeRenderPass(self.overlayLayer)
        MOAISim.removeRenderPass(self.layer)
        MOAISim.pushRenderPass(self.layer)
        MOAISim.pushRenderPass(self.overlayLayer)
            
        self.layer:clear()
        self.layer:insertProp(self.prop)
    end
    
    function Obj:setRect(left, top, right, bottom)
        self.quad:setRect(left, top, right, bottom)
    end
    
    function Obj:enableFullScreenEffects(bEnable)
        if nil == bEnable then
            bEnable = true
        end

        if bEnable ~= self.bEnableFullScreenEffects then
            MOAISim.removeRenderPass(self.overlayLayer)
            
            if bEnable then
                MOAISim.removeRenderPass(self.layer)
                
                Renderer.mBackgroundLayer:setFrameBuffer(Renderer.mBackgroundBuffer)
                Renderer.mGameLayer:setFrameBuffer(Renderer.mGameBuffer)
                Renderer.mUILayer:setFrameBuffer(Renderer.mUIBuffer)
                
                MOAISim.pushRenderPass(self.layer)
                MOAISim.pushRenderPass(self.overlayLayer)
            else
                Renderer.mBackgroundLayer:setFrameBuffer(nil)
                Renderer.mGameLayer:setFrameBuffer(nil)
                Renderer.mUILayer:setFrameBuffer(nil)
                MOAISim.removeRenderPass(self.layer)
                MOAISim.removeRenderPass(Renderer.mBackgroundLayer)
                MOAISim.removeRenderPass(Renderer.mGameLayer)
                MOAISim.removeRenderPass(Renderer.mUILayer)
                MOAISim.pushRenderPass(Renderer.mBackgroundLayer)
                MOAISim.pushRenderPass(Renderer.mGameLayer)
                MOAISim.pushRenderPass(Renderer.mUILayer)
                MOAISim.pushRenderPass(self.overlayLayer)
            end
            
            self.bEnableFullScreenEffects = bEnable

        end
        
        MOAISim.removeRenderPass(self.overlayLayer)
        MOAISim.pushRenderPass(self.overlayLayer)
    end
    
    local halfScreenX, halfScreenY = Renderer.GetHalfScreenSize()
    
    Obj.quad = MOAIGfxQuad2D.new ()
    Obj.quad:setUVRect ( 0, 0, 1, 1 )
    
    Obj:setRect(-halfScreenX, -halfScreenY, halfScreenX, halfScreenY)
    
    Obj.layer = Renderer.mPostLayer
    Obj.overlayLayer = Renderer.mPostOverlayLayer
    
    Obj.prop = MOAIProp2D.new ()
    Obj.prop:setDeck ( Obj.quad )
    Obj.prop:setLoc( 0, 0, 1 )
    Obj.layer:insertProp(Obj.prop)
    
    Obj:enableFullScreenEffects(false)
    
    return Obj
end
 
function Post.CreateBuffer(w, h)
    local Obj = MOAIFrameBuffer.new ()
    Obj:init (w, h)
    Obj:setClearColor(1, 1, 1, 1)
    Obj:setFilter(MOAITexture.GL_LINEAR, MOAITexture.GL_LINEAR)
    Obj.bInUse = false
    Obj.rViewport = Renderer.mGameViewport
    
    function Obj:SetInUse(bInUse)
        if nil == bInUse then bInUse = true end
        
        self.bInUse = bInUse
    end
    
    return Obj
end
 
function Post.CreateScratchLayer()
    local Obj = {}
    
    function Obj:setOutBuffer(rOutputFrameBuffer)
        MOAISim.removeRenderPass ( self.layer )
        
        if rOutputFrameBuffer then
            if rOutputFrameBuffer.rViewport then
                self.layer:setViewport(rOutputFrameBuffer.rViewport)
            end
            self.layer:setFrameBuffer(rOutputFrameBuffer)
            MOAISim.pushRenderPass ( self.layer )
        end
    end
    
    function Obj:ClearProps()
        self.layer:clear()
        if self.prop then
            self.layer:insertProp(self.prop)
        end
    end   
    
    function Obj:SetInUse(bInUse)
        if nil == bInUse then bInUse = true end
        
        self.bInUse = bInUse
    end
    
    function Obj:insertProp(prop)
        self.layer:insertProp(prop)
    end
    
    function Obj:removeProp(prop)
        self.layer:removeProp(prop)
    end
    
    function Obj:setCamera(rCamera)
        self.layer:setCamera(rCamera)
    end
    
    Obj.sLayerType = "ScratchLayer"
    
    Obj.layer = MOAILayer2D.new ()
    Obj.layer:setDebugName( Obj.sLayerType )
    Obj.layer:setViewport ( Renderer.mGameViewport )
    Obj.bInUse = false
    
    return Obj
end
 
function Post.CreateFilter(rInputFrameBuffer, sMaterial)
    local Obj = Post.CreateScratchLayer()

    Obj.sLayerType = "FilterLayer"
    
    local halfScreenX, halfScreenY = Renderer.GetHalfScreenSize()
    
    Obj.quad = MOAIGfxQuad2D.new ()
    Obj.quad:setTexture ( rInputFrameBuffer )
    Obj.quad:setRect ( -halfScreenX, -halfScreenY, halfScreenX, halfScreenY)
    Obj.quad:setUVRect ( 0, 0, 1, 1 )
    
    if nil == sMaterial then
        sMaterial =  "Materials/Blit.material" 
    end
           
    Obj.color = MOAIColor.new()
    
    function Obj:getOutBuffer()
        return self.rOutFrameBuffer
    end
    
    function Obj:setVisible(bVisible)
        self.prop:setVisible(bVisible)
    end
    
    function Obj:setBlendMode(tModeConstant)
        self.rMaterial:setBlendMode(tModeConstant[1], tModeConstant[2])
    end
    
    function Obj:setColor(r, g, b, a)
        self.rMaterial:setColor(r, g, b, a)
    end
    
    function Obj:setMaterial(sMaterial)
        self.layer:setDebugName(Obj.sLayerType .. " - " .. sMaterial)
    
        if self.rMaterial then
            DFGraphics.unloadMaterial(self.rMaterial)
        end
    
        self.rMaterial = DFGraphics.loadMaterial( sMaterial )
        self.prop:setMaterial(self.rMaterial)
    end
    
    function Obj:setShaderValue( sName, type, value )
        self.rMaterial:setShaderValue(sName, type, value)
    end
    
    function Obj:setInBuffer(rInputFrameBuffer)
        self.quad:setTexture ( rInputFrameBuffer )
    end

    Obj.prop = MOAIProp2D.new ()
    Obj.prop:setDeck ( Obj.quad )
    Obj.prop:setLoc( 0, 0, 1 )
    Obj.layer:insertProp(Obj.prop)

    Obj:setMaterial(sMaterial)
    Obj:setInBuffer(rInputFrameBuffer)
    Obj:setBlendMode(kBLEND_ALPHA)
    return Obj
end

function Post:Init()
    Post.tFilters = {}
    Post.tBuffers = {}
    Post.tHalfBuffers = {}
    Post.tQuarterBuffers = {}
    Post.tEighthBuffers = {}
    Post.tScratchLayers = {}
        
    if not Constants.UsePostEffects then return end

    Post.nCurFilter = 1
    Post.rWhiteTexture = DFGraphics.loadTexture( "White", true )
    Post.rBlackTexture = DFGraphics.loadTexture( "Effects/Textures/Black", true )
    
    if Constants.VisualQuality ~= Constants.VisualQualityLevelLow then
        Post.Preload()
    end
        
    for i=1,kNUM_POST_FILTERS do
        local filter = Post.CreateFilter(Renderer.mGameBuffer)
        table.insert(Post.tFilters, filter)
    end

    local screenX = Constants.ScreenBufferWidth
    local screenY = Constants.ScreenBufferHeight
    
    for i=1,kNUM_POST_BUFFERS do      
        local buffer = Post.CreateBuffer(screenX, screenY)
        buffer.rViewport = Renderer.mBufferViewport
        table.insert(Post.tBuffers, buffer)
    end
    
    for i=1,kNUM_POST_HALFBUFFERS do      
        local buffer = Post.CreateBuffer(screenX/2, screenY/2)
        buffer.rViewport = Renderer.mHalfSizeGameViewport
        table.insert(Post.tHalfBuffers, buffer)
    end
    
    for i=1,kNUM_POST_QUARTERBUFFERS do      
        local buffer = Post.CreateBuffer(screenX/4, screenY/4)
        buffer.rViewport = Renderer.mQuarterSizeGameViewport
        table.insert(Post.tQuarterBuffers, buffer)
    end
    
    for i=1,kNUM_POST_EIGHTHBUFFERS do      
        local buffer = Post.CreateBuffer(screenX/8, screenY/8)
        buffer.rViewport = Renderer.mEighthSizeGameViewport
        table.insert(Post.tEighthBuffers, buffer)
    end
    
    for i=1,kNUM_POST_SCRATCHLAYERS do
        local scratchLayer = Post.CreateScratchLayer()
        scratchLayer:setCamera( Renderer.mGameCamera )
        table.insert(Post.tScratchLayers, scratchLayer)
    end
    
    Post.rScreen = Post:CreateScreenQuad()
end

function Post.Preload()
    Post.rGradientTexture = DFGraphics.loadTexture( "Effects/Textures/Gradient", true )
    Post.rGradientTextureB = DFGraphics.loadTexture( "Effects/Textures/GradientB", true ) 
    Post.rGradientTextureC = DFGraphics.loadTexture( "Effects/Textures/GradientC", true ) 
    Post.rGradientTextureD = DFGraphics.loadTexture( "Effects/Textures/GradientD", true )     
    Post.rGradientTextureE = DFGraphics.loadTexture( "Effects/Textures/GradientE", true )     
    Post.rGradientTextureF = DFGraphics.loadTexture( "Effects/Textures/GradientF", true )     
    Post.rBackdropA = DFGraphics.loadTexture( "Effects/Textures/Backdrops/BackdropA", true )    
    
    Post.rFlowTexture = DFGraphics.loadTexture( "Effects/Textures/Flow", true ) 
    Post.rNoiseTexture = DFGraphics.loadTexture( "Effects/Textures/Backdrops/Noise", true ) 
    Post.rNoiseTexture:setWrap(true)
end

function Post:EnableFullscreenEffects(bEnable)
    if nil == bEnable then
        bEnable = true
    end
    
    Post.rScreen:enableFullScreenEffects(bEnable)
end

function Post:OnScreenResize(width, height)
    if Constants.UsePostEffects then
        Post.rScreen:setRect( -width/2, -height/2, width/2, height/2)
        
        for i,rFilter in ipairs(Post.tFilters) do
            rFilter.quad:setRect ( -width/2, -height/2, width/2, height/2)
        end

        width = Constants.ScreenBufferWidth
        height = Constants.ScreenBufferHeight
        
        for i,rBuffer in ipairs(Post.tBuffers) do
            rBuffer:init(width, height)
        end
        
        for i,rBuffer in ipairs(Post.tHalfBuffers) do
            rBuffer:init(width/2, height/2)    
        end
        
        for i,rBuffer in ipairs(Post.tQuarterBuffers) do
            rBuffer:init(width/4, height/4)    
        end
        
        for i,rBuffer in ipairs(Post.tEighthBuffers) do
            rBuffer:init(width/8, height/8)    
        end
    end
end

local function F(nIndex)
    return Post.tFilters[nIndex]
end

local function HalfB(nIndex)
    return Post.tHalfBuffers[nIndex]
end

local function QuarterB(nIndex)
    return Post.tQuarterBuffers[nIndex]
end

local function EighthB(nIndex)
    return Post.tEighthBuffers[nIndex]
end

local function B(nIndex)
    return Post.tBuffers[nIndex]
end

function Post:ClearFilters()
    for i,rFilter in ipairs(Post.tFilters) do
        rFilter:setInBuffer(Renderer.mGameBuffer)
        rFilter:setMaterial("Materials/Blit.material")
        rFilter:setOutBuffer(nil)
        rFilter:setBlendMode(kBLEND_ALPHA)
        rFilter:SetInUse(false)
    end
    
    for i,rBuffer in ipairs(Post.tBuffers) do
        rBuffer:setClearColor(0, 0, 0, 0)
        rBuffer:SetInUse(false)
    end
    
    for i,rBuffer in ipairs(Post.tHalfBuffers) do
        rBuffer:setClearColor(0, 0, 0, 0)
        rBuffer:SetInUse(false)
    end
    
    for i,rBuffer in ipairs(Post.tQuarterBuffers) do
        rBuffer:setClearColor(0, 0, 0, 0)
        rBuffer:SetInUse(false)
    end
    
    for i,rBuffer in ipairs(Post.tEighthBuffers) do
        rBuffer:setClearColor(0, 0, 0, 0)
        rBuffer:SetInUse(false)
    end
    
    for i,rScratchLayer in ipairs(Post.tScratchLayers) do
        rScratchLayer:ClearProps()
        rScratchLayer:SetInUse(false)
    end
    
    Post.fnUpdate = nil
    
    Post.nCurFilter = 1
    
    RadEffects.ClearInfiniteDurationEffects()
    
    if g_DebugDisplay then
        g_DebugDisplay:SetValue("FIL", (Post.nCurFilter - 1))
    end
end

function Post:ResetDrawColors()
    ColorManager.BGColor            = {0.2,0.2,0.2}
    ColorManager.Color1             = {0,1,0}
    ColorManager.Color2             = {0,0,1}
    ColorManager.Color3             = {1,1,0}
    ColorManager.SpecialColor = {0.95, 0.95, 0.95}
    ColorManager:SetColor()
    
    Renderer.mGameBuffer:setClearColor ( 0,0,0,0 )
	Renderer.mUIBuffer:setClearColor ( 0,0,0,0 )
    Renderer.mBackgroundBuffer:setClearColor ( 0, 0, 0, 1 )
end

function Post:ResetDrawFlavor()
    Grid:SetColor(1, 1, 1, 1)
    Grid:Enable(false)
    ChargeField:SetColor(1, 1, 1, 1)
    ChargeField:Enable(false)
    Wind:Enable(false)
    Wind:SetColor(1, 1, 1, 1)
    InputVisuals:ClearFingerEffects()
    AudioBars:Enable(false)
    AudioRing:Enable(false)
    GlowRing:Enable(false)
    HUDGlow:Enable(false)
end

-- Reset all rendering state between filters. 
function Post:Reset()
    Post:ClearFilters()
    Post:ResetDrawColors()
    Post:ResetDrawFlavor()
    RadShapes:ClearShapes()
end

function Post:VisualQualityChange()
    Post:Reset()
    -- TODO: destroy all old buffers
    -- TODO: create all buffers and stuff
    Post:OnScreenResize(Renderer.mTrueScreenWidth, Renderer.mTrueScreenHeight)
end

function Post:ShowBuffer(index)
    print("show buffer: " .. tostring(index))
    Post.rScreen:setInBuffer(B(index))
end

function Post:ShowHalfBuffer(index)
    print("show half buffer: " .. tostring(index))
    Post.rScreen:setInBuffer(Post.tHalfBuffers[index])
end

function Post:ShowQuarterBuffer(index)
    print("show quarter buffer: " .. tostring(index))
    Post.rScreen:setInBuffer(Post.tQuarterBuffers[index])
end

function Post:ShowEighthBuffer(index)
    print("show eighth buffer: " .. tostring(index))
    Post.rScreen:setInBuffer(Post.tEighthBuffers[index])
end

function Post:GetNextFilter()
    local filter = F(Post.nCurFilter)
    Post.nCurFilter = Post.nCurFilter + 1
    filter:SetInUse()
    
    if g_DebugDisplay then
        g_DebugDisplay:SetValue("FIL", (Post.nCurFilter - 1))
    end
    
    return filter
end

function Post:GetUnusedBuffer(tBufferList)
    if not tBufferList then tBufferList = Post.tBuffers end
    
    local out = nil
    
    for i,rBuffer in ipairs(tBufferList) do
        if not rBuffer.bInUse then
            rBuffer:SetInUse()
            out = rBuffer
            --print ("assigned buffer to index: " .. i)
            break
        end
    end
    
    return out
end

function Post:GetUnusedHalfBuffer()
    return Post:GetUnusedBuffer(Post.tHalfBuffers)
end

function Post:GetUnusedQuarterBuffer()
    return Post:GetUnusedBuffer(Post.tQuarterBuffers)
end

function Post:GetUnusedEighthBuffer()
    return Post:GetUnusedBuffer(Post.tEighthBuffers)
end

function Post:GetUnusedScratchLayer()
    local out = nil
    
    for i,rScratchLayer in ipairs(Post.tScratchLayers) do
        if not rScratchLayer.bInUse then
            out = rScratchLayer
            out:SetInUse()
            break
        end
    end
    
    return out
end

function Post:SetUpdateCallback(fnUpdate)
    Post.fnUpdate = fnUpdate
end

function Post:Update()
    if not Constants.UsePostEffects then return end

    MOAISim.profileEnterScope("LUA Post update")
    
    HUDGlow:Update()
    GlowRing:Update()
    
    if Post.fnUpdate then
       Post.fnUpdate(self)
    end
   
    MOAISim.profileLeaveScope("LUA Post update")
end

-----------------------------------------------------------------------------
-- Cool filters, bro
-----------------------------------------------------------------------------
function Post:BlurFilter(rInBuffer, rOutBuffer, amount, blendMode, tColor)
    local filter = Post:GetNextFilter()
    
    if not amount then
        amount = 0.001
    end

    -- TODO: make "amount" tolerant of buffer size?
    
    filter:setInBuffer(rInBuffer)

    -- in mobile we will always use the low quality blur
    filter:setMaterial("Materials/Blur_low.material")
    filter:setShaderValue( "g_fBlurWidth", MOAIMaterial.VALUETYPE_FLOAT, amount )
    filter:setOutBuffer(rOutBuffer)
       
    if tColor then
        filter:setColor(tColor[1], tColor[2], tColor[3], tColor[4])
    end
    
    if blendMode then
        filter:setBlendMode(blendMode)
    end
    
    return filter
end

function Post:FXAAFilter(rInBuffer, rOutBuffer)
    local filter = Post:GetNextFilter()
    
    filter:setInBuffer(rInBuffer)
    filter:setMaterial("Materials/FXAA.material")
    local tInvBuffSize = { 1.0 / Constants.ScreenBufferWidth, 1.0 / Constants.ScreenBufferHeight }
    filter:setShaderValue( "g_vTexCoordOffset", MOAIMaterial.VALUETYPE_VEC2, tInvBuffSize )
    filter:setOutBuffer(rOutBuffer)
    
    return filter
end

function Post:WarpFilter(rInBuffer, rOutBuffer, warpIntensity, blendMode, tColor, dampenWarpNearTop)
    local filter = Post:GetNextFilter()
    
    if not warpIntensity then
        warpIntensity = 0.4
    end
    
    if not dampenWarpNearTop then
        dampenWarpNearTop = 0.0
    end

    filter:setInBuffer(rInBuffer)
    filter:setMaterial("Materials/Warp.material")
    local tInvBuffSize = { 1.0 / Constants.ScreenBufferWidth, 1.0 / Constants.ScreenBufferHeight }
    filter:setShaderValue( "g_vTexCoordOffset", MOAIMaterial.VALUETYPE_VEC2, tInvBuffSize )
    filter:setShaderValue( "g_fWarpIntensity", MOAIMaterial.VALUETYPE_FLOAT,  warpIntensity )
    filter:setOutBuffer(rOutBuffer)
       
    if tColor then
        filter:setColor(tColor[1], tColor[2], tColor[3], tColor[4])
    end
    
    if blendMode then
        filter:setBlendMode(blendMode)
    end
    
    return filter
end

function Post:CrossBlurFilter(rInBuffer, rOutBuffer, amount, blendMode, tColor)
    local filter = Post:GetNextFilter()
    
    if not amount then
        amount = 0.001
    end

    filter:setInBuffer(rInBuffer)
    filter:setMaterial("Materials/CrossBlur.material")
    filter:setShaderValue( "g_fBlurWidth", MOAIMaterial.VALUETYPE_FLOAT, amount )
    filter:setOutBuffer(rOutBuffer)
       
    if tColor then
        filter:setShaderValue( "g_vColor", MOAIMaterial.VALUETYPE_VEC4, tColor )
    end
    
    if blendMode then
        filter:setBlendMode(blendMode)
    end
    
    return filter
end


function Post:DirectionalBlurFilter(rInBuffer, rOutBuffer, direction, blendMode, tColor)
    local filter = Post:GetNextFilter()
    
    if not direction then
        direction = {0.03, 0.00}
    end

    filter:setInBuffer(rInBuffer)
    filter:setMaterial("Materials/DirectionalBlur.material")
    filter:setShaderValue( "g_vBlurDirection", MOAIMaterial.VALUETYPE_VEC2, direction )
    filter:setOutBuffer(rOutBuffer)
       
    if tColor then
        filter:setShaderValue( "g_vColor", MOAIMaterial.VALUETYPE_VEC4, tColor )
    end
    
    if blendMode then
        filter:setBlendMode(blendMode)
    end
    
    return filter
end


function Post:BlitFilter(rInBuffer, rOutBuffer, blendMode, tColor, tScale, tOffset)
    local filter = Post:GetNextFilter()
    
    filter:setInBuffer(rInBuffer)

    if not tColor then tColor = {1,1,1,1} end
    if not tScale then tScale = {1,1} end
    if not tOffset then tOffset = {0,0} end
    
    filter:setShaderValue( "g_vColor", MOAIMaterial.VALUETYPE_VEC4, tColor)
    filter:setShaderValue( "g_vScale", MOAIMaterial.VALUETYPE_VEC2, tScale)
    filter:setShaderValue( "g_vOffset", MOAIMaterial.VALUETYPE_VEC2, tOffset)
    
    filter:setOutBuffer(rOutBuffer)
    
    if blendMode then
        filter:setBlendMode(blendMode)
    end
    
    return filter
end

function Post:CombineFilter(rInBuffer1, rInBuffer2, rOutBuffer, blendMode, tColor1, tScale1, tOffset1, tColor2, tScale2, tOffset2)
    local filter = Post:GetNextFilter()
    
    filter:setMaterial("Materials/Combine.material") 
    filter:setInBuffer(rInBuffer1)

    if not tColor1 then tColor1 = {1,1,1,1} end
    if not tScale1 then tScale1 = {1,1} end
    if not tOffset1 then tOffset1 = {0,0} end
    
    if not tColor2 then tColor2 = {1,1,1,1} end
    if not tScale2 then tScale2 = {1,1} end
    if not tOffset2 then tOffset2 = {0,0} end
    
    filter:setShaderValue( "g_samSampler2", MOAIMaterial.VALUETYPE_TEXTURE, rInBuffer2)
    filter:setShaderValue( "g_vColor1", MOAIMaterial.VALUETYPE_VEC4, tColor1)
    filter:setShaderValue( "g_vScale1", MOAIMaterial.VALUETYPE_VEC2, tScale1)
    filter:setShaderValue( "g_vOffset1", MOAIMaterial.VALUETYPE_VEC2, tOffset1)   
    filter:setShaderValue( "g_vColor2", MOAIMaterial.VALUETYPE_VEC4, tColor2)
    filter:setShaderValue( "g_vScale2", MOAIMaterial.VALUETYPE_VEC2, tScale2)
    filter:setShaderValue( "g_vOffset2", MOAIMaterial.VALUETYPE_VEC2, tOffset2)
    
    filter:setOutBuffer(rOutBuffer)
    
    if blendMode then
        filter:setBlendMode(blendMode)
    end
    
    return filter
end


function Post:SuperBlurFilter(rInBuffer, rOutBuffer, amount, blendMode)
    local scratch = Post:GetUnusedBuffer()
    if not amount then amount = 3.0 end
    
    local intensity = .3
    Post:BlurFilter(rInBuffer, scratch, .004 * amount, kBLEND_ADD, {intensity,intensity,intensity,1})
    intensity = .2
    Post:BlurFilter(rInBuffer, scratch, .005 * amount, kBLEND_ADD, {intensity,intensity,intensity,1})
    intensity = .2
    Post:BlurFilter(rInBuffer, scratch, .006 * amount, kBLEND_ADD, {intensity,intensity,intensity,1})
    intensity = .2
    Post:BlurFilter(rInBuffer, scratch, .007 * amount, kBLEND_ADD, {intensity,intensity,intensity,1})
    intensity = .15
    Post:BlurFilter(rInBuffer, scratch, .008 * amount, kBLEND_ADD, {intensity,intensity,intensity,1})
    
    Post:BlitFilter(scratch, rOutBuffer, blendMode)
    
    scratch:SetInUse(false)
end

function Post:GenerateBasicCompFilter(rOutBuffer)
    Post:BlitFilter(Renderer.mBackgroundBuffer, rOutBuffer)
    Post:BlitFilter(Renderer.mGameBuffer, rOutBuffer)
    Post:BlitFilter(Renderer.mUIBuffer, rOutBuffer)
end


function Post:InvertFilter(rInBuffer, rOutBuffer, blendMode, tColor)
    local filter = Post:GetNextFilter()
    
    filter:setInBuffer(rInBuffer)
    filter:setMaterial("Materials/Invert.material")
    if tColor then
        filter:setColor(tColor[1], tColor[2], tColor[3], tColor[4])
    end
    filter:setOutBuffer(rOutBuffer)
    
    if blendMode then
        filter:setBlendMode(blendMode)
    end
    
    return filter
end

function Post:BeatFilter(rInBuffer, rOutBuffer, tTime, blendMode, tColor)
    local filter = Post:GetNextFilter()
    
    filter:setInBuffer(rInBuffer)
    filter:setMaterial("Materials/Beat.material")
    if tColor then
        filter:setColor(tColor[1], tColor[2], tColor[3], tColor[4])
    end
    if not tTime then tTime = 0.0 end
   
    filter:setShaderValue( "g_fBeat", MOAIMaterial.VALUETYPE_FLOAT, tTime)
    filter:setOutBuffer(rOutBuffer)
    
    if blendMode then
        filter:setBlendMode(blendMode)
    end
    
    return filter
end


function Post:DifferenceFilter(rInBuffer1, rInBuffer2, rOutBuffer, blendMode, tColor)
    local filter = Post:GetNextFilter()
    
    filter:setInBuffer(rInBuffer1)
    filter:setMaterial("Materials/Difference.material")
    filter:setShaderValue( "g_samDifferenceFrame", MOAIMaterial.VALUETYPE_TEXTURE, rInBuffer2)

    if tColor then
        filter:setColor(tColor[1], tColor[2], tColor[3], tColor[4])
    end
    filter:setOutBuffer(rOutBuffer)
    
    if blendMode then
        filter:setBlendMode(blendMode)
    end
    
    return filter
end

function Post:EdgeFilter(rInBuffer, rOutBuffer, width, blendMode, tColor)
    local filter = Post:GetNextFilter()
    
    filter:setInBuffer(rInBuffer)
    filter:setMaterial("Materials/Edges.material")
    if tColor then
        filter:setShaderValue("g_vColor", MOAIMaterial.VALUETYPE_VEC4, {tColor[1], tColor[2], tColor[3], tColor[4]})
    end
    
    if not width then width = 1 end
    filter:setShaderValue( "g_fEdgeWidth", MOAIMaterial.VALUETYPE_FLOAT, width)
    
    filter:setOutBuffer(rOutBuffer)
    
    if blendMode then
        filter:setBlendMode(blendMode)
    end
    
    return filter
end

--Accum filter returns the accumulation on the scratch buffer. It must be blitted back into the accum buffer manually
--this is so that you can do additional operations to it prior to saving it out for the next frame such as an additional blur
function Post:AccumFilter(rInBuffer, rAccumBuffer, rScratchBuffer, tColor, tScale, tOffset)
    local filter = Post:GetNextFilter()
     
    filter:setInBuffer(rInBuffer)
    filter:setMaterial("Materials/Accum.material")
    if tColor then
        filter:setColor(tColor[1], tColor[2], tColor[3], tColor[4])
    end
    if not tScale then tScale = {1,1} end
    if not tOffset then tOffset = {0,0} end
    
    filter:setShaderValue( "g_samAccum", MOAIMaterial.VALUETYPE_TEXTURE, rAccumBuffer)
    filter:setShaderValue( "g_vScale", MOAIMaterial.VALUETYPE_VEC2, tScale)
    filter:setShaderValue( "g_vOffset", MOAIMaterial.VALUETYPE_VEC2, tOffset)
    filter:setOutBuffer(rScratchBuffer)
    filter:setBlendMode(kBLEND_ADD)
    
    return filter
end

--Accum filter returns the accumulation on the scratch buffer. It must be blitted back into the accum buffer manually
--this is so that you can do additional operations to it prior to saving it out for the next frame such as an additional blur
function Post:AccumFlowFilter(rInBuffer, rAccumBuffer, rScratchBuffer, rFlowTexture, tColor, tScale, tOffset, flowIntensity)
    local filter = Post:GetNextFilter()
     
    filter:setInBuffer(rInBuffer)
    filter:setMaterial("Materials/AccumFlow.material")
    if tColor then
        filter:setColor(tColor[1], tColor[2], tColor[3], tColor[4])
    end
    if not tScale then tScale = {1,1} end
    if not tOffset then tOffset = {0,0} end
    
    filter:setShaderValue( "g_samAccum", MOAIMaterial.VALUETYPE_TEXTURE, rAccumBuffer)
    filter:setShaderValue( "g_samFlow", MOAIMaterial.VALUETYPE_TEXTURE, rFlowTexture)
    filter:setShaderValue( "g_vScale", MOAIMaterial.VALUETYPE_VEC2, tScale)
    filter:setShaderValue( "g_vOffset", MOAIMaterial.VALUETYPE_VEC2, tOffset)
    filter:setShaderValue( "g_fFlowIntensity", MOAIMaterial.VALUETYPE_FLOAT, flowIntensity)
    
    filter:setOutBuffer(rScratchBuffer)
    filter:setBlendMode(kBLEND_ADD)
    
    return filter    
end

function Post:EnergyRippleFilter(rInBuffer, rOutBuffer, tTime, blendMode, tColor)
    local filter = Post:GetNextFilter()
     
    filter:setInBuffer(rInBuffer)
    filter:setMaterial("Materials/EnergyRipple.material")
    if tColor then
        filter:setColor(tColor[1], tColor[2], tColor[3], tColor[4])
    end
    if not tTime then tTime = 0.0 end
    if not tColor then tColor = {1.0,1.0,1.0,1.0} end
   
    filter:setShaderValue( "g_fTime", MOAIMaterial.VALUETYPE_FLOAT, tTime)
    filter:setShaderValue( "g_vColor", MOAIMaterial.VALUETYPE_VEC4, tColor)
    filter:setOutBuffer(rOutBuffer)
    
    if blendMode then
        filter:setBlendMode(blendMode)
    end
    
    return filter
end

function Post:TVNoiseFilter(rInBuffer, rOutBuffer, tTime, blendMode, tColor)
    local filter = Post:GetNextFilter()
     
    filter:setInBuffer(rInBuffer)
    filter:setMaterial("Materials/TVNoise.material")
    if tColor then
        filter:setColor(tColor[1], tColor[2], tColor[3], tColor[4])
    end
    if not tTime then tTime = 0.0 end
    if not tColor then tColor = {1.0,1.0,1.0,1.0} end
   
    filter:setShaderValue( "g_fTime", MOAIMaterial.VALUETYPE_FLOAT, tTime)
    filter:setShaderValue( "g_vColor", MOAIMaterial.VALUETYPE_VEC4, tColor)
    filter:setOutBuffer(rOutBuffer)
    
    if blendMode then
        filter:setBlendMode(blendMode)
    end
    
    return filter
end

function Post:RemapFilter(rInBuffer, rOutBuffer, rRemapTexture, scrollRate, blendMode, tColor)
    local filter = Post:GetNextFilter()
     
    filter:setInBuffer(rInBuffer)
    filter:setMaterial("Materials/Remap.material")
    if tColor then
        filter:setColor(tColor[1], tColor[2], tColor[3], tColor[4])
    end
    
    if not scrollRate then scrollRate = 0 end
    filter:setShaderValue( "g_fScrollRate", MOAIMaterial.VALUETYPE_FLOAT, scrollRate)
    filter:setShaderValue( "g_samGradient", MOAIMaterial.VALUETYPE_TEXTURE, rRemapTexture)
    filter:setOutBuffer(rOutBuffer)
    
    if blendMode then
        filter:setBlendMode(blendMode)
    end
    
    return filter    
end   

-----------------------------------------------------------------------------
-- Custom uber filters
-----------------------------------------------------------------------------

function Post:LeaderboardAFilter(rInBuffer1, rInBuffer2, rInBuffer3, rOutBuffer)
    local filter = Post:GetNextFilter()

    filter:setInBuffer(rInBuffer1)
    filter:setMaterial("Materials/LeaderboardA.material")
    filter:setShaderValue( "g_samSampler2", MOAIMaterial.VALUETYPE_TEXTURE, rInBuffer2)
    filter:setShaderValue( "g_samSampler3", MOAIMaterial.VALUETYPE_TEXTURE, rInBuffer3)
    
    filter:setOutBuffer(rOutBuffer)
    
    return filter    
end  

function Post:LeaderboardBFilter(rInBuffer1, rInBuffer2, rInBuffer3, rInBuffer4, rOutBuffer)
    local filter = Post:GetNextFilter()

    filter:setInBuffer(rInBuffer1)
    filter:setMaterial("Materials/LeaderboardB.material")
    filter:setShaderValue( "g_samSampler2", MOAIMaterial.VALUETYPE_TEXTURE, rInBuffer2)
    filter:setShaderValue( "g_samSampler3", MOAIMaterial.VALUETYPE_TEXTURE, rInBuffer3)
    filter:setShaderValue( "g_samSampler4", MOAIMaterial.VALUETYPE_TEXTURE, rInBuffer4)
    filter:setShaderValue( "g_samSampler5", MOAIMaterial.VALUETYPE_TEXTURE, Post.rNoiseTexture)
    
    filter:setOutBuffer(rOutBuffer)
    
    return filter    
end 

function Post:Level0AFilter(rInBuffer1, rInBuffer2, rOutBuffer)
    local filter = Post:GetNextFilter()

    filter:setInBuffer(rInBuffer1)
    filter:setMaterial("Materials/Level0A.material")
    filter:setShaderValue( "g_samSampler2", MOAIMaterial.VALUETYPE_TEXTURE, rInBuffer2)
    
    filter:setOutBuffer(rOutBuffer)
    
    return filter    
end  

function Post:Level0BFilter(rInBuffer1, rInBuffer2, rInBuffer3, rOutBuffer)
    local filter = Post:GetNextFilter()

    filter:setInBuffer(rInBuffer1)
    filter:setMaterial("Materials/Level0B.material")
    filter:setShaderValue( "g_samSampler2", MOAIMaterial.VALUETYPE_TEXTURE, rInBuffer2)
    filter:setShaderValue( "g_samSampler3", MOAIMaterial.VALUETYPE_TEXTURE, rInBuffer3)
    
    filter:setOutBuffer(rOutBuffer)
    
    return filter    
end 

function Post:Level1AFilter(rInBuffer1, rInBuffer2, rInBuffer3, rOutBuffer)
    local filter = Post:GetNextFilter()

    filter:setInBuffer(rInBuffer1)
    filter:setMaterial("Materials/Level1A.material")
    filter:setShaderValue( "g_samAccum", MOAIMaterial.VALUETYPE_TEXTURE, rInBuffer2)
    filter:setShaderValue( "g_samSampler2", MOAIMaterial.VALUETYPE_TEXTURE, rInBuffer3)
    
    filter:setOutBuffer(rOutBuffer)
    
    return filter    
end  

function Post:Level1BFilter(rInBuffer1, rInBuffer2, rOutBuffer)
    local filter = Post:GetNextFilter()

    filter:setInBuffer(rInBuffer1)
    filter:setMaterial("Materials/Level1B.material")
    filter:setShaderValue( "g_samSampler2", MOAIMaterial.VALUETYPE_TEXTURE, rInBuffer2)
    
    filter:setOutBuffer(rOutBuffer)
    
    return filter    
end  

function Post:Level1CFilter(rInBuffer1, rInBuffer2, rOutBuffer)
    local filter = Post:GetNextFilter()

    filter:setInBuffer(rInBuffer1)
    filter:setMaterial("Materials/Level1C.material")
    filter:setShaderValue( "g_samSampler2", MOAIMaterial.VALUETYPE_TEXTURE, rInBuffer2)
    
    filter:setOutBuffer(rOutBuffer)
    
    return filter    
end  

function Post:Level1DFilter(rInBuffer1, rInBuffer2, rInBuffer3, rInBuffer4, rOutBuffer)
    local filter = Post:GetNextFilter()

    filter:setInBuffer(rInBuffer1)
    filter:setMaterial("Materials/Level1D.material")
    filter:setShaderValue( "g_samSampler2", MOAIMaterial.VALUETYPE_TEXTURE, rInBuffer2)
    filter:setShaderValue( "g_samSampler3", MOAIMaterial.VALUETYPE_TEXTURE, rInBuffer3)
    filter:setShaderValue( "g_samSampler4", MOAIMaterial.VALUETYPE_TEXTURE, rInBuffer4)
    
    filter:setOutBuffer(rOutBuffer)
    
    return filter    
end  

function Post:Level1EFilter(rInBuffer1, rInBuffer2, rInBuffer3, rOutBuffer)
    local filter = Post:GetNextFilter()

    filter:setInBuffer(rInBuffer1)
    filter:setMaterial("Materials/Level1E.material")
    filter:setShaderValue( "g_samSampler2", MOAIMaterial.VALUETYPE_TEXTURE, rInBuffer2)
    filter:setShaderValue( "g_samSampler3", MOAIMaterial.VALUETYPE_TEXTURE, rInBuffer3)
    
    filter:setOutBuffer(rOutBuffer)
    
    return filter    
end  

function Post:Level1FFilter(rInBuffer1, rInBuffer2, rInBuffer3, rOutBuffer)
    local filter = Post:GetNextFilter()

    filter:setInBuffer(rInBuffer1)
    filter:setMaterial("Materials/Level1F.material")
    filter:setShaderValue( "g_samSampler2", MOAIMaterial.VALUETYPE_TEXTURE, rInBuffer2)
    filter:setShaderValue( "g_samSampler3", MOAIMaterial.VALUETYPE_TEXTURE, rInBuffer3)
    
    filter:setOutBuffer(rOutBuffer)
    
    return filter    
end  

function Post:Level1GFilter(rInBuffer1, rInBuffer2, rInBuffer3, rOutBuffer)
    local filter = Post:GetNextFilter()

    filter:setInBuffer(rInBuffer1)
    filter:setMaterial("Materials/Level1G.material")
    filter:setShaderValue( "g_samSampler2", MOAIMaterial.VALUETYPE_TEXTURE, rInBuffer2)
    filter:setShaderValue( "g_samSampler3", MOAIMaterial.VALUETYPE_TEXTURE, rInBuffer3)
    
    filter:setOutBuffer(rOutBuffer)
    
    return filter    
end  

function Post:Level1HFilter(rInBuffer1, rInBuffer2, rInBuffer3, rOutBuffer)
    local filter = Post:GetNextFilter()

    filter:setInBuffer(rInBuffer1)
    filter:setMaterial("Materials/Level1H.material")
    filter:setShaderValue( "g_samSampler2", MOAIMaterial.VALUETYPE_TEXTURE, rInBuffer2)
    filter:setShaderValue( "g_samSampler3", MOAIMaterial.VALUETYPE_TEXTURE, rInBuffer3)
    
    filter:setOutBuffer(rOutBuffer)
    
    return filter    
end 

function Post:Level2AFilter(rInBuffer1, rInBuffer2, rInBuffer3, rOutBuffer, tColor, tScale)
    local filter = Post:GetNextFilter()

    filter:setInBuffer(rInBuffer1)
    filter:setMaterial("Materials/Level2A.material")
    if not tColor then tColor = {1.0,1.0,1.0,1.0} end
    filter:setShaderValue( "g_vColor", MOAIMaterial.VALUETYPE_VEC4, tColor)
    if not tScale then tScale = {1,1} end
    filter:setShaderValue( "g_vScale", MOAIMaterial.VALUETYPE_VEC2, tScale)
    filter:setShaderValue( "g_samSampler2", MOAIMaterial.VALUETYPE_TEXTURE, rInBuffer2)
    filter:setShaderValue( "g_samSampler3", MOAIMaterial.VALUETYPE_TEXTURE, rInBuffer3)
    
    filter:setOutBuffer(rOutBuffer)
    
    return filter    
end  

function Post:Level2BFilter(rInBuffer1, rInBuffer2, rInBuffer3, rOutBuffer)
    local filter = Post:GetNextFilter()

    filter:setInBuffer(rInBuffer1)
    filter:setMaterial("Materials/Level2B.material")
    filter:setShaderValue( "g_samSampler2", MOAIMaterial.VALUETYPE_TEXTURE, rInBuffer2)
    filter:setShaderValue( "g_samSampler3", MOAIMaterial.VALUETYPE_TEXTURE, rInBuffer3)
    
    filter:setOutBuffer(rOutBuffer)
    
    return filter    
end  

function Post:Level2CFilter(rInBuffer1, rInBuffer2, rOutBuffer, amount, tColor, tColor2, tColor3, tColor4)
    local filter = Post:GetNextFilter()

    filter:setInBuffer(rInBuffer1)
    filter:setMaterial("Materials/Level2C.material")
    filter:setShaderValue( "g_samSampler2", MOAIMaterial.VALUETYPE_TEXTURE, rInBuffer2)
    if not amount then amount = 1 end
    filter:setShaderValue( "g_fBlurWidth", MOAIMaterial.VALUETYPE_FLOAT, amount )
    if not tColor then tColor = {1.0,1.0,1.0,1.0} end
    filter:setShaderValue( "g_vColor", MOAIMaterial.VALUETYPE_VEC4, tColor)
    if not tColor2 then tColor2 = {1.0,1.0,1.0,1.0} end
    filter:setShaderValue( "g_vColor2", MOAIMaterial.VALUETYPE_VEC4, tColor2)
    if not tColor3 then tColor3 = {1.0,1.0,1.0,1.0} end
    filter:setShaderValue( "g_vColor3", MOAIMaterial.VALUETYPE_VEC4, tColor3)
    if not tColor4 then tColor4 = {1.0,1.0,1.0,1.0} end
    filter:setShaderValue( "g_vColor4", MOAIMaterial.VALUETYPE_VEC4, tColor4)
    filter:setOutBuffer(rOutBuffer)
    
    return filter    
end 

function Post:Level3AFilter(rInBuffer1, rInBuffer2, rInBuffer3, rOutBuffer, tColor, tScale)
    local filter = Post:GetNextFilter()

    filter:setInBuffer(rInBuffer1)
    filter:setMaterial("Materials/Level3A.material")
    filter:setShaderValue( "g_samSampler2", MOAIMaterial.VALUETYPE_TEXTURE, rInBuffer2)
    filter:setShaderValue( "g_samSampler3", MOAIMaterial.VALUETYPE_TEXTURE, rInBuffer3)
    if not tColor then tColor = {1.0,1.0,1.0,1.0} end
    filter:setShaderValue( "g_vColor", MOAIMaterial.VALUETYPE_VEC4, tColor)
    if not tScale then tScale = {1,1} end
    filter:setShaderValue( "g_vScale", MOAIMaterial.VALUETYPE_VEC2, tScale)
    filter:setOutBuffer(rOutBuffer)
    
    return filter    
end 

function Post:Level3BFilter(rInBuffer1, rInBuffer2, rInBuffer3, rOutBuffer)
    local filter = Post:GetNextFilter()

    filter:setInBuffer(rInBuffer1)
    filter:setMaterial("Materials/Level3B.material")
    filter:setShaderValue( "g_samSampler2", MOAIMaterial.VALUETYPE_TEXTURE, rInBuffer2)
    filter:setShaderValue( "g_samSampler3", MOAIMaterial.VALUETYPE_TEXTURE, rInBuffer3)
    filter:setOutBuffer(rOutBuffer)
    
    return filter    
end 

function Post:Level3CFilter(rInBuffer1, rInBuffer2, rInBuffer3, rOutBuffer, tColor, tColor2, tColor3)
    local filter = Post:GetNextFilter()

    filter:setInBuffer(rInBuffer1)
    filter:setMaterial("Materials/Level3C.material")
    filter:setShaderValue( "g_samSampler2", MOAIMaterial.VALUETYPE_TEXTURE, rInBuffer2)
    filter:setShaderValue( "g_samSampler3", MOAIMaterial.VALUETYPE_TEXTURE, rInBuffer3)
    if not tColor then tColor = {1.0,1.0,1.0,1.0} end
    filter:setShaderValue( "g_vColor", MOAIMaterial.VALUETYPE_VEC4, tColor)
    if not tColor2 then tColor2 = {1.0,1.0,1.0,1.0} end
    filter:setShaderValue( "g_vColor2", MOAIMaterial.VALUETYPE_VEC4, tColor2)
    if not tColor3 then tColor3 = {1.0,1.0,1.0,1.0} end
    filter:setShaderValue( "g_vColor3", MOAIMaterial.VALUETYPE_VEC4, tColor3)
    filter:setOutBuffer(rOutBuffer)
    
    return filter    
end 

function Post:Level4AFilter(rInBuffer1, rInBuffer2, rOutBuffer, tScale)
    local filter = Post:GetNextFilter()

    filter:setInBuffer(rInBuffer1)
    filter:setMaterial("Materials/Level4A.material")
    filter:setShaderValue( "g_samSampler2", MOAIMaterial.VALUETYPE_TEXTURE, rInBuffer2)
    if not tScale then tScale = {1,1} end
    filter:setShaderValue( "g_vScale", MOAIMaterial.VALUETYPE_VEC2, tScale)
    filter:setOutBuffer(rOutBuffer)
    
    return filter    
end 

function Post:Level4BFilter(rInBuffer1, rInBuffer2, rInBuffer3, rOutBuffer)
    local filter = Post:GetNextFilter()

    filter:setInBuffer(rInBuffer1)
    filter:setMaterial("Materials/Level4B.material")
    filter:setShaderValue( "g_samSampler2", MOAIMaterial.VALUETYPE_TEXTURE, rInBuffer2)
    filter:setShaderValue( "g_samSampler3", MOAIMaterial.VALUETYPE_TEXTURE, rInBuffer3)
    filter:setOutBuffer(rOutBuffer)
    
    return filter    
end 

function Post:Level4CFilter(rInBuffer1, rInBuffer2, rInBuffer3, rOutBuffer)
    local filter = Post:GetNextFilter()

    filter:setInBuffer(rInBuffer1)
    filter:setMaterial("Materials/Level4C.material")
    filter:setShaderValue( "g_samSampler2", MOAIMaterial.VALUETYPE_TEXTURE, rInBuffer2)
    filter:setShaderValue( "g_samSampler3", MOAIMaterial.VALUETYPE_TEXTURE, rInBuffer3)
    filter:setOutBuffer(rOutBuffer)
    
    return filter    
end 

function Post:Level5Filter(rInBuffer1, rInBuffer2, rInBuffer3, rOutBuffer, tColor)
    local filter = Post:GetNextFilter()

    filter:setInBuffer(rInBuffer1)
    filter:setMaterial("Materials/Level5.material")
    filter:setShaderValue( "g_samSampler2", MOAIMaterial.VALUETYPE_TEXTURE, rInBuffer2)
    filter:setShaderValue( "g_samSampler3", MOAIMaterial.VALUETYPE_TEXTURE, rInBuffer3)
    if not tColor then tColor = {1.0,1.0,1.0,1.0} end
    filter:setShaderValue( "g_vColor", MOAIMaterial.VALUETYPE_VEC4, tColor)
    filter:setOutBuffer(rOutBuffer)
    
    return filter    
end 

function Post:Level6AFilter(rInBuffer1, rInBuffer2, rInBuffer3, rOutBuffer, tScale, tScale2)
    local filter = Post:GetNextFilter()

    filter:setInBuffer(rInBuffer1)
    filter:setMaterial("Materials/Level6A.material")
    if not tScale then tScale = {1,1} end
    filter:setShaderValue( "g_vScale", MOAIMaterial.VALUETYPE_VEC2, tScale)
    if not tScale2 then tScale2 = {1,1} end
    filter:setShaderValue( "g_vScale2", MOAIMaterial.VALUETYPE_VEC2, tScale2)
    filter:setShaderValue( "g_samSampler2", MOAIMaterial.VALUETYPE_TEXTURE, rInBuffer2)
    filter:setShaderValue( "g_samSampler3", MOAIMaterial.VALUETYPE_TEXTURE, rInBuffer3)
    
    filter:setOutBuffer(rOutBuffer)
    
    return filter    
end 

function Post:Level6BFilter(rInBuffer1, rInBuffer2, rOutBuffer)
    local filter = Post:GetNextFilter()

    filter:setInBuffer(rInBuffer1)
    filter:setMaterial("Materials/Level6B.material")
    filter:setShaderValue( "g_samSampler2", MOAIMaterial.VALUETYPE_TEXTURE, rInBuffer2)
    filter:setOutBuffer(rOutBuffer)
    
    return filter    
end  

function Post:Level7Filter(rInBuffer1, rInBuffer2, rInBuffer3, rOutBuffer, tScale, tScale2)
    local filter = Post:GetNextFilter()

    filter:setInBuffer(rInBuffer1)
    filter:setMaterial("Materials/Level7.material")
    if not tScale then tScale = {1,1} end
    filter:setShaderValue( "g_vScale", MOAIMaterial.VALUETYPE_VEC2, tScale)
    if not tScale2 then tScale2 = {1,1} end
    filter:setShaderValue( "g_vScale2", MOAIMaterial.VALUETYPE_VEC2, tScale2)
    filter:setShaderValue( "g_samSampler2", MOAIMaterial.VALUETYPE_TEXTURE, rInBuffer2)
    filter:setShaderValue( "g_samSampler3", MOAIMaterial.VALUETYPE_TEXTURE, rInBuffer3)
    
    filter:setOutBuffer(rOutBuffer)
    
    return filter    
end  

function Post:Level7BFilter(rInBuffer1, rInBuffer2, rOutBuffer, tColor, tScale, tBGstrength)
    local filter = Post:GetNextFilter()

    filter:setInBuffer(rInBuffer1)
    filter:setMaterial("Materials/Level7B.material")
    filter:setShaderValue( "g_samSampler2", MOAIMaterial.VALUETYPE_TEXTURE, rInBuffer2)
    if not tScale then tScale = {1,1} end
    filter:setShaderValue( "g_vScale", MOAIMaterial.VALUETYPE_VEC2, tScale)
    if not tColor then tColor = {1.0,1.0,1.0,1.0} end
    filter:setShaderValue( "g_vColor", MOAIMaterial.VALUETYPE_VEC4, tColor)
    if not tBGstrength then tBGstrength = 1 end
    filter:setShaderValue( "g_fBackgroundStrength", MOAIMaterial.VALUETYPE_FLOAT, tBGstrength)
    filter:setBlendMode(kBLEND_ADD)
    filter:setOutBuffer(rOutBuffer)
    
    return filter    
end 

function Post:Level7CFilter(rInBuffer1, rInBuffer2, rOutBuffer, tColor, tScale)
    local filter = Post:GetNextFilter()

    filter:setInBuffer(rInBuffer1)
    filter:setMaterial("Materials/Level7C.material")
    filter:setShaderValue( "g_samSampler2", MOAIMaterial.VALUETYPE_TEXTURE, rInBuffer2)
    if not tColor then tColor = {1.0,1.0,1.0,1.0} end
    filter:setShaderValue( "g_vColor", MOAIMaterial.VALUETYPE_VEC4, tColor)
    if not tScale then tScale = {1,1} end
    filter:setShaderValue( "g_vScale", MOAIMaterial.VALUETYPE_VEC2, tScale)
    filter:setOutBuffer(rOutBuffer)
    
    return filter    
end  

function Post:Level9Filter(rInBuffer1, rInBuffer2, rOutBuffer)
    local filter = Post:GetNextFilter()

    filter:setInBuffer(rInBuffer1)
    filter:setMaterial("Materials/Level9.material")
    filter:setShaderValue( "g_samSampler2", MOAIMaterial.VALUETYPE_TEXTURE, rInBuffer2)
    
    filter:setOutBuffer(rOutBuffer)
    
    return filter    
end  

function Post:Level10Filter(rInBuffer1, rInBuffer2, rInBuffer3, rOutBuffer)
    local filter = Post:GetNextFilter()

    filter:setInBuffer(rInBuffer1)
    filter:setMaterial("Materials/Level10.material")
    filter:setShaderValue( "g_samSampler2", MOAIMaterial.VALUETYPE_TEXTURE, rInBuffer2)
    filter:setShaderValue( "g_samSampler3", MOAIMaterial.VALUETYPE_TEXTURE, rInBuffer3)
    
    filter:setOutBuffer(rOutBuffer)
    
    return filter    
end  

-----------------------------------------------------------------------------
-- Oculus Rift filter
-----------------------------------------------------------------------------

function Post:OculusRiftFilter(rInBuffer1, rOutBuffer, rLensCenter, rScreenCenter, rScale, rScaleIn, rHmdWarpParam)
    local filter = Post:GetNextFilter()

    filter:setInBuffer(rInBuffer1)
    filter:setMaterial("Materials/OculusRift.material")   
    filter:setShaderValue( "LensCenter", MOAIMaterial.VALUETYPE_VEC2, rLensCenter)
    filter:setShaderValue( "ScreenCenter", MOAIMaterial.VALUETYPE_VEC2, rScreenCenter)
    filter:setShaderValue( "Scale", MOAIMaterial.VALUETYPE_VEC2, rScale)
    filter:setShaderValue( "ScaleIn", MOAIMaterial.VALUETYPE_VEC2, rScaleIn)
    filter:setShaderValue( "HmdWarpParam", MOAIMaterial.VALUETYPE_VEC4, rHmdWarpParam)
    filter:setOutBuffer(rOutBuffer)
    
    return filter    
end 

return Post
